/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_Graphics.c,v 1.35 2005/11/07 22:54:39 guich Exp $ */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#include "superwaba.h"
#include "SW_natives.h"
#include "SDL.h"
#include "SW_logging.h"


#ifndef max
#define max(x,y) (((x)>=(y))?(x):(y))
#define min(x,y) (((x)>=(y))?(y):(x))
#endif

#ifndef WIN32
#define abs(x) (((x)>0)?(x):-(x))
#endif

#define sgn(x) (((x)<0)?-1:(((x)>0)?1:0))

void drawHourGlass(Sint32 step);
void graphicsSwap(Sint32 on);
void graphicsStop();
void graphicsStart();
Uint8 graphicsInit(); // called by the vm to init the graphics
void graphicsSetup();
Uint8 loadAllFonts();
void releaseAllFonts();

Uint8 intersects(RectangleType *r1, RectangleType *r2);
#define makeRect(xx, yy, ww, hh, r) \
{ \
   r.topLeft.x = xx; \
   r.topLeft.y = yy; \
   r.extent.x = ww;  \
   r.extent.y = hh;  \
}
char *strcopy = "s@h78bc!";
////////////////////////////////////////////////////////////////////////////
static void selectGraphicsObj(WObject newGraphics);
void intersectWith(RectangleType *r1, RectangleType *r2);
Uint8 translateAndClip(Sint32 *x, Sint32 *y, Sint32 *width, Sint32 *height);
void setDrawWindow(GrHandle h);
void drawRow(Sint32 xStart, Sint32 xEnd, Sint32 y, Uint8 * rgbOrIndex, Uint32 * colorMap);
void graphicsStartImage(Uint32 * ret, Uint32 * args);
void graphicsEndImage(GrHandle oldG);

Sint32 transXY; // saves the translation coords x in the upper 16 bits and the y in the lower 16 bits
Sint32 clipX, clipY; // saves the clip coords. the lower 16 bits have x1/y1 and the upper 16 bits have x2/y2
Sint32 maxX, maxY;
GrHandle mainGrHandle;
GrHandle activeGrHandle;
WObject activeGraphics;
Pixel currentForePixel;
Pixel currentBackPixel;
WObject lastColor;
Pixel lastUsedPixel;
struct UserFont * currentUserFont;
WClass * coordClass;
WClass * imageClass;

// Note: Colors in PalmOS are stored as the index to the current palette
// and in WindowsCE are stored as the color itself
// For SDL, the color is either an index (bpp <= 8) , or a true color

/*
| Hashtable used to speedup the usualy slow process that converts
| a user Color in the corresponding device's color.
| In some PalmOS devices, this speeds up to 200%
*/
#define COLOR_HASHSIZE 0xFF // must be a 11..111 (binary) sequence
#define htHash(key) (unsigned)((key) & COLOR_HASHSIZE)
struct htElemColor
{
   Uint32 key;
   Sint32 value;
   struct htElemColor * next;
};

void drawRect(Sint32 x, Sint32 y, Sint32 w, Sint32 h, Pixel p); // can't be static
void fillRect(Sint32 x, Sint32 y, Sint32 w, Sint32 h, Pixel p); // can't be static
// static void FreeGrHandle(GrHandle h);  // defined in graphics.c

/** Variables used in this platform */
// Sint32 lastBackColor = -1;
// Sint32 lastForeColor = -1;

static struct htElemColor * colorHashTable[COLOR_HASHSIZE+1];
static Uint8 graphicsActive;

#define SDL_USEREVENT_REFRESH SDL_USEREVENT+1
#ifdef PALMOS
extern int use8bppEmulation; // guich: zire21 will create 8bpp offscreen and will let the OS do on-the-fly convertion to 4bpp
#else
int use8bppEmulation = 0;
#endif

/*-------------------------------------------------------createWebSafePalette-+
|                                                                             |
+----------------------------------------------------------------------------*/
static SDL_Color webpalette[256]; // guich: initialize the palette only once. make it avail for paletteGetCurrentPalette
static int webpalInitialized;
static void setWebSafePalette(SDL_Surface * screen) // guich@421_20
{
   if (!webpalInitialized) // initialize only once
   {
      int32 palAux[] =
      {
         0x111111,0x222222,0x444444,0x555555,0x777777,0x888888,0xAAAAAA,0xBBBBBB,0xDDDDDD,0xEEEEEE,
         0xC0C0C0,0x800000,0x800080,0x008000,0x008080
      };
      Sint32 i=0,r,g,b;
      SDL_Color *newpal = webpalette;
      Sint32 samples[] = {0x00, 0x33, 0x66, 0x99, 0xcc, 0xff};
      Sint32 steps[] = {120,114,108,12,6,0};
      xmemzero(webpalette,sizeof(webpalette));
      for (r = 5; r >=0; r--,i += 12)
         for (g = 5; g >=0; g--,i++)
            for (b = 5; b >= 0; b--)
            {
               SDL_Color *p = &newpal[i+steps[b]];
               p->r = samples[r];
               p->g = samples[g];
               p->b = samples[b];
            }
      for (r = 0; r < 15; r++)
      {
         SDL_Color *p = &newpal[215+r];
         p->r = palAux[r]>>16;
         p->g = palAux[r]>>8;
         p->b = palAux[r]&0xFF;
      }
      newpal[255].r = 0x06; newpal[255].g = 0x00; newpal[255].b = 0x03;
      webpalInitialized = 1;
   }
   if (screen)
      SDL_SetPalette(screen, use8bppEmulation ? SDL_LOGPAL : (SDL_LOGPAL|SDL_PHYSPAL), webpalette, 0, 256);
}
SDL_Color* getStaticWebPalette()
{
   return webpalette;
}

/*------------------------------------------------------------updateRectangle-+
| Add a new refresh rectangle event (or extends the current one)              |
+----------------------------------------------------------------------------*/
void updateRectangle(SDL_Surface * surface, Sint32 x, Sint32 y, Sint32 w, Sint32 h)
{
   if ((surface == SDL_GetVideoSurface()) && w && h && (x < (Sint32)(surface->w)) && (y < (Sint32)(surface->h)))
   {
      SDL_Event event;

      /* Clip with the Screen surface */
      if (x < 0) x = 0;
      if (y < 0) y = 0;
      if ((x+w) > (Sint32)(surface->w)) w = surface->w - x;
      if ((y+h) > (Sint32)(surface->h)) h = surface->h - y;

      /* Pull out the previous refresh event, if any, and update x, y, h, w */
      if (SDL_PeepEvents(&event, 1, SDL_GETEVENT, SDL_EVENTMASK(SDL_USEREVENT_REFRESH)) > 0)
      {
         Sint32 xP = (Sint32)(Sint16)((Sint32)event.user.data1);
         Sint32 yP = (Sint32)((Sint32)event.user.data1 >> 16);
         Sint32 wP = (Sint32)(Sint16)((Sint32)event.user.data2); // guich@503_11: fixed problem when updating UIGadget's vertical scrollbar (was Uint16, replaced by Sint16
         Sint32 hP = (Sint32)((Sint32)event.user.data2 >> 16);
         if (xP < x)
         {
            w = ((xP+wP) > (x+w)) ? wP : ((x+w) - xP);
            x = xP;
         }
         else
         if ((xP+wP) > (x+w))
            w = (xP+wP) - x;

         if (yP < y)
         {
            h = ((yP+hP) > (y+h)) ? hP : ((y+h) - yP);
            y = yP;
         }
         else
         if ((yP+hP) > (y+h))
            h = (yP+hP) - y;
      }
      else
      {
         if (x < 0) x = 0;
         if (y < 0) y = 0;
         event.type = SDL_USEREVENT_REFRESH;
         event.user.code = 0;
      }
      event.user.data1 = (void *)(x | ((int32)y  << 16));
      event.user.data2 = (void *)(w | ((int32)h  << 16));
      SDL_PushEvent(&event);
   }
}

/*----------------------------------------------------------------------htGet-+
| Description:                                                                |
|   Given a key, returns the htElemColor found.                               |
|   If key doesnt exist, NULL is returned                                     |
+----------------------------------------------------------------------------*/
struct htElemColor * htGet(struct htElemColor * ht[], Uint32 key);
struct htElemColor * htGet(struct htElemColor * ht[], Uint32 key)
{
   struct htElemColor *np;
   for (np = ht[htHash(key)]; np; np = np->next)
      if (np->key == key) return np;
   return NULL;
}

/*----------------------------------------------------------------------htPut-+
| Description:                                                                |
|   Stores a pair key/value in the given hashtable.                           |
|   Returns the structure created or NULL if out of memory                    |
+----------------------------------------------------------------------------*/
struct htElemColor * htPut(struct htElemColor * ht[], Uint32 key, Sint32 value);
struct htElemColor * htPut(struct htElemColor * ht[], Uint32 key, Sint32 value)
{
   struct htElemColor *np;
   unsigned hashval;
   if ((np = htGet(ht, key)) == NULL) // duplicated keys not allowed
   {
      np = (struct htElemColor *)(*vmGlobals->xmalloc)(sizeof(struct htElemColor));
      if (np)
      {
         hashval = htHash(key);
         np->next = ht[hashval];  /* point to last w/ this hash */
         np->key = key;
         np->value = value;
         ht[hashval] = np;       /* attach to hashtab array */
      }
   }
   return np;
}

/*-----------------------------------------------------------disposeHashtable-+
| Description:                                                                |
|   Releases the memory used by the ColorHashtable                            |
+----------------------------------------------------------------------------*/
void disposeHashtable(struct htElemColor *ht[], Uint8 deleteObj);
void disposeHashtable(struct htElemColor *ht[], Uint8 deleteObj)
{
   Sint32 i;
   for (i = 0; i <= COLOR_HASHSIZE; i++)
   {
      struct htElemColor *np;
      np = ht[i];
      while (np)
      {
         struct htElemColor *next;
         next = np->next;
         xfree(np);
         np = next;
      }
      ht[i] = NULL;
   }
}

/*-------------------------------------------------------------------getColor-+
| Description:                                                                |
|    Returns the color obtained in rgb2index.                                 |
|    Cache things in a hash table for performance.                            |
+----------------------------------------------------------------------------*/
static Pixel getColor(Sint16 red, Sint16 green, Sint16 blue)
{
   Uint32 key = ((Sint32)red << 16) | ((Sint32)green << 8) | blue;
   struct htElemColor * elem = htGet(colorHashTable, key);

   if (elem)
   {
      return (Pixel)elem->value;
   }
   else
   {
      Pixel val = SDL_MapRGB(SDL_GetVideoSurface()->format, (Uint8)red, (Uint8)green, (Uint8)blue);
      _LOG_DEBUG((LOG_CAT_GFX, "getColor=(0x%x,0x%x,0x%x)", (int)red, (int)green, (int)blue));
/* TODO handled in SDL_MapRGB now, tests required
#if defined(PALMOS) && defined(__arm__)
      if (sizeof(Pixel) == 2)
      {
         unsigned char *p = (unsigned char *)&val;
         _LOG_DEBUG((LOG_CAT_GFX, "rawcolor=(0x%x,0x%x,0x%x,0x%x)", (int)p[0], (int)p[1], (int)p[2], (int)p[3]));
         //val = (p[0] << 16) | (p[1] << 8) | (p[2]);
         val = p[0] << 8 | p[1];
         _LOG_DEBUG((LOG_CAT_GFX, "fixedcolor=0x%x", (int)val));
      }
#endif
*/
      htPut(colorHashTable, key, val);
      return val;
   }
}

/*-----------------------------------------------------------------setPalette-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Applies the given palette to the system. It caches the last               |
|   palette to avoid aplying two times the same palette (and to               |
|   speedup the process)                                                      |
+----------------------------------------------------------------------------*/
// static void setPalette(int32 * pal, int start, int end)
// {
//    int len = end-start+1;
//    SDL_Surface * surface = SDL_GetVideoSurface();
//    if (surface->format->BitsPerPixel == 8)
//    {
//       SDL_Color * colors = (SDL_Color *)xmalloc(len * sizeof (SDL_Color));
//       int i;
//
//       for (i=0; i < len; ++i)
//       {
//          colors[i].r = (pal[i] >> 24) & 0x0F;
//          colors[i].g = (pal[i] >> 16) & 0x0F;
//          colors[i].b = pal[i] & 0x0F;
//       }
//       SDL_SetPalette(surface, SDL_LOGPAL|SDL_PHYSPAL, colors, start, len);
//       disposeHashtable(colorHashTable, 0); // cached colors now invalid.
//       xfree(colors);
//    }
// }

/*------------------------------------------------------------------getPixel2-+
| Get a Pixel from the source into into the target format.                    |
| Bump the source pointer.                                                    |
+----------------------------------------------------------------------------*/
#ifdef SURFACES_ARE_DIFFERENT
//>>>PGR: ifdef'ed.  it's total nosense.
static Pixel getPixel2(void ** sourcePixels,
                       SDL_PixelFormat * sourceFormat,
                       SDL_PixelFormat * targetFormat)
{
   Uint32 pixel;
   Uint8 r, g, b, a;
   Uint8 *p = *sourcePixels;
   switch(sourceFormat->BytesPerPixel)
   {
      case 1:
         pixel = *p;
         *sourcePixels = p + 1;
         break;
      case 2:
         pixel = *(Uint16 *)p;
         *sourcePixels = p + 2;
         break;
      case 3:
         // TODO create asm optimized getUInt24 versions.
         if (SDL_BYTEORDER == SDL_BIG_ENDIAN) // a good compiler will strip the not used code block
            pixel = ((Sint32)p[0] << 16) | ((Sint32)p[1] <<  8) | p[2];
         else
            pixel = ((Sint32)p[2] << 16) | ((Sint32)p[1] <<  8) | p[0];
         *sourcePixels = p + 3;
         break;
      default:   // case 4:
         pixel = *(Uint32 *)p;
         *sourcePixels = p + 4;
         break;
   }
   SDL_GetRGBA(pixel, sourceFormat, &r, &g, &b, &a);
   return (Pixel)SDL_MapRGBA(targetFormat, r, g, b, a);
}
#endif

/*-----------------------------------------------------------------drawBitmap-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   This is the main routine that draws a bitmap in the active                |
|   GrHandle. It must implement all drawOp defined.                           |
|   Gets the area srcX, srcY, width, height of the given bitmap and places    |
|   it at position dstX, dstY                                                 |
+----------------------------------------------------------------------------*/
static void drawBitmap(GrHandle hSource,
                       Sint32 srcX, Sint32 srcY, Sint32 width, Sint32 height,
                       Sint32 dstX, Sint32 dstY,
                       Sint32 drawOp,
                       Pixel backPixel, Pixel forePixel,
                       Sint32 doClip)
{
   Sint32 i;
   SDL_Surface * srcSurf = (SDL_Surface *)GRHANDLE_WinHandle(hSource);
   SDL_Surface * dstSurf = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
   SDL_Rect srcRect = { srcX, srcY, width, height };
   SDL_Rect dstRect = { dstX + TRANSX, dstY + TRANSY, width, height };

   _LOG_INFO((LOG_CAT_GFX, "drawBitmap src=%ld,%ld size=%ld,%ld dst=%ld,%ld op=%ld backC=%04x foreC=%04x",
                srcX, srcY, width, height, dstX, dstY, drawOp, backPixel, forePixel));
   if (!doClip)
   {
      /*
      | Even if no clip is required, we still have to make sure that the
      | area of the bitmap that we want to copy is inside its area.
      */
      if ((srcX <= -width)  || (srcX >= GRHANDLE_Width(hSource)) ||
          (srcY <= -height) || (srcY >= GRHANDLE_Height(hSource))) return;
      dstX += TRANSX;
      dstY += TRANSY;
   }
   else

   #ifdef USE_SDL_VALIDATE_RECT
   //>>>PGR: This is just for future reference purposes.
   //        I did this code when sending the SDL_ValidateRect Patch to SDL org.
   //        It assumes that we have migrated all the clipping onto the SDL side
   //        But, at today (Aug 1st, 2004) we still use SW clipping (CLIPX1, etc...)
   {
      if (!SDL_ValidateRect(srcSurf, srcRect, dstSurf, dstRect, srcRect))
         return;
      // update with valid values.
      srcX = srcRect.x;
      srcY = srcRect.y;
      width = srcRect.w;
      height = srcRect.h;
      dstX = dstRect.x;
      dstY = dstRect.y;
   }

   #else
   {
      dstX += TRANSX;
      dstY += TRANSY;

      /* clip the source rectangle to the source surface */
      if (srcX < 0)
      {
         width += srcX;
         dstX -= srcX;
         srcX = 0;
      }
      i = GRHANDLE_Width(hSource) - srcX;
      if (width > i)
         width = i;
      if (srcY < 0)
      {
         height += srcY;
         dstY -= srcY;
         srcY = 0;
      }
      i = GRHANDLE_Height(hSource) - srcY;
      if (height > i)
         height = i;

      /* clip the destination rectangle against the clip rectangle */
      if (dstX < CLIPX1)
      {
         i = CLIPX1 - dstX;
         dstX = CLIPX1;
         srcX += i;
         width -= i;
      }
      if ((dstX + width) > CLIPX2)
         width = CLIPX2 - dstX;
      if (dstY < CLIPY1)
      {
         i = CLIPY1 - dstY;
         dstY = CLIPY1;
         srcY += i;
         height -= i;
      }
      if ((dstY + height) > CLIPY2)
         height = CLIPY2 - dstY;

      /* check the validity */
      if ((width <= 0) || (height <= 0))
         return;

      srcRect.x = srcX;
      srcRect.y = srcY;
      srcRect.w = width;
      srcRect.h = height;
      dstRect.x = dstX;
      dstRect.y = dstY;
      dstRect.w = width;
      dstRect.h = height;
   }
   #endif // USE_SDL_VALIDATE_RECT (50 lines of not-so-obvious code are gone)

   // lets first do the obvious and most frequent cases...
   _LOG_INFO((LOG_CAT_GFX, "drawBitmap, drawOp=%d", (int)drawOp));

   switch (drawOp)
   {
      case WIN_SPRITE:        // Dest replaced if source pixels != color
         _LOG_DEBUG((LOG_CAT_GFX, "do WIN_SPRITE"));
         if (GRHANDLE_VideoFlag(hSource) != VIDEOFLAG_DIFFERENT_OFFSCREEN)
         {
            /*
            | DIFFERENT_OFFSCREEN also means that this image was built
            | by SDL_Image.  In this case, just let DRAW_PAINT do the job.
            | It will magically apply transparency, alpha-blending, etc...
            */
            _LOG_DEBUG((LOG_CAT_GFX, "format %lx != offscreen", GRHANDLE_VideoFlag(hSource)));
            _LOG_DEBUG((LOG_CAT_GFX, "backPixel=%04x", backPixel));

            SDL_SetColorKey(srcSurf, SDL_SRCCOLORKEY, backPixel);
            SDL_LowerBlit(srcSurf, &srcRect, dstSurf, &dstRect);
            SDL_SetColorKey(srcSurf, 0, 0);  // reset
            break;
         }
         /* otherwise, fall thru */
      case WIN_PAINT:         // Dest replaced with source pixels (copy mode)
         _LOG_DEBUG((LOG_CAT_GFX, "do WIN_PAINT"));
         SDL_LowerBlit(srcSurf, &srcRect, dstSurf, &dstRect);
         break;
      case WIN_ERASE:         // Dest cleared where source pixels are off
         //>>>PGR: this is wrong, I know.  Read below.
         SDL_FillRect(dstSurf, &dstRect, backPixel);
         break;
      default:
         // not an obvious case.  most of this is not fully implemented yet.
         if (GRHANDLE_VideoFlag(hSource) == VIDEOFLAG_DIFFERENT_OFFSCREEN)
         {
            /*>>>PGR: dunno how to do this.
            | My deep thoughts: these methods are for Windozish bitmaps (except
            | for a few of them, like swapping colors.)  What would or'ing,
            | and'ing or any bit function mean for a colormapped surface?
            | Isn't alpha blending much more powerful?
            | Well, if needed...
            | I should convert to a COMPATIBLE_OFFSCREEN (and save it
            | as such, so not to do it twice.)  This is what SDL_DisplayFormat
            | is for.  Not today.  So just...
            | guich: when exactly this may happen?
            */
            return;
         }
         else
         {
            Pixel * srcPixels;
            Pixel * dstPixels;
            Sint32 srcPitch = srcSurf->pitch;
            Sint32 dstPitch = dstSurf->pitch;

            _LOG_DEBUG((LOG_CAT_GFX, "do DEFAULT draw"));
            SDL_LockSurface(srcSurf);
            SDL_LockSurface(dstSurf);
            srcPixels = (Pixel *)((char *)srcSurf->pixels + ((srcY * srcPitch) + (srcX * srcSurf->format->BytesPerPixel)));
            dstPixels = (Pixel *)((char *)dstSurf->pixels + ((dstY * dstPitch) + (dstX * dstSurf->format->BytesPerPixel)));
            for (i=height; i > 0 ; --i)
            {
               Pixel * pTgt = dstPixels;
               Pixel * pSrc = srcPixels;
               Pixel * pEnd = pTgt + width;
               switch (drawOp)
               {
                  case WIN_MASK:          // T & !S
                     for (; pTgt < pEnd; ++pTgt)
                        if (*pSrc++ == backPixel)
                           *pTgt = backPixel;
                     break;

                  case WIN_INVERT:        // T ^ S
                     for (; pTgt < pEnd; ++pTgt)
                        *pTgt ^= *pSrc++;
                     break;

                  case WIN_OVERLAY:       // T | S
                     for (; pTgt < pEnd; ++pTgt)
                        *pTgt |= *pSrc++;
                     break;

                  case WIN_PAINT_INVERSE: // Invert and paint
                     for (; pTgt < pEnd; ++pTgt)
                        *pTgt = ~*pSrc++;
                     break;

                  case WIN_REPLACE_COLOR: // Dest replaced with <color> if source pixels != 0
                     for (; pTgt < pEnd; ++pTgt)
                        if (backPixel != *pSrc++)
                           *pTgt = forePixel;
                     break;

                  case WIN_SWAP_COLORS:   // Dest replaced with <color> if source pixels != 0
                     for (; pTgt < pEnd; ++pTgt)
                     {
                        Pixel pixel = *pSrc++;
                        if (pixel == backPixel)
                           *pTgt = forePixel;
                        else
                        if (pixel == forePixel)
                           *pTgt = backPixel;
                     }
                     break;

                  default:
                     break;
               }
               srcPixels = (Pixel *)((char *)srcPixels + srcPitch);
               dstPixels = (Pixel *)((char *)dstPixels + dstPitch);
            }
            SDL_UnlockSurface(dstSurf);
            SDL_UnlockSurface(srcSurf);
         }
   }
   updateRectangle(dstSurf, dstX, dstY, width, height);
}

/*-------------------------------------------------------------------getPixel-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Gets the color value of the pixel, using the current translation          |
|   Returns -1 if error (out of clip bounds)                                  |
|   In PalmOS, returns the index to the current palette.                      |
|   In WindowsCE, returns the color itself                                    |
|   In SDL, returns what's appropriate ;-)                                    |
+----------------------------------------------------------------------------*/
static Sint32 getPixel(Sint32 x, Sint32 y)
{
   Sint32 pixel;
   x += TRANSX;
   y += TRANSY;
   if ((CLIPX1 <= x) && (x < CLIPX2) && (CLIPY1 <= y) && (y < CLIPY2))
   {
      SDL_Surface * active = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
      Sint32 bytesPerPixel = active->format->BytesPerPixel;
      Uint8 * p;

      SDL_LockSurface(active);
      p = (Uint8 *)active->pixels + (((Sint32)y * active->pitch) + ((Sint32)x * bytesPerPixel));
      switch (bytesPerPixel)
      {
         case 1:
            pixel = *p;
            break;
         case 2:
            {
              uint8 r,g,b; // guich@511_13: now returns a 24-bit RGB value.
#if defined(PALMOS) && defined(__arm__)
              uint16 v = *(Uint16 *)p;
              unsigned char *p = (unsigned char *)&v;
              pixel = p[0] << 8 | p[1];
#else
              pixel = *(uint16 *)p;
#endif
              SDL_GetRGB(pixel, active->format, &r, &g, &b);
              pixel = ((int32)r << 16) | ((int32)g << 8) | (int32)b;
            }
            break;
         case 3:
            // TODO: replace by an optimized getUnsigned24 asm routine
            if (SDL_BYTEORDER == SDL_BIG_ENDIAN) // a good compiler will remove the unused code block
               pixel = ((Sint32)p[0] << 16) | ((Sint32)p[1] << 8) | (p[2]);
            else
               pixel = ((Sint32)p[2] << 16) | ((Sint32)p[1] << 8) | (p[0]);
             break;
         default:
             pixel = *(Uint32 *)p;
             break;
      }
      SDL_UnlockSurface(active);
      return pixel;
   }
   return -1;
}

/*---------------------------------------------------------------changeColors-+
| Given a surface, replace a color by another one                             |
+----------------------------------------------------------------------------*/
void changeColors(SDL_Surface * surface,
                  Sint32 x, Sint32 y, Sint32 width, Sint32 height,
                  Pixel pixelFrom, Pixel pixelTo)
{
   // SDL_PixelFormat * format = surface->format;
   Sint32 pitch = surface->pitch;
   char * row;
   Sint32 i;

   SDL_LockSurface(surface);
   row = ((char *)surface->pixels + ((y * pitch) + (x * sizeof(Pixel))));
   for (i=height; i > 0; --i)
   {
      Pixel * p = (Pixel *)row;
      Pixel * pMax = (Pixel *)row + width;
      for (; p < pMax; ++p)
         if (*p == pixelFrom)
            *p = pixelTo;
      row += pitch;
   }
   SDL_UnlockSurface(surface);
}

/*-------------------------------------------------------------------setPixel-+
| Device specific routine.                                                    |
| Sets the pixel to the given color, translating and clipping                 |
+----------------------------------------------------------------------------*/
static void setPixel(Sint32 x, Sint32 y, Pixel pixel)
{
   x += TRANSX;
   y += TRANSY;
   if ((CLIPX1 <= x) && (x < CLIPX2) && (CLIPY1 <= y) && (y < CLIPY2))
   {
      SDL_Surface * target = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
      SDL_LockSurface(target);
      *(Pixel *)((char *)target->pixels + ((y * target->pitch) + (x * sizeof(Pixel)))) = pixel;
      SDL_UnlockSurface(target);
      updateRectangle(target, x, y, 1, 1);
   }
}

/*------------------------------------------------------------------drawHLine-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Draws a horizontal line from x to x+w-1, translating and clipping         |
+----------------------------------------------------------------------------*/
static void drawHLine(Sint32 x, Sint32 y, Sint32 width, Pixel pixel1, Pixel pixel2)
{
   x += TRANSX;
   y += TRANSY;
   /*
   | line must lie inside y clip bounds, must not end before clip x1
   | and must not start after clip x2
   */
   if ((CLIPY1 <= y) && (y < CLIPY2) && (CLIPX1 <= (x+width)) && (x < CLIPX2))
   {
      SDL_Surface * active = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
      Pixel pixel = pixel1;
      Pixel * pTgt;
      Pixel * pMax;
      if (x < CLIPX1)           // line start before clip x1
      {
         width -= CLIPX1-x;
         x = CLIPX1;
      }
      if ((x+width) > CLIPX2)   // line stops after clip x2
         width = CLIPX2-x;

      SDL_LockSurface(active);
      pTgt = ((Pixel *)((char *)active->pixels + (y * active->pitch))) + x;
      pMax = pTgt + width;
      if (pixel1 == pixel2) // same color?
      {
         while (pTgt < pMax)
            *pTgt++ = pixel;          // plot the pixel
      }
      else
      {
         while (pTgt < pMax)
         {
            *pTgt++ = pixel;          // plot the pixel
            pixel = (pixel == pixel1)? pixel2 : pixel1;
         }
      }
      SDL_UnlockSurface(active);
      updateRectangle(active, x, y, width, 1);
   }
}

/*------------------------------------------------------------------drawVLine-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Draws a vertical line from y to y+h-1 using the given color               |
+----------------------------------------------------------------------------*/
static void drawVLine(Sint32 x, Sint32 y, Sint32 height, Pixel pixel1, Pixel pixel2)
{
   x += TRANSX;
   y += TRANSY;
   /*
   | line must lie inside x clip bounds, must not end before clip y1
   | and must not start after clip y2
   */
   if ((CLIPX1 <= x) && (x < CLIPX2) && (CLIPY1 <= (y+height)) && (y < CLIPY2))
   {
      SDL_Surface * active = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
      Pixel pixel = pixel1;
      Pixel * pTgt;
      Pixel * pMax;
      Uint32 pitch = active->pitch;
      if (y < CLIPY1)           // line start before clip y1
      {
         height -= CLIPY1-y;
         y = CLIPY1;
      }
      if ((y+height) > CLIPY2)
         height = CLIPY2-y;           // line stops after clip y2

      SDL_LockSurface(active);
      pTgt = ((Pixel *)((char *)active->pixels + (y * active->pitch))) + x;
      pMax = ((Pixel *)((char *)pTgt + (height * active->pitch)));
      if (pixel1 == pixel2) // same color?
      {
         while (pTgt < pMax)
         {
            *pTgt = pixel;          // plot the pixel
            (char *)pTgt += pitch;
         }
      }
      else
      {
         while (pTgt < pMax)
         {
            *pTgt = pixel;          // plot the pixel
            pixel = (pixel == pixel1)? pixel2 : pixel1;
            (char *)pTgt += pitch;
         }
      }
      SDL_UnlockSurface(active);
      updateRectangle(active, x, y, 1, height);
   }
}

/*-------------------------------------------------------------drawDottedLine-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void drawDottedLine(Sint32 x1, Sint32 y1, Sint32 x2, Sint32 y2,
                           Pixel pixel1, Pixel pixel2)
{
    // guich@501_10: added pyInc and clipping support for this routine.
    // store the change in X and Y of the line endpoints
    Sint32 dX,dY;
    // DETERMINE "DIRECTIONS" TO INCREMENT X AND Y (REGARDLESS OF DECISION)
    Sint32 xInc,yInc,pyInc; // py incs by pixel, y incs by row
    // used in clipping
    Sint32 xMin,yMin;
    // compute the values
    if (x1 <= x2)
    {
       dX = x2-x1;
       xMin = x1;
       xInc = 1;
    }
    else
    {
       dX = x1-x2;
       xMin = x2;
       xInc = -1;
    }
    if (y1 <= y2)
    {
       dY = y2-y1;
       yMin = y1;
       pyInc = 1;
    }
    else
    {
       dY = y1-y2;
       yMin = y2;
       pyInc = -1;
    }


    // guich: if its a pixel, draw only the pixel
    if (!dX && !dY)
       setPixel(xMin,yMin,pixel1);
    else
    // guich: if they are vertical or horizontal lines, use specific methods
    if (!dY) // horizontal line?
       drawHLine(xMin,yMin,dX+1,pixel1,pixel2);
    else
    if (!dX) // vertical line?
       drawVLine(xMin,yMin,dY+1,pixel1,pixel2);
    else // diagonal
    {
       Sint32 dx, dy,currentX,currentY; // save to be used in updateRectangle
       Pixel *row;
       Pixel pixel = pixel1;
       SDL_Surface *active;
       Sint32 clipX1 = CLIPX1, clipY1 = CLIPY1,
              clipX2 = CLIPX2, clipY2 = CLIPY2;
       Uint8 dontClip = 1; // the most common will be draw lines that do not cross the clip bounds, so we may speedup a little

       xMin += TRANSX;
       yMin += TRANSY;

       if (xMin < clipX1 || (xMin+dX) >= clipX2 || yMin < clipY1 || (yMin+dY) >= clipY2)
          dontClip = 0;

       dx = dX; // save to be used in updateRectangle
       dy = dY;

       currentX = x1+TRANSX;
       currentY = y1+TRANSY;

       active = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
       SDL_LockSurface(active);
       row = (Pixel *)((char *)active->pixels
                     + ((pyInc>0 ? yMin : (yMin+dY)) * active->pitch)
                     + (sizeof(Pixel) * (xInc>0 ? xMin : (xMin+dX)))); // currentX    | else start from the max value

       yInc = pyInc>0 ? active->pitch : 0-active->pitch;  // now that we have active, correctly assign the y incrementing value

       // DETERMINE INDEPENDENT VARIABLE (ONE THAT ALWAYS INCREMENTS BY 1 (OR -1) )
       // AND INITIATE APPROPRIATE LINE DRAWING ROUTINE (BASED ON FIRST OCTANT
       // ALWAYS). THE X AND Y'S MAY BE FLIPPED IF Y IS THE INDEPENDENT VARIABLE.
       if (dX >= dY)   // if X is the independent variable
       {
          Sint32 dPr     = dY<<1;                        // amount to increment decision if right is chosen (always)
          Sint32 dPru    = dPr - (dX<<1);                // amount to increment decision if up is chosen
          Sint32 p       = dPr - dX;                     // decision variable start value

          if (pixel1 == pixel2) // quick optimization
             for (; dX >= 0; dX--)                       // process each point in the line one at a time (just use dX)
             {
                if (dontClip || (clipX1 <= currentX && currentX < clipX2 && clipY1 <= currentY && currentY < clipY2))
                   *row = pixel;                            // plot the pixel
                row      += xInc;                        // increment independent variable
                currentX += xInc;
                if (p > 0)                               // is the pixel going right AND up?
                {
                   currentY += pyInc;
                   p += dPru;                            // increment decision (for up)
                   row = (Pixel *)((char *)row + yInc); // increment dependent variable
                }
                else                                     // is the pixel just going right?
                   p += dPr;                             // increment decision (for right)
             }
          else
             for (; dX >= 0; dX--)                       // process each point in the line one at a time (just use dX)
             {
                if (dontClip || (clipX1 <= currentX && currentX < clipX2 && clipY1 <= currentY && currentY < clipY2))
                   *row = pixel;                            // plot the pixel
                pixel = (pixel == pixel1)? pixel2 : pixel1; // swap the pixel colors
                row += xInc;                             // increment independent variable
                currentX += xInc;
                if (p > 0)                               // is the pixel going right AND up?
                {
                   currentY += pyInc;
                   p += dPru;                            // increment decision (for up)
                   row = (Pixel *)((char *)row + yInc); // increment dependent variable
                }
                else                                     // is the pixel just going right?
                   p += dPr;                             // increment decision (for right)
             }
       }
       else            // if Y is the independent variable
       {
          Sint32 dPr     = dX<<1;                        // amount to increment decision if right is chosen (always)
          Sint32 dPru    = dPr - (dY<<1);                // amount to increment decision if up is chosen
          Sint32 p       = dPr - dY;                     // decision variable start value

          if (pixel1 == pixel2) // quick optimization
             for (; dY >= 0; dY--)                       // process each point in the line one at a time (just use dY)
             {
                if (dontClip || (clipX1 <= currentX && currentX < clipX2 && clipY1 <= currentY && currentY < clipY2))
                   *row = pixel;                            // plot the pixel
                row = (Pixel *)((char *)row + yInc);    // increment independent variable
                currentY += pyInc;
                if (p > 0)                               // is the pixel going up AND right?
                {
                   row += xInc;                          // increment dependent variable
                   currentX += xInc;
                   p += dPru;                            // increment decision (for up)
                }
                else                                     // is the pixel just going up?
                   p += dPr;                             // increment decision (for right)
             }
          else
             for (; dY >= 0; dY--)                       // process each point in the line one at a time (just use dY)
             {
                if (dontClip || (clipX1 <= currentX && currentX < clipX2 && clipY1 <= currentY && currentY < clipY2))
                   *row = pixel;                            // plot the pixel
                pixel = (pixel == pixel1)? pixel2 : pixel1; // swap the pixel colors
                row = (Pixel *)((char *)row + yInc);    // increment independent variable
                currentY += pyInc;
                if (p > 0)                               // is the pixel going up AND right?
                {
                   row += xInc;                          // increment dependent variable
                   currentX += xInc;
                   p += dPru;                            // increment decision (for up)
                }
                else                                     // is the pixel just going up?
                   p += dPr;                             // increment decision (for right)
             }
       }
       SDL_UnlockSurface(active);
       updateRectangle(active, xMin, yMin, dx, dy);
    }
}

/*-------------------------------------------------------------------drawLine-+
|                                                                             |
+----------------------------------------------------------------------------*/
_inline_ static void drawLine(Sint32 x1, Sint32 y1, Sint32 x2, Sint32 y2, Pixel pixel)
{
   drawDottedLine(x1, y1, x2, y2, pixel, pixel);
}

/*-------------------------------------------------------------------drawRect-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Draws a rectangle with the given color                                    |
+----------------------------------------------------------------------------*/
void drawRect(Sint32 x, Sint32 y, Sint32 width, Sint32 height, Pixel pixel)
{
   drawHLine(x, y, width, pixel, pixel);
   drawHLine(x, y+height-1, width, pixel, pixel);
   drawVLine(x, y, height, pixel, pixel);
   drawVLine(x+width-1, y, height, pixel, pixel);
}

/*-------------------------------------------------------------------fillRect-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Fills a rectangle with the given color                                    |
+----------------------------------------------------------------------------*/
void fillRect(Sint32 x, Sint32 y, Sint32 width, Sint32 height, Pixel pixel)
{
   x += TRANSX;
   y += TRANSY;
   if (x < CLIPX1)             // line starts before clip x1
   {
      width -= CLIPX1-x;
      x = CLIPX1;
   }
   if ((x+width) > CLIPX2)     // line stops after clip x2
      width = CLIPX2-x;

   if (y < CLIPY1)             // line starts before clip y1
   {
      height -= CLIPY1-y;
      y = CLIPY1;
   }
   if ((y+height) > CLIPY2)    // line stops after clip y2
      height = CLIPY2-y;

   if (height > 0 && width > 0)
   {
      SDL_Surface * active = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
      Sint32 pitch = active->pitch;
      Sint32 h = height;
      char * row;

      SDL_LockSurface(active);
      for (row = ((char *)active->pixels + ((y * pitch) + (x * sizeof(Pixel))));  h--;  row += pitch)
      {
         Pixel * p = (Pixel *)row;
         Pixel * pMax = (Pixel *)row + width;
         while (p < pMax)
            *p++ = pixel;
      }
      SDL_UnlockSurface(active);
      updateRectangle(active, x, y, width, height);
   }
}

/*----------------------------------------------------------------clearScreen-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Clears the whole area with the given color, ignoring translation and clip.|
| Returns:                                                                    |
|                                                                             |
| Exceptions:                                                                 |
|                                                                             |
+----------------------------------------------------------------------------*/
static void clearScreen(Pixel pixel)
{
   SDL_Surface * active = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
   SDL_Rect rect = { 0, 0, active->w, active->h };
   SDL_FillRect(active, &rect, pixel);
   updateRectangle(active, 0, 0, active->w, active->h);
}

/*-------------------------------------------------------------------drawText-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void drawText(Uint16 * text, Uint32 len, Sint32 x0, Sint32 y0, Pixel pixel)
{
   SDL_Surface * active = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
   static unsigned const char ands8[8] = {0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};
   struct UserFont *uf = currentUserFont;
   Uint8 isUff = IS_UNICODE_FONT(uf);
   Uint8 antialiased = IS_ANTIALIASED(uf);
   Sint32 rowWidthInBytes;
   Uint16 numberOfChars;
   Uint8 const * bitmapTable;
   Uint16 const * bitIndexTable;
   Sint32 const height = (Sint32)currentUserFont->fontP->fRectHeight;
   Uint16 firstChar;
   Pixel * row0;
   Sint32 xMin;
   Sint32 yMin;
   Sint32 xMax;
   Sint32 yMax;

   /*
   | Here is an SDL trick.  To compute the alpha-blending of the structure
   | requires a computation which, in theory, should be applied to each of
   | the 3 (R, G, B) | components.  For optimization purposes, it would be
   | better to avoid to do 3 times these computations, but rather to
   | work on the pixel value as a whole.  Can it be done?
   |
   | This computation has a left shift.
   |
   | On 32-bits pixels (nibbles: 00RRGGBB), the GG would be overwritten.
   | The pixel is split into 2 integers:  ..RR..BB, and ....GG..
   | (leading to 2 computations)
   |
   | On 16-bits pixels (bits: rrrrrggggggbbbbb), gggggg would be overwritten
   | as well.  We make it one Sint32:  ---- -ggg, ggg- ----, rrrr r---, ---b bbbb
   | resulting in only one computation.
   |
   | On 8 bits pixels... we don't do anti-aliasing on 8 bits pixels at today!
   | Optimization tricks remain to be invented (Web safe palette)
   */

   #if VIDEO_BITSPERPIXEL == 16
      Uint32 px;
      if (antialiased)
      {
         px = ((uint32)pixel | (uint32)pixel << 16) & 0x07e0f81fL;
      }
      #if defined(__GNUC__)
      else
      {
         px = 0;
      }
      #endif
   #elif VIDEO_BITSPERPIXEL == 32
      Uint32 pxRB;
      Uint32 pxG;
      if (antialiased)
      {
         pxRB = (Uint32)pixel & 0xff00ffL;
         pxG = (Uint32)pixel & 0xff00;
      }
      #if defined(__GNUC__)
      else
      {
         pxRB = 0;
         pxG = 0;
      }
      #endif
   #endif
   _LOG_INFO((LOG_CAT_GFX, "drawText(%s,%d,%d,%d,0x%x)", text, (int)len, (int)x0, (int)y0, (int)pixel));

   if (!isUff)
   {
      rowWidthInBytes = uf->rowWidthInBytes;
      bitmapTable = uf->bitmapTable;
      bitIndexTable = (Uint16 *)uf->bitIndexTable;
      firstChar = (Uint16)uf->fontP->firstChar;
      numberOfChars = (Uint16)uf->numberOfChars;
   }
   #if defined(__GNUC__)
   else
   {
      rowWidthInBytes = 0;
      bitmapTable = 0;
      bitIndexTable = 0;
      firstChar = 0;
      numberOfChars = 0;
   }
   #endif

   x0 += TRANSX;
   y0 += TRANSY;

   xMax = xMin = (x0 < CLIPX1)? CLIPX1 : x0;
   yMin = (y0 < CLIPY1)? CLIPY1 : y0;
   yMax = y0 + height;
   if (yMax >= CLIPY2) yMax = CLIPY2;
   row0 = (Pixel *)((char *)active->pixels + (y0 * active->pitch));

   SDL_LockSurface(active);
   for (; len--; ++text)
   {
      Pixel * row;
      Uint32 offset;
      Uint8 width;

      _LOG_DEBUG((LOG_CAT_GFX, " Char='%c'", *text));

      // guich@341_9: fixed <enter> problems with MultiEdit
      if (*text == 13) continue;

      if (isUff)
      {
         static struct CharBits bits;
         _LOG_DEBUG((LOG_CAT_GFX, " UFF font"));
         uffSetCharBits(*text, uf, &bits); // pgr@402_50
         bitmapTable = bits.bitmapTable;
         rowWidthInBytes = bits.rowWIB;
         offset = bits.offset;
         width = bits.width;
      }
      else
      {
         Uint32 index;
         _LOG_DEBUG((LOG_CAT_GFX, " firstChar=%d", (int)firstChar));
         _LOG_DEBUG((LOG_CAT_GFX, " numberOfChars=%d", (int)numberOfChars));

         if ((*text < firstChar) ||
             ((index = (*text - firstChar)) >= (Uint32)numberOfChars) ||
             (bitIndexTable[index+1] == bitIndexTable[index]))
         {
            _LOG_ERROR((LOG_CAT_GFX, " invalid char!"));
            offset = 0;
            width = 0;                     // invalid char
         }
         else
         {
            // Get the source x coordinate and width of the character
            offset = bitIndexTable[index];
            width = (bitIndexTable[index+1] - offset);
            _LOG_DEBUG((LOG_CAT_GFX, " source=%d, width=%d", (int)offset, (int)width));
         }
      }
      if (width == 0)                     // invalid char
      {
         xMax = x0 + UF_MCS_WIDTH; // guich@340_46: show the MCS char if no width (glyph not present)
      }
      else if (!antialiased)
      {
         Sint32 y;
         Uint8 const * start = bitmapTable + (offset >> 3);
         Uint8 startBit = offset & 7;
         Uint32 pitch = active->pitch;

         if ((xMax = x0 + width) > CLIPX2) xMax = CLIPX2;
         // draws the char
         for (y=y0, row=row0; y < yMax; ++y, start+=rowWidthInBytes, (char*)row += pitch) // draw each row
         {
            if (y >= CLIPY1)
            {
               Uint8 const * current = start;
               Uint8 currentBit = startBit;
               Uint8 const * ands = ands8 + currentBit;
               Sint32 x;
               if (xMax > CLIPX2) xMax = CLIPX2;
               for (x=x0; x < xMax; ++x)
               {
                  if ((*current & *ands++) && (x >= CLIPX1))
                  {
                     row[x] = pixel;
                  }
                  if (++currentBit == 8)   // finished this Uint8?
                  {
                     currentBit = 0;       // reset counter
                     ands = ands8;         // reset test bit pointer
                     ++current;            // inc current Uint8
                  }
               }
            }
         }
      }
      else // anti-aliasing, e.g.: alpha-blending computation
      {
         Sint32 y;
         Uint8 const * start = bitmapTable + (offset >> 1);
         Uint8 isNibbleStartingLow = offset & 1;
         Uint32 pitch = active->pitch;

         xMax = x0 + width;
         if (xMax > CLIPX2) xMax = CLIPX2;
         // draw the char
         for (y=y0, row=row0; y < yMax; ++y, start+=rowWidthInBytes, (char*)row += pitch) // draw each row
         {
            if (y >= CLIPY1)
            {
               Sint32 x;
               Uint8 const * current = start;
               Uint8 transparency;
               Uint8 isLowNibble = isNibbleStartingLow;
               for (x=x0; x < xMax; ++x)
               {
                  if (isLowNibble)
                  {
                     isLowNibble = 0;
                     transparency = *current++ & 0x0F;
                  }
                  else
                  {
                     isLowNibble = 1;
                     transparency = *current >> 4;
                  }
                  if (x >= CLIPX1)
                  {
                     switch (transparency)
                     {
                        case 0:
                           break;
                        case 0xF:
                           row[x] = pixel;
                           break;
                        default:
                           #if VIDEO_BITSPERPIXEL == 16
                           {
                              Uint32 i = row[x];
                              i = (i | i << 16) & 0x07e0f81fUL;
                              i = (i + (((px - i) * transparency) >> 4)) & 0x07e0f81fUL;
                              row[x] = i | (i >> 16);
                           }
                           #elif VIDEO_BITSPERPIXEL == 32
                           {
                              Uint32 i = row[x];
                              Uint32 j = i & 0xff00ffUL;
                              i &= 0xff00;
                              row[x] = ((j + (((pxRB - j) * transparency) >> 4)) & 0xff00ffUL)
                                     | ((i + (((pxG  - i) * transparency) >> 4)) & 0xff00);
                           }
                           #endif
                           break;
                     }
                  }
               }
            }
         }
      }
      if (xMax >= CLIPX2)
      {
         xMax = CLIPX2;
         break;
      }
      x0 = xMax;                           // next character
   }
   SDL_UnlockSurface(active);
   updateRectangle(active, xMin, yMin, (xMax - xMin), (yMax - yMin));
}

/*--------------------------------------------------------------graphicsSetup-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Setup the main variables. The following variables must be set in          |
|   this method: displayBPP, isColor, maxColors                               |
|   This function is called only once, before graphicsStart.                  |
+----------------------------------------------------------------------------*/
#if defined WINCE || defined PALMOS || defined __SYMBIAN32__
#define DEFAULT_SDL_SURFACE_FLAG SDL_NOFRAME
#else
#define DEFAULT_SDL_SURFACE_FLAG 0
#endif
void graphicsSetup()
{
#ifdef PALMOS
   Err err;
   UInt32 version;
#endif
   SDL_Surface * screen;
   Sint32 maxColors;
   Sint32 desired_bpp = VIDEO_BITSPERPIXEL;
   int ww,hh;

/**
 * fdie@ force a graphic mode (for my tests)
   vmGlobals->screenWidth = 320;
   vmGlobals->screenHeight = 320;
   desired_bpp = 16;
*/

#ifdef WIN32
   RECT r;
   SystemParametersInfo(SPI_GETWORKAREA,0,&r,0); // get the task bar height
   vmGlobals->taskBarHeight = r.top;
#endif

   /* Initialize the display */
   screen = SDL_SetVideoMode(vmGlobals->screenWidth, vmGlobals->screenHeight, desired_bpp,
                             vmGlobals->isFullScreen? SDL_FULLSCREEN : DEFAULT_SDL_SURFACE_FLAG);
   if (!screen)
   {
      (*vmGlobals->debug)("SDL_SetVideoMode failed:");
      (*vmGlobals->debug)(SDL_GetError());
      return;
   }
   ww = vmGlobals->screenWidth = screen->w;
   hh = vmGlobals->screenHeight = screen->h;

#ifdef WINCE
   if (!vmGlobals->isFullScreen)
      hh -= vmGlobals->taskBarHeight;
#endif
   vmGlobals->windowWidth = ww;
   vmGlobals->windowHeight = hh;

   // fdie@ store the current screen density required for later non SDL implemented PalmOS features.
   // FntWCharWidth(), to retrieve the current font width require to set the screen density before the call.
   vmGlobals->density = kCoordinatesStandard;

#if defined(PALMOS)
   err = FtrGet(sysFtrCreator, sysFtrNumWinVersion, &version);
   if (!err && version >= 4) // could WinGetCoordinateSystem() be called
   {
      vmGlobals->density = WinGetCoordinateSystem();
      // now forces the PalmOS standard coordinate system required by system dialogs and others
      WinSetCoordinateSystem(kCoordinatesStandard);
   }
#endif //PALMOS

   vmGlobals->screenBPP = screen->format->BitsPerPixel;
   switch (vmGlobals->screenBPP)
   {
      case 8:
         setWebSafePalette(screen);
         maxColors = 0x100;
         break;
      case 16:
         maxColors = 0x10000;
         break;
      default:   // guich: users are not prepared that maxColors can be 0 (even if documented).
                 // So, just return as if 24bpp, (which is pretty enough for most devices)
      //case 24:
         maxColors = 0x1000000;
         break;
      /*case 32:
         maxColors = 0;                   // means 2 ** 32 for waba
         break;*/
   }
   vmGlobals->maxColors = maxColors;   // used in settings
   // fill the screen with white
   if (!use8bppEmulation)
   {
      SDL_Rect r = { 0, 0, ww, hh};
      SDL_FillRect(SDL_GetVideoSurface(), &r, 0xFFFFFF);
      //SDL_UpdateRect(SDL_GetVideoSurface(), 0,0,ww,hh); - do not update now
   }
}

/*--------------------------------------------------------------graphicsStart-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Starts the special graphics mode. If the device only supports             |
|   one mode, nothing special is needed.                                      |
|   This function is called only once.                                        |
|   For PalmOS, we enter in grayscale mode here.                              |
|   In WindowsCE, nothing is done.                                            |
+----------------------------------------------------------------------------*/
void graphicsStart()
{
   graphicsActive = 1;
   // should call SDL_Start so that a Palm OS driver can implement it
}

/*---------------------------------------------------------------graphicsStop-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Stops the special graphics mode. Also, cleanup some variables             |
|   This function is called only once.                                        |
|   For PalmOS, we leave grayscale mode.                                      |
|   In WinCE, nothing is done.                                                |
+----------------------------------------------------------------------------*/
void graphicsStop()
{
   if (graphicsActive)
   {
      xfree(mainGrHandle);
      mainGrHandle = NULL;
      disposeHashtable(colorHashTable,0);
      graphicsActive = 0;
      // fill in with additional cleanup when the user is exiting the program
      // should call SDL_Stop so that a Palm OS driver can implement it
   }
   releaseAllFonts();
}

/*---------------------------------------------------------------graphicsSwap-+
| Description:                                                                |
|   Device specific routine.                                                  |
|   Swaps between the original graphics mode and our special mode.            |
|   This function may be callod many times.                                   |
|   In PalmOS, the graymode is closed and restored the default mode.          |
|   In WindowsCE, nothing is done                                             |
+----------------------------------------------------------------------------*/
void graphicsSwap(Sint32 on)
{
   // should call SDL_Swap so that a Palm OS driver can implement it
   if (on && !graphicsActive && !vmGlobals->errorWinDisplayed)
      graphicsActive = 1;
   else
   if (!on && graphicsActive)
      graphicsActive = 0;
}

/*---------------------------------------------------------------makeGrHandle-+
| videoFlag is one of:                                                        |
| VIDEOFLAG_MAINSCREEN                                                        |
| VIDEOFLAG_COMPATIBLE_OFFSCREEN                                              |
| VIDEOFLAG_DIFFERENT_OFFSCREEN                                               |
+----------------------------------------------------------------------------*/
GrHandle makeGrHandle(SDL_Surface * surface, int32 videoFlag)
{
   GrHandle hWin = (GrHandle)(*vmGlobals->xmalloc)(sizeof(GraphicsContext));
   if (hWin)
   {
      GRHANDLE_WidthInBPP(hWin) = surface->pitch;
      GRHANDLE_VideoFlag(hWin) = videoFlag;
      GRHANDLE_WinHandle(hWin) = surface;
      GRHANDLE_MemAddr(hWin) = 0;
      GRHANDLE_TransX(hWin)  = 0;
      GRHANDLE_TransY(hWin)  = 0;
      GRHANDLE_Width(hWin)   = surface->w;
      GRHANDLE_Height(hWin)  = surface->h;
      GRHANDLE_ClipX1(hWin)  = 0;
      GRHANDLE_ClipY1(hWin)  = 0;
      GRHANDLE_ClipX2(hWin)  = surface->w;
      GRHANDLE_ClipY2(hWin)  = surface->h;
      GRHANDLE_PaletteSize(hWin) = 0;
      GRHANDLE_Palette(hWin) = 0;
   }
   return hWin;
}

int charWidth(int16 ch, struct UserFont *uf);
static void centerText(int16 *s, int y, Pixel c, int *xmin)
{
   int32 w = 0,i,x;
   for (i = 0; s[i] != -255; i++)
   {
      s[i] += 65;
      w += (int32)charWidth((uint16)s[i], currentUserFont);
   }
   x = (vmGlobals->screenWidth-w)/2;
   drawText((uint16*)s,i,x,y,c);
   if (x < *xmin) *xmin = x;
}

static void drawCopyright() // ** removing this VIOLATES SECTION 2C OF THE GPL LICENSE **
{
   int16 SUPERWABA_VM[] ={18,20,15,4,17,22,0,1,0,-33,21,12,-255};
   int16 COMMUNITY_GPL_VERSION[] ={2,14,12,12,20,13,8,19,24,-33,6,15,11,-33,21,4,17,18,8,14,13,-255};
   int16 Copyright_C_2000_2005[] ={2,46,47,56,49,40,38,39,51,-33,-25,2,-24,-33,-15,-17,-17,-17,-20,-15,-17,-17,-12,-255};
   int16 SuperWaba_Ltda[] ={18,52,47,36,49,22,32,33,32,-33,11,51,35,32,-19,-255};
   int16 All_rights_reserved[] ={0,43,43,-33,49,40,38,39,51,50,-33,49,36,50,36,49,53,36,35,-255};

   int ww = vmGlobals->screenWidth;
   int hh = vmGlobals->screenHeight;
   int xmin = 9999;
#ifdef __SYMBIAN32__
   int y0 = 0;
#else
   int y0 = hh==160 ? 10 : 30;
#endif
   int g = hh==160 ? 2 : 5;
   int y=y0,h;
   Pixel br,pr;
   br = getColor(255,255,255);
   pr = getColor(0,0,0);

   fillRect(0,0,ww,hh,br);

   currentUserFont = &vmGlobals->allFonts[3];
   h = currentUserFont->fontP->ascent + currentUserFont->fontP->descent+3;

              centerText(SUPERWABA_VM,y,pr,&xmin);
   y += h;    centerText(COMMUNITY_GPL_VERSION,y,pr,&xmin);

   currentUserFont = &vmGlobals->allFonts[1];
   h = currentUserFont->fontP->ascent + currentUserFont->fontP->descent+3;

   y += h+(g<<2); centerText(Copyright_C_2000_2005,y,pr,&xmin);
   y += h;        centerText(SuperWaba_Ltda,y,pr,&xmin);
   y += h;        centerText(All_rights_reserved,y,pr,&xmin);

   xmin -= 5;
   y0 -= 3;
   drawRect(xmin,y0,ww-(xmin<<1),y+h-y0,pr);

   SDL_UpdateRect(SDL_GetVideoSurface(), 0,0,ww,hh);
   currentUserFont = 0;
   SDL_Delay(1500);
}

////  Methods that must be implemented by the Platforms ////

Uint8 graphicsInit() // called by the vm to init the graphics
{
   SDL_Surface * screen;

   _LOG_INFO((LOG_CAT_GFX, "graphicsSetup..."));
   graphicsSetup();     // call the port routines

   _LOG_INFO((LOG_CAT_GFX, "loadAllFonts..."));
   if (!loadAllFonts())
      return 0;
   _LOG_INFO((LOG_CAT_GFX, "graphicsStart..."));
   graphicsStart();

   vmGlobals->graphicsStartImage = graphicsStartImage; // petrus@402_15
   vmGlobals->graphicsEndImage = graphicsEndImage;     // petrus@402_15
   vmGlobals->makeGrHandle = makeGrHandle;
   screen = SDL_GetVideoSurface();
   mainGrHandle = makeGrHandle(screen, VIDEOFLAG_MAINSCREEN);
   // initialize the global values for describing the active graphics
   activeGrHandle = mainGrHandle;
   transXY = 0;
   maxX = clipX = MAKE32(0, screen->w);
   maxY = clipY = MAKE32(0, screen->h);
   SDL_EventState(SDL_SYSWMEVENT, SDL_ENABLE);  // Catch WM unhandled messages
   SDL_EventState(SDL_ACTIVEEVENT, SDL_ENABLE); // Tell me if I turn active

#if defined(GPL_VERSION)
   drawCopyright(); // ** removing this VIOLATES SECTION 2C OF THE GPL LICENSE **
#endif

   _LOG_INFO((LOG_CAT_GFX, "graphicsInit end!"));
   return 1;
}

static Sint32 surfaceGetType(WObject surface)
{
   // cache class pointers for performance
   if (!imageClass) imageClass = vmGlobals->getClass(vmGlobals->createUtfString("waba/fx/Image"),1);
   return (surface != NULL && (*vmGlobals->compatible)(WOBJ_class(surface), imageClass)) ? SURF_IMAGE : SURF_CONTROL;
}

////////////////////////////////////////////////////////////////////////////
_inline_ static void quadPixel(Sint32 xc, Sint32 yc, Sint32 x, Sint32 y, Pixel c)
{
    // draw 4 points using symetry
    setPixel(xc + x, yc + y, c);
    setPixel(xc + x, yc - y, c);
    setPixel(xc - x, yc + y, c);
    setPixel(xc - x, yc - y, c);
}

_inline_ static void quadLine(Sint32 xc, Sint32 yc, Sint32 x, Sint32 y, Pixel c)
{
    Sint32 w = x+x+1; // plus 1 for the drawHLine (draws to width-1)
    // draw 2 lines using symetry
    drawHLine(xc - x, yc - y, w, c, c);
    drawHLine(xc - x, yc + y, w, c, c);
}

// draws an ellipse incrementally
static void ellipseDrawAndFill (Sint32 xc, Sint32 yc, Sint32 rx, Sint32 ry, Pixel c, Uint8 fill)
{
   // intermediate terms to speed up loop
   long t1 = rx*rx, t2 = t1<<1, t3 = t2<<1;
   long t4 = ry*ry, t5 = t4<<1, t6 = t5<<1;
   long t7 = rx*t5, t8 = t7<<1, t9 = 0L;
   long d1 = t2 - t7 + (t4>>1);    // error terms
   long d2 = (t1>>1) - t8 + t5;
   Sint32 x = rx;      // ellipse points
   Sint32 y = 0;       // ellipse points
   if (rx < 0 || ry < 0) // guich@501_13
      return;

   while (d2 < 0)          // til slope = -1
   {
      if (fill)
         quadLine(xc,yc,x,y,c);
      else
         quadPixel(xc,yc,x,y,c);
      y++;          // always move up here
      t9 += t3;
      if (d1 < 0)   // move straight up
      {
         d1 += t9 + t2;
         d2 += t9;
      }
      else        // move up and left
      {
         --x;
         t8 -= t6;
         d1 += t9 + t2 - t8;
         d2 += t9 + t5 - t8;
      }
   }

   do             // rest of top right quadrant
   {
      // draw 4 points using symmetry
      if (fill)
         quadLine(xc,yc,x,y,c);
      else
         quadPixel(xc,yc,x,y,c);
      --x;        // always move left here
      t8 -= t6;
      if (d2 < 0)  // move up and left
      {
         ++y;
         t9 += t3;
         d2 += t9 + t5 - t8;
      }
      else d2 += t5 - t8; // move straight left
   } while (x >= 0);
}

////////////////////////////////////////////////////////////////////////////
static void drawDottedRect(Sint32 x, Sint32 y, Sint32 w, Sint32 h, Pixel c1, Pixel c2)
{
   Sint32 x2 = x+w-1;
   Sint32 y2 = y+h-1;
   drawDottedLine(x, y, x2,y, c1, c2);
   drawDottedLine(x, y, x ,y2, c1, c2);
   drawDottedLine(x2, y, x2, y2, c1, c2);
   drawDottedLine(x, y2, x2, y2, c1, c2);
}
////////////////////////////////////////////////////////////////////////////
// Generalized Polygon Fill * Implementation by Curt McDowell,
struct edge
{
   struct edge *next;
   Sint32 yTop, yBot;
   Sint32 xNowWhole, xNowNum, xNowDen, xNowDir;
   Sint32 xNowNumStep;
};

// updateActive first removes any edges which curY has entirely passed by. The removed edges are freed.
// It then removes any edges from the edge table at curY and places them on the active list.
static struct edge * updateActive(struct edge *active, struct edge *edgeTable[], Sint32 curY)
{
   struct edge *e, **ep;
   for (ep = &active, e = *ep; e != NULL; e = *ep)
   {
      if (e->yBot < curY)
      {
         *ep = e->next;
         xfree(e);
      }
      else ep = &e->next;
   }
   *ep = edgeTable[curY];
   return active;
}

static void fillPolygon(Sint32 *xPoints, Sint32 *yPoints, Sint32 nPoints, Pixel c)
{
   struct edge **edgeTable;
   struct edge *active;
   Sint32 curY;
   Sint32 *xCoords; // guich@400_33
   Sint32 i, j;
   struct edge *e;
   Sint32 *p = yPoints;
   Sint32 maxy = *p,miny=*p; // guich@421_37: calculates also miny to fix problem when the polygon has negative y values - guich@501_11: added again this code that was forgotten during the development of swsdl
   for (p++,i = nPoints; --i > 0; p++) // guich@350_26: now we compute the real polygon height, and also add 2 to the table size - guich@502_1: increment p in the first iteration
   {
      if (*p > maxy) maxy = *p;
      if (*p < miny) miny = *p;
   }
   if (maxy == miny) return; // guich@501_13: nothing to be drawn.
   xCoords = (Sint32 *)((*vmGlobals->xmalloc)(vmGlobals->screenHeight/3*sizeof(Sint32))); // guich@500_10: depends on the screen height from now on
   if (!xCoords) return; // guich@501_13
   edgeTable = (*vmGlobals->xmalloc)(((maxy-miny)+2) * sizeof(struct edge *));
   if (!edgeTable)
   {
      xfree(xCoords); // guich@402_25
      (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "In native fillPolygon 1"); // guich@200b4_78
      return;
   }
   edgeTable -= miny; // allow the edgetable to store negative values
   for (i = maxy+2; --i >= miny;)
      edgeTable[i] = NULL;

   for (i=0; i < nPoints; ++i)
   {
      Sint32 p1x,p1y,p2x,p2y,p3y;
      p1y = yPoints[i];
      p2y = yPoints[(i + 1) % nPoints];
      if (p1y == p2y)
          continue;   // Skip horiz. edges
      p1x = xPoints[i];
      p2x = xPoints[(i + 1) % nPoints];

      // Find next vertex not level with p2
      for (j = (i + 2) % nPoints; ; j = (j + 1) % nPoints)
      {
         p3y = yPoints[j];
         if (p2y != p3y) break;
      }
      e = (*vmGlobals->xmalloc)(sizeof(struct edge));
      if (!e)
      {
         (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "In native fillPolygon 2"); // guich@200b4_78 - note that we will not free some chunks
         goto end;
      }
      e->xNowNumStep = abs(p1x - p2x);
      if (p2y > p1y)
      {
         e->yTop = p1y;
         e->yBot = p2y;
         e->xNowWhole = p1x;
         e->xNowDir = sgn(p2x - p1x);
         e->xNowDen = e->yBot - e->yTop;
         e->xNowNum = (e->xNowDen >> 1);
         if (p3y > p2y) e->yBot--;
      }
      else
      {
         e->yTop = p2y;
         e->yBot = p1y;
         e->xNowWhole = p2x;
         e->xNowDir = sgn(p1x - p2x);
         e->xNowDen = e->yBot - e->yTop;
         e->xNowNum = (e->xNowDen >> 1);
         if (p3y < p2y)
         {
            e->yTop++;
            e->xNowNum += e->xNowNumStep;
            while (e->xNowNum >= e->xNowDen)
            {
               e->xNowWhole += e->xNowDir;
               e->xNowNum -= e->xNowDen;
            }
         }
      }
      e->next = edgeTable[e->yTop];
      edgeTable[e->yTop] = e;
   }

   for (curY=miny; edgeTable[curY] == NULL; ++curY)
       if (curY == maxy - 1) goto end;
   // draw runs
   for (active=NULL; (active = updateActive(active, edgeTable, curY)) != NULL; ++curY)
   {
      /*
      | DrawRuns first uses an insertion sort to order the X
      | coordinates of each active edge.  It updates the X coordinates
      | for each edge as it does this.
      | Then it draws a run between each pair of coordinates,
      | using the specified fill pattern.
      */
      Sint32 numCoords = 0;
      for (e = active; e != NULL; e = e->next)
      {
         for (i = numCoords; i > 0 && xCoords[i - 1] > e->xNowWhole; --i)
            xCoords[i] = xCoords[i-1];
         xCoords[i] = e->xNowWhole;
         if (++numCoords >= 50)
         {
            (*vmGlobals->debug)("too much points in fill polygon. please inform support@superwaba.com.br");
            goto end; // too much points! maybe a fatal error?
         }
         e->xNowNum += e->xNowNumStep;
         while (e->xNowNum >= e->xNowDen)
         {
            e->xNowWhole += e->xNowDir;
            e->xNowNum -= e->xNowDen;
         }
      }
      if (numCoords & 1)   /* Protect from degenerate polygons */
      {
         xCoords[numCoords] = xCoords[numCoords - 1], /** !!! */
         numCoords++;
      }

      for (i = 0; i < numCoords; i += 2)
      {
         Sint32 w = xCoords[i+1]-xCoords[i]+1;
         if (w > 0) drawHLine(xCoords[i], (Sint32)curY, w, c, c);
      }
   }
end:
    edgeTable += miny;
    xfree(xCoords);
    xfree(edgeTable);
}

////////////////////////////////////////////////////////////////////////////
// draws a polygon. if the polygon is not closed, close it

static void drawPolygon(Sint32 *xPoints, Sint32 *yPoints, Sint32 nPoints,
                        Pixel pixel, Uint8 close)
{
   Sint32 i;
   if (nPoints < 2) return;
   for (i=1; i < nPoints; i++)
   {
      drawLine(xPoints[i-1], yPoints[i-1], xPoints[i], yPoints[i], pixel);
   }
   if (close)
   {
      --nPoints;
      if (xPoints[0] != xPoints[nPoints] || yPoints[0] != yPoints[nPoints])
         drawLine(xPoints[0], yPoints[0], xPoints[nPoints], yPoints[nPoints], pixel);
   }
}
////////////////////////////////////////////////////////////////////////////
// draw an elliptical arc from startAngle to endAngle.
// c is the fill color and c2 is the outline color
// (if in fill mode - otherwise, c = outline color)
static void arcPiePointDrawAndFill(Sint32 xc, Sint32 yc, Sint32 rx, Sint32 ry,
                                   float startAngle, float endAngle,
                                   Pixel c, Pixel c2,
                                   Uint8 fill, Uint8 pie,
                                   Sint32 *startAnglePoint)
{
   static Sint32 * xPoints = 0;
   static Sint32 * yPoints = 0;
   static Sint32 nPoints = 0;
   static Sint32 lastRX =-1;
   static Sint32 lastRY =-1;
   static Sint32 lastXC = -1;
   static Sint32 lastYC = -1;
   static Sint32 lastSize=0;
   static float lastPPD=0;
   Sint32 limitx1 = CLIPX1-TRANSX; // guich@501_13: store these for faster computation
   Sint32 limitx2 = CLIPX2-TRANSX;
   Sint32 limity1 = CLIPY1-TRANSY;
   Sint32 limity2 = CLIPY2-TRANSY;
   // this algorithm was created by Guilherme Campos Hazan
   float ppd;
   Sint32 startIndex;
   Sint32 endIndex;
   Sint32 index;
   Sint32 i;
   Sint32 nq;
   Sint32 size=0;
   Sint32 oldX=0;
   Sint32 oldY=0;
   Uint8 checkClipX = (Uint8)((xc+rx) > limitx2 || (xc-rx) < limitx1); // guich@340_3 - guich@401_2: added TRANSX/Y
   Uint8 checkClipY = (Uint8)((yc+ry) > limity2 || (yc-ry) < limity1);

   if (rx < 0 || ry < 0) // guich@501_13
      return;
   // step 0: correct angle values
   if (startAngle < 0.1 && endAngle > 359.9) // full circle? use the fastest routine instead
   {
      if (fill)
         ellipseDrawAndFill(xc, yc, rx, ry, c, 1);
      ellipseDrawAndFill(xc, yc, rx, ry, (Pixel)(fill? c2:c), (Pixel)0);
      return;
   }

   // step 0: if possible, use cached results
   if (xc != lastXC || yc != lastYC || rx != lastRX || ry != lastRY)
   {
      // step 1: computes how many points the circle has (computes only 45 degrees and mirrors the rest)
      // intermediate terms to speed up loop
      Sint32 t1 = rx*rx, t2 = t1<<1, t3 = t2<<1;
      Sint32 t4 = ry*ry, t5 = t4<<1, t6 = t5<<1;
      Sint32 t7 = rx*t5, t8 = t7<<1, t9 = 0L;
      Sint32 d1 = t2 - t7 + (t4>>1);    // error terms
      Sint32 d2 = (t1>>1) - t8 + t5;
      Sint32 x = rx;                 // ellipse points
      Sint32 y = 0;                  // ellipse points
      while (d2 < 0)              // til slope = -1
      {
         ++y;                     // always move up here
         t9 += t3;
         if (d1 < 0)             // move straight up
         {
            d1 += t9 + t2;
            d2 += t9;
         }
         else                   // move up and left
         {
            --x;
            t8 -= t6;
            d1 += t9 + t2 - t8;
            d2 += t9 + t5 - t8;
         }
         ++size;
      }

      do             // rest of top right quadrant
      {
         --x;         // always move left here
         t8 -= t6;
         if (d2 < 0)  // move up and left
         {
            y++;
            t9 += t3;
            d2 += t9 + t5 - t8;
         }
         else d2 += t5 - t8;  // move straight left
         ++size;
      } while (x >= 0);
      nq = size;
      size *= 4;
      // step 2: computes how many points per degree
      ppd = (float)size / 360.0f;
      // step 3: create space in the buffer so it can save all the circle
      size+=2;
      if (nPoints < size)
      {
         Uint32 temp = sizeof(Sint32)*size;
         xPoints = (*vmGlobals->xrealloc)(xPoints, &temp);
         if (!xPoints)
         {
            (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "in ArcPiePointDrawAndFill 1"); // guich@200b4_78
            return;
         }
         yPoints = (*vmGlobals->xrealloc)(yPoints, &temp);
         if (!yPoints)
         {
            xfree(xPoints);
            (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "in ArcPiePointDrawAndFill 2"); // guich@200b4_78
            return;
         }
      }
      nPoints = size;
      // step 4: stores all the circle in the array. the odd arcs are drawn in reverse order
      // intermediate terms to speed up loop
      t2 = t1<<1;
      t3 = t2<<1;
      t8 = t7<<1;
      t9 = 0L;
      d1 = t2 - t7 + (t4>>1); // error terms
      d2 = (t1>>1) - t8 + t5;

      x = rx;
      y = 0;  // ellipse points

      i=0;
      while (d2 < 0)          // til slope = -1
      {
         // save 4 points using symmetry
         // guich@340_3: added clipping
         #define doClipX(x) ( ((x) < limitx1) ? limitx1 : ((x) >= limitx2) ? limitx2 : (x) ) // guich@401_2: added TRANSX/Y
         #define doClipY(y) ( ((y) < limity1) ? limity1 : ((y) >= limity2) ? limity2 : (y) )

         index = nq*0+i;      // 0/3
         xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
         yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

         index = nq*2-i-1;    // 1/3
         xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
         yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

         index = nq*2+i;      // 2/3
         xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
         yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);

         index = nq*4-i-1;    // 3/3
         xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
         yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);

         i++;
         y++;        // always move up here
         t9 += t3;
         if (d1 < 0)  // move straight up
         {
             d1 += t9 + t2;
             d2 += t9;
         }
         else      // move up and left
         {
             x--;
             t8 -= t6;
             d1 += t9 + t2 - t8;
             d2 += t9 + t5 - t8;
         }
      }

      do             // rest of top right quadrant
      {
         // save 4 points using symmetry
         // guich@340_3: added clipping
         index = nq*0+i;    // 0/3
         xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
         yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

         index = nq*2-i-1;  // 1/3
         xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
         yPoints[index]=checkClipY?doClipY(yc-y):(yc-y);

         index = nq*2+i;    // 2/3
         xPoints[index]=checkClipX?doClipX(xc-x):(xc-x);
         yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);

         index = nq*4-i-1;  // 3/3
         xPoints[index]=checkClipX?doClipX(xc+x):(xc+x);
         yPoints[index]=checkClipY?doClipY(yc+y):(yc+y);

         ++i;
         --x;        // always move left here
         t8 -= t6;
         if (d2 < 0)  // move up and left
         {
            ++y;
            t9 += t3;
            d2 += t9 + t5 - t8;
         }
         else d2 += t5 - t8;   // move straight left
      } while (x >= 0);
      // save last arguments
      lastXC = xc;
      lastYC = yc;
      lastRX = rx;
      lastRY = ry;
      lastPPD = ppd;
      lastSize = size;
   }
   else ppd = lastPPD;
   // step 5: computes the start and end indexes that will become part of the arc
   startIndex = (Sint32)(ppd * startAngle);
   endIndex = (Sint32)(ppd * endAngle);
   // step 5 1/2: if only computing the point, return it
   if (startAnglePoint)
   {
      startAnglePoint[0] = xPoints[startIndex];
      startAnglePoint[1] = yPoints[startIndex];
      return;
   }
   if (endIndex == (lastSize-2)) // 360?
      endIndex--;
   // step 6: fill or draw the polygons
   endIndex++;
   if (pie)
   {
      oldX = xPoints[endIndex];
      oldY = yPoints[endIndex];
      xPoints[endIndex] = xc;
      yPoints[endIndex] = yc;
      endIndex++;
   }
   if (fill)
      fillPolygon(xPoints+startIndex, yPoints+startIndex, endIndex-startIndex, c2);  // guich@401_3
   drawPolygon(xPoints+startIndex, yPoints+startIndex, endIndex-startIndex, c, pie); // (Pixel)(fill? c2 : c), fill // guich@402_57 and guich@402_58: see waba.fx.Graphics.java
   if (pie)  // restore saved points
   {
      xPoints[endIndex-1] = oldX;
      yPoints[endIndex-1] = oldY;
   }

}
////////////////////////////////////////////////////////////////////////////
// calls arcPiePointDrawAndFill twice if the angles cross the 0
static void preArcPiePointDrawAndFill(Sint32 xc, Sint32 yc, Sint32 rx, Sint32 ry,
                                      float startAngle, float endAngle,
                                      Pixel c, Pixel c2,
                                      Uint8 fill, Uint8 pie)
{
   // make sure the values are -359 <= x <= 359
   while (startAngle <= -360) startAngle += 360;
   while (endAngle   <= -360) endAngle   += 360;
   while (startAngle >   360) startAngle -= 360;
   while (endAngle   >   360) endAngle   -= 360;

   if (startAngle == endAngle) // guich@501_13
      return;
   if (startAngle > endAngle) // eg 235 to 45
      startAngle -= 360; // set to -45 to 45 so we can handle it correctly
   if (startAngle >= 0 && endAngle <= 0) // eg 135 to -135
      endAngle += 360; // set to 135 to 225

   if (startAngle >= 0 && endAngle >= 0)
      arcPiePointDrawAndFill(xc, yc, rx, ry, startAngle, endAngle, c, c2, fill, pie, 0);
   else
   if (startAngle <= 0 && endAngle >= 0)  // eg -45 to 45
   {
      startAngle += 360;
      arcPiePointDrawAndFill(xc, yc, rx, ry, startAngle, 360, c, c2, fill, pie, 0);
      arcPiePointDrawAndFill(xc, yc, rx, ry, 0, endAngle, c, c2, fill, pie, 0);
   }
   else (*vmGlobals->debug)("arc/pie/point could not be filled with given angles");
}

////////////////////////////////////////////////////////////////////////////
// for a angle, computes its point. useful if you want to place a label in the pie and dont want to use sin/cos. index 0 is x ad index 1 is y
static void getAnglePoint(Sint32 xc, Sint32 yc, Sint32 rx, Sint32 ry,
                          float angle, Sint32 * twoInts)
{
   while (angle >= 360) angle -= 360;
   while (angle < 0) angle += 360;
   arcPiePointDrawAndFill(xc, yc, rx, ry, angle, angle, 0, 0, 0, 0, twoInts);
}

////////////////////////////////////////////////////////////////////////////
static void drawRoundRect(Sint32 x, Sint32 y, Sint32 width, Sint32 height, Sint32 r, Pixel c)
{
   Sint32 x1, y1, x2, y2, dec, xx, yy;
   Sint32 w, h;
   if (r > (width/2) || r > (height/2)) r = min(width/2,height/2); // guich@200b4_6: correct bug that crashed the device.
   w = width - 2*r;
   h = height - 2*r;
   x1 = x+r;
   y1 = y+r;
   x2 = x+width-r-1;
   y2 = y+height-r-1;
   dec = 3-2*r;

   drawHLine(x+r, y, w, c, c); // top
   drawHLine(x+r, y+height-1, w, c, c); // bottom
   drawVLine(x, y+r, h, c, c); // left
   drawVLine(x+width-1, y+r, h, c, c); // right

   // draw the round rectangles.
   for (xx = 0, yy = r; xx <= yy; xx++)
   {
      setPixel(x2+xx, y2+yy, c);
      setPixel(x2+xx, y1-yy, c);
      setPixel(x1-xx, y2+yy, c);
      setPixel(x1-xx, y1-yy, c);

      setPixel(x2+yy, y2+xx, c);
      setPixel(x2+yy, y1-xx, c);
      setPixel(x1-yy, y2+xx, c);
      setPixel(x1-yy, y1-xx, c);
      if (dec >= 0)
         dec += -4*(yy--)+4;
      dec += 4*xx+6;
   }
}
////////////////////////////////////////////////////////////////////////////
static void fillRoundRect(Sint32 x, Sint32 y, Sint32 width, Sint32 height, Sint32 r, Pixel c)
{
   Sint32 x1, y1, x2, y2, xx, yy, dec;
   if (r > (width/2) || r > (height/2)) r = min(width/2,height/2); // guich@200b4_6: correct bug that crashed the device.

   x1 = x+r;
   y1=y+r;
   x2=x+width-r-1;
   y2=y+height-r-1;
   dec = 3-2*r;

   height -= 2*r;
   y += r;
   while (height--)
      drawHLine(x, y++, width, c, c);
   // fill the round rectangles
   for (xx = 0, yy = r; xx <= yy; xx++)
   {
      drawLine(x1-xx, y1-yy, x2+xx, y1-yy, c);
      drawLine(x1-xx, y2+yy, x2+xx, y2+yy, c);

      drawLine(x1-yy, y1-xx, x2+yy, y1-xx, c);
      drawLine(x1-yy, y2+xx, x2+yy, y2+xx, c);

      if (dec >= 0)
         dec += -4*(yy--)+4;
      dec += 4*xx+6;
   }
}
////////////////////////////////////////////////////////////////////////////
// sets the clip rect to (x,y,x+w-1,y+h-1), translated to the current translated origin
static void setClip(Sint32 x, Sint32 y, Sint32 w, Sint32 h)
{
   Sint32 clipX1 = x+TRANSX;
   Sint32 clipY1 = y+TRANSY;
   Sint32 clipX2 = clipX1+w;
   Sint32 clipY2 = clipY1+h;

   if (clipX1 <    0) clipX1 = 0;
   if (clipY1 <    0) clipY1 = 0;
   if (clipX1 > maxX) clipX1 = maxX;
   if (clipY1 > maxY) clipY1 = maxY;

   if (clipX2 <    0) clipX2 = 0;
   if (clipY2 <    0) clipY2 = 0;
   if (clipX2 > maxX) clipX2 = maxX;
   if (clipY2 > maxY) clipY2 = maxY;

   GRHANDLE_ClipX1(activeGrHandle) = clipX1;
   GRHANDLE_ClipY1(activeGrHandle) = clipY1;
   GRHANDLE_ClipX2(activeGrHandle) = clipX2;
   GRHANDLE_ClipY2(activeGrHandle) = clipY2;

   clipX = MAKE32(clipX1,clipX2);
   clipY = MAKE32(clipY1,clipY2);
}

////////////////////////////////////////////////////////////////////////////
// translates the origin of the drawing area by dx and dy. the final translation coords cannot be lesser than 0.
static void translate(Sint32 dx, Sint32 dy)
{
   dx += TRANSX;
   dy += TRANSY;

   GRHANDLE_TransX(activeGrHandle) = dx;
   GRHANDLE_TransY(activeGrHandle) = dy;
   transXY = MAKE32(dx,dy);
}

/*-----------------------------------------------------------------intersects-+
| returns 1 if the coords intersects the clip rectangle                       |
+----------------------------------------------------------------------------*/
Uint8 intersects(RectangleType *r1, RectangleType *r2)
{
   return (r1->topLeft.x + r1->extent.x) > r2->topLeft.x &&
          (r1->topLeft.y + r1->extent.y) > r2->topLeft.y &&
          (r2->topLeft.x + r2->extent.x) > r1->topLeft.x &&
          (r2->topLeft.y + r2->extent.y) > r1->topLeft.y;
}

/*--------------------------------------------------------------intersectWith-+
| store in r1 the intersection between r1 and r2                              |
+----------------------------------------------------------------------------*/
void intersectWith(RectangleType * r1, RectangleType * r2)
{
   if (r1->topLeft.x > r2->topLeft.x)
   {
      if ((r2->topLeft.x + r2->extent.x) < (r1->topLeft.x + r1->extent.x))
         r1->extent.x = r2->topLeft.x + r2->extent.x - r1->topLeft.x;
   }
   else
   {
      if ((r2->topLeft.x + r2->extent.x) <= (r1->topLeft.x + r1->extent.x))
         r1->extent.x = r2->extent.x;
      else
         r1->extent.x = r1->topLeft.x + r1->extent.x - r2->topLeft.x;
      r1->topLeft.x = r2->topLeft.x;
   }

   if (r1->topLeft.y > r2->topLeft.y)
   {
      if ((r2->topLeft.y + r2->extent.y) < (r1->topLeft.y + r1->extent.y))
         r1->extent.y = r2->topLeft.y + r2->extent.y - r1->topLeft.y;
   }
   else
   {
      if ((r2->topLeft.y + r2->extent.y) <= (r1->topLeft.y + r1->extent.y))
         r1->extent.y = r2->extent.y;
      else
         r1->extent.y = r1->topLeft.y + r1->extent.y - r2->topLeft.y;
      r1->topLeft.y = r2->topLeft.y;
   }
}

/*------------------------------------------------------------collisionDetect-+
| returns true if the two bitmaps collids, ie,                                |
| if there are any points != backcolor intersected                            |
+----------------------------------------------------------------------------*/
static Uint8 collisionDetect(GrHandle bmp1, RectangleType rect1,
                            GrHandle bmp2, RectangleType rect2,
                            Pixel backPixel)
{
   Sint32 x1,y1,x2,y2;
   Sint32 w,h;

   if (rect1.topLeft.x > rect2.topLeft.x)
   {
      x1 = 0;
      x2 = rect1.topLeft.x - rect2.topLeft.x;
      if ((rect1.topLeft.x + rect1.extent.x) < (rect2.topLeft.x + rect2.extent.x))
         w = rect1.extent.x;
      else
         w = rect2.extent.x - x2;
   }
   else
   {
      x1 = rect2.topLeft.x - rect1.topLeft.x;
      x2 = 0;
      if ((rect1.topLeft.x + rect1.extent.x) < (rect2.topLeft.x + rect2.extent.x))
         w = rect1.extent.x - x1;
      else
         w = rect2.extent.x;
   }
   if (w <= 0) return 0;

   if (rect1.topLeft.y > rect2.topLeft.y)
   {
      y1 = 0;
      y2 = rect1.topLeft.y - rect2.topLeft.y;
      if ((rect1.topLeft.y + rect1.extent.y) < (rect2.topLeft.y + rect2.extent.y))
         h = rect1.extent.y;
      else
         h = rect2.extent.y - y2;
   }
   else
   {
      y1 = rect2.topLeft.y - rect1.topLeft.y;
      y2 = 0;
      if ((rect1.topLeft.y + rect1.extent.y) < (rect2.topLeft.y + rect2.extent.y))
         h = rect1.extent.y - y1;
      else
         h = rect2.extent.y;
   }
   if (h > 0)
   {
      SDL_Surface * surface1 = (SDL_Surface *)GRHANDLE_WinHandle(bmp1);
      SDL_Surface * surface2 = (SDL_Surface *)GRHANDLE_WinHandle(bmp2);
      Sint32 i, j;
      Pixel const * pixels1;
      Pixel const * pixels2;
      Sint32 pitch;
      Sint32 bytesPerPixel;

      SDL_LockSurface(surface1);
      SDL_LockSurface(surface2);
      pitch = surface1->pitch;
      bytesPerPixel = surface1->format->BytesPerPixel;
      pixels1 = (Pixel const *)((char const *)surface1->pixels + ((y1 * pitch) + (x1 * bytesPerPixel)));
      pixels2 = (Pixel const *)((char const *)surface2->pixels + ((y2 * pitch) + (x2 * bytesPerPixel)));
      for (i=0; i < h; ++i)
      {
         Pixel const * p1 = pixels1;
         Pixel const * p2 = pixels2;
         for (j=0; j < w; ++j)
            if ((*p1++ != backPixel) && (*p2++ != backPixel))
            {
               // found a touching pixel.
               SDL_UnlockSurface(surface2);
               SDL_UnlockSurface(surface1);
               return 1;
            }
         pixels1 = (Pixel const *)((char const *)pixels1 + pitch);
         pixels2 = (Pixel const *)((char const *)pixels2 + pitch);
      }
      SDL_UnlockSurface(surface2);
      SDL_UnlockSurface(surface1);
   }
   return 0;
}

/*-----------------------------------------------------------translateAndClip-+
| Translates the given coords and returns the intersection between            |
| the clip rect and the coords passed.                                        |
|                                                                             |
| Returns: 1 if OK, 0 if the coords are outside the clip rect                 |
+----------------------------------------------------------------------------*/
Uint8 translateAndClip(Sint32 *pX, Sint32 *pY, Sint32 *pWidth, Sint32 *pHeight)
{
   Sint32 x = *pX;
   Sint32 y = *pY;
   Sint32 w = *pWidth;
   Sint32 h = *pHeight;
   x += TRANSX;
   y += TRANSY;
   if (x < CLIPX1)
   {
      if ((x+w) > CLIPX2)
         w = CLIPX2 - CLIPX1;
      else
      {
         w -= CLIPX1-x;
         if (w <= 0) return 0;
      }
      x = CLIPX1;
   }
   else
   if ((x+w) > CLIPX2)
   {
      w = CLIPX2 - x;
      if (w <= 0) return 0;
   }
   if (y < CLIPY1)
   {
      if ((y+h) > CLIPY2)
         h = CLIPY2 - CLIPY1;
      else
      {
         h -= CLIPY1-y;
         if (h <= 0) return 0;
      }
      y = CLIPY1;
   }
   else
   if ((y+h) > CLIPY2)
   {
      h = CLIPY2 - y;
      if (h <= 0) return 0;
   }
   *pX      = x;
   *pY      = y;
   *pWidth  = w;
   *pHeight = h;
   return 1;
}

/*--------------------------------------------------------------setDrawWindow-+
|                                                                             |
+----------------------------------------------------------------------------*/
void setDrawWindow(GrHandle newGrHandle)
{
   if (activeGrHandle != newGrHandle)
   {
      // saves back the values in the object
      GRHANDLE_TransX(activeGrHandle) = TRANSX;
      GRHANDLE_TransY(activeGrHandle) = TRANSY;
      GRHANDLE_ClipX1(activeGrHandle) = CLIPX1;
      GRHANDLE_ClipX2(activeGrHandle) = CLIPX2;
      GRHANDLE_ClipY1(activeGrHandle) = CLIPY1;
      GRHANDLE_ClipY2(activeGrHandle) = CLIPY2;
      activeGrHandle = newGrHandle;
      transXY = MAKE32(GRHANDLE_TransX(newGrHandle), GRHANDLE_TransY(newGrHandle));
      clipX = MAKE32(GRHANDLE_ClipX1(newGrHandle), GRHANDLE_ClipX2(newGrHandle));
      clipY = MAKE32(GRHANDLE_ClipY1(newGrHandle), GRHANDLE_ClipY2(newGrHandle));
      maxX = GRHANDLE_Width(newGrHandle);
      maxY = GRHANDLE_Height(newGrHandle);
   }
}

/*----------------------------------------------------------selectGraphicsObj-+
| Set the current graphics global values if the Graphics object changes       |
+----------------------------------------------------------------------------*/
static void selectGraphicsObj(WObject newGraphics)
{
   GrHandle newGrHandle = WOBJ_ControlImageGraphicsGrHandle(newGraphics);
   if (newGraphics != activeGraphics)
   {
      _LOG_INFO((LOG_CAT_GFX, "selectGraphicsObj !="));
      if (activeGraphics && WOBJ_ControlImageGraphicsGrHandle(activeGraphics))
      {
         WOBJ_GraphicsForeRGB(activeGraphics) = currentForePixel;
         WOBJ_GraphicsBackRGB(activeGraphics) = currentBackPixel;
         WOBJ_GraphicsFont(activeGraphics) = currentUserFont;
         _LOG_DEBUG((LOG_CAT_GFX, "oldgfx backC=%04x foreC=%04x", currentForePixel, currentBackPixel));
      }
      activeGraphics = newGraphics;
      currentForePixel = WOBJ_GraphicsForeRGB(newGraphics);
      currentBackPixel = WOBJ_GraphicsBackRGB(newGraphics);
      currentUserFont = WOBJ_GraphicsFont(newGraphics);
      _LOG_DEBUG((LOG_CAT_GFX, "newgfx backC=%04x foreC=%04x", currentForePixel, currentBackPixel));
   }
   if (activeGrHandle != newGrHandle)
      setDrawWindow(newGrHandle);
}

////////////////////////////////////////////////////////////////////////////
//         INTERFACE OF THE GRAPHICS CLASS METHODS WITH THE C CODE        //
////////////////////////////////////////////////////////////////////////////

/*--------------------------------------------------------------drawHourGlass-+
|                                                                             |
+----------------------------------------------------------------------------*/
void drawHourGlass(Sint32 step)
{
   if (graphicsActive)
   {
      Pixel c1, c2, c3, c4, c5;
      Sint32 x = vmGlobals->screenWidth / 2;
      Sint32 y = vmGlobals->screenHeight / 2;
      Sint32 maxColors = vmGlobals->maxColors;

      if ((maxColors == 0) || (maxColors >= 256))
      {
         c1 = getColor(0, 0, 0);
         c2 = getColor(0, 0, 255);
         c3 = getColor(0, 255, 0);
         c4 = getColor(255, 0, 0);
      }
      else
      if (maxColors == 16)
      {
         c1 = 0xF;
         c2 = 0x2;
         c3 = 0x6;
         c4 = 0xA;
      }
      else
      {
         c1 = 3;
         c2 = 1;
         c3 = 2;
         c4 = 3;
      }
      c5 = getColor(255,255,255);
      if (step == 0)
      {
         // draw hourglass
         fillRect(x-4, y-7, 9, 14, c1);
         // top line
         drawLine(x-3, y-6, x+3, y-6, c5);
         // left indent
         drawLine(x-4, y-3, x-4, y+2, c5);
         drawLine(x-3, y-2, x-3, y+1, c5);
         drawLine(x-2, y-1, x-2, y, c5);
         // right indent
         drawLine(x+2, y-1, x+2, y, c5);
         drawLine(x+3, y-2, x+3, y+1, c5);
         drawLine(x+4, y-3, x+4, y+2, c5);
         // dots
         drawLine(x, y, x, y, c5);
         drawLine(x, y+2, x, y+2, c5);
         drawLine(x-1, y+3, x-1, y+3, c5);
         drawLine(x+1, y+3, x+1, y+3, c5);
         // top line
         drawLine(x-3, y+5, x+3, y+5, c5);
         fillRect(x-10, y+10, 5, 8, c4);
         fillRect(x-5, y+10, 5, 8, c3);
         fillRect(x, y+10, 5, 8, c2);
         drawRect(x-10, y+10, 20, 8, c1);
      }
      else fillRect(x-10, y+10, 20, 8, c5);  // erase
   }
}

//////////////////////////////////////////////////////////////////////////
//              Functions called by the Java routines                   //
//////////////////////////////////////////////////////////////////////////

/*
| A Graphics is created from an Isurface: either a Control or an Image.
|
| Graphics, Image, and Control have all a pointer to a GraphicsContext
| structure.  This structure groups miscellaneous fields pertinent to
| the surface.  The pointer to the GraphicsContext is named a GrHandle,
| x_mallocated (in makeGrHandle) and x_freed.
|
| For SDL, the GraphicsContext structure points to the SDL_Surface.
|
| When an Image is created, the GraphicsContext is allocated
| and a SDL_Surface is created as well (through: createSurface)
| The SDL_Surface is either:
| - a video-compatible surface
|   GRHANDLE_VideoFlag is VIDEOFLAG_COMPATIBLE_OFFSCREEN)
| or:
| - a surface with different characteristics (RGB, pitch, ...)
|   GRHANDLE_VideoFlag is VIDEOFLAG_DIFFERENT_OFFSCREEN
|
| When a Graphics is created for an Image, its GrHandle is copied
| from the GrHandle of the Image.  Therefore, it should not be freed when
| the Graphics is destroyed.
|
| When a Graphics is created from a Control, the GraphicsContext
| is allocated, the existing Video surface is used (but with a new clip
| and translate variables) and GRHANDLE_VideoFlag is VIDEOFLAG_MAINSCREEN.
|
| Hence:
| - GraphicsCreate never creates an SDL_Surface,
| - graphicsDestroy never frees an SDL_Surface
|
| Only imageDestroy(WObject image) can frre an SDL_Surface,
| and, finally, the main screen SDL_Surface is freed by SDL_Quit.
*/

/*-------------------------------------------------------------GraphicsCreate-+
|                                                                             |
+----------------------------------------------------------------------------*/
// waba/fx/Graphics void create()
void GraphicsCreate(Var stack[])
{
   WObject gr = stack[0].obj;
   WObject surface = WOBJ_GraphicsSurface(gr);
   Sint32 stype = surfaceGetType(surface);
   GrHandle h;

   if (stype == SURF_IMAGE)
   {
      h = WOBJ_ControlImageGraphicsGrHandle(surface);
   }
   else
   {
      // same bitmap as MainWin, but with its own clip and translate
      h = makeGrHandle(SDL_GetVideoSurface(), VIDEOFLAG_MAINSCREEN);
   }
   WOBJ_GraphicsSurfType(gr) = stype;
   WOBJ_ControlImageGraphicsGrHandle(gr) = h;
   WOBJ_GraphicsFont(gr) = vmGlobals->allFonts+0;
   WOBJ_ControlImageGraphicsGrHandle(surface) = h; // guich@402_28
}

/*------------------------------------------------------------graphicsDestroy-+
|                                                                             |
+----------------------------------------------------------------------------*/
void graphicsDestroy(WObject obj)
{
   if (obj != NULL)
   {
      GrHandle h = WOBJ_ControlImageGraphicsGrHandle(obj);
      if ((h != NULL) && (h != mainGrHandle))
      {
         if (h == activeGrHandle && mainGrHandle != NULL)  // guich@200b3: is this the active GrHandle?
            setDrawWindow(mainGrHandle); // switch to other handle if the handle is valid
         if (WOBJ_GraphicsSurfType(obj) != SURF_IMAGE) // guich@200b4_165: avoid freeing graphics pertinent to Images
         {
            if (GRHANDLE_Palette(h)) xfree(GRHANDLE_Palette(h));
            (*vmGlobals->xfree_)(h);
            WOBJ_ControlImageGraphicsGrHandle(obj) = NULL;
         }
      }
      if (obj == activeGraphics)
         activeGraphics = NULL;
   }
}

//////////////////////////////////////////////////////////////////////////
//                       Native Methods Interface                       //
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void free()
void GraphicsFree(Var stack[]) // ()V
{
   graphicsDestroy(stack[0].obj); // class object
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics static void nativeApplyPalette(int []palrgb, int startIndex, int endIndex)
void Graphics_applyPalette(Var stack[]) // ([III)V
{
   WObject intArray;
   Sint32 *pal;
   Sint32 start, end, len;
   SDL_Surface * surface;

   intArray = stack[0].obj;
   start = stack[1].intValue;
   end = stack[2].intValue;
   pal = (Sint32 *)WOBJ_arrayStart(intArray);

   len = end-start+1;
   surface = SDL_GetVideoSurface();
   if (surface->format->BitsPerPixel == 8)
   {
      SDL_Color * colors = (SDL_Color *)xmalloc(len * sizeof (SDL_Color));
      Sint32 i;

      for (i=0; i < len; ++i)
      {
         colors[i].r = (pal[i] >> 24) & 0x0F;
         colors[i].g = (pal[i] >> 16) & 0x0F;
         colors[i].b = pal[i] & 0x0F;
      }
      SDL_SetPalette(surface, SDL_LOGPAL|SDL_PHYSPAL,
                     colors, start, len);
      disposeHashtable(colorHashTable, 0); // cached colors now invalid.
      xfree(colors);
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void eraseRect(int x, int y, int w, int h)
void GraphicsEraseRect(Var stack[]) // (IIII)V
{
   Sint32 x = stack[1].intValue;
   Sint32 y = stack[2].intValue;
   Sint32 width = stack[3].intValue;
   Sint32 height = stack[4].intValue;
   selectGraphicsObj(stack[0].obj);
   if (translateAndClip(&x, &y, &width, &height)) {
      SDL_Surface * surf = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
      changeColors(surf, x, y, width, height,
                   currentForePixel, currentBackPixel); // pixelFrom, pixelTo
      updateRectangle(surf, x, y, width, height);
   }
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawEllipse(int xc, int yc, int rx, int ry)
void GraphicsDrawEllipse(Var stack[]) // (IIII)V
{
   selectGraphicsObj(stack[0].obj);
   ellipseDrawAndFill(stack[1].intValue, stack[2].intValue, // xc, yc
                      stack[3].intValue, stack[4].intValue, // rx, ry
                      currentForePixel, 0);                 // pixel, fill
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillEllipse(int xc, int yc, int rx, int ry)
void GraphicsFillEllipse(Var stack[]) // (IIII)V
{
   selectGraphicsObj(stack[0].obj);
   ellipseDrawAndFill(stack[1].intValue, stack[2].intValue, // xc, yc
                      stack[3].intValue, stack[4].intValue, // rx, ry
                      currentBackPixel, 1);                 // pixel, fill  - guich@340_41: replaced currentForePixel by currentBackPixel
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawArc(int xc, int yc, int r, float startAngle, float endAngle)
void GraphicsDrawArc(Var stack[]) // (IIIFF)V
{
   selectGraphicsObj(stack[0].obj);
   preArcPiePointDrawAndFill(stack[1].intValue, stack[2].intValue,      // xc, yc
                             stack[3].intValue, stack[3].intValue,      // rx, ry
                             stack[4].floatValue, stack[5].floatValue,  // startAngle, endAngle
                             currentForePixel, currentForePixel,        // c, c2
                             0, 0);                                     // fill, pie
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawPie(int xc, int yc, int r, float startAngle, float endAngle)
void GraphicsDrawPie(Var stack[]) // (IIIFF)V
{
   selectGraphicsObj(stack[0].obj);
   preArcPiePointDrawAndFill(stack[1].intValue, stack[2].intValue,       // xc, yc
                             stack[3].intValue, stack[3].intValue,       // rx, ry
                             stack[4].floatValue, stack[5].floatValue,   // startAngle, endAngle
                             currentForePixel, currentForePixel,         // c, c2
                             0, 1);                                      // fill, pie
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillPie(int xc, int yc, int r, float startAngle, float endAngle)
void GraphicsFillPie(Var stack[]) // (IIIFF)V
{
   selectGraphicsObj(stack[0].obj);
   preArcPiePointDrawAndFill(stack[1].intValue, stack[2].intValue,       // xc, yc
                             stack[3].intValue, stack[3].intValue,       // rx, ry
                             stack[4].floatValue, stack[5].floatValue,   // startAngle, endAngle
                             currentForePixel, currentBackPixel,         // c, c2
                             1, 1);                                      // fill, pie
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawEllipticalArc(int xc, int yc, int rx, int ry, float startAngle, float endAngle)
void GraphicsDrawEllipticalArc(Var stack[]) // (IIIIFF)V
{
   selectGraphicsObj(stack[0].obj);
   preArcPiePointDrawAndFill(stack[1].intValue, stack[2].intValue,       // xc, yc
                             stack[3].intValue, stack[4].intValue,       // rx, ry
                             stack[5].floatValue, stack[6].floatValue,   // startAngle, endAngle
                             currentForePixel, 0,                        // c, c2
                             0, 0);                                      // fill, pie
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawEllipticalPie(int xc, int yc, int rx, int ry, float startAngle, float endAngle)
void GraphicsDrawEllipticalPie(Var stack[]) // (IIIIFF)V
{
   selectGraphicsObj(stack[0].obj);
   preArcPiePointDrawAndFill(stack[1].intValue, stack[2].intValue,       // xc, yc
                             stack[3].intValue, stack[4].intValue,       // rx, ry
                             stack[5].floatValue, stack[6].floatValue,   // startAngle, endAngle
                             currentForePixel, 0,                        // c, c2
                             0, 1);                                      // fill, pie    - guich@402_57: was 0, 1, 0
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillEllipticalPie(int xc, int yc, int rx, int ry, float startAngle, float endAngle)
void GraphicsFillEllipticalPie(Var stack[]) // (IIIIFF)V
{
   selectGraphicsObj(stack[0].obj);
   preArcPiePointDrawAndFill(stack[1].intValue, stack[2].intValue,       // xc, yc
                             stack[3].intValue, stack[4].intValue,       // rx, ry
                             stack[5].floatValue, stack[6].floatValue,   // startAngle, endAngle
                             currentForePixel, currentBackPixel,         // c, c2
                             1, 1);                                      // fill, pie    - guich@402_57: was 1, 0
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawCircle(int xc, int yc, int r)
void GraphicsDrawCircle(Var stack[]) // (III)V
{
   selectGraphicsObj(stack[0].obj);
   ellipseDrawAndFill(stack[1].intValue, stack[2].intValue, // xc, yc
                      stack[3].intValue, stack[3].intValue, // rx, ry
                      currentForePixel, 0);                 // pixel, fill
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillCircle(int xc, int yc, int r)
void GraphicsFillCircle(Var stack[]) // (III)V
{
   selectGraphicsObj(stack[0].obj);
   ellipseDrawAndFill(stack[1].intValue, stack[2].intValue, // xc, yc
                      stack[3].intValue, stack[3].intValue, // rx, ry
                      currentBackPixel, 1);                 // pixel, fill
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics waba.fx.Coord getAnglePoint(int xc, int yc, int rx, int ry, float angle)
void GraphicsGetAnglePoint(Var stack[]) // (IIIIF)waba.fx.Coord
{
   WObject coord;
   if (!coordClass) coordClass = vmGlobals->getClass(vmGlobals->createUtfString("waba/fx/Coord"),1);
   coord = vmGlobals->createObject(coordClass);
   if (coord && ((*vmGlobals->pushObject)(coord) != -1)) // save the object, since getAnglePoint can trigger the GC
   {
      Sint32 ints[2];
      selectGraphicsObj(stack[0].obj);
      getAnglePoint(stack[1].intValue, stack[2].intValue,  // xc, yc
                    stack[3].intValue, stack[4].intValue,  // rx, ry
                    stack[5].floatValue,                   // angle
                    ints);                                 // twoInts
      // store in the object the result
      WOBJ_CoordX(coord) = ints[0];
      WOBJ_CoordY(coord) = ints[1];
      (*vmGlobals->popObject)(); // coord
      RETURN32(coord)
   }
   RETURN32(0);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics int getPixel (int x, int y)
void GraphicsGetPixel(Var stack[]) // (II)I
{
   Sint32 p;
   selectGraphicsObj(stack[0].obj);
   p = getPixel(stack[1].intValue,stack[2].intValue);
   RETURN32(p)
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void setPixel (int x, int y)
void GraphicsSetPixel(Var stack[]) // (II)V
{
   selectGraphicsObj(stack[0].obj);
   setPixel(stack[1].intValue, stack[2].intValue, currentForePixel);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawLine(int Ax, int Ay, int Bx, int By)
void GraphicsDrawLine(Var stack[]) // (IIII)V
{
   selectGraphicsObj(stack[0].obj);
   drawLine(stack[1].intValue, stack[2].intValue, stack[3].intValue, stack[4].intValue, currentForePixel);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void clearScreen()
void GraphicsClearScreen(Var stack[]) // ()V
{
   selectGraphicsObj(stack[0].obj);
   clearScreen(currentBackPixel);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawDots(int Ax, int Ay, int Bx, int By)
void GraphicsDrawDots(Var stack[]) // (IIII)V
{
   selectGraphicsObj(stack[0].obj);
   drawDottedLine(stack[1].intValue, stack[2].intValue, // x1, y1
                  stack[3].intValue, stack[4].intValue, // x2, y2
                  currentForePixel, currentBackPixel);  // c1, c2
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawCursor(int x, int y, int w, int h)
void GraphicsDrawCursor(Var stack[]) // (IIII)V
{
   Sint32 x = stack[1].intValue;
   Sint32 y = stack[2].intValue;
   Sint32 width = stack[3].intValue;
   Sint32 height = stack[4].intValue;

   selectGraphicsObj(stack[0].obj);
   if (translateAndClip(&x, &y, &width, &height))
   {
      SDL_Surface * active = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
      Sint32 pitch = active->pitch;
      char * row;
      Sint32 h = height+1;

      SDL_LockSurface(active);
      for (
         row = ((char *)active->pixels + ((y * pitch) + (x * sizeof(Pixel))));
         --h;
         row += pitch
      ) {
         Pixel * p = (Pixel *)row;
         Pixel * pMax = p + width;
         while (p < pMax)
            *p++ ^= (Pixel)(0xFFFFFFFF);
      }
      SDL_UnlockSurface(active);
      updateRectangle(active, x, y, width, height);
   }
}
////////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawDottedCursor(int x, int y, int w, int h)
void GraphicsDrawDottedCursor(Var stack[]) // (IIII)V
{
   Sint32 x = stack[1].intValue;
   Sint32 y = stack[2].intValue;
   Sint32 width = stack[3].intValue;
   Sint32 height = stack[4].intValue;

   selectGraphicsObj(stack[0].obj);
   {
      SDL_Surface * active = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
      Sint32 pitch = active->pitch;
      Sint32 h = height-1;
      Sint32 i,j,x2;
      Uint8 * row;
      x += TRANSX;
      y += TRANSY;
      x2 = x+width-1;

      // TODO a better clipping algotithm
      // NOTE: translateAndClip cannot be used for this routine!
      SDL_LockSurface(active);
      row = ((char *)active->pixels + ((y * pitch) + (x * sizeof(Pixel))));
      for (i=0; ; ++i)
      {
         if ((i==0) || (i==h)) // draw the horizontal rows
         {
            Pixel * p = (Pixel *)row;
            int yy = i + y;
            if (CLIPY1 <= yy && yy < CLIPY2)
               for (j=x; j < x2; j+=2)
               {
                  if (CLIPX1 <= j && j < CLIPX2)
                     *p ^= (Pixel)(0xFFFFFFFF);
                  p+=2;
               }
            if (i == h) break;
         }
         else
         if (i&1 && CLIPY1 <= i && i < CLIPY2)
         {
            if (CLIPX1 <= x && x < CLIPX2)
               *(Pixel *)row ^= (Pixel)(0xFFFFFFFF);
            if (CLIPX1 <= x2 && x2 < CLIPX2)
               *((Pixel *)row + (width-1)) ^= (Pixel)(0xFFFFFFFF);
         }
         row += pitch;
      }
      SDL_UnlockSurface(active);
      updateRectangle(active, x, y, width, height);
   }
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawCursorOutline(int x, int y, int w, int h)
void GraphicsDrawCursorOutline(Var stack[]) // (IIII)V
{
   Sint32 x = stack[1].intValue;
   Sint32 y = stack[2].intValue;
   Sint32 width = stack[3].intValue;
   Sint32 height = stack[4].intValue;

   selectGraphicsObj(stack[0].obj);
   if (translateAndClip(&x, &y, &width, &height))
   {
      SDL_Surface * active = (SDL_Surface *)GRHANDLE_WinHandle(activeGrHandle);
      Sint32 pitch = active->pitch;
      Sint32 h = height-1;
      Sint32 i;
      Uint8 * row;

      SDL_LockSurface(active);
      row = ((char *)active->pixels + ((y * pitch) + (x * sizeof(Pixel))));
      for (i=0; ; ++i)
      {
         if ((i==0) || (i==h))
         {
            Pixel * p = (Pixel *)row;
            Pixel * pMax = (Pixel *)row + width;
            while (p < pMax)
            {
               *p++ ^= (Pixel)(0xFFFFFFFF);
            }
            if (i == h) break;
         }
         else
         {
            *(Pixel *)row ^= (Pixel)(0xFFFFFFFF);
            *((Pixel *)row + (width-1)) ^= (Pixel)(0xFFFFFFFF);
         }
         row += pitch;
      }
      SDL_UnlockSurface(active);
      updateRectangle(active, x, y, width, height);
   }
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawRect(int x, int y, int w, int h)
void GraphicsDrawRect(Var stack[]) // (IIII)V
{
   selectGraphicsObj(stack[0].obj);
   drawRect(stack[1].intValue,stack[2].intValue,stack[3].intValue, stack[4].intValue, currentForePixel);
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillRect(int x, int y, int w, int h)
void GraphicsFillRect(Var stack[]) // (IIII)V
{
   selectGraphicsObj(stack[0].obj);
   if ((stack[3].intValue == maxX) && (stack[4].intValue == maxY) &&
       (transXY == 0) && (stack[1].intValue == 0) && (stack[2].intValue == 0) &&
       (CLIPX1 == 0) && (CLIPY1 == 0) && (CLIPX2 == maxX) && (CLIPY2 == maxY)) // guich@210_6: forgot to test against clip boundaries.
   {
      clearScreen(currentBackPixel); // guich@200b4_10 - fix wrong window paint
   }
   else
   {
      fillRect(stack[1].intValue, stack[2].intValue, stack[3].intValue, stack[4].intValue, currentBackPixel);
   }
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawDottedRect(int x, int y, int w, int h)
void GraphicsDrawDottedRect(Var stack[]) // (IIII)V
{
   selectGraphicsObj(stack[0].obj);
   drawDottedRect(stack[1].intValue, stack[2].intValue,  // x1, y1
                  stack[3].intValue, stack[4].intValue,  // x2, y2
                  currentForePixel, currentBackPixel);   // c1, c2
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillPolygon(int []xPoints, int []yPoints, int nPoints)
void GraphicsFillPolygon(Var stack[]) // ([I[II)V
{
   WObject xPoints = stack[1].obj;
   WObject yPoints = stack[2].obj;
   Sint32 nPoints = stack[3].intValue;
   // fdie@ the vm has a 4bytes pointer!
   Sint32 * xp = (Sint32 *)WOBJ_arrayStart(xPoints);
   Sint32 * yp = (Sint32 *)WOBJ_arrayStart(yPoints);

   selectGraphicsObj(stack[0].obj);
   if (!(*vmGlobals->arrayRangeCheck)(xPoints, 0, nPoints) ||
       !(*vmGlobals->arrayRangeCheck)(yPoints, 0, nPoints))
      (*vmGlobals->VmQuickError)(ERR_IndexOutOfRange,"In fillPolygon");
   else
      fillPolygon(xp, yp, nPoints, currentBackPixel);
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawPolygon(int []xPoints, int []yPoints, int nPoints)
void GraphicsDrawPolygon(Var stack[]) // ([I[II)V
{
   WObject xPoints,yPoints;
   Sint32 *xp;
   Sint32 *yp;
   Sint32 nPoints;

   selectGraphicsObj(stack[0].obj);
   xPoints = stack[1].obj;
   yPoints = stack[2].obj;
   nPoints = stack[3].intValue;
   // fdie@ the vm has a 4bytes pointer!
   xp = (Sint32 *)WOBJ_arrayStart(xPoints);
   yp = (Sint32 *)WOBJ_arrayStart(yPoints);

   if (!(*vmGlobals->arrayRangeCheck)(xPoints, 0, nPoints) || !(*vmGlobals->arrayRangeCheck)(yPoints, 0, nPoints))
      (*vmGlobals->VmQuickError)(ERR_IndexOutOfRange, "In drawPolygon");
   else
      drawPolygon(xp, yp, nPoints, currentForePixel,1);
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawPolyline(int []xPoints, int []yPoints, int nPoints)
void GraphicsDrawPolyline(Var stack[]) // ([I[II)V
{
   WObject xPoints,yPoints;
   Sint32 *xp;
   Sint32 *yp;
   Sint32 nPoints;

   selectGraphicsObj(stack[0].obj);
   xPoints = stack[1].obj;
   yPoints = stack[2].obj;
   nPoints = stack[3].intValue;
   xp = (Sint32 *)WOBJ_arrayStart(xPoints);
   yp = (Sint32 *)WOBJ_arrayStart(yPoints);

   if (!(*vmGlobals->arrayRangeCheck)(xPoints, 0, nPoints) || !(*vmGlobals->arrayRangeCheck)(yPoints, 0, nPoints))
      (*vmGlobals->VmQuickError)(ERR_IndexOutOfRange, "In drawPolyline");
   else
      drawPolygon(xp, yp, nPoints, currentForePixel, 0);
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawRoundRect(int x, int y, int width, int height, int r)
void GraphicsDrawRoundRect(Var stack[]) // (IIIII)V
{
   selectGraphicsObj(stack[0].obj);
   drawRoundRect(stack[1].intValue, stack[2].intValue,  // xc, yc
                 stack[3].intValue, stack[4].intValue,  // width, height
                 stack[5].intValue, currentForePixel);  // r, pixel
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void fillRoundRect(int x, int y, int width, int height, int r)
void GraphicsFillRoundRect(Var stack[]) // (IIIII)V
{
   selectGraphicsObj(stack[0].obj);
   fillRoundRect(stack[1].intValue, stack[2].intValue,  // xc, yc
                 stack[3].intValue, stack[4].intValue,  // width, height
                 stack[5].intValue, currentBackPixel);  // r, pixel
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawText(String text, int x, int y)
void GraphicsDrawText(Var stack[]) // (Ljava/lang/String;II)V
{
   if (stack[1].obj)            // kdarabi@400_84: null Strings lead to reset
   {
      WObject charArray = WOBJ_StringCharArrayObj(stack[1].obj);
      if (charArray)
      {
         // guich@200b4_91: use the Uint16* charArray directly
         Uint16 * charStart = (Uint16 *)WOBJ_arrayStart(charArray);
         Uint32 len = WOBJ_arrayLen(charArray);
         selectGraphicsObj(stack[0].obj);
         drawText(charStart, len,                        // text, len
                  stack[2].intValue, stack[3].intValue,  // x,y
                  currentForePixel);                     // pixel
      }
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics static void drawText(char []chars, int start, int count, int x, int y)
void GraphicsDrawTextChars(Var stack[]) // ([CIIII)V  - guich@402_29
{
   if (stack[1].obj)  // kdarabi@400_84: null Strings lead to reset
   {
      WObject charArray = stack[1].obj;
      // guich@200b4_91: use the Uint16* charArray directly
      Uint16 * charStart = (Uint16 *)WOBJ_arrayStart(charArray);
      Uint32 strLen = WOBJ_arrayLen(charArray);
      Uint32 start = stack[2].intValue;
      Uint32 len = stack[3].intValue;
      if ((start+len) <= strLen)
      {
         selectGraphicsObj(stack[0].obj);
         drawText(charStart+start, len,                    // text, len
                  stack[4].intValue, stack[5].intValue,    // x, y
                  currentForePixel);                       // color
      }
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void setClip(int x, int y, int w, int h)
void GraphicsSetClip(Var stack[]) // (IIII)V
{
   selectGraphicsObj(stack[0].obj);
   setClip(stack[1].intValue, stack[2].intValue,    // x, y
           stack[3].intValue, stack[4].intValue);   // w, h
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics waba.fx.Rect getClip(waba.fx.Rect r)
void GraphicsGetClip(Var stack[]) // (Lwaba/fx/Rect;)Lwaba/fx/Rect;
{
   WObject rect = stack[1].obj;
   if (rect)
   {
      selectGraphicsObj(stack[0].obj);
      WOBJ_RectX(rect) = CLIPX1-TRANSX;
      WOBJ_RectY(rect) = CLIPY1-TRANSY;
      WOBJ_RectWidth(rect)  = CLIPX2-CLIPX1; // guich@300_36: removed +1 here and line below
      WOBJ_RectHeight(rect) = CLIPY2-CLIPY1;
      RETURN32(rect)
   }
   RETURN32(0)
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void clearClip()
void GraphicsClearClip(Var stack[]) // ()V
{
   selectGraphicsObj(stack[0].obj);
   setClip(0, 0, GRHANDLE_Width(activeGrHandle), GRHANDLE_Height(activeGrHandle));
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics boolean clip(waba.fx.Rect r)
void GraphicsClip(Var stack[]) // (Lwaba/fx/Rect;)Z
{
   WObject rect = stack[1].obj;
   Sint32 x,y;
   Sint32 w,h;
   Sint32 b=0;

   if (rect)
   {
      selectGraphicsObj(stack[0].obj);
      x = WOBJ_RectX(rect);
      y = WOBJ_RectY(rect);
      w = WOBJ_RectWidth(rect);
      h = WOBJ_RectHeight(rect);
      b = translateAndClip(&x,&y,&w,&h);
      if (b)   // intersect?
      {
         WOBJ_RectX(rect) = x-TRANSX;
         WOBJ_RectY(rect) = y-TRANSY;
         WOBJ_RectWidth(rect) = w;
         WOBJ_RectHeight(rect) = h;
      }
   }
   RETURN32(b)
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void translate(int dx, int dy)
void GraphicsTranslate(Var stack[]) // (II)V
{
   Sint32 dx = stack[1].intValue;
   Sint32 dy = stack[2].intValue;
   selectGraphicsObj(stack[0].obj);
   if (dx || dy)
      translate(dx,dy);
}

//////////////////////////////////////////////////////////////////////////
// used by the drawImage method
// waba/fx/Graphics void copyRect(waba.fx.ISurface surface, int x, int y, int width, int height, int dstX, int dstY)
void GraphicsCopyRect(Var stack[]) // (Lwaba/fx/ISurface;IIIIII)V
{
   WObject surfOrig = stack[1].obj;
   WObject surfDest = stack[0].obj;
   GrHandle hOrig = WOBJ_ControlImageGraphicsGrHandle(surfOrig);
   GrHandle hDest = WOBJ_ControlImageGraphicsGrHandle(surfDest);

   _LOG_INFO((LOG_CAT_GFX, "GraphicsCopyRect"));
   if (hOrig && hDest)
   {
      selectGraphicsObj(surfDest);
      drawBitmap(hOrig,                                 // GrHandle hSource
                 stack[2].intValue, stack[3].intValue,  // x, y
                 stack[4].intValue, stack[5].intValue,  // width, height
                 stack[6].intValue, stack[7].intValue,  // dstX, dstY
                 WOBJ_GraphicsDrawOp(surfDest),         // drawOp
                 currentBackPixel, currentForePixel,    // backPixel, forePixel
                 1);                                    // Sint32 doClip
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void copyScreen(waba.fx.ISurface src, int ySrc, int yDst, int height)
void GraphicsCopyScreen(Var stack[]) // (Lwaba/fx/ISurface;III)V
{
   WObject surfSource = stack[1].obj;
   WObject surfTarget = stack[0].obj;
   GrHandle hSource = WOBJ_ControlImageGraphicsGrHandle(surfSource);
   GrHandle hTarget = WOBJ_ControlImageGraphicsGrHandle(surfTarget);

   _LOG_INFO((LOG_CAT_GFX, "GraphicsCopyScreen"));
   if (hSource && hTarget)
   {
      Sint32 width = GRHANDLE_Width(hSource);
      if (width == GRHANDLE_Width(hTarget))  // do a basic check
      {
         Sint32 ySource = stack[2].intValue;
         Sint32 yTarget = stack[3].intValue;
         Sint32 height  = stack[4].intValue;
         SDL_Rect sourceRect = { 0, ySource, width, height };
         SDL_Rect targetRect = { 0, yTarget, width, height };
         SDL_Surface * targetSurface = (SDL_Surface *)GRHANDLE_WinHandle(hTarget);
         SDL_Surface * sourceSurface = (SDL_Surface *)GRHANDLE_WinHandle(hSource);
         SDL_BlitSurface(sourceSurface, &sourceRect, targetSurface, &targetRect);
         updateRectangle(targetSurface, 0, yTarget, width, height);
      }
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void drawImage(waba.fx.Image image, int x, int y, int drawOp, waba.fx.Color backColor, boolean doClip)
void GraphicsDrawImageSprite(Var stack[]) // (Lwaba/fx/Image;IIILwaba/fx/Color;Z)V
{
   WObject surfOrig = stack[1].obj;
   WObject surfDest = stack[0].obj;
   GrHandle hOrig = surfOrig? WOBJ_ControlImageGraphicsGrHandle(surfOrig) : 0;
   GrHandle hDest = surfDest? WOBJ_ControlImageGraphicsGrHandle(surfDest) : 0;

   _LOG_INFO((LOG_CAT_GFX, "GraphicsDrawImageSprite"));
   if (hOrig && hDest)
   {
      Pixel transpPixel = (Pixel)(WOBJ_ImageTranspPixel(surfOrig));
      Sint32 drawOp = stack[4].intValue;
      WObject transpColor = stack[5].obj;
      _LOG_DEBUG((LOG_CAT_GFX, "orig transp=%04x", transpPixel));
      selectGraphicsObj(surfDest);
      // if we had changed the background color, convert the color to the real pixel value
      if ((drawOp != 0) && (transpColor) && (transpColor != lastColor))
      {
         lastColor = transpColor;
         transpPixel = lastUsedPixel = getColor(WOBJ_ColorRedP(transpColor), WOBJ_ColorGreenP(transpColor), WOBJ_ColorBlueP(transpColor));
      }
      drawBitmap(hOrig,                                                 // hSource
                 0, 0,                                                  // x, y
                 WOBJ_ImageWidth(surfOrig), WOBJ_ImageHeight(surfOrig), // width, height
                 stack[2].intValue, stack[3].intValue,                  // dstX, dstY
                 drawOp,                                                // int drawOp
                 transpPixel, currentForePixel,                         // backPixel, forePixel
                 stack[6].intValue);                                    // doClip
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void copyImageRect(waba.fx.Image image, int x, int y, int width, int height, int drawOp, waba.fx.Color backColor, boolean doClip)
void GraphicsCopyImageRect(Var stack[]) // (Lwaba/fx/Image;IIIIILwaba/fx/Color;Z)V
{
   WObject surfOrig = stack[1].obj;
   WObject surfDest = stack[0].obj;
   GrHandle hOrig = surfOrig ? WOBJ_ControlImageGraphicsGrHandle(surfOrig) : 0;
   GrHandle hDest = surfDest ? WOBJ_ControlImageGraphicsGrHandle(surfDest) : 0;

   _LOG_INFO((LOG_CAT_GFX, "GraphicsCopyImageRect"));
   if (hOrig && hDest)
   {
      Pixel transpPixel = (Pixel)(WOBJ_ImageTranspPixel(surfOrig));
      Sint32 drawOp = stack[6].intValue;
      WObject transpColor = stack[7].obj;
      _LOG_DEBUG((LOG_CAT_GFX, "orig transp=%04x", transpPixel));
      selectGraphicsObj(surfDest);
      if ((drawOp != 0) && (transpColor) && (transpColor != lastColor))
      {
         lastColor = transpColor;
         transpPixel = lastUsedPixel = getColor(WOBJ_ColorRedP(transpColor), WOBJ_ColorGreenP(transpColor), WOBJ_ColorBlueP(transpColor));
      }

      drawBitmap(hOrig,                                 // hSource
                 stack[2].intValue, stack[3].intValue,  // x, y
                 stack[4].intValue, stack[5].intValue,  // width, height
                 0, 0,                                  // dstX, dstY
                 drawOp,                                // drawOp
                 transpPixel, currentForePixel,         // backPixel, forePixel
                 stack[8].intValue);                    // doClip
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void graphicsSetForeColorC(Graphics obj, waba.fx.Color c)
void GraphicsSetForeColor(Var stack[]) // (Lwaba/fx/Graphics;Lwaba/fx/Color;)V
{
   Var *c = stack[1].obj; // class object
   _LOG_INFO((LOG_CAT_GFX, "setForeC=(%02x,%02x,%02x)", WOBJ_ColorRedP(c), WOBJ_ColorGreenP(c), WOBJ_ColorBlueP(c)));
   selectGraphicsObj(stack[0].obj);
   currentForePixel = getColor(WOBJ_ColorRedP(c), WOBJ_ColorGreenP(c), WOBJ_ColorBlueP(c));
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void graphicsSetBackColorC(Graphics obj, waba.fx.Color c)
void GraphicsSetBackColor(Var stack[]) // (Lwaba/fx/Graphics;Lwaba/fx/Color;)V
{
   Var *c = stack[1].obj; // class object
   _LOG_INFO((LOG_CAT_GFX, "setBackC=(%02x,%02x,%02x)", WOBJ_ColorRedP(c), WOBJ_ColorGreenP(c), WOBJ_ColorBlueP(c)));
   selectGraphicsObj(stack[0].obj);
   currentBackPixel = getColor(WOBJ_ColorRedP(c), WOBJ_ColorGreenP(c), WOBJ_ColorBlueP(c));
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics void setFont(waba.fx.Font font)
void GraphicsSetFont(Var stack[]) // (Lwaba/fx/Font;)V
{
   selectGraphicsObj(stack[0].obj);
   currentUserFont = WOBJ_FontUserFont(stack[1].obj);
   WOBJ_GraphicsLastFont(stack[0].obj) = stack[1].obj;
   WOBJ_GraphicsFont(stack[0].obj) = currentUserFont;
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics boolean collisionDetect(waba.fx.ISurface surface, int x1, int y1, waba.fx.Image img, int x2, int y2)
void GraphicsCollisionDetect(Var stack[]) // (Lwaba/fx/ISurface;IILwaba/fx/Image;II)Z
{
   WObject surface;
   GrHandle h1,h2;
   RectangleType r1,r2;
   Sint32 b=0;

   surface = stack[1].obj;
   selectGraphicsObj(stack[0].obj);
   r1.topLeft.x = stack[2].intValue;
   r1.topLeft.y = stack[3].intValue;
   r2.topLeft.x = stack[5].intValue;
   r2.topLeft.y = stack[6].intValue;

   if (!surface)
      h1 = mainGrHandle;
   else
   {
      h1 = WOBJ_ControlImageGraphicsGrHandle(surface);
      if (surfaceGetType(surface) == SURF_IMAGE)
      {
         r1.extent.x = WOBJ_ImageWidth(surface);
         r1.extent.y = WOBJ_ImageHeight(surface);
      }
      else
      {
         r1.extent.x = GRHANDLE_Width(h1)  - r1.topLeft.x; // guich@200b2: sub the width
         r1.extent.y = GRHANDLE_Height(h1) - r1.topLeft.y; // guich@200b2: sub the height
      }
   }
   surface = stack[4].obj;
   h2 = WOBJ_ControlImageGraphicsGrHandle(surface);
   r2.extent.x = WOBJ_ImageWidth(surface);
   r2.extent.y = WOBJ_ImageHeight(surface);

   b = collisionDetect(h1, r1, h2, r2, currentBackPixel);
   RETURN32(b)
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Graphics waba.fx.Coord getTranslation()
void GraphicsGetTranslation(Var stack[]) // ()Lwaba/fx/Coord;
{
   WObject coord;
   if (!coordClass) coordClass = vmGlobals->getClass(vmGlobals->createUtfString("waba/fx/Coord"),1);
   coord = vmGlobals->createObject(coordClass);
   if (coord)
   {
      selectGraphicsObj(stack[0].obj);
      WOBJ_CoordX(coord) = TRANSX;
      WOBJ_CoordY(coord) = TRANSY;
      RETURN32(coord)
   }
   RETURN32(0)
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/Palette int []getCurrentPalette()
void PaletteGetCurrentPalette(Var stack[]) // ()[I
{
   WObject intArray=0;
   SDL_Palette * palette = SDL_GetVideoSurface()->format->palette;
   Sint32 nColors;
   SDL_Color * pColor;

   if (!palette) // use the default websafe palette
   {
      if (!webpalInitialized) setWebSafePalette(0);
      nColors = 256;
      pColor = webpalette;
   }
   else
   {
      nColors = palette->ncolors;
      pColor = palette->colors;
   }
   // move to java's palette
   intArray = (*vmGlobals->createArrayObject)(10, nColors);
   if (intArray)
   {
      Sint32 * ints = (Sint32 *)WOBJ_arrayStart(intArray);
      while (nColors--)
      {
         *ints++ = ((((Uint32)pColor->r) << 16) |  // was 24 and 16. changed to 16 and 8
                     (((Uint32)pColor->g) << 8) |
                     (((Uint32)pColor->b)));
         ++pColor;
      }
   }
   RETURN32(intArray)
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void graphicsDrawHatchedRect(Graphics obj, int x, int y, int width, int height, boolean top, boolean bottom)
void GraphicsDrawHatchedRect(Var stack[]) // (Lwaba/fx/Graphics;IIIIZZ)V
{
   Sint32 x = stack[1].intValue;
   Sint32 y = stack[2].intValue;
   Sint32 w = stack[3].intValue;
   Sint32 h = stack[4].intValue;
   Uint8 top = (Uint8)stack[5].intValue;
   Uint8 bottom = (Uint8)stack[6].intValue;
   Pixel c = currentForePixel;
   Sint32 x2 = x+w-1;
   Sint32 y2 = y+h-1;

   selectGraphicsObj(stack[0].obj);

   if (top && bottom)
   {
      drawLine(x, y+2, x, y2-2, c); // left
      drawLine(x+2, y2, x2-2, y2, c); // bottom
      drawLine(x2, y+2, x2, y2-2, c); // right
      drawLine(x+2, y, x2-2, y, c); // top
      setPixel(x+1, y+1, c); // top left
      setPixel(x2-1, y+1, c); // top right
      setPixel(x+1, y2-1, c); // bottom left
      setPixel(x2-1, y2-1, c); // bottom right
   }
   else
   if (top && !bottom)
   {
      drawLine(x, y+2, x, y2, c); // left
      drawLine(x, y2, x2, y2, c); // bottom
      drawLine(x2, y+2, x2, y2, c); // right
      drawLine(x+2, y, x2-2, y, c); // top
      setPixel(x+1, y+1, c); // top left
      setPixel(x2-1, y+1, c); // top right
   }
   else
   if (!top && bottom)
   {
      drawLine(x, y, x, y2-2, c); // left
      drawLine(x+2, y2, x2-2, y2, c); // bottom
      drawLine(x2, y, x2, y2-2, c); // right
      drawLine(x, y, x2, y, c); // top
      setPixel(x+1, y2-1, c); // bottom left
      setPixel(x2-1, y2-1, c); // bottom right
   }
}
//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void graphicsFillHatchedRect(Graphics obj, int x, int y, int width, int height, boolean top, boolean bottom)
void GraphicsFillHatchedRect(Var stack[]) // (Lwaba/fx/Graphics;IIIIZZ)V
{
   Sint32 x = stack[1].intValue;
   Sint32 y = stack[2].intValue;
   Sint32 w = stack[3].intValue;
   Sint32 h = stack[4].intValue;
   Uint8 top = (Uint8)stack[5].intValue;
   Uint8 bottom = (Uint8)stack[6].intValue;
   Pixel c = currentBackPixel;
   Sint32 x2 = x+w-1;
   Sint32 y2 = y+h-1;

   selectGraphicsObj(stack[0].obj);
   if (top && bottom)
   {
      fillRect(x, y+2, w, h-4, c);  // middle
      drawLine(x+2,y, x2-2, y, c);           // 1st line
      drawLine(x+1,y+1, x2-1, y+1, c);       // 2nd line
      drawLine(x+1, y2-1, x2-1, y2-1, c);    // last-1 line
      drawLine(x+2, y2, x2-2, y2, c);        // last line
   }
   else
   if (top && !bottom)
   {
      drawLine(x+2, y, x2-2, y, c);           // 1st line
      drawLine(x+1, y+1, x2-1, y+1, c);       // 2nd line
      fillRect(x, y+2, w, h-2, c);   // middle
   }
   else
   if (!top && bottom)
   {
      fillRect(x, y, w, h-2, c);     // middle
      drawLine(x+1,y2-1, x2-1, y2-1, c);      // last-1 line
      drawLine(x+2, y2, x2-2, y2, c);         // last line
   }
}

//////////////////////////////////////////////////////////////////////////
// waba/fx/NativeMethods void graphicsDrawHighLightFrame(Graphics obj, int x, int y, int w, int h, waba.fx.Color topLeft, waba.fx.Color bottomRight, boolean yMirror)
void GraphicsDrawHighLightFrame(Var stack[]) // (Lwaba/fx/Graphics;IIIILwaba/fx/Color;Lwaba/fx/Color;Z)V
{
   Sint32 x = stack[1].intValue;
   Sint32 y = stack[2].intValue;
   Sint32 w = stack[3].intValue;
   Sint32 h = stack[4].intValue;
   WObject topLeft = stack[5].obj;
   WObject bottomRight = stack[6].obj;
   Sint32 yMirror = stack[7].intValue;
   Sint32 x2 = x+w-1;
   Sint32 y2 = y+h-1;

   selectGraphicsObj(stack[0].obj);
   if (topLeft)
   {
      Pixel c = getColor(WOBJ_ColorRedP(topLeft), WOBJ_ColorGreenP(topLeft), WOBJ_ColorBlueP(topLeft));
      drawLine(x, y, x, y2, c);
      if (!yMirror)
         drawLine(x, y, x2, y, c);
      else
         drawLine(x, y2, x2, y2, c);
   }
   if (bottomRight)
   {
      Pixel c = getColor(WOBJ_ColorRedP(bottomRight), WOBJ_ColorGreenP(bottomRight), WOBJ_ColorBlueP(bottomRight));
      drawLine(x2, y2, x2, y, c);
      if (!yMirror)
         drawLine(x2, y2, x, y2, c);
      else
         drawLine(x2, y, x, y, c);
   }
}

/*===========================================================================*/
