/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>, All Rights Reserved    *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
#include <PalmOS.h>						// Standard Palm stuff
#include <PalmCompatibility.h>		// use new headers
#include <SysEvtMgr.h>
#include "resources.h"

#define MIN_VERSION              0x02000000 // palm 2.0
#define NEW_GLOBALS_AND_UI_APP   (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)

static Err RomVersionCompatible(DWord requiredVersion, Word launchFlags)
{
   UInt32 romVersion;

   // See if we're on in minimum required version of the ROM or later.
   FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
   if (romVersion < requiredVersion)
   {
      if ((launchFlags & NEW_GLOBALS_AND_UI_APP) == NEW_GLOBALS_AND_UI_APP)
      {
         // PALMOS 1.0 will continuously relaunch this app unless we switch to another safe one.
         if (romVersion < MIN_VERSION)
         {
            AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
         }
      }
      return sysErrRomIncompatible;
   }
   return 0;
}

Ptr userCmdLine;

// ATTENTION: do not modify this string!
char launchString[] =
// guich@200: expanded from 80 to 200 chars. codewarrior puts strange chars in string lengths > 127, so we need to be careful
"L2345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890";
//        1         2         3         4         5         6         7         8         9         10        11        12        13        14        15        16        17        18        19        20

static Err launchWabaApp()
{
   Err err;
   UInt cardNo;
   LocalID dbID;
   DmSearchStateType searchState;
   int len1,len0;
   CharPtr s;

   // guich@200b2: corrected the param line problem.

   // guich@200 compute the total commandline length
   len0 = StrLen(launchString);
   len1 = userCmdLine ? StrLen(userCmdLine) + 1 : 0; // plus 1 for the space

   s = MemPtrNew(len0+len1 + 1);
   StrCopy(s, launchString);
   if (len1)
   {
      s[len0] = ' ';
      StrCopy(s+len0+1,userCmdLine);
   }
   s[len0+len1] = 0;
   MemPtrSetOwner(s, 0);

   // now launch the vm
   err = DmGetNextDatabaseByTypeCreator(true, &searchState, 'appl', 'SWAB', true, &cardNo, &dbID);
   if (err) return err;
   return SysUIAppSwitch(cardNo, dbID, 40404, s);
}

static Boolean FormHandleEvent(EventPtr eventP)
{
   switch (eventP->eType)
   {
      case frmUpdateEvent:
      case frmOpenEvent:
      {
         FrmDrawForm(FrmGetActiveForm());
         return true;
      }
      default:
   }
   return false;
}

UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
   EventType event;
   int formId;
   FormPtr formPtr;
   Err err;

   if (cmd != sysAppLaunchCmdNormalLaunch)
   {
      return 0;
   }

   err = RomVersionCompatible (MIN_VERSION, launchFlags);
   if (err) return (err);

   userCmdLine = cmdPBP;

   err = launchWabaApp(); // this call should not return, if it does, there is a VM launch error
   if (!err) return err;

   FrmGotoForm(idMissingForm);

   for (;;)
   {
      EvtGetEvent(&event, 100);
      if (SysHandleEvent(&event))
      {
         continue;
      }
      else if (MenuHandleEvent((void *)0, &event, &err))
      {
         continue;
      }
      else
      {
         switch (event.eType)
         {
            case ctlSelectEvent:
               FrmDispatchEvent(&event);
               if (event.data.ctlSelect.controlID != idMissingCloseButton) break;
            case appStopEvent:
               return 0;
            case frmLoadEvent:
               formId = event.data.frmLoad.formID;
               formPtr = FrmInitForm(formId);
               FrmSetActiveForm(formPtr);
               FrmSetEventHandler(formPtr, FormHandleEvent);
               break;
            default:
               FrmDispatchEvent(&event);
               break;
         }
      }
   }

   return 0;
}
