// POOMAppointment.cpp: implementation of the POOMAppointment class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "POOMAppointment.h"

#include "PimalLongHelper.h"
#include "PimalDateHelper.h"
#include "PimalBoolHelper.h"

#include "POOMRecurrence.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// extractViewData
// extracts only certain fields from this contact and puts it into a String formatted like this:
// [ID on device]+DATASET_SEP
// ...
//////////////////////////////////////////////////////////////////////

CString POOMAppointment::extractViewData()
{
	CString retVal("");
	// get the information needed out of the actual item and store it into temp variables
	long id;

	appointment->get_Oid(&id);
	PimalLongHelper idH(id);

	// attach a new line with the data accomplished to the return value
	retVal = idH.getCString() + DATASET_SEP;

	return retVal;
}

//////////////////////////////////////////////////////////////////////
// extractViewData
// extracts only certain fields from this contact and puts it into a String formatted like this:
// [ID on device]|~
// [ID on device]|~
// ...
//////////////////////////////////////////////////////////////////////

CString POOMAppointment::extractAllData()
{
	CString retVal("");
	// get the information needed out of the actual item and store it into temp variables
	LPTSTR subject, location, categories, reminderSoundFile, note;
	DATE startDate, endDate;
	long id, duration, meetingStatus, sensitivity, busyStatus, reminderOptions, reminderMinutesBeforeStart;
	VARIANT_BOOL allDayEvent, isRecurring, reminderSet;
	IRecipients *recipients;
	IRecurrencePattern *recurrence;

	appointment->get_Oid(&id);
	PimalLongHelper idH(id);
	appointment->get_Subject(&subject);
	appointment->get_Location(&location);
	appointment->get_Categories(&categories);
	appointment->get_ReminderSoundFile(&reminderSoundFile);
	appointment->get_Body(&note);
	appointment->get_Start(&startDate);
	PimalDateHelper startDateH(startDate);
	appointment->get_End(&endDate);
	PimalDateHelper endDateH(endDate);
	appointment->get_Duration(&duration);
	PimalLongHelper durationH(duration);
	appointment->get_MeetingStatus(&meetingStatus);
	PimalLongHelper meetingStatusH(meetingStatus);
	appointment->get_Sensitivity(&sensitivity);
	PimalLongHelper sensitivityH(sensitivity);
	appointment->get_BusyStatus(&busyStatus);
	PimalLongHelper busyStatusH(busyStatus);
	appointment->get_ReminderOptions(&reminderOptions);
	PimalLongHelper reminderOptionsH(reminderOptions);
	appointment->get_ReminderMinutesBeforeStart(&reminderMinutesBeforeStart);
	PimalLongHelper reminderMinutesBeforeStartH(reminderMinutesBeforeStart);
	appointment->get_AllDayEvent(&allDayEvent);
	PimalBoolHelper allDayEventH(allDayEvent);
	appointment->get_IsRecurring(&isRecurring);
	appointment->get_ReminderSet(&reminderSet);
	PimalBoolHelper reminderSetH(reminderSet);
	appointment->get_Recipients(&recipients);
	appointment->GetRecurrencePattern(&recurrence);
	POOMRecurrence recurrenceH(isRecurring, recurrence);

	// attach a new line with the data accomplished to the return value
	retVal = idH.getCString() + D1_SEP + (CString)subject + D1_SEP + (CString)location + D1_SEP + (CString)categories + D1_SEP + (CString)reminderSoundFile + D1_SEP + (CString)note + D1_SEP + startDateH.getGenericDateString() + D1_SEP + endDateH.getGenericDateString() + D1_SEP + durationH.getCString() + D1_SEP + meetingStatusH.getCString() + D1_SEP + sensitivityH.getCString() + D1_SEP + busyStatusH.getCString() + D1_SEP + reminderOptionsH.getCString() + D1_SEP + reminderMinutesBeforeStartH.getCString() + D1_SEP + allDayEventH.getCString() + D1_SEP + recurrenceH.getGenericRecurrenceString() + D1_SEP + reminderSetH.getCString() + D1_SEP + extractRecipients(recipients) + DATASET_SEP;

	return retVal;
}

//////////////////////////////////////////////////////////////////////
// extractRecipients(IRecipients *recipients) convert the Recipients into a CString
//////////////////////////////////////////////////////////////////////

CString POOMAppointment::extractRecipients(IRecipients *recipients)
{
	int count, i;
	recipients->get_Count(&count);
	IRecipient *recipient;
	CString retVal("");

	for (i=1; i <= count; i++)
	{
		recipients->Item(i, &recipient);
		BSTR nameBSTR, emailBSTR;
		recipient->get_Name(&nameBSTR);
		recipient->get_Address(&emailBSTR);
		CString name(nameBSTR), email(emailBSTR);

		retVal = retVal + name + D3_SEP + email + D2_SEP;
	}

	return retVal;
}

//////////////////////////////////////////////////////////////////////
// createRecipients(BSTR recipients) convert Recipients from BSTR to IRecipients*
//////////////////////////////////////////////////////////////////////
void POOMAppointment::rebuildRecipients(BSTR reclist)
{
	IRecipients *recipients;
	IRecipient *recipient;
	CString list(reclist);

	appointment->get_Recipients(&recipients);
	
	// delete all the existing recipients first
	int recipientCount;
	recipients->get_Count (&recipientCount);
	while (recipientCount > 0)
	{
		recipients->Remove(1);
		recipients->get_Count (&recipientCount);
	}

	// now that all recipients are deleted we are setting the new ones
	
	//first we divide the list in single recipients
	CString singleRecipient("");

	for (int i=0; i<list.GetLength(); i++)
	{
		if ( i+1 == list.GetLength() || (list.GetAt(i) == D1_SEP1 && list.GetAt(i+1) == D1_SEP2 && list.GetAt(i+2) == D1_SEP3))
		{
			//we have found one complete single recipient
			
			CString tmp("");
			for (int j=0; j<singleRecipient.GetLength(); j++)
			{
				if (singleRecipient.GetAt(j) == D2_SEP1 && singleRecipient.GetAt(j+1) == D2_SEP2 && singleRecipient.GetAt(j+2) == D2_SEP3)
				{
					// we have completet parsing the Name
					recipients->Add(tmp.AllocSysString(), &recipient);
					tmp = "";
					j = j+2;
				}
				else 
				{
					tmp += singleRecipient.GetAt(j);
				}
			}
			recipient->put_Address(tmp.AllocSysString());

			// clearing the variable
			singleRecipient = "";
			// incrementing iso that ist points to new recipient
			i = i+2;
		}
		else
		{
			singleRecipient += list.GetAt(i);
		}
	}

	recipients->Release();
	recipient->Release();
}

//////////////////////////////////////////////////////////////////////
// save() saves the stored IAppointment (all changes made are stored on the device
// and refreshes the oid
//////////////////////////////////////////////////////////////////////

BOOL POOMAppointment::save()
{
	HRESULT h = appointment->Save();
	long id;
	CString strId;
	appointment->get_Oid(&id);
	strId.Format(_T("%ld"), id);

	oid = strId.AllocSysString();
   return h == 0;
}

//////////////////////////////////////////////////////////////////////
// getOid() returns the Oid of the IAppointment as a BSTR
//////////////////////////////////////////////////////////////////////

BSTR POOMAppointment::getOid()
{
	return oid;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

POOMAppointment::POOMAppointment(IAppointment* tmp)
{
	appointment = tmp;
}

POOMAppointment::POOMAppointment(IAppointment* tmp, BSTR subject)
{
	appointment = tmp;
   appointment->put_Subject(subject);
}

POOMAppointment::POOMAppointment(IAppointment* tmp, BSTR subject, BSTR location, BSTR categories, BSTR reminderSoundFile, BSTR note, BSTR startDate, BSTR endDate, BSTR duration, BSTR meetingStatus, BSTR sensitivity, BSTR busyStatus, BSTR reminderOptions, BSTR reminderMinutesBeforeStart, BSTR allDayEvent, BSTR isRecurring, BSTR reminderSet, BSTR recurrenceType, BSTR occurrences, BSTR interval, BSTR dayOfWeek, BSTR dayOfMonth, BSTR weekOfMonth, BSTR monthOfYear, BSTR patternStartDate, BSTR patternEndDate, BSTR startTime, BSTR endTime, BSTR noEndDate, BSTR recipients)
{
	appointment = tmp;
	
	if (subject && subject != UNCHANGED) 
      appointment->put_Subject(subject);
	if (location && location != UNCHANGED) 
      appointment->put_Location(location);
	if (categories && categories != UNCHANGED) 
      appointment->put_Categories(categories);
	if (reminderSoundFile && reminderSoundFile != UNCHANGED) 
      appointment->put_ReminderSoundFile(reminderSoundFile);
	if (note && note != UNCHANGED) appointment->put_Body(note);
	if ((startDate || endDate) && startDate != UNCHANGED) 
	{
		PimalDateHelper startDateH(startDate);
		appointment->put_Start(startDateH.getDATE());
	}
	if ((startDate || endDate) && endDate != UNCHANGED) 
	{
		PimalDateHelper endDateH(endDate);
		appointment->put_End(endDateH.getDATE());
	}
	if (duration && duration != UNCHANGED)
	{
		PimalLongHelper durationH(duration);
		appointment->put_Duration(durationH.getLong());
	}
	if (sensitivity && sensitivity != UNCHANGED)
	{
		PimalLongHelper sensitivityH(sensitivity);
		appointment->put_Sensitivity(sensitivityH.getLong());
	}
	if (busyStatus && busyStatus != UNCHANGED)
	{
		PimalLongHelper busyStatusH(busyStatus);
		appointment->put_BusyStatus(busyStatusH.getLong());
	}
	if (reminderOptions && reminderOptions != UNCHANGED)
	{
		PimalLongHelper reminderOptionsH(reminderOptions);
		appointment->put_ReminderOptions(reminderOptionsH.getLong());
	}
	if (reminderMinutesBeforeStart && reminderMinutesBeforeStart != UNCHANGED)
	{
		PimalLongHelper reminderMinutesBeforeStartH(reminderMinutesBeforeStart);
		appointment->put_ReminderMinutesBeforeStart(reminderMinutesBeforeStartH.getLong());
	}
	if (allDayEvent && allDayEvent != UNCHANGED)
	{
		PimalBoolHelper allDayEventH(allDayEvent);
		appointment->put_AllDayEvent(allDayEventH.getVARIANT_BOOL());
	}
	if (reminderSet && reminderSet != UNCHANGED)
	{
		PimalBoolHelper reminderSetH(reminderSet);
		appointment->put_ReminderSet(reminderSetH.getVARIANT_BOOL());
	}
	if (recipients && recipients != UNCHANGED)
	{
		rebuildRecipients(recipients);
	}
	if (isRecurring && isRecurring != UNCHANGED)
	{
		PimalBoolHelper isRecurringH(isRecurring);
		if (isRecurringH.getVARIANT_BOOL())
		{
			IRecurrencePattern* recurrencePattern;
			appointment->GetRecurrencePattern(&recurrencePattern);
			POOMRecurrence recurrenceH(recurrencePattern, isRecurring, recurrenceType, occurrences, duration, interval, dayOfWeek, dayOfMonth, weekOfMonth, monthOfYear, patternStartDate, patternEndDate, startTime, endTime, noEndDate);
		}
		else
		{
			appointment->ClearRecurrencePattern();
		}
		
	}
}
