// POOMRecurrence.cpp: implementation of the POOMRecurrence class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "POOMRecurrence.h"

#include "PimalDateHelper.h"
#include "PimalLongHelper.h"
#include "PimalBoolHelper.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// getGenericRecurrenceString() returns the recurrence as String
//////////////////////////////////////////////////////////////////////

CString POOMRecurrence::getGenericRecurrenceString()
{
	return genericRecurrenceString;
}

//////////////////////////////////////////////////////////////////////
// getIRecurrencePattern() returns the recurrence as IRecurrencePattern*
//////////////////////////////////////////////////////////////////////
IRecurrencePattern* POOMRecurrence::getIRecurrencePattern()
{
	return recurrence;
}

//////////////////////////////////////////////////////////////////////
// getDayOfWeekAsString(long day) : returns the name of the day of week that fits to this
//////////////////////////////////////////////////////////////////////
CString POOMRecurrence::getDayOfWeekAsString(long day)
{
	CString retVal;

	if		(day == 1) retVal = "sunday";
	else if (day == 2) retVal = "monday";
	else if (day == 4) retVal = "tuesday";
	else if (day == 8) retVal = "wednesday";
	else if (day == 16) retVal = "thursday";
	else if (day == 32) retVal = "friday";
	else if (day == 64) retVal = "saturday";
	else				retVal = "unknownDay";

	return retVal;
}

//////////////////////////////////////////////////////////////////////
// getDayOfWeekAsLong(BSTR day) : returns the long of the day of week that fits to this
//////////////////////////////////////////////////////////////////////
long POOMRecurrence::getDayOfWeekAsLong(BSTR day)
{
	long retVal;
	

	if		(day == _T("sunday")) retVal = 1;
	else if (day == _T("monday")) retVal = 2;
	else if (day == _T("tuesday")) retVal = 4;
	else if (day == _T("wednesday")) retVal = 8;
	else if (day == _T("thursday")) retVal = 16;
	else if (day == _T("friday")) retVal = 32;
	else if (day == _T("saturday")) retVal = 64;
	else	retVal = 0;

	return retVal;
}

//////////////////////////////////////////////////////////////////////
// getRecurrenceTypeAsString(long recType) : returns the RecurrenceType as String
//////////////////////////////////////////////////////////////////////
CString POOMRecurrence::getRecurrenceTypeAsString(long recType)
{
	CString retVal;
	if		(recType == 0) retVal = "daily";
	else if (recType == 1) retVal = "weekly";
	else if (recType == 2) retVal = "monthly";
	else if (recType == 3) retVal = "monthlyNth";
	else if (recType == 5) retVal = "yearly";
	else if (recType == 6) retVal = "yearlyNth";
	else				   retVal = "unknownRecurrence";

	return retVal;
}

//////////////////////////////////////////////////////////////////////
// getRecurrenceTypeAsLong(BSTR recType) : returns the RecurrenceType as long
//////////////////////////////////////////////////////////////////////
long POOMRecurrence::getRecurrenceTypeAsLong(BSTR recType)
{
	long retVal;

	if		(recType == _T("daily")) retVal = 0;
	else if (recType == _T("weekly")) retVal = 1;
	else if (recType == _T("monthly")) retVal = 2;
	else if (recType == _T("monthlyNth")) retVal = 3;
	else if (recType == _T("yearly")) retVal = 5;
	else if (recType == _T("yearlyNth")) retVal = 6;
	else	retVal = 0;

	return retVal;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
POOMRecurrence::POOMRecurrence(IRecurrencePattern *retVal, BSTR isRecurrent, BSTR recurrenceType, BSTR occurrences, BSTR duration, BSTR interval, BSTR dayOfWeek, BSTR dayOfMonth, BSTR weekOfMonth, BSTR monthOfYear, BSTR patternStartDate, BSTR patternEndDate, BSTR startTime, BSTR endTime, BSTR noEndDate)
{
	recurrence = retVal;

	PimalBoolHelper isRecurrentH(isRecurrent);
	long recurrenceTypeH = getRecurrenceTypeAsLong(recurrenceType);
	PimalLongHelper occurrencesH(occurrences);
	PimalLongHelper durationH(duration);
	PimalLongHelper intervalH(interval);
	long dayOfWeekH = getDayOfWeekAsLong(dayOfWeek);
	PimalLongHelper dayOfMonthH(dayOfMonth);
	PimalLongHelper weekOfMonthH(weekOfMonth);
	PimalLongHelper monthOfYearH(monthOfYear);

	PimalDateHelper patternStartDateH(patternStartDate);
	PimalDateHelper patternEndDateH(patternEndDate);
	PimalDateHelper startTimeH(startTime);
	PimalDateHelper endTimeH(endTime);

	PimalBoolHelper noEndDateH(noEndDate);

	if (!isRecurrentH.getVARIANT_BOOL()) ;
	else
	{
		recurrence->put_RecurrenceType(recurrenceTypeH);

		if (recurrenceTypeH == 0)
		{
			recurrence->put_Interval(intervalH.getLong());
		}
		else if (recurrenceTypeH == 1)
		{
			recurrence->put_DayOfWeekMask(dayOfWeekH);
		}
		else if (recurrenceTypeH == 2)
		{
			recurrence->put_Interval(intervalH.getLong());
			recurrence->put_DayOfMonth(dayOfMonthH.getLong());
		}
		else if (recurrenceTypeH == 3)
		{
			recurrence->put_Interval(intervalH.getLong());
			recurrence->put_DayOfWeekMask(dayOfWeekH);
			recurrence->put_Instance(weekOfMonthH.getLong());
		}
		else if (recurrenceTypeH == 5)
		{
			recurrence->put_DayOfMonth(dayOfMonthH.getLong());
			recurrence->put_MonthOfYear(monthOfYearH.getLong());
		}
		else if (recurrenceTypeH == 6)
		{
			recurrence->put_Interval(intervalH.getLong());
			recurrence->put_DayOfWeekMask(dayOfWeekH);
			recurrence->put_Instance(weekOfMonthH.getLong());
			recurrence->put_MonthOfYear(monthOfYearH.getLong());
		}

		recurrence->put_PatternStartDate(patternStartDateH.getDATE());

		recurrence->put_NoEndDate(noEndDateH.getVARIANT_BOOL());
		if (!noEndDateH.getVARIANT_BOOL()) 
		{
			recurrence->put_PatternEndDate(patternEndDateH.getDATE());
			recurrence->put_Occurrences(occurrencesH.getLong());
		}

		recurrence->put_StartTime(startTimeH.getDATE());
		recurrence->put_EndTime(endTimeH.getDATE());
		recurrence->put_Duration(durationH.getLong());
	}
}
			



POOMRecurrence::POOMRecurrence(VARIANT_BOOL isRecurrent, IRecurrencePattern *recurrenceTmp)
{
	recurring = isRecurrent;
	recurrence = recurrenceTmp;

	CString retVal("");

	if (recurring == FALSE) 
	{
		retVal = "noRecurrence";
	}
	else
	{
		long recurrenceType, occurrences, duration, interval, dayOfWeek, dayOfMonth , weekOfMonth, monthOfYear;
		DATE patternStartDate, patternEndDate, startTime, endTime;
		VARIANT_BOOL noEndDate;

		recurrence->get_Interval(&interval);
		PimalLongHelper intervalH(interval);
		recurrence->get_DayOfWeekMask(&dayOfWeek);
		recurrence->get_DayOfMonth(&dayOfMonth);
		PimalLongHelper dayOfMonthH(dayOfMonth);
		recurrence->get_Instance(&weekOfMonth);
		PimalLongHelper weekOfMonthH(weekOfMonth);
		recurrence->get_MonthOfYear(&monthOfYear);
		PimalLongHelper monthOfYearH(monthOfYear);

		recurrence->get_RecurrenceType(&recurrenceType);
		recurrence->get_Occurrences(&occurrences);
		PimalLongHelper occurrencesH(occurrences);
		recurrence->get_Duration(&duration);
		PimalLongHelper durationH(duration);

		recurrence->get_PatternStartDate(&patternStartDate);
		PimalDateHelper patternStartDateH(patternStartDate);
		recurrence->get_PatternEndDate(&patternEndDate);
		PimalDateHelper patternEndDateH(patternEndDate);
		recurrence->get_StartTime(&startTime);
		PimalDateHelper startTimeH(startTime);
		recurrence->get_EndTime(&endTime);
		PimalDateHelper endTimeH(endTime);

		recurrence->get_NoEndDate(&noEndDate);
		
		// recurrence is daily
		if		(recurrenceType == 0)
		{
			CString daily("daily");
			retVal = daily + D3_SEP + intervalH.getCString();
			
		}

		// recurrence is weekly
		else if (recurrenceType == 1)
		{
			CString weekly("weekly");
			retVal = weekly + D3_SEP + getDayOfWeekAsString(dayOfWeek);
		}

		// recurrence is monthly
		else if (recurrenceType == 2)
		{
			CString monthly("monthly");
			retVal = monthly + D3_SEP + intervalH.getCString() + D3_SEP + dayOfMonthH.getCString();
		}

		// recurrence is monthlyNth
		else if (recurrenceType == 3)
		{
			CString monthlyNth("monthlyNth");
			retVal = monthlyNth + D3_SEP + intervalH.getCString() + D3_SEP + getDayOfWeekAsString(dayOfWeek) + D3_SEP + weekOfMonthH.getCString();
		}

		// recurrence is yearly
		else if (recurrenceType == 5)
		{
			CString yearly("yearly");
			retVal = yearly + D3_SEP + dayOfMonthH.getCString() + D3_SEP + monthOfYearH.getCString();
		}

		// recurrence is yearlyNth
		else if (recurrenceType == 6)
		{
			CString yearlyNth("yearlyNth");
			retVal = yearlyNth + D3_SEP + intervalH.getCString() + D3_SEP + getDayOfWeekAsString(dayOfWeek) + D3_SEP + weekOfMonthH.getCString() + D3_SEP + monthOfYearH.getCString();
		}

		else
		{
			CString unknownRecurrence("unknownRecurrence:");
			PimalLongHelper recurrenceTypeH(recurrenceType);
			retVal = unknownRecurrence + recurrenceTypeH.getCString();
		}

		retVal += D2_SEP + patternStartDateH.getGenericDateString() + D3_SEP;
		
		if (noEndDate == TRUE) 
		{
			CString noEnd("noEnd");
			CString oh("0");
			retVal += noEnd + D3_SEP + oh;
		}
		else retVal += patternEndDateH.getGenericDateString() + D3_SEP + occurrencesH.getCString();

		retVal += D3_SEP + startTimeH.getGenericTimeString() + D3_SEP + endTimeH.getGenericTimeString() + D3_SEP + durationH.getCString();



	}

	genericRecurrenceString = retVal;

}

POOMRecurrence::~POOMRecurrence()
{

}
