/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

#include "superwaba.h"
#include "CustomLib.h"
#include "ScanMgrDef.h" // guich@200
#include "ScanMgrStruct.h" // guich@200
#include "ScanMgr.h" // guich@200
//#include "SWLib.c"

#ifdef __SWNATIVE_CUSTOM_LIB_H
You included the wrong CustomLib.h file. Be sure it is NOT the
SWNatives\CustomLib.h. If you are using CodeWarrior, Specify in the "Access path"
the directory where your CustomLib.h file resides, moving it to the top. This must
be done because a project (like SuperWaba) can have multiple CustomLib.h
files in different directories, and as default the compiler gets it from
the first directory found.
#endif

#define TRIGGERING_PARAM             1
#define LINEAR_SECURITY_LEVEL_PARAM  2
#define SUPPLEMENTALS_PARAM          3
#define TRANSMIT_CHECKDIGIT_PARAM    4
#define PREAMBLE_PARAM               5
#define MSI_PLESSEY_CHECKDIGIT_PARAM 7
#define MSI_PLESSEY_OPTIONS_PARAM    8
#define MSI_PLESSEY_ALGORITHMS_PARAM 9
#define TRANSMIT_CODEID_PARAM        10
#define SCAN_DATA_TRANSMISSION_PARAM 12
#define SCAN_ANGLE_PARAM             13

void ScannerActivate(Var stack[]);
void ScannerSetBarcodeParam(Var stack[]);
void ScannerSetParam(Var stack[]);
void ScannerSetBarcodeLength(Var stack[]);
void ScannerCommitBarcodeParams(Var stack[]);
void ScannerGetData(Var stack[]);
void ScannerGetScanManagerVersion(Var stack[]);
void ScannerGetScanPortDriverVersion(Var stack[]);
void ScannerDeactivate(Var stack[]);

uint32 nativeHashes[10] =
{
	1261220242,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerActivate_()Z
	1261207899,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerSetBarcodeParam_(IZ)Z
	1261232981,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerSetParam_(III)Z
	1261223390,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerSetBarcodeLength_(IIII)Z
	1261225117,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerCommitBarcodeParams_()Z
	1261182306,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerGetData_()Ljava/lang/String;
	1261209712,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerGetScanManagerVersion_()Ljava/lang/String;
	1261231603,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerGetScanPortDriverVersion_()Ljava/lang/String;
	1261233108,  // superwaba/ext/xplat/io/scanner/NativeMethods_ScannerDeactivate_()Z
	0  // Must end with 0
};

// build at runtime the nativeFuncs array
NativeFunc nativeFuncs[10];
void buildNativeFuncArray()
{
	nativeFuncs[  0] = ScannerActivate;
	nativeFuncs[  1] = ScannerSetBarcodeParam;
	nativeFuncs[  2] = ScannerSetParam;
	nativeFuncs[  3] = ScannerSetBarcodeLength;
	nativeFuncs[  4] = ScannerCommitBarcodeParams;
	nativeFuncs[  5] = ScannerGetData;
	nativeFuncs[  6] = ScannerGetScanManagerVersion;
	nativeFuncs[  7] = ScannerGetScanPortDriverVersion;
	nativeFuncs[  8] = ScannerDeactivate;
	nativeFuncs[  9] = 0;  // Must end with 0
}

// Called by the Lib when a public method is invoked by the vm or someone else. This defines a interface that lets you add custom methods to interface with the vm without modifying the basic architecture of SW Libs
Err callPublic(int index, uint32 *ret, uint32 *args)
{
   /* Example of implementation for callPublic:
   switch (index)
   {
      case PUBLIC_DummyFunc  : *ret = dummyFunc();      break;
      case PUBLIC_DummyFunc2 : dummyFunc2((int)*args);  break;
   }
   */
   return 0; // nothing here to call
}

/* Example of implementation for getClassHooks:
ClassHook classHooks[] =
{
   {  NULL, PUBLIC_ScannerDestroy, 0 },
      NULL
};
*/

ClassHook *getClassHooks()
{
/* Example (continued)
   classHooks[0].className = "superwaba/ext/xplat/NativeMethods";
   return classHooks; */

   return 0; // nothing here to call
}

// Called by the vm to let the native libraries handle particular events. Must return SWErrEventHandled or SWErrEventNotHandled.
Err handleEvent(EventPtr eventP)
{
   static WClass *scannerClass=0;
   static WClassMethod *onEventMethod=0;
   static WObject scannerObj=0;
   Var params[1];

   switch (eventP->eType)
   {
      case scanDecodeEvent:        // guich@200b2: forgot to add those events!
      case scanTriggerEvent:
      case scanBatteryErrorEvent:
         if (!onEventMethod)
         {
            WClassMethod *constructor;
            UtfString utf = (*vmGlobals->createUtfString)("superwaba/ext/xplat/io/scanner/Scanner");
            ///////////////////////// get the class /////////////////////////////
            scannerClass = (*vmGlobals->getClass)(utf,1);
            if (!scannerClass)
               return SWErrEventNotHandled;

            ///////////////////////// create an instance ////////////////////////
            scannerObj = (*vmGlobals->createObject)(scannerClass);
            if (!scannerObj || (*vmGlobals->pushObject)(scannerObj) == -1) // guich@200b4: is a good idea push the scannerObj? i dont know...
               return SWErrEventNotHandled;

            ///////////////////////// call constructor //////////////////////////
            constructor = (*vmGlobals->getMethod)(scannerClass, (*vmGlobals->createUtfString)("<init>"), (*vmGlobals->createUtfString)("()V"), NULL);
            if (!constructor)
               return SWErrEventNotHandled;
            params[0].obj = scannerObj;

            ///////////////////////// get the method ////////////////////////////
            (*vmGlobals->executeMethod)(scannerClass, constructor, params, 1);
            onEventMethod = (*vmGlobals->getMethod)(scannerClass, (*vmGlobals->createUtfString)("_onEvent"), (*vmGlobals->createUtfString)("(I)V"), NULL);
         }

         if (onEventMethod) // dont make this an else or it won't be called in the first time...
         {
            int32 type = (eventP->eType == scanBatteryErrorEvent) ? 2 : // BATTERY_ERROR
                         (eventP->eType == scanDecodeEvent)       ? 1 : // SCANNED
                                                                    0 ; // TRIGGERED
            //params[0].obj = scannerObj; note: the method Scanner._onEvent is static, so it don't have the obj parameter
            params[0].intValue = type; // type - index 0 since this method is static
            (*vmGlobals->executeMethod)(scannerClass, onEventMethod, params, 1);
            return SWErrEventHandled; // Hey, VM, don't let anyone process this event anymore!
         }
   }
   return SWErrEventNotHandled; // Hey, VM, this event was not handled by me, Continue with normal event processing...
}

int scannerActive;

// called when this lib is about to be closed. The library must free any allocated resources.
void onClose()
{
   if (scannerActive && ScanCmdScanDisable() == STATUS_OK)
      ScanCloseDecoder();
}

//
// Scanner
// Tom Cuthill - Palm Solutions
// 16/3/00
//
// superwaba/ext/xplat/NativeMethods static boolean scannerActivate()
void ScannerActivate(Var stack[]) // ()Z
{
    int32 status;
    if (!ScanIsPalmSymbolUnit())
        RETURN32(0)
    status = ScanOpenDecoder();
    if (status != STATUS_OK)
        RETURN32(0)
    status = ScanCmdScanEnable();
    if (status != STATUS_OK)
        RETURN32(0)
    status = ScanSetTriggeringModes(LEVEL);
    RETURN32(status == STATUS_OK)
}
//////////////////////////////////////////////////////////////////////////
// superwaba/ext/xplat/NativeMethods static boolean scannerSetBarcodeParam(int barcodeType, boolean enable)
void ScannerSetBarcodeParam(Var stack[]) // (IZ)Z
{
    BarType barcodeType = (BarType) stack[0].intValue;
    int32 enabled     = stack[1].intValue;
    int32 status;
    ScanSetBarcodeEnabled(barcodeType, enabled);

    /*
     * If the barcode if of the following types, a length must be set. We will check and
     * set it to decode any length barcodes. Thomas M. Roehr 011113
     *
     * Barcodes needing length parameter:
     *
     *          barCODABAR
     *          barCODE39
     *          barCODE93
     *          barD25
     *          barI2OF5
     *          barMSI_PLESSY
     */

    /* only if turning on the barcode...  trying brute force for the moment */
    status = ScanSetBarcodeLengths(barcodeType, ANY_LENGTH, 0, 0);
    //ScanSetDecodeUpcEanSupplementals
    RETURN32(status == STATUS_OK)
/*
   if(enabled)
   {
      switch(barcodeType)
      {
         barCODABAR:
         barCODE39:
         barCODE93:
         barD25:
         barI2OF5:
         barMSI_PLESSY:
         status = ScanSetBarcodeLengths(barcodeType, ANY_LENGTH, 0, 0);
         v.intValue = status == STATUS_OK;
      }
   }
*/
}
//////////////////////////////////////////////////////////////////////////
// superwaba/ext/xplat/io/scanner/NativeMethods boolean scannerSetParam(int type, int barcodeType, int value)
void ScannerSetParam(Var stack[]) // (III)Z
{
	Err err = 1;
	int32 type = stack[1].intValue;
	BarType barcode = (BarType)stack[2].intValue;
	uint16 value = (uint16)stack[3].intValue;
	switch (type)
	{
      case TRIGGERING_PARAM            :
         err = ScanSetTriggeringModes(value);
         break;
      case LINEAR_SECURITY_LEVEL_PARAM :
         err = ScanSetLinearCodeTypeSecurityLevel(value);
         break;
      case SUPPLEMENTALS_PARAM         :
         err = ScanSetDecodeUpcEanSupplementals(value);
         break;
      case TRANSMIT_CHECKDIGIT_PARAM   :
         err = ScanSetTransmitCheckDigit(barcode, value);
         break;
      case PREAMBLE_PARAM              :
         err = ScanSetUpcPreamble(barcode, value);
         break;
      case MSI_PLESSEY_CHECKDIGIT_PARAM:
      case MSI_PLESSEY_OPTIONS_PARAM   :
         err = ScanSetMsiPlesseyCheckDigits(value);
         break;
      case MSI_PLESSEY_ALGORITHMS_PARAM:
         err = ScanSetMsiPlesseyCheckDigitAlgorithm(value);
         break;
      case TRANSMIT_CODEID_PARAM       :
         err = ScanSetTransmitCodeIdCharacter(value);
         break;
      case SCAN_DATA_TRANSMISSION_PARAM:
         err = ScanSetScanDataTransmissionFormat(value);
         break;
      case SCAN_ANGLE_PARAM            :
         err = ScanSetAngle(value);
         break;
   }
	RETURN32(err == errNone)
}
//////////////////////////////////////////////////////////////////////////
// superwaba/ext/xplat/io/scanner/NativeMethods boolean scannerSetBarcodeLength(int barcodeType, int lengthType, int length1, int length2 )
void ScannerSetBarcodeLength(Var stack[]) // (IIII)Z
{
	BarType type = (BarType)stack[1].intValue;
	uint16 lengthType = stack[2].intValue;
	uint16 length1 = stack[3].intValue;
	uint16 length2 = stack[4].intValue;
	RETURN32(ScanSetBarcodeLengths(type, lengthType, length1, length2) == 0);
}
//////////////////////////////////////////////////////////////////////////
// superwaba/ext/xplat/NativeMethods static boolean scannerCommitBarcodeParams()
void ScannerCommitBarcodeParams(Var stack[]) // ()Z
{
    RETURN32(ScanCmdSendParams(No_Beep) == STATUS_OK)
}
//////////////////////////////////////////////////////////////////////////
// superwaba/ext/xplat/NativeMethods static String scannerGetData()
void ScannerGetData(Var stack[]) // ()Ljava/lang/String;
{
	 MESSAGE msg;
    if (ScanGetDecodedData(&msg) == STATUS_OK)
    {
        msg.data[msg.length] = '\0';
        RETURN32((*vmGlobals->createString)((char *)msg.data))
    }
    else RETURN32((*vmGlobals->createString)(""))
}
//////////////////////////////////////////////////////////////////////////
// superwaba/ext/xplat/NativeMethods static String scannerGetScanManagerVersion()
void ScannerGetScanManagerVersion(Var stack[]) // ()Ljava/lang/String;
{
    char  version[20];
    int32 status = ScanGetScanManagerVersion(version, sizeof(version));

    if (status != STATUS_OK) // ericfrik@200b4_14
       version[0] = '\0';
    RETURN32((*vmGlobals->createString)(version))
}
//////////////////////////////////////////////////////////////////////////
// superwaba/ext/xplat/NativeMethods static String scannerGetScanPortDriverVersion()
void ScannerGetScanPortDriverVersion(Var stack[]) // ()Ljava/lang/String;
{
    char version[20];

    if (ScanGetScanPortDriverVersion(version, sizeof(version)) != STATUS_OK) // ericfrik@200b4_14
        version[0] = '\0';
    RETURN32((*vmGlobals->createString)(version))
}
//////////////////////////////////////////////////////////////////////////
// superwaba/ext/xplat/NativeMethods static boolean scannerDeactivate()
void ScannerDeactivate(Var stack[]) // ()Z
{
    if(ScanCmdScanDisable() != STATUS_OK)
       RETURN32(0)
    scannerActive = 0;
    RETURN32(ScanCloseDecoder() == STATUS_OK)
}
