/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#include "superwaba.h"

#define __DEFINE_FUNCTIONS
#include "SW_String.h"
#include "SW_utils.h"
#include "SW_Class.h"
#include "SW_logging.h"

// from SuperWaba.c:
_inline_ WObject createObject(WClass *wclass);
WObject throwException(char *exceptionClass inA0, char *message inA1);
UtfString getClassName(WObject obj);
WClass * getClass(UtfString className, byte throwEx);
int pushObject(WObject obj);
WObject popObject();
void dumpStackTrace();
byte callMethod(WObject classObj, char *methodName, char *methodDesc, Var *params, int paramCount);
int compatible(WClass *wclass inA0, WClass *target inA1);
void swError(uint16 errNum, WClass *iclass, UtfString *desc1, UtfString *desc2);

// Frank@ StrNCompare doesn't work indeed but StrNCompareAscii is the recommanded strncmp implementation
//#ifdef PALMOS
//#ifdef xstrncmp
//#undef xstrncmp
//#endif
////>>>PGR: no idea why we use this method (previously called isostrncmp)
////        i suspect that guich found a bug ...  just for safety, then:
//static int xstrncmp(uchar *s1, uchar *s2, uint16 n)
//{
//   uint16 i;
//
//   for (i = 0; i < n; i++)
//   {
//      if (s1[i] != s2[i])
//         return s1[i] - s2[i];
//   }
//   return 0;
//}
//#endif

/*----------------------------------------------------------------lockWarpRec-+
|                                                                             |
+----------------------------------------------------------------------------*/
uchar * lockWarpRec(char *path, uint16 pathLen, uint32 *size)
{
   uchar *p;
#if defined(PALMOS)
   VoidHand recHandle;
#else
   uint32 recSize;
#endif
   uint16 top, bot, mid, numRecs;
   uint16 i, nameLen, minLen;
   int cmp;

   _LOG_DEBUG((LOG_CAT_CLASSLOADER, "lockWarpRec(%s)", path));
   for (i = 0; vmGlobals->classDbList[i]; i++)
   {
      // binary search to find class
      numRecs = DmNumRecords(vmGlobals->classDbList[i]);
      if (numRecs == 0)
         continue;
      top = 0;
      bot = numRecs;
      while (1)
      {
         mid = (bot + top) / 2;
#if defined(PALMOS)
         recHandle = DmQueryRecord(vmGlobals->classDbList[i], mid);
         p = (uchar *)MemHandleLock(recHandle);
#else
         recSize = pathLen+2; // read only enough for the name
         p = (uchar *)DmQueryRecord(vmGlobals->classDbList[i], mid, &recSize);
#endif
         nameLen = READ_BE_UINT16(p);
         minLen = (pathLen > nameLen) ? nameLen : pathLen;
         cmp = xstrncmp((uchar *)path, (uchar *)p+2, minLen); // guich: note that palm os compiler issues a warning if using char* instead of uchar*
         #if 0//defined(ENABLE_LOGGING)
         {
            uchar buf[128];
            boundedCopyN(buf, &buf[127], p+2, minLen);
            _LOG_DEBUG((LOG_CAT_CLASSLOADER, "cmp:%s", buf));
         }
         #endif
#if !defined(PALMOS)
         xfree(p);
#endif
         if (!cmp)
         {
            cmp = pathLen - nameLen;
            if (cmp == 0)
            {
#if defined(PALMOS)
               if (size)
               {
                  *size = MemHandleSize(recHandle);
                  *size -= (nameLen + 2);
               }
#else
               // read the full record
               recSize = 0;
               p = (uchar *)DmQueryRecord(vmGlobals->classDbList[i], mid, &recSize);
               if (!p) return 0;
               if (size)
               {
                  *size = recSize - (pathLen+2);
               }
#endif
               _LOG_DEBUG((LOG_CAT_CLASSLOADER, "size = %d", size ? (int)*size : -1));
               return p + pathLen+2;
            }
         }
#if defined(PALMOS)
         MemHandleUnlock(recHandle);
#endif
         if (mid == top)
            break; // not found
         if (cmp < 0)
            bot = mid;
         else
            top = mid;
      }
   }
   return 0;
}

/*--------------------------------------------------------------unlockWarpRec-+
|                                                                             |
+----------------------------------------------------------------------------*/
void unlockWarpRec(uchar *ptr, uint16 pathLen)
{
   // NOTE: The pointer given is "inside" the record. To unlock the record
   // we need to find the original pointer locked which is at the front
   // of the path.
#ifdef PALMOS
   MemPtrUnlock(ptr - pathLen - 2);
#else
   xfree_(ptr - pathLen - 2);
#endif
}

/*------------------------------------------------------------nativeLoadClass-+
|                                                                             |
+----------------------------------------------------------------------------*/
uchar * nativeLoadClass(UtfString className, uint32 *size)
{
   uchar *p;
   char *path;
   uint16 pathLen;

   // make full path by appending ".class"
   pathLen = (uint16)(className.len + 6);
   path = xmalloc(pathLen+1); // guich@350_29: now xmalloc the buffer
   xstrncpy(path, className.str, className.len);
   xstrncpy(&path[className.len], ".class", 6);
   path[pathLen] = 0;
   // NOTE: We unlock the record locked here in stopApp()
   p = lockWarpRec(path, pathLen, size);
   if (p != NULL && READ_BE_UINT32(p) != (uint32)0xCAFEBABE)
   {
      swError(ERR_BadClass, NULL, &className, NULL);
      unlockWarpRec(p, pathLen);
      p = NULL;
   }
   _LOG_INFO((LOG_CAT_CLASSLOADER, "loadClass(%s)=%d", path, (int)(p && size) ? *size : 0));
   xfree(path);
   return p;
}

/*---------------------------------------------------------------releaseClass-+
|                                                                             |
+----------------------------------------------------------------------------*/
void releaseClass(UtfString className, uchar *p)
{
   unlockWarpRec(p, (uint16)(className.len+6));
}

/*------------------------------------------------------------releaseAndClose-+
|                                                                             |
+----------------------------------------------------------------------------*/
void releaseAndClose(DmOpenRef dmRef, byte removeRecord)
{
#ifdef PALMOS
   UInt16 attrP;
#else
   byte attrP;
#endif
   int32 count = DmNumRecords(dmRef);
   while (--count >= 0)
   {
#ifdef PALMOS
      DmRecordInfo(dmRef, count, &attrP, NULL, NULL); // guich@300_32: changed 0 always to count
#else
      DmGetRecordAttributes(dmRef, (uint16)count, &attrP);
#endif
      // is the record locked?
      if (attrP & dmRecAttrBusy)
      {
#ifdef PALMOS
         MemHandleUnlock(DmQueryRecord(dmRef,count));
#endif
         DmReleaseRecord(dmRef,count,0); // guich@300_46: moved to inside this if
      }
      if (removeRecord)
         DmRemoveRecord(dmRef,count); // guich@200b4_103: fix "records left locked in closed unprotected database"
   }
   // delete the database
   DmCloseDatabase(dmRef); // guich@200b2: removed from before the last block to correct the "address error" from pose
}

//////////////////////////////////////////////////////////////////////////
// java/lang/Object java.lang.Class getClass()
void ObjectGetClass(Var stack[]) // ()Ljava/lang/Class;
{
   WClass *wclass;               // guich@340_8
   WObject obj;
   UtfString className, objName;

   obj = stack[0].obj;
   objName = getClassName(obj);
   className = createUtfString("java/lang/Class");
   wclass = WOBJ_class(obj);
   if (wclass == NULL)
      RETURN32(0)

   obj = createObject(getClass(className,1));
   if (pushObject(obj) == -1) // guich@402_25
      RETURN32(0)
   else
   {
      WOBJ_ClassThis(obj) = wclass;
      WOBJ_ClassName(obj) = createStringFromUtf(objName);
      popObject();
      RETURN32(obj)
   }
}
//////////////////////////////////////////////////////////////////////////
// java/lang/Throwable void printStackTraceNative()
void ThrowablePrintStackTraceNative(Var stack[]) // ()V
{
   dumpStackTrace();
}
//////////////////////////////////////////////////////////////////////////
// java/lang/Class java.lang.Class forName(String className)
void ClassForName(Var stack[]) // (Ljava/lang/String;)Ljava/lang/Class;
{
   WObject name,obj;
   name = stack[0].obj;
   if (name == NULL)
      RETURN32(throwException("java/lang/NullPointerException","In Class.forName"))
   else
   {
      UtfString className = stringToUtf(name, STU_USE_STATIC | STU_NULL_TERMINATE);
      uint32 i;
      WClass *wclass;
      // replace all . to /
      for (i = 0; i<className.len; i++) if (className.str[i] == '.') className.str[i] = '/';

      wclass = getClass(className,0);

      // change back / to .
      for (i = 0; i<className.len; i++) if (className.str[i] == '/') className.str[i] = '.';

      if (wclass != NULL)
      {
         obj = createObject(getClass(createUtfString("java/lang/Class"),1));
         if (pushObject(obj) == -1) // guich@402_25
            RETURN32(0)
         else
         {
            WOBJ_ClassThis(obj) = wclass;
            WOBJ_ClassName(obj) = createStringFromUtf(className);
            popObject();
            RETURN32(obj)
         }
      }

      RETURN32(throwException("java/lang/ClassNotFoundException",className.str))
   }
}
//////////////////////////////////////////////////////////////////////////
// java/lang/Class Object newInstance()
void ClassNewInstance(Var stack[]) // ()Ljava/lang/Object;
{
   WObject myClass;
   WClass *desiredClass;
   WObject obj;
   Var params[1];
   byte called;

   myClass = stack[0].obj; // class object
   desiredClass = WOBJ_ClassThis(myClass);
   if (WCLASS_isInterface(desiredClass) || WCLASS_isAbstract(desiredClass))
      RETURN32(throwException("java/lang/InstantiationException",NULL))

   // allocate the memory for the object and call the default constructor
   obj = createObject(desiredClass);
   if (obj == NULL || pushObject(obj) == -1)
      RETURN32(0);
   params[0].obj = obj;
   called = callMethod(obj,"<init>","()V",params,1);
   popObject();
   if (!called)
      RETURN32(throwException("java/lang/IllegalAccessException",NULL))
   RETURN32(obj)
}
//////////////////////////////////////////////////////////////////////////
// java/lang/Class boolean isInstance(Object obj)
void ClassIsInstance(Var stack[]) // (Ljava/lang/Object;)Z
{
   WObject thisObj = stack[0].obj;
   WObject otherObj = stack[1].obj;
   WClass *myClass = WOBJ_ClassThis(thisObj);
   WClass *otherClass;
   if (otherObj == 0)
      RETURN32(throwException("java/lang/NullPointerException","In Class.isInstance"))
   else
   {
      otherClass = WOBJ_class(otherObj);
      RETURN32(compatible(otherClass,myClass))
   }
}
