/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_defines.h,v 1.15 2005/09/11 19:20:21 guich Exp $ */

#ifndef SW_DEFINES_H_INCLUDED
#define SW_DEFINES_H_INCLUDED

#ifdef WINCE
#define WIN32 1
#endif

#ifdef PALMOS_AUTO_RESET
#ifndef PALMOS5
#define PALMOS5
#endif
#endif

#ifdef PALMOS5
#ifndef PALMOS
#define PALMOS
#endif
#endif

#ifdef linux
#define LINUX
#endif

#include "SW_types.h"
#include "dbtypes.h"

#if defined(PALMOS)
extern DmSearchStateType state;
/**
 * LOCK_MEMORY has to be called to grant an exclusive write to the PALMOS storage heap.
 * At the opposite, UNLOCK_MEMORY has to be called to release this lock.
 * 
 * Darrin Massena wrote:
 * Very interesting! The first 32K of the Pilot's address space is RAM that is duplicated 
 * (simultaneously mapped to) the range $1000000-$10007fff. 
 * The memory from $0000000-$00007fff can be read from, written to, and executed from but 
 * the memory starting at $10000000 can only be read or executed -- no writing without 
 * gaining permission from an API (I later discovered) called MemSemaphoreReserve. 
 * The Pilot appears to support at least a primitive form of hardware memory protection 
 * which is used to separate "Dynamic" (Palm's term) from "Storage" RAM and keep rogue 
 * programs from trashing persistent system data structures, user data, and other 
 * programs accidentally. This is not real data security, just a very sensible precaution 
 * that should protect from the majority of random software happenings.
 */
#define LOCK_MEMORY     { MemSemaphoreReserve(true); }
#define UNLOCK_MEMORY   { MemSemaphoreRelease(true); }
#else
#define LOCK_MEMORY
#define UNLOCK_MEMORY
#endif // PALMOS

/**
 * Define PalmOS types used to write PalmOS like portable source code.
 */
#if !defined(PALMOS)

typedef struct
{
   int32 high;
   int32 low;
} FlpDouble;

typedef void *VoidHand;
typedef void * FormPtr;
typedef int16 Coord; // screen/window coordinate
typedef void *VoidPtr;
#if defined(LINUX) || defined(__SYMBIAN32__) || defined(__arm__)
typedef void *EventPtr;
#else // this struct is WIN32 only
typedef struct // Mimics windows 32 events
{
  	HWND hwnd;			/* The window for the message */
	UINT message;			/* The type of message */
	WPARAM wParam;			/* WORD message parameter */
	LPARAM lParam;			/* LONG message parameter */
} Event, *EventPtr;
#endif
typedef uint16 Err;
typedef void * WinHandle;
typedef void * MemHandle;

typedef struct AbsRectType
{
  Coord left;
  Coord top;
  Coord right;
  Coord bottom;
} AbsRectType;

typedef struct PointType
{
  Coord x;
  Coord y;
} PointType;

typedef struct RectangleType
{
  PointType  topLeft;
  PointType  extent;
} RectangleType;

typedef RectangleType *RectanglePtr;

typedef struct FontTag
{
   int16 fontType;         // font type
   int16 firstChar;        // ASCII code of first character
   int16 lastChar;         // ASCII code of last character
   int16 maxWidth;         // maximum character width
   int16 kernMax;          // negative of maximum character kern
   int16 nDescent;         // negative of descent
   int16 fRectWidth;       // width of font rectangle
   int16 fRectHeight;      // height of font rectangle
   int16 owTLoc;           // offset to offset/width table
   int16 ascent;           // ascent
   int16 descent;          // descent
   int16 leading;          // leading
   int16 rowWords;         // row width of bit image / 2
} FontType;

typedef FontType *FontPtr;

#endif //!PALMOS

// Symbian (on the device) rounds the size of the previous structure
// to a multiple of 4 (e.g.: 28 vs 26)
// On the other systems, make sure that the size is OK.
// (except for this poor microsoft preproc that doesn't catch it)
#define SIZEOF_FONTTYPE 26
#if (!defined __SYMBIAN32__ || defined __WINS__) && defined __GCC32__
  #if SIZEOF_FONTTYPE !=  sizeof(struct FontTag)
      #error FontType size is wrong
  #endif
#endif

typedef union  // added double bits to here
{
   double d;     // for easy assignment of values
   FlpDouble fd;    // for calling New Floating point manager routines
   uint32 ul[2]; // for accessing upper and lower longs
   int64  ll;
   long   l[2]; // signed
} Comp64;

//
// Register definitions for method parameters
//

#define inA0
#define inA1
#define inD0
#define inD1
#define inD2

//
// type converters
//

#if defined(PALMOS) && !defined(__arm__)

#if defined(__GNUC__)
#define getUInt32(b) \
({ \
   uchar *__b = (b); \
   uint32 __rtn; \
   asm __volatile__ \
   ( \
         "move.b (%1),(%2)       \n\t" \
         "move.b 1(%1),1(%2)     \n\t" \
         "move.b 2(%1),2(%2)     \n\t" \
         "move.b 3(%1),3(%2)     \n\t" \
         "move.l (%2),%0         \n\t" \
            : "=d" (__rtn) \
            : "a" (__b), "a" (&__rtn) \
   ); \
   __rtn; \
})

#define getUInt16(b) \
({ \
   uchar *__b = (b); \
   uint16 __rtn; \
   asm __volatile__ \
   ( \
         "move.b (%1),%0     \n\t" \
         "lsl.w  #8,%0       \n\t" \
         "move.b 1(%1),%0    \n\t" \
            : "=d" (__rtn) \
            : "a" (__b) \
   ); \
   __rtn; \
})

#define getInt32(b) \
({ \
   uchar *__b = (b); \
   uint32 __rtn; \
   asm __volatile__ \
   ( \
         "move.b (%1),(%2)       \n\t" \
         "move.b 1(%1),1(%2)     \n\t" \
         "move.b 2(%1),2(%2)     \n\t" \
         "move.b 3(%1),3(%2)     \n\t" \
         "move.l (%2),%0         \n\t" \
            : "=d" (__rtn) \
            : "a" (__b), "a" (&__rtn) \
   ); \
   __rtn; \
})

#define getInt16(b) \
({ \
   uchar *__b = (b); \
   int16 __rtn; \
   asm __volatile__ \
   ( \
         "move.b (%1),%0      \n\t" \
         "lsl.w  #8,%0        \n\t" \
         "move.b 1(%1),%0     \n\t" \
            : "=d" (__rtn) \
            : "a" (__b) \
   ); \
   __rtn; \
})

#else // __GNUC__

_inline_asm_ static uint32 getUInt32(uchar *b: __A0):__D0
{
   lea temp32, a1
   move.b (a0),(a1)     // don't change to (a0)+,(a1)+ ! code smaller but slowest!
   move.b 1(a0),1(a1)
   move.b 2(a0),2(a1)
   move.b 3(a0),3(a1)
   move.l (a1),d0
}
_inline_asm_ static int32 getInt32(uchar *b: __A0):__D0
{
   lea temp32, a1
   move.b (a0),(a1)
   move.b 1(a0),1(a1)
   move.b 2(a0),2(a1)
   move.b 3(a0),3(a1)
   move.l (a1),d0
}
_inline_asm_ static uint16 getUInt16(uchar *b: __A0):__D0
{
   move.b    (a0)+,d0 // 01
   lsl.w     #8,d0    // 10
   move.b    (a0),d0  // 12
}
_inline_asm_ static int16 getInt16(uchar *b: __A0):__D0
{
   move.b    (a0)+,d0 // 01
   lsl.w     #8,d0    // 10
   move.b    (a0),d0  // 12
}

float32 getFloat32(uchar *buf);
double get64bits(uchar *buf);
_inline_asm_ static uint32 ushr(uint32 number:__D0, int32 count:__D1):__D0
{
   lsr.l d1,d0
} 

#endif // __GNUC__

// these macros are now deprecated.  here for "legacy"
#else

#define getUInt32 READ_BE_UINT32
#define getInt32  READ_BE_INT32
#define getUInt16 READ_BE_UINT16
#define getInt16  READ_BE_INT16

#endif // defined(PALMOS) && !defined(__arm__)

//
// miscellaneous
//

uint32 ushr(uint32 number, int32 count); // can't use inline in wince!
void *xmalloc(uint32 size);
void xfree_(void *p);
int16 xstrcaselesscmp(const char *s1, const char *s2);
void *xrealloc(void *p, uint32 *newSize);
uint32 xFreeMemInfo();

#endif
/*===========================================================================*/
