/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_libDynam.h,v 1.5 2005/11/13 02:14:57 fdie Exp $ */

#ifndef __SW_LIB_DYNAM_H__
#define __SW_LIB_DYNAM_H__

// If this header is being used by the library code itself, then
// BUILDING_CUSTOM_LIB will be defined

// dave@400 Removed __declspec(dllexport) and placed exports in SWLib.def
//  If we use __declspec(dllexport) and the compile defaults to __stdcall (option /Gz),
//  which it does for WCE x86em (windows emulation), then the exported names are decorated
//  to things like _SWCallPublic@12, which cannot be found using
//  GetProcAddress(h,TEXT("SWCallPublic"));.  Using a .def file means the function is
//  actually exported as SWCallPublic.

#if defined(__SYMBIAN32__)
#define SWLIB_API EXPORT_C
#else
#define SWLIB_API
#endif

/********************************************************************
 * Library Specific Constants
 ********************************************************************/

#ifdef BUILDING_CUSTOM_LIB
#include "CustomLib.h" // Defined by each native library
#endif

/********************************************************************
 * Public functions
 ********************************************************************/

// Standard library open, close, sleep and wake functions

#if defined(__cplusplus)
extern "C" {
#endif

SWLIB_API Err SWLibOpen(LibRefnum inRefnum, void *inGlobalsPtr);
SWLIB_API Err SWLibClose(LibRefnum inRefnum);

// Custom library functions

// Invoke the native method
SWLIB_API Err SWInvokeNative(LibRefnum inRefnum, int32 hash, Var stack[]);
// Invoke a public method
SWLIB_API Err SWCallPublic(LibRefnum inRefnum, int index, uint32 *ret, uint32 *args);
// Let the native lib handle events if it want so
SWLIB_API Err SWHandleEvent(LibRefnum inRefnum, EventPtr eventP);
// Gets the NativeFunc given the hash value
SWLIB_API Err SWGetNativeFunc(LibRefnum inRefnum, int32 hash, NativeFunc *func);
// Invoke the native method given the native func (faster)
SWLIB_API Err SWInvokeNativeFunc(LibRefnum inRefnum, NativeFunc func, Var stack[]);

#if defined(__cplusplus)
};
#endif

#endif   // __SW_LIB_H__
