/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id$ */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#include "superwaba.h"
#include "SW_utils.h"
#if !defined(PALMOS) || defined(__arm__)
#include "SW_defines.h"
#endif
#include "SW_logging.h"

#if defined(ENABLE_LOGGING)

#define MAXLINE_LENGTH  384

//TODO make debugging levels configurable

#ifndef DEFAULT_LEVEL
#define DEFAULT_LEVEL	LOG_LEVEL_ERROR
#endif

static struct
{
   const char *name;
   LogLevel prio;
}
categories[]=
{
   { "----", LOG_LEVEL_MAX },
   { "posx", DEFAULT_LEVEL },
   { "sdl_", DEFAULT_LEVEL },
   { "init", DEFAULT_LEVEL },
   { "grfx", DEFAULT_LEVEL },
   { "img_", DEFAULT_LEVEL },
   { "ser_", DEFAULT_LEVEL },
   { "sock", DEFAULT_LEVEL },
   { "time", DEFAULT_LEVEL },
   { "file", DEFAULT_LEVEL },
   { "snd_", DEFAULT_LEVEL },
   { "font", DEFAULT_LEVEL },
   { "evts", DEFAULT_LEVEL },
   { "_vm_", DEFAULT_LEVEL },
   { "clsL", DEFAULT_LEVEL },
   { "dbmg", DEFAULT_LEVEL },
   { "msgs", DEFAULT_LEVEL },
   { "math", DEFAULT_LEVEL },
   { "nlib", DEFAULT_LEVEL },
};

static void logging(LogLevel level, LogCategory category, const char *message);

#if defined(PALMOS) && defined(__arm__)

#include <stdarg.h>
#include <stdio.h>

/* TODO no overrun check */
#if defined(__OS5DBG__) && (PALMOS_SDK_VERSION >= 0x0541)
#define __palmos__ // required or AdnDebug instruction are discarded
#include <AdnDebugMgr.h>
#undef __palmos__
#define LOG_LINEFEED(line) \
   strcat(line, "\n"); \
   AdnDebugMessage(line); \
   line[0] = '\0'; \
   volatile register short *watchdog = (volatile short *)0xFFFEC800; \
   watchdog[4] = 0xf5; \
   watchdog[4] = 0xa0;
#else
extern void debug(const char *line);
#define LOG_LINEFEED(line) \
   debug(line); \
   line[0] = '\0';
#endif

#elif defined(LINUX)

#define LOG_LINEFEED(line)   { printf("%s\n", line); *line = '\0'; }

#elif defined(__SYMBIAN32__)

#define LOG_LINEFEED(line)   { log_write(line); *line = '\0'; }

#else

extern void debug(const char *line);
#define LOG_LINEFEED(line) \
   debug(line); \
   line[0] = '\0';

#endif

#if HAVE_STDARG_H
#include <stdarg.h>

void log_error(LogCategory category, const char *format, ...)
{
   char msg[MAXLINE_LENGTH];
   va_list va;
   va_start(va, format);
   vsprintf(msg, format, va);
   logging(LOG_LEVEL_ERROR, category, msg);
   va_end(va);
}

void log_info(LogCategory category, const char *format, ...)
{
   char msg[MAXLINE_LENGTH];
   va_list va;
   va_start(va, format);
   vsprintf(msg, format, va);
   logging(LOG_LEVEL_INFO, category, msg);
   va_end(va);
}

void log_debug(LogCategory category, const char *format, ...)
{
   char msg[MAXLINE_LENGTH];
   va_list va;
   va_start(va, format);
   vsprintf(msg, format, va);
   logging(LOG_LEVEL_DEBUG, category, msg);
   va_end(va);
}

#else

void log_debug(LogCategory category, const char *format, ...)
{
}
void log_info(LogCategory category, const char *format, ...)
{
}
void log_error(LogCategory category, const char *format, ...)
{
}

#endif

static void logging(LogLevel level, LogCategory category, const char *message)
{
   if (category < LOG_CAT_MAX && level >= categories[category].prio)
   {
      char linebuf[MAXLINE_LENGTH];
      char *endp = &linebuf[MAXLINE_LENGTH-1];
      char *ptr = boundedCopy(linebuf, endp, level == LOG_LEVEL_ERROR ? "E/" : (level == LOG_LEVEL_INFO ? "I/" : "D/"));
      ptr = boundedCopy(ptr, endp, categories[category].name);
      ptr = boundedCopy(ptr, endp, ": ");
      ptr = boundedCopy(ptr, endp, message);
      LOG_LINEFEED(linebuf)
   }
}

#else

void log_debug(LogCategory category, const char *format, ...)
{
}
void log_info(LogCategory category, const char *format, ...)
{
}
void log_error(LogCategory category, const char *format, ...)
{
}

#endif //ENABLE_LOGGING

/*===========================================================================*/
