/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id$ */

#ifndef __SW_LOGGING_H
#define __SW_LOGGING_H

/**
 * logging levels.
 */
typedef enum
{
  LOG_LEVEL_DEBUG,
  LOG_LEVEL_INFO,
  LOG_LEVEL_ERROR,
  LOG_LEVEL_MAX
} LogLevel;

/**
 * logging categories.
 */
typedef enum
{
  LOG_CAT_NONE,
  LOG_CAT_PALMPOSIX,
  LOG_CAT_SDL,
  LOG_CAT_INIT,
  LOG_CAT_GFX,
  LOG_CAT_IMG,
  LOG_CAT_SERIAL,
  LOG_CAT_SOCK,
  LOG_CAT_TIME,
  LOG_CAT_FILE,
  LOG_CAT_SOUND,
  LOG_CAT_FONT,
  LOG_CAT_EVENT,
  LOG_CAT_VM,
  LOG_CAT_CLASSLOADER,
  LOG_CAT_DBM,
  LOG_CAT_MSG,
  LOG_CAT_MATH,
  LOG_CAT_NLIB,
  LOG_CAT_MAX
} LogCategory;

void log_error (LogCategory category, const char *format, ...);
void log_info  (LogCategory category, const char *format, ...);
void log_debug (LogCategory category, const char *format, ...);

#if defined(ENABLE_LOGGING)

#undef _LOG_DEBUG
#undef _LOG_INFO
#undef _LOG_ERROR

#define _LOG_ERROR(x) if (GLOBAL_VAR(vmGlobals)!=NULL) (VM_CALL(logError) x)
#if defined(PALMOS) && !defined(__GCC32__) // in Palm OS 68k, ignore the other log options or we'll run out of memory in some libraries
 #define _LOG_INFO(x)
#else
 #define _LOG_INFO(x)  if (GLOBAL_VAR(vmGlobals)!=NULL) (VM_CALL(logInfo) x)
#endif
 #define _LOG_DEBUG(x) if (GLOBAL_VAR(vmGlobals)!=NULL) (VM_CALL(logDebug) x)

#else

#define _LOG_DEBUG(x)
#define _LOG_INFO(x)
#define _LOG_ERROR(x)

#endif

#endif //__SW_LOGGING_H
