/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id$ */

#ifndef SW_PALMDB_PRIVATE_H
#define SW_PALMDB_PRIVATE_H

#ifndef PALMOS
 #ifndef SW_DEFINES_H_INCLUDED
  typedef uint16 Err;
 #endif
#endif

#if defined(PALMOS)
 typedef uint32 _HANDLE;
#else 
 typedef FILE *_HANDLE;
#endif

#ifdef PALMOS
#define BYTE_BUF_LEN 2048 // if this value is changed, replace it also in sw_palmdb_vfs.c
#else
#define BYTE_BUF_LEN 8192
#endif

#ifndef WIN32
 #define TCHAR char                          
#endif 

// The implementations of these function are in _vfs/_posix/_ce/_rapi
_HANDLE _myCreateFile(TCHAR * fullPath, byte createIt, byte readOnly);
void _myCloseFile(_HANDLE h);
byte _myRead(_HANDLE h, void *buf, uint32 size);
byte _myWrite(_HANDLE h, void *buf, uint32 size);
byte _myReadAt(_HANDLE h, void *buf, uint32 size, uint32 at);
byte _myWriteAt(_HANDLE h, void *buf, uint32 size, uint32 at);
uint32 _myGetFileSize(_HANDLE h);
byte _myGrowFileSize(_HANDLE h, int32 oldSize, int32 growSize);
uint32 _myGetNow();
void _myRename(TCHAR *oldName, TCHAR *newName);
void _myRemove(TCHAR *name);
void _myListDatabasesIn(TCHAR *path, HandlePDBSearchProcType proc, void *userVars, byte recursive);
char* getAppPath();
TCHAR* getSWPath();
extern int32 ldinPdbCount; // used in myListDatabasesIn
extern byte ldinStopSearch;


// these ones are in SW_palmDb.c
TCHAR *replaceDataPath(TCHAR *fullPath);
void setPalmDBProcs(struct VMGlobals * vmGlobals);
void createFileName(char *appPath, char *file, TCHAR *path, char *ext);
byte endsWithPDB(TCHAR *fName);

// the internal implementations of the data manager
// Note: in Palm OS, DmOpenRef will be replaced by CatalogPtr inside the routines
#if !defined(PALMOS) && !defined(LINUX) && !defined(__SYMBIAN32__)
typedef struct t_Catalog* DmOpenRef;
#endif

DmOpenRef myDmFindAndOpenDB(char *name, uint32 creator, uint32 type, byte readOnly, byte onCard);
char* myDmGetName(DmOpenRef db, char *buf);
DmOpenRef myDmOpenDatabase(TCHAR *fileName, uint32 creator, uint32 type);
DmOpenRef myDmOpenDatabaseReadOnly(TCHAR *fileName, uint32 creator, uint32 type);
Err myDmCloseDatabase(DmOpenRef db);
Err myDmCreateDatabase(char *name, uint32 creator, uint32 type);
Err myDmDeleteDatabase(DmOpenRef db);
Err myDmGetAttributes(DmOpenRef db, int16 *attrs);
Err myDmGetDatabaseTypeCreatorCount(TCHAR *fullPath, uint32 *type, uint32 *creator, uint16 *recCount);
Err myDmGetRecordAttributes(DmOpenRef db, uint16 index, byte *attrs);
Err myDmReleaseRecord(DmOpenRef db, int32 index, byte dirty);
Err myDmRemoveRecord(DmOpenRef db, uint16 index);
Err myDmRenameDatabase(DmOpenRef db, uint32 newType, uint32 newCreator, char *newName);
Err myDmSetAttributes(DmOpenRef db, int16 attrs);
Err myDmSetBackupAttribute(DmOpenRef db);
Err myDmSetRecordAttributes(DmOpenRef db, uint16 index, byte attrs);
Err myDmWrite(DmOpenRef db, uint16 index, uint32 offset, const void *srcP, uint32 bytes);
int32 myDmGetRecordSize(DmOpenRef db, uint16 index);
int32 myDmListDatabases(HandlePDBSearchProcType proc, void *userVars, byte recursive);
int32 myDmListDatabasesIn(char *path, HandlePDBSearchProcType proc, void *userVars, byte recursive);
MemHandle myDmGetRecord(DmOpenRef db, uint16 index);
MemHandle myDmNewRecord(DmOpenRef db, uint16 *atP, uint32 size);
MemHandle myDmQueryRecord(DmOpenRef db, uint16 index, int32 *size);
Err myDmQueryRecordPart(DmOpenRef db, uint16 index, int32 size, int32 ofs, byte* buf);
MemHandle myDmResizeRecord(DmOpenRef db, uint16 index, uint32 size);
uint16 myDmNumRecords(DmOpenRef db);

#endif  // SW_PALM_DB_H_INCLUDED
/*=============================================================================*/

