/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_palmDb_posix.c,v 1.20 2005/09/21 23:05:20 fdie Exp $ */

#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef WIN32
  #include <io.h>
  #include <sys/timeb.h>
  #include <tchar.h>
  #include "win32/dirent.h"
#else
  #include <sys/time.h>
  #include <unistd.h>
  #include <dirent.h>
  #define _tfopen fopen
#endif

#ifdef __ECOS
#include "romfs.h"
#include <cyg/fileio/fileio.h>
#endif

#include "superwaba.h"
#include "SDL_endian.h"
#include "SW_palmDbPrivate.h"

byte byteBuf[BYTE_BUF_LEN];

_HANDLE _myCreateFile(TCHAR * fullPath, byte createIt, byte readOnly)
{
   if (readOnly)
   {
      // createIt and  readOnly can't coexist: readOnly takes precedence
      // Open for read binary mode
      return _tfopen(replaceDataPath(fullPath), TEXT("rb"));
   }
   else
   if (createIt)
   {
      // Open for update binary mode, erasing previous data
      return _tfopen(replaceDataPath(fullPath), TEXT("wb+"));
   }
   else
   {
      // Open for update binary mode, without erasing
      return _tfopen(replaceDataPath(fullPath), TEXT("rb+"));
   }
}
void _myRename(TCHAR *oldName, TCHAR *newName)
{
   rename(oldName, newName);
}
void _myRemove(TCHAR *name)
{
   remove(name);
}
void _myCloseFile(_HANDLE h)
{
   fclose(h);
}
byte _myRead(_HANDLE h, void *buf, uint32 size)
{
   return fread(buf, 1, size, h);
}
byte _myWrite(_HANDLE h, void *buf, uint32 size)
{
   return fwrite(buf, 1, size, h);
}
byte _myReadAt(_HANDLE h, void *buf, uint32 size, uint32 at)
{
   if ((fseek(h, at, SEEK_SET) != 0) || (size != fread(buf, 1, size, h)))
      return PALM_ERROR;
   else
      return PALM_SUCCESS;
}
byte _myWriteAt(_HANDLE h, void *buf, uint32 size, uint32 at)
{
    if ((fseek(h, at, SEEK_SET) != 0) || (size != fwrite(buf, 1, size, h)))
       return PALM_ERROR;
    else
    {
       //fflush(h);
       return PALM_SUCCESS;
    }
}
uint32 _myGetFileSize(FILE * h)
{
   // Return 0 if errors
   uint32 pos,size;
   // Record the current position
   pos = ftell(h);
   if ((int)pos == -1) return 0;

   // Seek to the end
   if (fseek(h, 0, SEEK_END) != 0) return 0;

   // Re-get position; it's the size
   size = ftell(h);

   // Restore the stream pointer
   fseek(h, pos, SEEK_SET);

   return ((int32)size == -1)? 0 : size;
}

byte _myGrowFileSize(_HANDLE h, int32 oldSize, int32 growSize)
{
#ifdef WIN32
   int32 newSize = oldSize + growSize;
   if (growSize < 0) // guich@520_11: if shrinking, use chsize
   {
      _HANDLE osh = (_HANDLE)_get_osfhandle(_fileno((FILE*)h)); // guich: this let us use low level with high level apis!
      return ((SetFilePointer(osh, (LONG)newSize, NULL, FILE_BEGIN) != 0xFFFFFFFFL) && (SetEndOfFile(osh) != 0))? PALM_SUCCESS : PALM_ERROR; // wince does not have chsize. doh!
   }
   else
   {
      int32 cur = ftell(h);
      if (fseek(h,newSize,SEEK_SET) == 0)
      {
         if (cur != -1L)
            fseek(h, cur, SEEK_SET);
         return PALM_SUCCESS;
      }
      return PALM_ERROR;
   }
#else
   // Exact behavior expected ???
   // Write operations extend the file automatically
   //>>>PGR: could seek?
   if (growSize < 0)
   {
      #if !defined __ECOS && !defined __SYMBIAN32__
         return (ftruncate(fileno(h), growSize + oldSize) == 0)? PALM_SUCCESS : PALM_ERROR;
      #else
         return PALM_SUCCESS;
      #endif
   }
   else
   {
      uint32 pos = ftell(h);
      if (fseek(h, 0, SEEK_END) == 0)
      {
         int done = 0;
         // write any bytes
         do
         {
            uint32 actualSize = (growSize > BYTE_BUF_LEN)? BYTE_BUF_LEN : growSize;
            growSize -= actualSize;
            if (_myWrite(h, byteBuf, actualSize) == PALM_ERROR)
               return PALM_ERROR;
         } while (growSize > 0);
         fseek(h, pos, SEEK_SET);
         //fflush(h);
         return PALM_SUCCESS;
      }
      else
      {
         return PALM_ERROR;
      }
   }
#endif
}

void _myListDatabasesIn(TCHAR* path, HandlePDBSearchProcType proc, void *userVars, byte recursive)
{
   TCHAR fullPath[MAX_PATH];
   int const len = lstrlen(path);
   struct dirent * entry;
   DIR * targetDirectory = opendir(path);

   if (!targetDirectory) return;

   lstrcpy(fullPath, path);
   while (!ldinStopSearch && (entry = readdir(targetDirectory)))
   {
      int isDir;
      TCHAR const * name = entry->d_name;

      #if defined WIN32 // and BSD as well - LINUX (I think)
         isDir = (entry->d_type & DT_DIR);
      #else  // EPOC and SYMBIAN do not have the d_myype field.
      {
         struct stat statData;
         lstrcpy(fullPath+len, entry->d_name);
         if (stat(fullPath, &statData) != 0) continue;  /* error: skip this */
         #ifdef __ECOS
         isDir = (statData.st_mode & __stat_mode_DIR);
         #else
         isDir = (statData.st_mode & S_IFDIR);
         #endif
      }
      #endif
      if (isDir)
      {
         if (recursive &&          /* warning: order matters! */
            ((name[0] != '.') || ((name[1] != '\0') && ((name[1] != '.') || (name[2] != '\0')))))
         {
            #if defined WIN32  // or also, BSD
            lstrcpy(fullPath+len, entry->d_name);
            lstrcat(fullPath+len,TEXT("\\"));
            #endif
            _myListDatabasesIn(fullPath, proc, userVars, 1);
         }
      }
      else
      { // not a dir
         if (endsWithPDB((TCHAR*)entry->d_name))
         {
            ++ldinPdbCount;
            if (proc)
            {
               #if defined WIN32  // or also, BSD
               lstrcpy(fullPath+len, entry->d_name);
               #endif
               ldinStopSearch = (*proc)(fullPath, userVars);
            }
         }
      }
   }
   closedir(targetDirectory);
}

uint32 _myGetNow()
{
   #if defined WIN32
      struct _timeb tb;
      _ftime(&tb);
      return (uint32)tb.time;
   #else
      struct timeval tv;
      struct timezone tz;
      gettimeofday(&tv, &tz);
      return (uint32)tv.tv_sec;
   #endif
}

TCHAR* getAppPath()
{
	return vmGlobals->appPath;
}

TCHAR* getSWPath()
{
	return vmGlobals->swPath;
}
