/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_settings_win32.c,v 1.4 2005/03/09 23:51:38 guich Exp $ */

/* This file is WIN32 only */

#include "superwaba.h"
#define __DEFINE_FUNCTIONS
#include "SW_String.h"
#include "SW_utils.h"
#include "SW_Settings.h"

int16 lastOn;
int16 lastOntop;

// from superwaba.c
WClass * getClass(UtfString className, byte throwEx);

// the crid replaces the last 4 characters of the given string
// @pre buf contains a string of at least 4 chars
static TCHAR *createRegistryKey(TCHAR *buf, uint32 crid)
{
   TCHAR *cur = buf + lstrlen(buf) - 4;

   // set the key name to the creator id of the application
   *cur++ = (TCHAR)      (crid >> 24);
   *cur++ = (TCHAR)LOWER8(crid >> 16);
   *cur++ = (TCHAR)LOWER8(crid >> 8);
   *cur++ = (TCHAR)LOWER8(crid);

   return buf;
}

static void deleteAppPreferences(uint32 crid)
{
   TCHAR buf[40];

   ascii2unicode(buf,"Software\\SuperWaba\\appSettings\\1234",sizeof(buf));

   RegDeleteKey(HKEY_CURRENT_USER,createRegistryKey(buf,crid));
}

static void setAppPreferences(uint32 crid, UtfString utf)
{
   HKEY handle;
   DWORD disp;
   long ret;
   TCHAR buf[40];

   ascii2unicode(buf,"Software\\SuperWaba\\appSettings\\1234",sizeof(buf));

   ret = RegCreateKeyEx(HKEY_CURRENT_USER,createRegistryKey(buf, crid),0,NULL,0,KEY_ALL_ACCESS,NULL,&handle,&disp);
   if (ret == ERROR_SUCCESS)
   {
      uint32 len = utf.len+1;
      ret = RegSetValueEx(handle,TEXT("Len"),0,REG_DWORD,(byte*)&len,4); // store the length
      ret = RegSetValueEx(handle,TEXT("Value"),0,REG_BINARY,(byte*)utf.str,len); // store the \0
      RegCloseKey(handle);
   }
}

// don't forget to free the allocated buffer
static char *getAppPreferences(uint32 crid)
{
   HKEY handle;
   long ret;
   DWORD len,size, type;
   TCHAR buf[40];
   char *temp = NULL;

   ascii2unicode(buf,"Software\\SuperWaba\\appSettings\\1234",sizeof(buf));

   ret = RegOpenKeyEx(HKEY_CURRENT_USER,createRegistryKey(buf, crid),0,KEY_READ,&handle);
   if (ret == ERROR_SUCCESS)
   {
      type = REG_DWORD;
      size = 4;
      ret = RegQueryValueEx(handle,TEXT("Len"),NULL,&type,(byte *)&len,&size);
      temp = xmalloc(len);
      if (!temp) return NULL;
      type = REG_BINARY;
      ret = RegQueryValueEx(handle,TEXT("Value"),NULL,&type,(byte*)temp, &len);
      RegCloseKey(handle);
   }
   return temp;
}

static byte queryRegistry(HKEY key, TCHAR *subkey, TCHAR *name, char *buf, uint32 size)
{
   HKEY handle;
   long ret;
   DWORD type;

   ret = RegOpenKeyEx(key,subkey,0,KEY_READ,&handle);
   if (ret == ERROR_SUCCESS) // error or success? :-P
   {
      type = REG_DWORD;
      RegQueryValueEx(handle,name,NULL,&type,(byte*)buf,(LPDWORD) &size);
      RegCloseKey(handle);
      return 1;
   }
   return 0;
}

static uint32 getSecretKeyCreator(uint32 crtr)
{
   uint32 c0 = ((crtr >> 24) & 0xFF)+64;
   uint32 c1 = ((crtr >> 16) & 0xFF)+64;
   uint32 c2 = ((crtr >> 8 ) & 0xFF)+64;
   uint32 c3 = ((crtr      ) & 0xFF)+64;
   return (c0 << 24) | (c1 << 16) | (c2 << 8) | c3;
}

static Var* getSettingsField(char *name, char *desc)
{
   UtfString fieldName = createUtfString(name);
   UtfString fieldDesc = createUtfString(desc);
   WClassField *field = vmGlobals->getField(vmGlobals->prefClass, &fieldName, &fieldDesc);
   return &field->var.staticVar;
}

#ifdef WINCE

// bcao@421_53 smartphone detection
#define NO_OEMID               0
#define MOTOROLLA_OEMID        1
#define SAMSUNG_OEMID          2
#define SMARTPHONE_TYPE_OEMID  4

// bcao@421_53 added to perform system check on device.
static int CheckSystemParameterInfo(HWND hWnd)
{
   int oemId = NO_OEMID;
   TCHAR buf[MAX_PATH+1];
   buf[0] = (TCHAR)' ';

   if (SystemParametersInfo(SPI_GETOEMINFO, MAX_PATH, buf+1, 0))
   {
      if (lstrcmp(buf, TEXT(" Motorola, Inc."))==0) // Motorola smartphone
         oemId = MOTOROLLA_OEMID;
      else
      if (lstrcmp(buf, TEXT(" MITS"))==0) // Samsung smartphone
         oemId = SAMSUNG_OEMID;
   }
   if (SystemParametersInfo(SPI_GETPLATFORMTYPE, MAX_PATH, buf+1, 0))
   {
      if (lstrcmp(buf, TEXT(" SmartPhone"))==0) // Smartphone running on Microsoft Mobile
         oemId |= SMARTPHONE_TYPE_OEMID;
   }
   return oemId;
}

// note: the buffer passed in has 64 char16s
static WObject GetSerialNumberPocketPC2002() // see http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnnetcomp/html/retrievedeviceid.asp
{
    char OutputBuffer[256];
    int32 BytesReturned=0;
    int32 PresetIDOffset;
    int32 PlatformIDOffset;
    int32 PlatformIDSize;
    char buf[64];
    char *c=buf;
    int i;
    *buf = 0;
    if (KernelIoControl(IOCTL_HAL_GET_DEVICEID, 0, 0, OutputBuffer, 256, &BytesReturned) == 0)
       return 0;

    // Examine the OutputBuffer byte array to find the start of the
    // Preset ID and Platform ID, as well as the size of the
    // PlatformID.
    // PresetIDOffset - The number of bytes the preset ID is offset
    //                  from the beginning of the structure
    // PlatformIDOffset - The number of bytes the platform ID is
    //                    offset from the beginning of the structure
    // PlatformIDSize - The number of bytes used to store the
    //                  platform ID
    // Use BitConverter.ToInt32() to convert from byte[] to int
    PresetIDOffset = READ_BE_INT32(&OutputBuffer+4);
    PlatformIDOffset = READ_BE_INT32(&OutputBuffer+12);
    PlatformIDSize = READ_BE_INT32(&OutputBuffer+16);
    if (PresetIDOffset < 0 || PlatformIDOffset < 0 || PlatformIDSize < 0)
       return 0; // sometimes the KernelIoControl function does not get valid results

    // Convert the Preset ID segments into a string so they can be
    // displayed easily.
    {
       char *b;
       int32 i1;
       int16 i2,i3,i4;
       b = &OutputBuffer[0]+PresetIDOffset;   i1 = READ_BE_INT32(b);
       b += 4; i2 = READ_BE_INT16(b);
       b += 2; i3 = READ_BE_INT16(b);
       b += 2; i4 = READ_BE_INT16(b);
       xstrprintf(buf,"%4X-%2X-%2X-%2X-", i1,i2,i3,i4);
    }

    // Break the Platform ID down into 2-digit hexadecimal numbers
    // and append them to the Preset ID. This will result in a
    // string-formatted Device ID
    c += 14;
    for (i = PlatformIDOffset; i < (PlatformIDOffset + PlatformIDSize); i++)
    {
        xstrprintf(c,"%2X", OutputBuffer[i]);
        c += 2;
    }
    *c = 0;
    return vmGlobals->createString(buf);
}

static WObject GetSerialNumberPocketPC2000() // see http://www.pocketpcdn.com/articles/serial_number.html
{
   // Start CreateAssetFile.exe
   HANDLE hInFile;
   DWORD dwBytesRead;
   PROCESS_INFORMATION pi;
   TCHAR strSN[65];

   // Read data from cpqAssetData.dat file
   hInFile = CreateFile(TEXT("\\windows\\cpqAssetData.dat"), GENERIC_READ,
      FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);

   if (hInFile == INVALID_HANDLE_VALUE) // still does not exist?
   {
      if (!CreateProcess(TEXT("\\windows\\CreateAssetFile.exe"),
         NULL, NULL, NULL, FALSE, 0, NULL, NULL, NULL, &pi))
         return 0;

      // Wait until CreateAssetFile.exe will be finished
      WaitForSingleObject(pi.hProcess, INFINITE);

      hInFile = CreateFile(TEXT("\\windows\\cpqAssetData.dat"), GENERIC_READ,
         FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);

      if (hInFile == INVALID_HANDLE_VALUE) // not created? fail
         return 0;
   }

   SetFilePointer(hInFile, 976, NULL, FILE_BEGIN);
   memset(strSN, 0, 64 * sizeof(TCHAR));
   ReadFile(hInFile, &strSN, 64, &dwBytesRead, NULL);
   CloseHandle(hInFile);
   return createStringFromUnicode(strSN,-1);
}
#endif

/*----------------------------------------------------------------getSettings-+
|                                                                             |
+----------------------------------------------------------------------------*/
void getSettings()
{
   OSVERSIONINFO osvi;
   int date;
   char dateSep;
   int weekStart;
   int time24h;
   char timeSep;
   char ts; // thousandsSeparator;
   char ds; // decimalSeparator;
   WObject userName=0;
   int32 maxColors;
   int isColor;
   WObject platform=0;
   int screenWidth,screenHeight;
   int32 romVersion;
   WObject appSettings = 0;
   char buf[50];
   WObject appCreatorId=0;
   TCHAR wcbuf[36];
   char *c;
   int haveVirtualKeyboard;
   WObject appSecretKey=NULL;
   TIME_ZONE_INFORMATION tzi;
   int daylightSavings=0;
   int32 timeZone=0;
   WObject romSerialNumber=0;

   // OS version
   osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
   GetVersionEx(&osvi);

   romVersion   = osvi.dwMajorVersion * 100 + osvi.dwMinorVersion; // 2 * 100 + 11 = 2.11 //vmGlobals->romVersion;
   maxColors    = vmGlobals->maxColors;
   screenWidth  = vmGlobals->windowWidth;
   screenHeight = vmGlobals->windowHeight;
   isColor      = maxColors > 16; // isColor guich@200b3: now returns the real status. In PalmOS, if <= 16 colors, then its grayScale
#ifdef WINCE
   if ((romVersion) >= 300)
   {
      int oemId = CheckSystemParameterInfo(NULL); //test
      if (oemId & SMARTPHONE_TYPE_OEMID) {
         if (oemId & MOTOROLLA_OEMID) {
            c = "MS_SmartPhone_Motorola";
         }else if (oemId & SAMSUNG_OEMID) {
            c = "MS_SmartPhone_Samsung";
         }else {
            c = "MS_SmartPhone";
         }
      }else {
         c = "PocketPC";
      }
   }else {
      c = "WindowsCE";
   }
   platform = createString(c);
   haveVirtualKeyboard = 1;
#else
   platform = createString("Win32");
   haveVirtualKeyboard = 0;
#endif
#ifdef WINCE
   // get the name entered in the control panel
   if (queryRegistry(HKEY_CURRENT_USER, TEXT("ControlPanel\\Owner"), TEXT("Owner"), (char *)wcbuf, 36*sizeof(TCHAR)))
      userName = createStringFromUnicode(wcbuf, (uint32)-1);
#else
   // guich@502_9: return the logged on user.   
   if (queryRegistry(HKEY_CURRENT_USER, TEXT("Software\\Microsoft\\Windows\\CurrentVersion\\Explorer"), TEXT("Logon User Name"), (char *)wcbuf, 36*sizeof(TCHAR)) // first, try as a winnt machine
    || queryRegistry(HKEY_LOCAL_MACHINE, TEXT("Network\\Logon"), TEXT("Username"), (char *)wcbuf, 36*sizeof(TCHAR))) // else, try as on windows 98
      userName = createStringFromUnicode(wcbuf, (uint32)-1);
#endif

   // guich@230_25: get the preferences of the Welcome app and set the accordingly flags
   if ((c = getAppPreferences('SWAB')) != NULL) // guich@241_20: now storing in saved referrences database
   {
      vmGlobals->audibleGC = c[0] == '1';
      vmGlobals->dumpMemStats = c[1] == '1';
      xfree(c);
   }

   // decimal separator
   if (!GetLocaleInfo(LOCALE_USER_DEFAULT,LOCALE_SDECIMAL,wcbuf,1))
      ds = '.';
   else
      ds = (char)wcbuf[0];
   // thousands separator
   if (!GetLocaleInfo(LOCALE_USER_DEFAULT,LOCALE_STHOUSAND,wcbuf,2))
      ts = ',';
   else
      ts = (char)wcbuf[0];
   if (ds == ts) // guich@421_12: make sure they differ
      ds = (ts=='.')?',':'.';
   // date separator
   if (!GetLocaleInfo(LOCALE_USER_DEFAULT,LOCALE_SDATE,wcbuf,2))
      dateSep = '/';
   else
      dateSep = (char)wcbuf[0];
   // time separator
   if (!GetLocaleInfo(LOCALE_USER_DEFAULT,LOCALE_STIME,wcbuf,2))
      timeSep = ':';
   else
      timeSep = (char)wcbuf[0];
   // week start
   if (!GetLocaleInfo(LOCALE_USER_DEFAULT,LOCALE_IFIRSTDAYOFWEEK,wcbuf,2))
      weekStart = 0;
   else
      weekStart = ((char)wcbuf[0]-'0' + 1) % 7; // for SW, 0 is sunday; for Win, 6 is sunday
   // time24h
   if (!GetLocaleInfo(LOCALE_USER_DEFAULT,LOCALE_ITIME,wcbuf,2))
      time24h = 1;
   else
      time24h = (char)wcbuf[0] == '1';
   // date format
   if (!GetLocaleInfo(LOCALE_USER_DEFAULT,LOCALE_IDATE,wcbuf,2))
      date = 1;
   else
      date = (char)wcbuf[0]-'0'+1; // MDY, DMY, YMD

   // guich@230_22: get the size then the app preferrences, if any
   if ((c = getAppPreferences(vmGlobals->appCreatorId)) != NULL) // guich@241_20: now storing in saved referrences database
   {
      appSettings = (*vmGlobals->createString)(c);
      xfree(c);
   }
   // guich@330_47: get the size then the app secret key, if any
   if ((c = getAppPreferences(getSecretKeyCreator(vmGlobals->appCreatorId))) != NULL) // guich@241_20: now storing in saved referrences database
   {
      appSecretKey = (*vmGlobals->createString)(c);
      xfree(c);
   }

   // guich@340_33: timezone and daylight savings
   if (GetTimeZoneInformation(&tzi) != TIME_ZONE_ID_UNKNOWN)
   {
      //>>>PGR: FIXME!  GMT lags are not always multibles of hours.
      //        That should be minutes (timeZone -= tzi.Bias)
      timeZone = tzi.Bias / -60; // for gmt-3 it returns 180. So i divide by -60 to get -3.
      daylightSavings = tzi.DaylightBias != 0;
   }

   // guich@230_22: get the creator id
   buf[0] = (char)((vmGlobals->appCreatorId >> 24) & 0xFF);
   buf[1] = (char)((vmGlobals->appCreatorId >> 16) & 0xFF);
   buf[2] = (char)((vmGlobals->appCreatorId >> 8 ) & 0xFF);
   buf[3] = (char)((vmGlobals->appCreatorId      ) & 0xFF);
   buf[4] = 0;
   appCreatorId = (*vmGlobals->createString)(buf);

   // store the preferences in the waba.sys.Settings class
   vmGlobals->prefClass = getClass(createUtfString("waba/sys/Settings"),1);

   // guich@421_16: get the serial number
#ifdef WINCE
   romSerialNumber = GetSerialNumberPocketPC2000();
   if (!romSerialNumber && romVersion >= 300)
      romSerialNumber = GetSerialNumberPocketPC2002();
#endif
   // guich@421_17: don't call the constructor; just set directly the values
   getSettingsField("dateFormat","B")->intValue = date;
   getSettingsField("dateSeparator","C")->intValue = dateSep;
   getSettingsField("weekStart","B")->intValue = weekStart;
   getSettingsField("is24Hour","Z")->intValue = time24h;
   getSettingsField("timeSeparator","C")->intValue = timeSep;
   getSettingsField("thousandsSeparator","C")->intValue = ts;
   getSettingsField("decimalSeparator","C")->intValue = ds;
   getSettingsField("screenWidth","I")->intValue = screenWidth;
   getSettingsField("screenHeight","I")->intValue = screenHeight;
   getSettingsField("isColor","Z")->intValue = isColor;
   getSettingsField("maxColors","I")->intValue = maxColors;
   getSettingsField("onDevice","Z")->intValue = 1;
   getSettingsField("romVersion","I")->intValue = romVersion;
   getSettingsField("platform","Ljava/lang/String;")->obj = platform;
   getSettingsField("userName","Ljava/lang/String;")->obj = userName;
   getSettingsField("appSettings","Ljava/lang/String;")->obj = appSettings;
   getSettingsField("appCreatorId","Ljava/lang/String;")->obj = appCreatorId;
   getSettingsField("virtualKeyboard","Z")->intValue = haveVirtualKeyboard;
   getSettingsField("appSecretKey","Ljava/lang/String;")->obj = appSecretKey;
   getSettingsField("daylightSavings","Z")->intValue = daylightSavings;
   getSettingsField("timeZone","I")->intValue = timeZone;
   getSettingsField("romSerialNumber","Ljava/lang/String;")->obj = romSerialNumber;
   getSettingsField("isHighColor","Z")->intValue = vmGlobals->maxColors > 256 || vmGlobals->screenBPP > 8; // guich@421_35
}

/*---------------------------------------------------------storeSettingsEntry-+
|                                                                             |
+----------------------------------------------------------------------------*/
static void storeSettingsEntry(char *name, uint32 crtr)
{
   WObject obj = getSettingsField(name,"Ljava/lang/String;")->obj;
   if (obj == NULL || WOBJ_StringCharArrayObj(obj) == NULL) // if string null, delete it - guich@240_3: first condition added.
      deleteAppPreferences(crtr);
   else
      setAppPreferences (crtr, stringToUtf(obj, STU_USE_STATIC | STU_NULL_TERMINATE));
}

/*--------------------------------------------------------------storeSettings-+
|                                                                             |
+----------------------------------------------------------------------------*/
void storeSettings() // guich@230_22
{
   storeSettingsEntry("appSettings",vmGlobals->appCreatorId);
   storeSettingsEntry("appSecretKey",getSecretKeyCreator(vmGlobals->appCreatorId)); // guich@330_47
}

/*--------------------------------------------------------------------showSip-+
| show, hide or move the sip                                                  |
+----------------------------------------------------------------------------*/
void showSip(int16 on, int16 ontop)
{
#ifdef HAS_SIP
   // teste3
   CLSID Clsid;

   if (SipStatus() == SIP_STATUS_UNAVAILABLE)
      return;

   if (on    == SIP_RESTORE_STATE) on = lastOn;
   if (ontop == SIP_RESTORE_STATE) ontop = lastOntop;
   if (on == SIP_TURN_OFF || on == SIP_TO_DEFAULT)
   {
#ifdef __AYGSHELL_H__
      if (vmGlobals->mainHWnd) SHFullScreen(vmGlobals->mainHWnd, SHFS_HIDESIPBUTTON); // only supported in 3.0
#endif
      if (on == SIP_TO_DEFAULT)
      {
         RECT r = vmGlobals->sipRect;
         SipSetDefaultRect(&r);
         SipGetCurrentIM(&Clsid);
         SipSetCurrentIM(&Clsid);
      }
      SipShowIM(SIPF_OFF); // cannot be moved from here!
      lastOn = SIP_TURN_OFF;
   }
   else
   {
      if (ontop != SIP_RESTORE_STATE)
      {
         RECT r = vmGlobals->sipRect;
         if (ontop)
         {
            r.top = vmGlobals->taskBarHeight;
            r.bottom = r.top + (vmGlobals->sipRect.bottom - vmGlobals->sipRect.top);
         }

         SipSetDefaultRect(&r);
      }
#ifdef __AYGSHELL_H__
      SHFullScreen(vmGlobals->mainHWnd, SHFS_SHOWSIPBUTTON);
#endif
      SipGetCurrentIM(&Clsid);
      SipSetCurrentIM(&Clsid);
      SipShowIM(SIPF_ON);
      lastOn = SIP_TURN_ON;
      lastOntop = ontop;
   }
#endif
}

/*===========================================================================*/
