/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_utils.c,v 1.5 2005/09/25 12:56:34 guich Exp $ */

#include "SW_types.h"
#include "SW_utils.h"

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

/*--------------------------------------------------------------ascii2unicode-+
| This is misnamed.                                                           |
| We copy src, possibly multibyte, to either                                  |
| - byte (e.g.: memcpy); or,                                                  |
| - unicode dst, if unicode is enabled.                                       |
+----------------------------------------------------------------------------*/
int32 ascii2unicode(TCHAR * dst, char const * src, int32 dstSize)
{
   int32 len;
#ifdef UNICODE
   len = mbstowcs(0, src, 0);
   if (len < 0)  // MS Visual bug: wcstombs(dst, src, 0) returns... -1!
      len = 0;
   else
   {
      if (dstSize <= len) len = dstSize-1;
      len = mbstowcs(dst, src, len);
   }
#else
   len = xstrlen(src);
   if (dstSize <= len) len = dstSize-1;
   xstrncpy(dst, src, len);
#endif
   dst[len] = 0;
   return len;
}

/*--------------------------------------------------------------unicode2ascii-+
| This is misnamed.                                                           |
| We copy either:                                                             |
| - byte (e.g.: memcpy); or,                                                  |
| - unicode dst, if unicode is enabled,                                       |
| to possibly multibyte dst.                                                  |
+----------------------------------------------------------------------------*/
int32 unicode2ascii(char * dst, TCHAR const * src, int32 dstSize)
{
   int32 len;

#ifdef UNICODE
   len = wcslen(src);
   if (len < 0)  // in case they have the same bug as above
      len = 0;
   else
   {
      if (dstSize <= len) len = dstSize-1;
      len = wcstombs(dst, src, len);
   }
#else
   len = xstrlen(src);
   if (dstSize <= len) len = dstSize-1;
   xstrncpy(dst, src, len);
#endif
   dst[len] = 0;
   return len;
}

/*----------------------------------------------------------------pathConvert-+
| calls asciiToUnicode() and strips lead and trailing quotes                  |
+----------------------------------------------------------------------------*/
// static void pathConvert(char *src, TCHAR *dst, int max)
// {
//    int start, len;
//
//    start = 0;
//    if (src[0] == '"')
//       start = 1;
//    len = ascii2unicode(dst, &src[start], max);
//    if (len > 0 && dst[len - 1] == '"')
//       dst[len - 1] = 0;
// }

/*--------------------------------------------------------------------my_atoi-+
| String to int.                                                              |
| The string is read until a non-numeric character is found.                  |
+----------------------------------------------------------------------------*/
//>>PGR: I'm quite fed up with all these libraries missing atoi!  what a mess.
#if !HAVE_GETOPT_LONG
int32 my_atoi(char const * c_p)
{
   long i = 0L;
   unsigned int k;

   --c_p;
   /* Transform the lengthed string into an int */
   while (((k = *++c_p) <= '9') && (k >= '0'))
   {
      if ((i>=((unsigned long)~0)/10) && ((i > ((unsigned long)~0)/10) || (k > '5')))
         return 0;
      i = ((i +(i<<2))<<1) + k - '0';    /* i*10 + k - '0' */
   }
   return i;
}
#endif

/*----------------------------------------------------------------boundedCopy-+
| Safe copy of the asciiz string src to dst.                                  |
| dstEnd points the 1st byte which is not part of dst (i.e.: dst + sizeof dst)|
| dst is guaranteed to have a zero sentinel, except if there is not even      |
| one byte available.                                                         |
| Returns a pointer to the ending sentinel, or to the end of dst if none.     |
+----------------------------------------------------------------------------*/
char * boundedCopy(char * dst, char * dstEnd, char const * src)
{
   return boundedCopyN(dst, dstEnd, src, -1);
}

/*---------------------------------------------------------------boundedCopyN-+
| Safe copy of the N cars of an asciiz string src to dst.                     |
| dstEnd points the 1st byte which is not part of dst (i.e.: dst + sizeof dst)|
| dst is guaranteed to have a zero sentinel, except if there is not even      |
| one byte available.                                                         |
| Returns a pointer to the ending sentinel, or to the end of dst if none.     |
+----------------------------------------------------------------------------*/
char * boundedCopyN(char * dst, char * dstEnd, char const * src, int len)
{
   if (dst <= --dstEnd)     // at least one byte is required
   {
      while (dst < dstEnd && len-- != 0)
         if ((*dst++ = *src++) == '\0')   // ending sentinel
            return dst-1;  // return the '\0' position
      *dst = '\0';
   }
   return dst;
}

/*-----------------------------------------------------------------boundedCat-+
| Safely concatenate the asciiz string src to dst.                            |
| dstEnd points the 1st byte which is not part of dst (i.e.: dst + sizeof dst)|
| If dst is no zero ended, or if no room is left, then nothing is done.       |
| Otherwise, the dst is appended as much as possible, with a zero sentinel.   |
| Returns a pointer to the ending sentinel, or to the end of dst if none.     |
+----------------------------------------------------------------------------*/
char * boundedCat(char * dst, char * dstEnd, char const * src)
{
   return boundedCatN(dst, dstEnd, src, -1);
}

/*-----------------------------------------------------------------boundedCat-+
| Safely concatenate the asciiz string src to dst.                            |
| dstEnd points the 1st byte which is not part of dst (i.e.: dst + sizeof dst)|
| If dst is no zero ended, or if no room is left, then nothing is done.       |
| Otherwise, the dst is appended as much as possible, with a zero sentinel.   |
| Returns a pointer to the ending sentinel, or to the end of dst if none.     |
+----------------------------------------------------------------------------*/
char * boundedCatN(char * dst, char * dstEnd, char const * src, int len)
{
   while (dst < dstEnd)
   {
      if (*dst == '\0')   // ending sentinel
         return boundedCopyN(dst, dstEnd, src, len);
      ++dst;
   }
   return dst;
}

/*==========================================================================*/
