/* $Id: epoc_sound.cpp,v 1.9 2004/09/30 11:05:59 petrus Exp $*/

#include "audiosvr/AudioServer.h"
#include "epoc_sound.h"

#include <e32base.h>

TUint const maxConcurrentRequests = 5;

/*-------------------------------------------------------class ServerSession -+
|                                                                             |
+----------------------------------------------------------------------------*/
class ServerSession : public RSessionBase
{
public:
   void init();
   int execute(TInt audioFunction, TAny * args);
   void quit();
private:
   int m_isOpen;
   RSemaphore m_semAudioIdle;
   TInt open();
   void close();
};

static ServerSession server;

/*--------------------------------------------------------ServerSession::open-+
|                                                                             |
+----------------------------------------------------------------------------*/
TInt ServerSession::open()
{
   TInt err = startAudioServer();
   if ((err == KErrNone) || (err == KErrAlreadyExists)) {
      err = CreateSession(
         KAudioServerName,
         TVersion(
            KAudioServerMajorVersionNumber,
            KAudioServerMinorVersionNumber,
            KAudioServerBuildVersionNumber
         ),
         maxConcurrentRequests
      );
   }
   if (err == KErrNone) {
      err = m_semAudioIdle.OpenGlobal(KAudioServerName);
      if (err != KErrNone) {
         User::Panic(KAudioServerName, err);
      }
      m_isOpen = 1;
   }else {
      m_isOpen = 0;
   }
   return err;
}

/*-------------------------------------------------------ServerSession::close-+
|                                                                             |
+----------------------------------------------------------------------------*/
void ServerSession::close()
{
   if (m_isOpen) {
      m_isOpen = 0;
      TAny * p[KMaxMessageArguments];
      SendReceive(EAudioServerCloseSession, p);
      RHandleBase::Close();
   }
}

/*-----------------------------------------------------ServerSession::execute-+
|                                                                             |
+----------------------------------------------------------------------------*/
int ServerSession::execute(TInt audioFunction, TAny * args)
{
#if 1   // synchronous
   if (m_isOpen || (KErrNone == open())) {
      SendReceive(audioFunction, args);
      m_semAudioIdle.Wait();           // wait until done
      return 1;
   }else {
      return 0;
   }
#else   // asynchronous
   if (m_isOpen) {
      m_semAudioIdle.Wait();           // wait for previous to be done
   }else if (KErrNone != open()) {
      return 0;
   }
   SendReceive(audioFunction, args);
   return 1;
#endif
}

/*--------------------------------------------------------ServerSession::quit-+
|                                                                             |
+----------------------------------------------------------------------------*/
void ServerSession::quit()
{
   if (m_isOpen) {
      TAny * p[KMaxMessageArguments];
      execute(EAudioServerEndServer, p);
      m_semAudioIdle.Close();
      close();
   }
}

/*----------------------------------------------------------epoc_sound_enable-+
|                                                                             |
+----------------------------------------------------------------------------*/
void epoc_sound_enable(int doIt)
{
   TAny * p[KMaxMessageArguments];
   p[0]= (TAny*)doIt;
   server.execute(EAudioServerEnable, p);
}

/*--------------------------------------------------------epoc_sound_playTone-+
|                                                                             |
+----------------------------------------------------------------------------*/
int epoc_sound_playTone(int frequence, int duration)
{
   TAny * p[KMaxMessageArguments];
   p[0] = (TAny *)frequence;
   p[1] = (TAny *)duration;
   return server.execute(EAudioServerPlayTone, p);
}

/*---------------------------------------------------epoc_sound_makeBufHandle-+
| Return newly created handle, 0 if bad                                       |
+----------------------------------------------------------------------------*/
TUint epoc_sound_makeBufHandle(char * data)
{
   TAny * p[KMaxMessageArguments];
   TPckgBuf<TUint> pkgbuf;

   p[0] = (TAny *)data;
   p[1] = (TAny *)&pkgbuf;
   server.execute(EAudioServerMakeBufHandle, p);
   return pkgbuf();
}

/*--------------------------------------------------epoc_sound_makeFileHandle-+
| Return newly created handle, 0 if bad                                       |
+----------------------------------------------------------------------------*/
TUint epoc_sound_makeFileHandle(unsigned short * path, int len)
{
   TAny * p[KMaxMessageArguments];
   TPckgBuf<TUint> pkgbuf;

   p[0] = (TAny *)path;
   p[1] = (TAny *)len;
   p[2] = (TAny *)&pkgbuf;
   server.execute(EAudioServerMakeFileHandle, p);
   return pkgbuf();
}

/*------------------------------------------------------epoc_sound_playHandle-+
|                                                                             |
+----------------------------------------------------------------------------*/
void epoc_sound_playHandle(TAny * handle)
{
   TAny * p[KMaxMessageArguments];

   p[0] = (TAny *)handle;
   server.execute(EAudioServerPlayHandle, p);
}

/*------------------------------------------------------epoc_sound_freeHandle-+
|                                                                             |
+----------------------------------------------------------------------------*/
void epoc_sound_freeHandle(TAny * handle)
{
   TAny * p[KMaxMessageArguments];

   p[0] = (TAny *)handle;
   server.execute(EAudioServerFreeHandle, p);
}

/*------------------------------------------------------------epoc_sound_quit-+
|                                                                             |
+----------------------------------------------------------------------------*/
void epoc_sound_quit()
{
   server.quit();
}

/*===========================================================================*/
