/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: superwaba.h,v 1.31 2005/11/18 03:03:14 fdie Exp $ */

#ifndef __SUPERWABA_H
#define __SUPERWABA_H

// #define ENABLE_TRACE                                            /*@@*/
#ifdef __cplusplus                                                 /*@@*/
extern "C" {                                                       /*@@*/
#endif                                                             /*@@*/
void showMsg(int no, char * msg1, char * msg2);                    /*@@*/
void showError(int error);                                         /*@@*/
#ifdef __cplusplus                                                 /*@@*/
}                                                                  /*@@*/
#endif                                                             /*@@*/

///////////////////  PLATFORM DEFINES ////////////////////
// IMPORTANT NOTICE: To compile the SuperWaba VM, you need to define:
//
// For PalmOS < 5.0
//     PALMOS;SDL;VIDEO_BITSPERPIXEL=??
//
// For PalmOS >= 5.0
//     PALMOS5;SDL;VIDEO_BITSPERPIXEL=??
//
// For Windows 98, NT, 2000 or similar
//     WIN32;SDL;VIDEO_BITSPERPIXEL=??;_WINDOWS
//
// For Windows CE and Pocket PC
//     WIN32;SDL;VIDEO_BITSPERPIXEL=??;_WINDOWS;WINCE
//
// For CodeWarrior, these defines are found in palm\sysdefs.h
// For all other compilers, define it on the compilation command.
#include "sysdefs.h"   // relevant for CodeWarrior only
#include "SW_defines.h"
#include "SW_logging.h"

#include "SW_palmDb.h"

/**
 * The following macros help for the support of components which do not support globals.
 * Indeed, some systems do not support globals, sometimes only for some special software 
 * components like dynamic libraries. To write code and use globals seemlessly, one should
 * use these macros and implement another way to store data.
 * On Symbian, which doesn't support globals in dynamic libraries, we use TLS (Thread Local Storage) 
 * which is a common technique to solve this issue. An existing implementation is provided through 
 * a class helper named EpocStat (see http://www.peroon.co.il/epocstat.html) described 
 * in detail here: http://www.newlc.com/article.php3?id_article=206
 **/
#if defined(__SYMBIAN32__) && defined(BUILDING_CUSTOM_LIB)
#include <e32std.h>
#define GLOBAL_VAR(x) (((SStaticRepository*)GetStaticsRepository(StaticRepositoryCreateFunction, KStaticsServerName))->x)
#define DECLARE_VAR(type, name)                     /* nothing */
#define DECLARE_INITED_VAR(type, name, init_val)    /* nothing */
#define DECLARE_ARRAY(type, name, size)             /* nothing */
#define CACHE_VMGLOBALS struct VMGlobals *vmGlobals = GLOBAL_VAR(vmGlobals);
#else
// non-Symbian - everything works as usual
#define GLOBAL_VAR(x) (x)
#define DECLARE_VAR(type, name)                     type name
#define DECLARE_INITED_VAR(type, name, init_val)    type name = init_val
#define DECLARE_ARRAY(type, name, size)             type name[size]
#define CACHE_VMGLOBALS
#endif

#define VM_CALL(x)      (*GLOBAL_VAR(vmGlobals)->x)
#define VM_VAR(x)       GLOBAL_VAR(vmGlobals)->x

/**
 * SuperWaba VM data structures
 **/
typedef union VarT
{
    int32         intValue;
    float32       floatValue;
    void *        classRef;
    uchar *       pc;
    uchar *       oldpc;
    void *        refValue;
    union VarT *  obj;
    uint32        half64; // guich@200 - added native support for 64 bit doubles and longs - they are stored as 2 consecutives vars
} Var;

// Var[-3]     = Next WObject
// Var[-2]     = Order
// Var[-1]     = Temp
// Var[0]      = ptr to class A (or null if Array)
// Var[1 to n] = primitives, arrays and objects fields in class A

typedef Var *WObject;

#define WOBJ_class(o)    (o)[0].classRef
#define WOBJ_var(o, idx) (o)[idx + 1]   // nopt

// guich@500_5: values are now returned in the libReturnedValue, which is pushed by the caller.
// for 32 bit, use stack[-1].
// for 64 bit, use stack[-2] = v[0] and stack[-1] = v[1];
// or the macros below
// to avoid confusions, the method returns immediately.
// Note that, in RETURN64, x cannot be a call to a function, but a temporary var instead
// guich@501_15: now, to correctly return float values, use RETURN32_FLOAT.
// note that the refValue convertion will result in a compiler error. This is proposital.

#define RETURN32(x) { \
   stack[-2].refValue = (void*)(x); \
   return; \
}

#define RETURN32_FLOAT(x) { \
   stack[-2].floatValue = (float32)(x); \
   return; \
}

#define RETURN32_FROM_BYTES(bp) { \
    int32 _v_ = READ_BE_INT32(bp); \
    RETURN32(_v_) \
}

/*
| store a double (or int64) onto the stack in native
| byte-endianness and word-endianness
*/
static __inline void store64(void * ip, Var * stack)
{
   stack[-2].intValue = *(uint32*)ip;
   stack[-1].intValue = *((uint32*)ip+1);
}

/*
| guich@200 - transforms 2 consecutive vars in one double, respecting the
| byte-endianness and word-endianness.
| the first var stores the lower or upper 32 bits
| the 2nd var stores the upper or lower 32 bits of the double or int64
*/
static __inline void copy64(Var *src, Var *dest)
{
#if defined(PALMOS) && defined(__arm__)
   if (((uint32)src & 3) != 0 || ((uint32)dest & 3) != 0) // cannot move unaligned DWORDs
   {
      uint16 *s = (uint16 *)src;
      uint16 *d = (uint16 *)dest;
      *d++ = *s++;
      *d++ = *s++;
      *d++ = *s++;
      *d = *s;
      return;
   }
#endif
   *dest = *src;
   *(dest+1) = *(src+1);
}

// should be: RETURN64
#define RETURN64_FROM_VALUE(v) \
{ \
    store64(&v, stack); \
    return; \
}

// should be: RETURN64_FROM_BYTES
#define RETURN64(bp) \
{ \
    int64 _v_ = READ_BE_INT64(bp); \
    RETURN64_FROM_VALUE(_v_) \
}

// should be: RETURN_DOUBLE_FROM_BYTES
#define RETURN_DOUBLE(bp) \
{ \
    int64 _v_ = READ_BE_DOUBLE(bp); \
    RETURN64_FROM_VALUE(_v_) \
}

typedef void (* NativeFunc)(Var stack[]);

typedef union
{
    // Code is either pointer to bytecode or pointer to native function
    // NOTE: If accessFlags(method) & ACCESS_NATIVE then nativeFunc
    // is set, otherwise codeAttr is set. Native methods don't have
    // maxStack, maxLocals so it is OK to merge the codeAttr w/nativeFunc.
    uchar *codeAttr;
    uint32 hash;
    NativeFunc nativeFunc;
} Code;

typedef struct
{
    char *className;
    int16 destroyFuncNum; // guich@200b4: not more ObjDestroyFunc destroyFunc;
    uint16  libRefnum;         // guich@200b4: the number of the library that owns this destroy func
} ClassHook;

typedef struct UtfStringStruct
{
    uint32 len;
    char *str;
} UtfString;

typedef struct UtfString16Struct
{
    uint32 len;
    uint16 *str;
} UtfString16;

//
// types and accessors
//

typedef union
{
    // FieldVar is either a reference to a static class variable (staticVar)
    // or an offset of a local variable within an object (varOffset)
    Var staticVar;
    uint32 varOffset; // computed var offset in object
} FieldVar;

typedef struct WExceptionHandlerStruct  // guich@230_14
{
   uint16 startPc, endPc, handlerPc, catchType;
} WExceptionHandler;

typedef struct WClassFieldStruct
{
    FieldVar var;  // 3
    FieldVar var2; // 4 guich@200 - one more int32 to hold 64 bit values
    uint32 hash;    // 2
    byte is64;     // 5 Extra flags value.
    byte isStatic;
    uchar *header; // 1
} WClassField;

typedef struct WClassMethodStruct // size = 36
{
   uint32 hash;    // 9 guich@200b2
   Code code;     // 8
   uchar *header; // 7
   uint16 numParams; // 6
   byte maxStack;
   byte maxLocals;
   byte returnedValueCount; // 1
   byte isInit;       // 2
   byte isNative;   // 3
   byte libRefNum;     // 4    libref + 1, 0 means not bound
   byte localsStack2; // 5
   byte dummy;
   byte *codeStart;
   byte *lineNumberTable; // fdie@500_13

   // guich@230_14
   uint16 numHandlers; // 10
   WExceptionHandler *handlers; // 11
} WClassMethod;

// The VM keeps an array of constant offsets for each constant in a class
// in the runtime class structure (see WClassStruct). For each constant,
// the offset is an offset from the start of the bytes defining the class.
//
// When a constant offset is "bound", instead of the offset being
// an offset into the constant pool, it is (with the exception of methods)
// a pointer offset from the start of the class heap to the actual data
// the constant refers to.
//
// For example, when a field constant is bound, it contains an offset
// from the start of the class heap to the actual WClassField * structure
// for the field. For class offsets, it is an offset to the WClass *
// structure. For method offsets, the offset is a virtual method number
// and class index. Only class, field and methods can be bound.
//
// A bound offset will only be bound if the offset of the actual structure
// in the class heap is within the range that can fit in the offset. For
// example, in a small memory model, if a WClassField * structure exists
// beyond 32K from the start of the class heap, its offset can't be bound.
// If that happens, the offset simply won't be bound and will retain
// an offset into the constant pool (known now as an "adaptive bind").
//
// Binding of constants (adaptive quickbind) will only be performed if
// QUICKBIND is defined.
// guich@200b4: now QUICKBIND is definetely used.
// When an offset is bound, it's CONST_boundBit
// will be set to 1.
// guich@200b4r5: no more SMALL MEM, everything now grow as needed.

typedef uint32 BindedConst;
typedef uint32 ConsOffset;

typedef struct
{
   void *c;  // class     - guich@400_16: replaced indexes by direct pointers - improves String methods
   void *m;  // method
} VMapValue;

typedef struct
{
   uint16 mapSize; // 2 size of map = number of inherited methods
   VMapValue *mapValues; // 1 maps virtual method number to class, virtual method index
   uint16 numVirtualMethods; // 3 number of new virtual methods in the class
} VirtualMethodMap;

// NOTE: In the following structure, a constant offset can either be
// bound (by having boundBit set) in which case it is an offset into
// the classHeap directly or unbound in which case it is an offset into
// the byteRep of the class
typedef struct WClassStruct
{
    ConsOffset *constantOffsets; // 7
    BindedConst *bindedConstants;
    WClassMethod **bindedMethods;
    struct WClassStruct **superClasses; // 1. array of this classes superclasses
    uint16 numSuperClasses; // 2
    uchar *byteRep; // 4 pointer to class representation in memory (bytes)
    uchar *attrib2; // 5 pointer to area after constant pool (accessFlags)
    uint16 numConstants; // 6
    uint16 numFields;  // 8
    WClassField *fields; // 9
    uint16 numMethods;   // 10
    WClassMethod *methods; // 11
    VirtualMethodMap vMethodMap; // 12
    uint16 numVars; // 13 computed number of object variables
    // guich@200b4: changed
    int16   objDestroyFuncNum;  // 14
    uint16  objDestroyFuncLibRefnum; // 15
    byte  isInterface;
    byte  isAbstract;
    byte  is12;                            // guich@310_14
    byte  nothing;
    UtfString className;  // guich@400_16
    struct WClassStruct *nextClass; // 16 next class in hash table linked list
} WClass;

// search types for getMethodMapNum()
#define CLASS_SEARCH_ALL 0
#define CLASS_SEARCH_INHERITED 1
#define CLASS_SEARCH_THISCLASS 2

#define WCLASS_accessFlags(wc) READ_BE_UINT16(wc->attrib2)
#define WCLASS_thisClass(wc) READ_BE_UINT16(&wc->attrib2[2])
#define WCLASS_superClass(wc) READ_BE_UINT16(&wc->attrib2[4])
#define WCLASS_numInterfaces(wc) READ_BE_UINT16(&wc->attrib2[6])
#define WCLASS_interfaceIndex(wc, idx) READ_BE_UINT16(&wc->attrib2[8 + ((idx) << 1)])
#define WCLASS_objectSize(wc) ((wc->numVars + 1) * sizeof(Var))
#define WCLASS_isInterface(wc) (wc->isInterface)
#define WCLASS_isAbstract(wc)  (wc->isAbstract)
#define WCLASS_hasDestroyFunc(wc) (wc->objDestroyFuncNum > 0) // guich@200b4

#define FIELD_is64(f) ((f)->is64) // mlb@200b3 : fixed support for double and longs
#define FIELD_accessFlags(f) READ_BE_UINT16(f->header)
#define FIELD_nameIndex(f) READ_BE_UINT16(&f->header[2])
#define FIELD_descIndex(f) READ_BE_UINT16(&f->header[4])
#define FIELD_isStatic(f) (f->isStatic)
#define FIELD_is(f,acc) ((FIELD_accessFlags(field) & acc) > 0)

#define METH_accessFlags(m) READ_BE_UINT16(m->header)
#define METH_nameIndex(m) READ_BE_UINT16(&m->header[2])
#define METH_descIndex(m) READ_BE_UINT16(&m->header[4])
#define METH_maxStack(m) READ_BE_UINT16(&m->code.codeAttr[6])
#define METH_maxLocals(m) READ_BE_UINT16(&m->code.codeAttr[8])
#define METH_codeCount(m) READ_BE_UINT32(&m->code.codeAttr[10])
#define METH_code(m) &m->code.codeAttr[14]
#define METH_isNative(m) (m->isNative)
#define METH_localsStack2(m) (m->localsStack2)

#define CONST_bindedMeth(wc, idx) wc->bindedMethods[idx]
#define CONST_binded(wc, idx) wc->bindedConstants[idx] // used for methods, fields and classes only
// #define CONST_offset(wc, idx) wc->constantOffsets[idx]  // guich@200b4: optimized idx-1
#define CONST_ptr(wc, idx)   ((byte *)wc->constantOffsets[idx])
#define CONST_ptr1(wc, idx) &((byte *)wc->constantOffsets[idx])[1]
#define CONST_ptr3(wc, idx) &((byte *)wc->constantOffsets[idx])[3]
#define CONST_tag(wc, idx) CONST_ptr(wc, idx)[0]
#define CONST_utfLen(wc, idx) READ_BE_UINT16(CONST_ptr1(wc, idx))
#define CONST_utfStr(wc, idx) CONST_ptr3(wc, idx)
#define CONST_integer(wc, idx) READ_BE_INT32(CONST_ptr1(wc, idx))
#define CONST_stringIndex(wc, idx) READ_BE_UINT16(CONST_ptr1(wc, idx))
#define CONST_classIndex(wc, idx) READ_BE_UINT16(CONST_ptr1(wc, idx))
#define CONST_nameAndTypeIndex(wc, idx) READ_BE_UINT16(CONST_ptr3(wc, idx))
#define CONST_nameIndex(wc, idx) READ_BE_UINT16(CONST_ptr1(wc, idx))
#define CONST_typeIndex(wc, idx) READ_BE_UINT16(CONST_ptr3(wc, idx))

// Access flags
#define ACCESS_PUBLIC       0x0001
#define ACCESS_PRIVATE      0x0002
#define ACCESS_PROTECTED    0x0004
#define ACCESS_STATIC       0x0008
#define ACCESS_FINAL        0x0010
#define ACCESS_SYNCHRONIZED 0x0020
#define ACCESS_VOLATILE     0x0040
#define ACCESS_TRANSIENT    0x0080
#define ACCESS_NATIVE       0x0100
#define ACCESS_INTERFACE    0x0200
#define ACCESS_ABSTRACT     0x0400

// Constant Pool tags
#define CONSTANT_Utf8               1
#define CONSTANT_Integer            3
#define CONSTANT_Float              4
#define CONSTANT_Long               5
#define CONSTANT_Double             6
#define CONSTANT_Class              7
#define CONSTANT_String             8
#define CONSTANT_Fieldref           9
#define CONSTANT_Methodref          10
#define CONSTANT_InterfaceMethodref 11
#define CONSTANT_NameAndType        12

typedef struct
{
    uint16 errNum;
    char className[128];
    char methodName[80];
    char arg1[100];
    char arg2[100];
} ErrorStatus;

// flags for stringToUtf()
#define STU_NULL_TERMINATE 1
#define STU_USE_STATIC     2
#define STU_CREATE         3 // guich@350: bypass buffer, memory will be reclaimed by the GC

//
// more types and accessors
//

// NOTE: These get various values in objects at defined offsets.
// If the variables in the base classes change, these offsets will
// need to be recomputed. For example, the first (StringCharArray)
// get the character array var offset in a String object.
#define WOBJ_StringCharArrayObj(o)  (o)[1].obj
#define WOBJ_StringBufferStrings(o) (o)[1].obj
#define WOBJ_StringBufferCount(o)   (o)[2].intValue

#define ARRAY_BIG_BUT_FAST // big: 787 objects until GC (test runs in 51430ms). small: 821 objects until GC (runs in 51890ms).
#ifdef ARRAY_BIG_BUT_FAST

#define WOBJ_arrayType(o)  (o)[1].intValue
#define WOBJ_arrayLen(o)   (o)[2].intValue
#define WOBJ_arrayStart(o) (&((o)[3]))

// for faster access
#define WOBJ_arrayTypeP(objPtr)  (objPtr)[1].intValue
#define WOBJ_arrayLenP(objPtr)   (objPtr)[2].intValue
#define WOBJ_arrayStartP(objPtr) (&(objPtr[3]))

#else
typedef struct
{
   uint16 len;
   uint16 type;
} ArrayTypeLen; // guich@400_15

#define WOBJ_arrayType(o)  ((ArrayTypeLen *)(o+1))->type
#define WOBJ_arrayLen(o)   ((ArrayTypeLen *)(o+1))->len
#define WOBJ_arrayStart(o) (&((o)[2]))

// for faster access
#define WOBJ_arrayTypeP(objPtr)  ((ArrayTypeLen *)(objPtr+1))->type
#define WOBJ_arrayLenP(objPtr)   ((ArrayTypeLen *)(objPtr+1))->len
#define WOBJ_arrayStartP(objPtr) (&(objPtr[2]))
#endif

// fatal errors
#define ERR_SanityCheckFailed        1
#define ERR_IncompatibleDevice       2
#define ERR_CantAccessCoreClasses    3
#define ERR_CantAccessAppClasses     4
#define ERR_CantAllocateMemory       5
#define ERR_OutOfClassMem            6
#define ERR_OutOfObjectMem           7
#define ERR_NativeStackOverflow      8
#define ERR_NativeStackUnderflow     9
#define ERR_StackOverflow            10

// program errors
#define ERR_BadClass                 11
#define ERR_BadOpcode                12
#define ERR_CantFindClass            13
#define ERR_CantFindMethod           14
#define ERR_CantFindField            15
#define ERR_NullObjectAccess         16
#define ERR_NullArrayAccess          17
#define ERR_ClassTooLarge            18
#define ERR_DivideByZero             19
#define ERR_ClassCastException       20
#define ERR_IndexOutOfRange          21

// superwaba new errors
#define ERR_CantAllocateObjectMemory 22
#define ERR_BadUtfString             23
#define ERR_NoFontsFound             24
#define ERR_MethodCalledOnArray      25
#define ERR_CantFindNativeMethod     26
#define ERR_UncaughtException        27
#define ERR_ClassInWrongPackage      28
#define ERR_CantInitializeApp        29

// other fatal errors (from main)
#define ERR_NoMainWindowObj          30
#define ERR_NoMemory                 31
#define ERR_SdlInitFailed            32
#define ERR_SdlNetInitFailed         33
#define ERR_BadOption                34
#define ERR_NoRunRequired            35

//
// Rect
//
// var[0] = Class
// var[1] = int x
// var[2] = int y
// var[3] = int width
// var[4] = int height

#define WOBJ_RectX(o)      (o)[1].intValue
#define WOBJ_RectY(o)      (o)[2].intValue
#define WOBJ_RectWidth(o)  (o)[3].intValue
#define WOBJ_RectHeight(o) (o)[4].intValue

//
// Time
//
// var[0] = Class
// var[1] = int year
// var[2] = int month
// var[3] = int day
// var[4] = int hour
// var[5] = int minute
// var[6] = int second
// var[7] = int millis

#define WOBJ_TimeYear(o)   (o)[1].intValue
#define WOBJ_TimeMonth(o)  (o)[2].intValue
#define WOBJ_TimeDay(o)    (o)[3].intValue
#define WOBJ_TimeHour(o)   (o)[4].intValue
#define WOBJ_TimeMinute(o) (o)[5].intValue
#define WOBJ_TimeSecond(o) (o)[6].intValue
#define WOBJ_TimeMillis(o) (o)[7].intValue

//
// Class
//
// var[0] = Class
// var[1] = Object thisClass
// var[2] = String name

#define WOBJ_ClassThis(o) (o)[1].classRef
#define WOBJ_ClassName(o) (o)[2].obj

//
// Coord
//
// var[0] = Class
// var[1] = int x
// var[2] = int y

#define WOBJ_CoordX(o) (o)[1].intValue
#define WOBJ_CoordY(o) (o)[2].intValue

//
// Control
//
// var[0] = Class
// var[1] = int x
// var[2] = int y
// var[3] = int width
// var[4] = int height

#define WOBJ_ControlX(o)      (o)[1].intValue
#define WOBJ_ControlY(o)      (o)[2].intValue
#define WOBJ_ControlWidth(o)  (o)[3].intValue
#define WOBJ_ControlHeight(o) (o)[4].intValue
//#define WOBJ_ControlImageGraphicsGrHandle(o) (o)[6].refValue; for compatibility purposes, the ControlGrHandle will be in the same position (6) of GraphicsGrHandle

//
// SerialPort
//
// var[0] = Class
// var[1] = lastError
// var[2] = serial library reference number
// var[3] = read timeout in millis
// var[4] = port number
// var[5] = new serial mgr used
// var[6] = direct hardware access

//// new SerialPort by Mathias Ringwald - added in SW 2.0 beta 3

#define WOBJ_SerialPortLastErr(o)         (o)[1].intValue
#ifdef PALMOS
 #define WOBJ_SerialPortRefNum(o)            (o)[2].intValue
#else
 #define WOBJ_SerialPortRefNum(o)            (o)[2].obj
#endif
#define WOBJ_SerialPortTimeout(o)         (o)[3].intValue
#define WOBJ_SerialPortNumber(o)          (o)[4].intValue
#define WOBJ_SerialPortNewSerialMgr(o)   (o)[5].intValue
#define WOBJ_SerialPortDirectHardware(o)  (o)[6].intValue

//
// Socket
//
// var[0] = Class
// var[1] = lastError
// var[2] = refreshBeforeEachRead
// var[3] = _isOpen (desktop) / SocketRef (device)
// var[4] = _socket (desktop) / SocketTimeOut (device)
//

#define WOBJ_SocketLastErr(o) (o)[1].intValue
#define WOBJ_SocketRefreshBeforeEachRead(o) (o)[2].intValue // guich@340_37
#define WOBJ_SocketRef(o)     (o)[3].intValue
#define WOBJ_SocketTimeout(o) (o)[4].intValue

//
// File
//
// var[0] = Class
// var[1] = String name
// var[2] = int mode
// var[3] = handle

#define WOBJ_FileName(o)       (o)[1].obj
#ifdef PALMOS
 #define WOBJ_FileHandle(o)     (o)[2].intValue
#else
 #define WOBJ_FileHandle(o)     (o)[2].refValue
#endif
#define WOBJ_FileSlot(o)       (o)[3].intValue // guich@552_15
#define WOBJ_FileLastError(o)  (o)[4].intValue

//
// Surface
//

//#define SURF_MAINWIN 0 // guich@200: one more type - 0 now means invalid
#define SURF_CONTROL 1
#define SURF_IMAGE   2

//
// Image
//
// var[0] = Class
// var[1] = int width
// var[2] = int height
// var[3] = int transparentPixel /* petrus@402_01 */
// var[4] = int[] palRGBs
// var[6] = int pixels[][] (desktop) / GrHandle (device)

#define WOBJ_ImageWidth(o)       (o)[1].intValue
#define WOBJ_ImageHeight(o)      (o)[2].intValue
#define WOBJ_ImageTranspPixel(o) (o)[3].intValue /* define transparency in pixel unit (index or true color) rather than waba.fx.Color - petrus@402_01 */
#define WOBJ_ImagePalette(o)     (o)[4].obj
//#define WOBJ_ControlImageGraphicsGrHandle(o)    (o)[6].refValue // guich@402_28: now synchronized with WOBJ_GraphicsGrHandle

//
// Color
//
// var[0] = Class
// var[1] = int red
// var[2] = int green
// var[3] = int blue
// var[4] = int rgb   - only if device has color
// var[5] = int gray  - only if device is mono
// var[6] = int equ
//

#define WOBJ_ColorRed(o)      (o)[1].intValue
#define WOBJ_ColorGreen(o)    (o)[2].intValue
#define WOBJ_ColorBlue(o)     (o)[3].intValue
#define WOBJ_ColorRGB(o)      (o)[4].intValue
#define WOBJ_ColorGray(o)     (o)[5].intValue
#define WOBJ_ColorEqu(o)      (o)[6].intValue
#define WOBJ_ColorRedP(o)      o[1].intValue
#define WOBJ_ColorGreenP(o)    o[2].intValue
#define WOBJ_ColorBlueP(o)     o[3].intValue

//
// Font
//
// var[0] = Class
// var[1] = String name
// var[2] = int size
// var[3] = int style
// var[4] = UserFont structure
//

#define WOBJ_FontName(o)     (o)[1].obj
#define WOBJ_FontStyle(o)    (o)[2].intValue
#define WOBJ_FontSize(o)     (o)[3].intValue
#define WOBJ_FontUserFont(o) (o)[4].refValue

#define Font_PLAIN           0
#define Font_BOLD            1
#define SWSTANDARD_FONT      0
#define SWSTANDARD_BOLD_FONT 1
#define SWLARGE_FONT         2
#define SWLARGE_BOLD_FONT    3

//
// FontMetrics    - guich@200b2: now uses user defined fonts
//
// var[0] = Class
// var[1] = Font
// var[2] = int ascent
// var[3] = int descent
// var[4] = int leading
//

#define WOBJ_FontMetricsFont(o)    (o)[1].obj
#define WOBJ_FontMetricsAscent(o)  (o)[2].intValue
#define WOBJ_FontMetricsDescent(o) (o)[3].intValue
#define WOBJ_FontMetricsLeading(o) (o)[4].intValue

//
// Graphics of SuperWaba 2.0
//
// var[0] = Class
// var[1] = lastFont
// var[2] = drawing op
// var[3] = Surface
// var[4] = tempColor
// var[5] = SurfType
// var[6] = GrHandle
// var[7] = RGB of ForeColor
// var[8] = RGB of BackColor
// var[9] = Font ID
//
// font drawOp foreColor backColor transX, transY, clipRect, winRect font

#define WOBJ_GraphicsLastFont(o) (o)[1].obj
#define WOBJ_GraphicsDrawOp(o)   (o)[2].intValue
#define WOBJ_GraphicsSurface(o)  (o)[3].obj
#define WOBJ_GraphicsSurfType(o) (o)[5].intValue
#define WOBJ_ControlImageGraphicsGrHandle(o) (o)[6].refValue // guich@402_28: now its always stored in the same position
#define WOBJ_GraphicsForeRGB(o)  (o)[7].intValue
#define WOBJ_GraphicsBackRGB(o)  (o)[8].intValue
#define WOBJ_GraphicsFont(o)     (o)[9].refValue

//
// Catalog
//
// var[0] = Class
// var[1] = bag (desktop) / DmOpenRef (device - null if no open db)
// var[2] = current record pos (or -1 if none)
// var[3] = VoidHand current record handle
// var[4] = VoidPtr current record pointer
// var[5] = current offset in record
// var[6] = length of current record
// var[7] = 1 if record has been modified, 0 otherwise
// var[8] = 1 if some record has been modified; used to set the backup bit of the database - guich@200
#define WOBJ_CatalogDmRef(o)          (o)[1].refValue
#define WOBJ_CatalogCurRecPos(o)      (o)[2].intValue
#define WOBJ_CatalogCurRecHandle(o)   (o)[3].refValue
#define WOBJ_CatalogCurRecPtr(o)      (o)[4].refValue
#define WOBJ_CatalogCurRecOffset(o)   (o)[5].intValue
#define WOBJ_CatalogCurRecLen(o)      (o)[6].intValue
#define WOBJ_CatalogCurRecModified(o) (o)[7].intValue
#define WOBJ_CatalogSetBackup(o)      (o)[8].intValue

#define Catalog_READ_ONLY  1
#define Catalog_WRITE_ONLY 2
#define Catalog_READ_WRITE 3
#define Catalog_CREATE     4

#define MAKE32(l,u) (int32)(((((int32)(u)) << 16) & 0xFFFF0000) + (((int32)(l)) & 0x0000FFFF)) // guich@200b4_68
#define UPPER16(u) (((u) >> 16) & 0x0000FFFF)
#define LOWER16(l) ((uint16)(l)) // changed! removed & 0xFFFF
#define LOWER8(l) ((byte)(l))

#define NET_NOT_READY_FOR_OPEN 0
#define NET_READY_FOR_OPEN     1
#define NET_IS_OPEN            2
#define NET_OPEN_FAILED        3

// used in devices that support virtual keyboard
#define SIP_RESTORE_STATE -1
#define SIP_UNUSED 0
// values for first parameter of showSip (on)
#define SIP_TURN_ON 1
#define SIP_TURN_OFF 0
#define SIP_TO_DEFAULT -2
// values for second parameter of showSip (ontop)
#define SIP_PLACE_ON_BOTTOM 0
#define SIP_PLACE_ON_TOP 1

#define GET_RED(rgb)      (LOWER8(rgb >> 16))
#define GET_GREEN(rgb)    (LOWER8(rgb >>  8))
#define GET_BLUE(rgb)     (LOWER8(rgb))

// this is kinda hacky, but better than to change user fonts
#define IS_UNICODE_FONT(uf) ((uf->uff) > (struct UffUserFont *)1)  // pgr@403_01
#define IS_ANTIALIASED(uf) ((uf->uff) == (struct UffUserFont *)1)
#define UF_MCS_WIDTH (uf->fontP->maxWidth>>1) // guich@402_55

/// Pierre, Alsan and Guich added support for CJK @402_50
struct CharBits         // pgr@402_06 - describe the bitmap for a given character
{
   int32 rowWIB;          // width in bytes
   byte *bitmapTable;
   int32 offset;          // offset relative to the bitmap table
   int32 width;
};

// holds the bitmap info for each uff range
struct UserFontRange
{
   int32 rowWIB;
   byte* bitmapTable; // glyph pixel data
   int32* bitIndexTable; // where to find one char in the bitmap soup -  guich: not sure if this can be an uint16
   VoidHand recH;
};

struct UffUserFont // extends UserFont
{
   // the ranges
   uint16 nParts; // number of parts following (without MCS)
   uint16* recNo; // record number, where is the range?
   uint16* uffFirstChar; // encoding of first char in this range
   uint16* uffLastChar; // encoding of last char in this range
   struct UserFontRange* ranges; // the bitmap
   // cache
   int32 recentRangeIndex;
   uint16 recentFirstChar;
   uint16 recentLastChar;
};

struct UserFont
{
   char name[41]; // guich@340_47: grewed from 32 to 41 to be able to store a font name with its style
   DmOpenRef dmRef;
   VoidHand recH;
   uint16 fid;    // no PalmOS FontID (enum), bc the font id is an array index (entries 0-3 are the system fonts, internal or not)
   FontPtr fontP;
   uint16 recIndex;
   uint16 rowWidthInBytes;
   int32 numberOfChars;
   byte  *bitmapTable;
   int16 *bitIndexTable;
   struct UffUserFont *uff; // guich@402_50
};

void uffSetCharBits(char16 ch, struct UserFont *uf, struct CharBits *bits);

#ifdef __DEFINE_FUNCTIONS
   // nm<platform>_defines.h
   void *xmalloc(uint32 size);
   void *xrealloc(void *p, uint32 *newSize);
   void xfree_(void *p);

   // nm<platform>_bridge.c
   char *int2str(char *msg, int32 ii);
   void debug(char *msg); // guich: to debug the vm
   void showCommand(int32 show); // guich  // ds@110 change Boolean to int to run on Win32
   void tone(int32 freq, int32 dur);

   // superwaba.c
   void dumpStackTrace();
   uint32 locateSourceLine(WClassMethod *method, uint16 pc); // fdie@500_13
   void gc();
   int32 gc_(); // also calls the deleteUnusedBlocks
#ifndef WIN32
_inline_
#endif
   WObject createArrayObject(int32 type inD0, int32 len inD1);
   int32 pushObject(WObject obj);
   WObject popObject();
   int32 arrayRangeCheck(WObject array, int32 start, int32 count);
   void swQuickError(uint16 errNum, char *str);
   UtfString getClassName(WObject obj);
   WClass * getClass(UtfString className, byte throwEx);
   int32 compatible(WClass *wclass inA0, WClass *target inA1);
   Var* executeMethod(WClass *wclass, WClassMethod *method,Var params[], uint32 numParams);
   WClassMethod *getMethodByMapNum(WClass *wclass inA0, WClass **vclass inA1, uint16 mapNum inD0);
   int16 getMethodMapNum(WClass *wclass, UtfString *name, UtfString *desc, byte searchType, WClassMethod **imethod, WClass **iclass);
   uchar *loadClassField(WClass *wclass, WClassField *field, uchar *p);
   Var constantToVar(WClass *wclass inA0, uint16 idx inD0);
   uchar *loadClassMethod(WClass *wclass, WClassMethod *method, uchar *p);
#ifndef WIN32
_inline_
#endif
   WObject createObject(WClass *wclass);
   int32 arraySize(int32 type, int32 len);
   uint16 arrayType(char c);
   WObject createMultiArray(int32 ndim, char *desc, Var *sizes);
   void copyArray(Var stack[]);
   WClassField *getField(WClass *wclass, UtfString *name, UtfString *desc);
   WClass *getClassByIndex(WClass *wclass inA0, uint16 classIndex inD0, byte throwEx inD1);
   WClassMethod *getMethod(WClass *wclass, UtfString name, UtfString desc, WClass **vclass);
   int32 countMethodParams(UtfString desc);
   int32 compatibleArray(WObject obj, UtfString arrayName);
   WObject allocObject(int32 size);
   char *expandArray(char *array, int32 itemSize, int32 increment, int32 *insertIndex);
   int32 arrayCount(char *array, int32 itemSize);
   byte callMethod(WObject classObj, char *methodName, char *methodDesc, Var *params, int32 paramCount);
   WObject throwException(char *exceptionClass inA0, char *message inA1);

   #define xfree(p) {xfree_(p); p = NULL;} //void xfree(void *p);
#else
   #define xfree(p) {VM_CALL(xfree_)(p); p = NULL;} //void xfree(void *p);
#endif // __DEFINE_FUNCTIONS

// Graphics.c
typedef struct
{
  byte *             memAddr; // 2 used if not internal. note: even if the bitmaps are not handled directly, this must be different of null
  uint16             widthInBPP; // 1
  VoidPtr            winHandle; // 3 used if is internal. note: this can be a WinHandle if in PalmOS or a HWnd if in Windows CE. just cast the pointer to what it should be
  VoidPtr            bmpHandle; // 4 a bitmap handle if this is a offscreen window.  on SDL this is just a flag.
  RectangleType      windowBounds; // 5
  AbsRectType        clippingBounds; // 6
  int32              paletteSize; // 7
  VoidPtr            palette;    // 8 the palette itself
  VoidPtr            bmpHandleAux; // guich@340_47
} GraphicsContext;

typedef GraphicsContext * GrHandle;

typedef void (*DrawRowProcType)(int32 xStart, int32 xEnd, int32 y, uchar *rgbOrIndex, uint32 *colorMap);

// nm<platform>_defines.h
typedef void * (*xmallocProcType)(uint32 size);
typedef void * (*xreallocProcType)(void *p, uint32 *newSize);
typedef void (*xfree_ProcType)(void *p);
// nm<platform>_b.c
typedef int32 (*eventIsAvailableProcType)();
typedef int32 (*pumpEventsProcType)(int32 eventWait);
typedef int32 (*getTimeStampProcType)();
typedef uint32 (*millisToTicksProcType)(int32 millis);
typedef void (*postStopEventProcType)(int32 exitCode);
typedef char * (*int2strProcType)(char *msg, int32 ii);
typedef void (*debugProcType)(char *msg); // guich: to debug the vm
typedef void (*showCommandProcType)(int32 show); // guich  // ds@110 change Boolean to int to run on Win32
typedef void (*toneProcType)(int32 freq, int32 dur);
typedef uchar * (*lockWarpRecProcType)(char *path, uint16 pathLen, uint32 *size);
typedef void (*unlockWarpRecProcType)(uchar *ptr, uint16 pathLen);
typedef void (*executePostEventProcType)(int32 type, int32 key, int32 x, int32 y, int32 modifiers, int32 timeStamp);

// xplat_javalang.c
typedef void (*standardNMCallProcType)(Var stack[]);
typedef void (*unsigned2hexProcType)(int32 b, int32 places, char *c);
typedef int32 (*toLUCaseProcType)(int32 c);

// superwaba.c
typedef void (*doubleToStrProcType)(char dest[], double val, int32 totalLength, int32 decimal, int32 options);
typedef WObject (*float2strProcType)(float f);
typedef WObject (*int642strProcType)(int64 i);
typedef WObject (*double2strProcType)(double d, int32 places);
typedef int32 (*str2int32ProcType)(char *str);
typedef double (*str2doubleProcType)(char *str);
typedef char *(*intToStringProcType)(int32 i);
typedef int64 (*str2int64ProcType)(char *str);

typedef UtfString16 (*stringToUtf16ProcType)(WObject string);
typedef void (*dumpStackTraceProcType)();
typedef WObject (*createStringProcType)(char *buf);
typedef UtfString (*stringToUtfProcType)(WObject str, byte flags);
typedef void (*gcProcType)();
typedef byte (*callMethodProcType)(WObject classObj, char *methodName, char *methodDesc, Var *params, int32 paramCount);
typedef WObject (*createArrayObjectProcType)(int32 type inD0, int32 len inD1);
typedef int32 (*pushObjectProcType)(WObject obj);
typedef WObject (*popObjectProcType)();
typedef int32 (*arrayRangeCheckProcType)(WObject array, int32 start, int32 count);
typedef void (*VmQuickErrorProcType)(uint16 errNum, char *str);
typedef WObject (*throwExceptionProcType)(char *exceptionClass inA0, char *message inA1);
typedef WObject (*createStringFromCharPtrProcType)(char *s, int32 len);
typedef WObject (*createStringObjectProcType)();
typedef UtfString (*getClassNameProcType)(WObject obj);
typedef WClass * (*getClassProcType)(UtfString className, byte throwEx);
typedef int32 (*compatibleProcType)(WClass *wclass inA0, WClass *target inA1);
typedef Var* (*executeMethodProcType)(WClass *wclass, WClassMethod *method,Var params[], uint32 numParams);
typedef WClassMethod * (*getMethodByMapNumProcType)(WClass *wclass inA0, WClass **vclass inA1, uint16 mapNum inD0);
typedef int16 (*getMethodMapNumProcType)(WClass *wclass, UtfString *name, UtfString *desc, byte searchType, WClassMethod **imethod, WClass **iclass);
typedef void (*printToBufProcType)(char *buf, int32 maxbuf, UtfString *s1, UtfString *s2);
typedef uchar * (*nativeLoadClassProcType)(UtfString className, uint32 *size);
typedef uchar * (*loadClassFieldProcType)(WClass *wclass, WClassField *field, uchar *p);
typedef Var (*constantToVarProcType)(WClass *wclass inA0, uint16 idx inD0);
typedef uchar * (*loadClassMethodProcType)(WClass *wclass, WClassMethod *method, uchar *p);
typedef UtfString (*createUtfStringProcType)(char *buf);
typedef UtfString (*getUtfStringProcType)(WClass *wclass inA0, uint16 idx inD0);
typedef WObject (*createObjectProcType)(WClass *wclass);
typedef int32 (*arraySizeProcType)(int32 type, int32 len);
typedef uint16 (*arrayTypeProcType)(char c);
typedef WObject (*createMultiArrayProcType)(int32 ndim, char *desc, Var *sizes);
typedef WObject (*createStringFromUtfProcType)(UtfString s);
typedef void (*copyArrayProcType)(Var stack[]);
typedef WClassField * (*getFieldProcType)(WClass *wclass, UtfString *name, UtfString *desc);
typedef WClass * (*getClassByIndexProcType)(WClass *wclass inA0, uint16 classIndex inD0, byte throwEx inD1);
typedef WClassMethod * (*getMethodProcType)(WClass *wclass, UtfString name, UtfString desc, WClass **vclass);
typedef int32 (*countMethodParamsProcType)(UtfString desc);
typedef int32 (*compatibleArrayProcType)(WObject obj, UtfString arrayName);
typedef WObject (*allocObjectProcType)(int32 size);
typedef char * (*expandArrayProcType)(char *array, int32 itemSize, int32 increment, int32 *insertIndex);
typedef int32 (*arrayCountProcType)(char *array, int32 itemSize);
typedef void (*showSipProcType)(int16 on, int16 ontop);

#if !defined DEPRECATED_PALMOS
typedef int32 (*ascii2unicodeProcType)(TCHAR *dst, char const * src, int32 max);
typedef int32 (*unicode2asciiProcType)(char * dst, TCHAR const * src, int32 max);
typedef void (*registerHotkeysProcType)();
typedef void (*unregisterHotkeysProcType)();
typedef WObject (*createStringFromUnicodeProcType)(TCHAR const *s, uint32 len);
#define swapUInt32(n) SDL_SwapBE32(n)
#define swapUInt16(n) SDL_SwapBE16(n)
#define swapInt32(n) SDL_SwapBE32((uint32)(n))
#define swapInt16(n) SDL_SwapBE16((uint16)(n))

#if defined(__SYMBIAN32__) || (defined(PALMOS) && defined(__arm__))
typedef uint32 LibRefnum;
#else
typedef uint16 LibRefnum;
#endif

typedef Err (*SWLibOpenProcType)(LibRefnum inRefNum, void *inGlobalsPtr); // guich@502_11: now all lib types must include the librefnum in all methods
typedef Err (*SWLibCloseProcType)(LibRefnum inRefnum);
typedef Err (*SWInvokeNativeProcType)(LibRefnum inRefnum, int32 hash, Var stack[]);
typedef Err (*SWCallPublicProcType)(LibRefnum inRefnum, int32 index, uint32 *ret, uint32 *args);
typedef Err (*SWHandleEventProcType)(LibRefnum inRefnum, EventPtr eventP);
typedef Err (*SWGetNativeFuncProcType)(LibRefnum inRefnum, int32 hash, NativeFunc *func);
typedef Err (*SWInvokeNativeFuncProcType)(LibRefnum inRefnum, NativeFunc func, Var stack[]);

typedef void (*LogDebugProcType)(LogCategory category, const char *format, ...);
typedef void (*LogInfoProcType)(LogCategory category, const char *format, ...);
typedef void (*LogErrorProcType)(LogCategory category, const char *format, ...);

#if defined(PALMOS_NATIVES_SYSLIB)
   typedef LibRefnum NativeLibPtr;
#else
typedef struct
{
   SWLibOpenProcType SWLibOpen;
   SWLibCloseProcType SWLibClose;
   SWCallPublicProcType SWCallPublic;
   SWHandleEventProcType SWHandleEvent;
   SWInvokeNativeProcType SWInvokeNative;
   SWGetNativeFuncProcType SWGetNativeFunc;
   SWInvokeNativeFuncProcType SWInvokeNativeFunc;
   LibRefnum ref;
} NativeLib, *NativeLibPtr;
#endif

typedef void (*GraphicsStartImageProcType)(uint32 *ret, uint32 *args);
typedef void (*GraphicsEndImageProcType)(GrHandle oldG);
struct SDL_Surface;  // required by gcc
typedef GrHandle (*MakeGrHandleProcType)(struct SDL_Surface * surface, int32 videoFlag);


struct VMGlobals
{
   NativeLibPtr *nativeLibs; // guich@500_5: since now returnedValue was removed, I ellected nativeLibs to be the first one of the structure
   int stateX,stateY;
   uint16 globalSocketLibRefNum;

   struct UserFont *allFonts;
   uint32 romVersion; // palm os 2 = 0x02000000 - DWord
   // guich@200: now these vars are globals. (taken from executeMethod)
   WObject globalMainWin;
   byte errorWinDisplayed;
   byte globalNetState;// = NET_NOT_READY_FOR_OPEN = 0;
   int32 globalTimerInterval;
   int32 globalTimerStart;
   DmOpenRef *classDbList;
   uint32 appCreatorId;
   int32 postEventMethodMapNum;
   int32 onTimerTickMethodMapNum;
   char *commandLine;
   uint32 interceptSystemKey;
   uint16 defSoundAmp; // guich@200b3 - UInt16
   ClassHook *classHooks;
   // systemOut
   #define SYSOUT_CLOSED 0
   #define SYSOUT_NOTFOUND 1
   #define SYSOUT_OPENED 2
   #define SYSOUT_COULDNOTCREATE 3
   #define SYSOUT_ERASED 4 // guich@420_18
   byte stateSystemOut;
   byte isFullScreen;
   byte isNetStarted;       // int win32WSAStarted;
   byte useMinimizeButton;
#ifdef PALMOS
   unsigned short posSystemOut; // index of record being appended
   DmOpenRef dmRefSystemOut;
#endif
   char * appTitle;
   void * mainHWnd;
   char * appPath;
   TCHAR * swPath;
 #ifdef WIN32
   RECT sipRect;            // bounds of the SIP
   int32 taskBarHeight;
 #endif
   void * hSystemOut;
   int desiredX,desiredY; // guich@400_62
#endif

   uint32 maxMemoLen;
   int threadCount; // guich@200b4_84
   // guich@200b4_140
   VoidPtr evtGetPenTrapPtr;
   VoidPtr sysTrapWinSetActiveWindowPtr;
   VoidPtr sysTrapFrmCustomAlertPtr;
   uint16 SWLibRefNum;  // UInt
   //
   int screenWidth,screenHeight;
   int windowWidth,windowHeight; // excludes the height of the task bar
   uint32 maxColors;

   byte reboot; // guich@401_42
   byte isSony;
   byte audibleGC;
   byte dumpMemStats;
   uint16 density;
   byte wants16bpp; // guich@421_50 - palmos5: we normally force 8bpp b/c of buggy colors
   byte quitRightNow; // guich@504_2: before the main even loop starts, if the user calls exit, we have to find another way to quit the vm.

   char *className;
   uint32 screenBPP; // guich@421_33: moved to here (from graphics.c)
   WClass *prefClass; // guich@230_22

   // methods

   // nm<platform>_defines.h
   xmallocProcType xmalloc;
   xreallocProcType xrealloc;
   xfree_ProcType xfree_;
   // nm<platform>_bridge.c
   eventIsAvailableProcType eventIsAvailable;
   pumpEventsProcType pumpEvents;
   dumpStackTraceProcType dumpStackTrace;
   getTimeStampProcType getTimeStamp;
   millisToTicksProcType millisToTicks;
   postStopEventProcType postStopEvent;
   int2strProcType int2str;
   debugProcType debug;
   showCommandProcType showCommand;
   toneProcType tone;
   lockWarpRecProcType lockWarpRec;
   unlockWarpRecProcType unlockWarpRec;
   executePostEventProcType executePostEvent;
   // xplat_javalang.c
   standardNMCallProcType StringToLowerCase;
   standardNMCallProcType StringToUpperCase;
   standardNMCallProcType StringBufferEnsureCapacity;
   standardNMCallProcType StringBufferSetLength;
   standardNMCallProcType appendString;
   standardNMCallProcType appendCharArray;
   standardNMCallProcType appendCharArrayOffLen;
   standardNMCallProcType appendChar;
   standardNMCallProcType appendInt;
   standardNMCallProcType appendFloat;
   standardNMCallProcType appendDouble;
   standardNMCallProcType valueOfDouble;
   standardNMCallProcType StringCopyChars;
   standardNMCallProcType StringEquals;
   standardNMCallProcType StringCompareTo;
   standardNMCallProcType indexOfChar;
   standardNMCallProcType indexOfCharStart;
   standardNMCallProcType StringIndexOf;
   standardNMCallProcType StringHashCode;
   standardNMCallProcType ObjectToStringNative;
   standardNMCallProcType ObjectGetClass;
   standardNMCallProcType ThrowablePrintStackTraceNative;
   standardNMCallProcType ClassForName;
   standardNMCallProcType ClassNewInstance;
   standardNMCallProcType ClassIsInstance;
   // superwaba.c
   doubleToStrProcType doubleToStr;
   float2strProcType float2str;
   int642strProcType int642str;
   double2strProcType double2str;
   str2int32ProcType str2int32;
   str2doubleProcType str2double;
   intToStringProcType intToString;
   str2int64ProcType str2int64;

   ErrorStatus *vmStatus;
   stringToUtf16ProcType stringToUtf16;
   createStringProcType createString;
   stringToUtfProcType stringToUtf;
   gcProcType gc;
   callMethodProcType callMethod;
   createArrayObjectProcType createArrayObject;
   pushObjectProcType pushObject;
   popObjectProcType popObject;
   arrayRangeCheckProcType arrayRangeCheck;
   VmQuickErrorProcType VmQuickError;
   throwExceptionProcType throwException;
   createStringFromCharPtrProcType createStringFromCharPtr;
   createStringObjectProcType createStringObject;
   getClassNameProcType getClassName;
   getClassProcType getClass;
   compatibleProcType compatible;
   executeMethodProcType executeMethod;
   getMethodByMapNumProcType getMethodByMapNum;
   getMethodMapNumProcType getMethodMapNum;
   printToBufProcType printToBuf;
   nativeLoadClassProcType nativeLoadClass;
   loadClassFieldProcType loadClassField;
   constantToVarProcType constantToVar;
   loadClassMethodProcType loadClassMethod;
   createUtfStringProcType createUtfString;
   getUtfStringProcType getUtfString;
   createObjectProcType createObject;
   arraySizeProcType arraySize;
   arrayTypeProcType arrayType;
   createMultiArrayProcType createMultiArray;
   createStringFromUtfProcType createStringFromUtf;
   copyArrayProcType copyArray;
   getFieldProcType getField;
   getClassByIndexProcType getClassByIndex;
   getMethodProcType getMethod;
   countMethodParamsProcType countMethodParams;
   compatibleArrayProcType compatibleArray;
   allocObjectProcType allocObject;
   expandArrayProcType expandArray;
   arrayCountProcType arrayCount;
   showSipProcType showSip;

   registerHotkeysProcType registerHotkeys;
   unregisterHotkeysProcType unregisterHotkeys;
   ascii2unicodeProcType ascii2unicode;
   unicode2asciiProcType unicode2ascii;
   createStringFromUnicodeProcType createStringFromUnicode;

   // used in Native Libraries
   DmCloseDatabaseProcType _DmCloseDatabase;
   DmCreateDatabaseProcType _DmCreateDatabase;
   DmOpenDatabaseProcType _DmOpenDatabase;
   DmOpenDatabaseReadOnlyProcType _DmOpenDatabaseReadOnly;
   DmRenameDatabaseProcType _DmRenameDatabase;
   DmDeleteDatabaseProcType _DmDeleteDatabase;
   DmListDatabasesProcType _DmListDatabases;
   DmListDatabasesInProcType _DmListDatabasesIn;
   DmNumRecordsProcType _DmNumRecords;
   DmWriteProcType _DmWrite;
   DmReleaseRecordProcType _DmReleaseRecord;
   DmNewRecordProcType _DmNewRecord;
   DmResizeRecordProcType _DmResizeRecord;
   DmRemoveRecordProcType _DmRemoveRecord;
   DmGetRecordProcType _DmGetRecord;
   DmQueryRecordProcType _DmQueryRecord;
   DmQueryRecordPartProcType _DmQueryRecordPart;
   DmGetRecordAttributesProcType _DmGetRecordAttributes;
   DmSetRecordAttributesProcType _DmSetRecordAttributes;
   DmSetAttributesProcType _DmSetAttributes;
   DmGetAttributesProcType _DmGetAttributes;
   DmSetBackupAttributeProcType _DmSetBackupAttribute;
   DmGetRecordSizeProcType _DmGetRecordSize;
   DmGetDatabaseTypeCreatorCountProcType _DmGetDatabaseTypeCreatorCount;
   DmGetNameProcType _DmGetName;
   DmFindAndOpenDBProcType findAndOpenDB;

   splitPathProcType _splitPath;
   GraphicsStartImageProcType graphicsStartImage;
   GraphicsEndImageProcType graphicsEndImage;
   MakeGrHandleProcType makeGrHandle;

   toLUCaseProcType toLowerCase;
   toLUCaseProcType toUpperCase;
   unsigned2hexProcType unsigned2hex;

   LogDebugProcType logDebug;
   LogInfoProcType  logInfo;
   LogErrorProcType logError;
   
   // add new variables to here (end of struct), to maintain compatibility with old versions
   uint16 tempLibInsertIndex; // guich@551_8: a hack to avoid a worst hack
};

extern struct VMGlobals *vmGlobals; // setted by the SWInvokeNative

// nm<platform>_xxx.c
// void SerialPortDestroy(WObject port);
// void CatalogDestroy(WObject cat);
// void SocketDestroy(WObject socket);
// void FileDestroy(WObject file);
void systemOutClose();
// graphics.c
// void ImageDestroy(WObject image);
#define TRANSX ((int16)LOWER16(transXY))     // guich@200b4_122
#define TRANSY ((int16)UPPER16(transXY))
#define CLIPX1 ((int16)LOWER16(clipX))
#define CLIPX2 ((int16)UPPER16(clipX))
#define CLIPY1 ((int16)LOWER16(clipY))
#define CLIPY2 ((int16)UPPER16(clipY))

#define VIDEOFLAG_MAINSCREEN 0
#define VIDEOFLAG_COMPATIBLE_OFFSCREEN 1   // same Pixel format as main screen
#define VIDEOFLAG_DIFFERENT_OFFSCREEN 2    // Pixel format differs from main screen

#define GRHANDLE_WidthInBPP(h) (h->widthInBPP)
#define GRHANDLE_MemAddr(h) ((byte *)h->memAddr)
#define GRHANDLE_WinHandle(h) (h->winHandle)
#define GRHANDLE_VideoFlag(h) ((uint32)(h->bmpHandle))
#define GRHANDLE_BmpHandleAux(h) (h->bmpHandleAux)
#define GRHANDLE_TransX(h) (h->windowBounds.topLeft.x)
#define GRHANDLE_TransY(h) (h->windowBounds.topLeft.y)
#define GRHANDLE_Width(h) (h->windowBounds.extent.x)
#define GRHANDLE_Height(h) (h->windowBounds.extent.y)
#define GRHANDLE_ClipX1(h) (h->clippingBounds.left)
#define GRHANDLE_ClipY1(h) (h->clippingBounds.top)
#define GRHANDLE_ClipX2(h) (h->clippingBounds.right)
#define GRHANDLE_ClipY2(h) (h->clippingBounds.bottom)
#define GRHANDLE_PaletteSize(h) (h->paletteSize)
#define GRHANDLE_Palette(h) (h->palette)
#define GRHANDLE_IsBitmap(h) (h->bmpHandle != NULL)

#define WIN_PAINT         0  // Destination replaced with source pixels (copy mode).
#define WIN_ERASE         1  // Destination cleared where source pixels are off (AND mode).
#define WIN_MASK          2  // Destination cleared where source pixels are on (AND NOT mode).
#define WIN_INVERT        3  // Destination inverted where source pixels are on (XOR mode).
#define WIN_OVERLAY       4  // Destination set only where source pixels are on (OR mode).
#define WIN_PAINT_INVERSE 5  // Destination replaced with inverted source (copy NOT mode).
#define WIN_SPRITE        6  // Destination replaced if source pixels != color (simulating transparent color - added by guich)
#define WIN_REPLACE_COLOR 7  // Destination replaced with <color> if source pixels != 0
#define WIN_SWAP_COLORS   8  // Destination replaced with <color> if source pixels != 0

///////////////////////////////////////////////////

#ifdef WIN32
 #define DIRSEP (TCHAR)'\\'
#else
 #define DIRSEP (TCHAR)'/'
#endif

// standard interface to call native methods
#if defined(BUILDING_CUSTOM_LIB) || defined(PALMOS_SWNATIVES_SYSLIB)
  #include "SW_libDynam.h"
#else
  #include "SW_libStatic.h"
#endif

#if defined(PALMOS_NATIVES_SYSLIB)
  #include "palm/SW_libDynam.h"
#endif

#endif // __SUPERWABA_H
