/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

/*
    WinMain function calling SuperWaba's main() function
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <windows.h>
#include <malloc.h>     /* For _alloca() */

#if defined(WIN32_PLATFORM_PSPC)
#include <aygshell.h>                      // Add Pocket PC includes
#pragma comment( lib, "aygshell" )         // Link Pocket PC library
#endif

#include "SDL.h"
#include "SDL_main.h"                      // SDL main definition header
#include "SW_events.h"

#ifdef _WIN32_WCE
#include <notify.h>
#endif

#if defined main && !defined _WIN32_WCE_EMULATION
#undef main
#endif

#if defined(_MSC_VER) && !defined(_WIN32_WCE)
/* The VC++ compiler needs main defined */
int main(int argc, char *argv[]) { return 0; }
#endif

static int parseCommandLine(char * cmdline, char ** argv, int * argcTitle);

/*
*/
#ifdef _WIN32_WCE
// static char defaultCmdLine[] = "SuperWaba /t \"SuperWaba\" /F /w 240 /h 320 /p \\SuperWaba\\";
static char defaultCmdLine[] = "SuperWaba /t \"SuperWaba\" /p \\SuperWaba\\";
#else
static char defaultCmdLine[] = "SuperWaba /t \"SuperWaba\" /p .\\";
#endif

static char defaultParams[] = "SuperWaba /t \"SuperWaba\" /n SuperWaba.pdb ";
#define WND_CLASS_NAME "SuperWabaWndClass"

extern char *errorMessages[]; // guich@501_9: add a better error message

static void verifyCurDir() // guich@510_27: make sure that we're on a folder named SuperWaba
{
#ifndef WINCE
   TCHAR curDir[MAX_PATH+1];
   GetCurrentDirectory(sizeof curDir, curDir);
   if (strstr(curDir,"SuperWaba") == 0) 
      SetCurrentDirectory(TEXT("\\SuperWaba"));
#endif
}
/*--------------------------------------------------------------------WinMain-+
| For vmSDL, this is what the SuperWaba main requires:                        |
|                                                                             |
| /p  superwaba directory:  required.  must end with the directory separator  |
| /t  title: required.                                                        |
| /F  full screen:  optional.                                                 |
| application class name:  required.                                          |
| application class path:                                                     |
|    required.   must NOT end with the directory separator (legacy)           |
|    if it starts with '/', '\', or ?: (as in C:) it is an absolute           |
|    path.  otherwise, it is relative to the SW path.                         |
|                                                                             |
| Fortunately, we make the caller's life easier, filling in default           |
| values, so the user can just enter "Application Application CrId"           |
|                                                                             |
| - If no arguments are passed, we use the default command line above.        |
| - If arguments are passed they are catenated to the default params above,   |
| and the default SuperWaba directory is the current directory.               |
+----------------------------------------------------------------------------*/
#ifdef _WIN32_WCE
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPWSTR szCmdLine, int sw)
#else
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
#endif
{
   char * cmdLineCopy;
   char **argv;
   int argc;
   int argcTitle;
   int rc;
   /* Get the command line */
   if ((szCmdLine == 0) || (szCmdLine[0] == 0))
   {
      cmdLineCopy = defaultCmdLine;
#if defined(WIN32) && !defined(WINCE)
      verifyCurDir(); // guich@510_27
#endif
   }
   else
   {
      char * p;
      int cmdLen;
      int dirLen;

#ifdef _WIN32_WCE
      TCHAR curDir[] = TEXT("\\SuperWaba");
      cmdLen = wcstombs(0, szCmdLine, 0);
#else
      TCHAR curDir[MAX_PATH+1];
#if defined(WIN32) && !defined(WINCE)
      verifyCurDir(); // guich@510_27
#endif
      GetCurrentDirectory(sizeof curDir, curDir);
      cmdLen = strlen(szCmdLine);
#endif
#ifdef UNICODE
      dirLen = wcstombs(0, curDir, 0);
#else
      dirLen = strlen(curDir);
#endif
      cmdLineCopy = (char *)alloca(cmdLen + dirLen + sizeof defaultParams + 7);
      memcpy(cmdLineCopy, defaultParams, sizeof defaultParams-1);
      p = cmdLineCopy + sizeof defaultParams-1;
      *p++ = '/';
      *p++ = 'p';
      *p++ = ' ';
      *p++ = '\"';
#ifdef UNICODE
      wcstombs(p, curDir, dirLen);
#else
      memcpy(p, curDir, dirLen);
#endif
      p += dirLen;
      *p++ = '\\';
      *p++ = '\"';
      *p++ = ' ';
#ifdef _WIN32_WCE
      wcstombs(p, szCmdLine, cmdLen);
      p[cmdLen] = 0;
#else
      memcpy(p, szCmdLine, cmdLen+1);
#endif
   }

   /* Parse it into argv and argc */
   argc = parseCommandLine(cmdLineCopy, NULL, &argcTitle);
   argv = (char **)alloca(argc * (sizeof *argv));
   if (argv == 0)
      return 0;
   parseCommandLine(cmdLineCopy, argv, &argcTitle);

   /* For WINCE, allow only one copy of the program to run at a time */
   #ifdef _WIN32_WCE
   {
      HWND appWin;
      TCHAR title[50];
      // make sure the title exists! (avoid "/t" with nothing after)
      if (argcTitle == argc)
      {
         --argc;  // void the ending "/t"
         argcTitle = 2;  // back to the 2nd parameter of the default
      }
      mbstowcs(title, argv[argcTitle], sizeof title - 1);
      appWin = FindWindow(TEXT(WND_CLASS_NAME), title);
      if (appWin != NULL)
      {
         ShowWindow(appWin, SW_SHOWNORMAL);

         /* hide that $#!^%#$ WinCE taskbar in fullscreen... */
//       SHFullScreen(appWin, SHFS_HIDESTARTICON | SHFS_HIDETASKBAR /* | SHFS_HIDESIPBUTTON*/);
//       ShowWindow(FindWindow(TEXT("HHTaskBar"), NULL), SW_HIDE);
         SetActiveWindow(appWin);
         BringWindowToTop(appWin);
         SetForegroundWindow(appWin); // guich@400_74
         return 0;
      }
   }
   #endif

   /* Register our class */
   #ifndef CS_BYTEALIGNCLIENT
   #define CS_BYTEALIGNCLIENT 0
   #endif
   SDL_RegisterApp(WND_CLASS_NAME, CS_BYTEALIGNCLIENT, 0);


   /*
   | Start up DDHELP.EXE before opening any files, so DDHELP doesn't
   | keep them open.  This is a hack.. hopefully it will be fixed
   | someday.  DDHELP.EXE starts up the first time DDRAW.DLL is loaded.
   */
   {
      HINSTANCE handle = LoadLibrary(TEXT("DDRAW.DLL"));
      if (handle != NULL)
         FreeLibrary(handle);
   }

   /* Run the application main() code */
   rc = SDL_main(argc, argv);
   if (rc > 0)    // -1 is expected
   {
      TCHAR temp[400];
      #ifdef UNICODE
      wsprintf(temp, TEXT("SuperWaba error\n'%S'\nPlease notify this program's author.\nFor more information, see the DebugConsole.txt\nfile inside the program's folder."), errorMessages[rc-1]); // guich@501_9: show better error messages
      #else
      sprintf (temp, TEXT("SuperWaba error\n'%s'\nPlease notify this program's author.\nFor more information, see the DebugConsole.txt\nfile inside the program's folder."), errorMessages[rc-1]);
      #endif

      MessageBox(NULL, temp, TEXT("Program Error"), MB_ICONEXCLAMATION|MB_OK|MB_TOPMOST);
   }

   return swExitCode;
}

/*-----------------------------------------------------------parseCommandLine-+
| Parse a command line buffer into arguments                                  |
+----------------------------------------------------------------------------*/
static int parseCommandLine(char * cmdline, char ** argv, int * argcTitle)
{
   char *bufp;
   int argc;
   int isQuoted;

   argc = 0;
   for (bufp = cmdline; *bufp;)
   {
      /* Skip leading whitespace */
      while (*bufp <= ' ') // isspace does not exist on some platforms
         ++bufp;
      /* Skip over argument */
      isQuoted = *bufp == '"';
      if (isQuoted)
         ++bufp;
      if (*bufp)
      {
         if (argv)
         {
            argv[argc] = bufp;
            if ((bufp[0] == '/') && (bufp[1]== 't'))
               *argcTitle = argc+1;
         }
         ++argc;
      }

      if (isQuoted)         /* Skip over quoted */
      {
         while (*bufp && (*bufp != '"'))
            ++bufp;
      }
      else
      {                     /* Skip over word */
         while (*bufp && (*bufp > ' '))
            ++bufp;
      }
      if (*bufp)
      {
         if (argv) *bufp = '\0';
         ++bufp;
      }
   }
   return argc;
}

/*===========================================================================*/
