/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>, All Rights Reserved    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

// NOTE: General rule for native functions.. don't hold a pointer across any
// call that could garbage collect. For example, this isn't good:
//
// ptr = WOBJ_arrayStart(array)
// ...
// string = createString(..)
// ptr[0]
//
// since the createString() could GC, the ptr inside of array could be invalid
// after the call since a GC would move memory around. Instead, use:
//
// ptr = WOBJ_arrayStart(array)
// ...
// string = createString(..)
// ...
// ptr = WOBJ_arrayStart(array)
// ptr[0]
//
// to recompute the pointer after the possible GC

/********************************************************************
 * Private macros
 ********************************************************************/

#define BUILDING_SW_STATIC
#include "superwaba.h" // inside this header file, the PALMOS or WIN32 is defined
#include "native_methods.h"

#if defined(PALMOS)
	// If we're built for the device, then we need to provide for
	// setting up and restoring register A4 on entry to/exit from
	// the library routines. Otherwise we're being built for the
	// simulator, and thus are running as part of a normal 68K Mac
	// application (better not mess with A4 in that case).
	#define EnterLibrary_(refnum) long oldA4 = SetCurA4(refnum);
	#define ExitLibrary_() RestoreA4(oldA4);

	// now uses 3.5 headers with palm compability
	#include <Graffiti.h> // guich
	#include <SerialMgrOld.h> // Palm OS 3.5 Support
	#include "HsCreators.h" // Handspring Support
	#include "UsbMgr.h"// Handspring Support
	#include "HsExt.h"
#endif

#ifndef max
#define max(x,y) (((x)>=(y))?(x):(y))
#define min(x,y) (((x)>=(y))?(y):(x))
#endif
extern uint32 maxColors; // ps: 0 means 2^32
extern int MaxScreenX;
extern int MaxScreenY;
byte loadAllFonts();
void releaseAllFonts();
void graphicsStop();
void drawHourGlass(int i);
#ifndef PALMOS
void initMainGrHandle();
#endif
void freeGrResources();
void graphicsSwap(int i);
byte graphicsInit();
void graphicsReset();
void GraphicsDestroy(WObject obj);
void graphicsStartImage(uint32 * ret, uint32 * args);
void graphicsEndImage(GrHandle oldG);

/** guich@200b4: unfortunately, libraries cannot declare arrays such the old NativeMethods one, because the compiler give link errors. So, the funcs array will be built at runtime. */
uint32 nativeHashes[201] =
{
	483830802UL ,  // java/lang/StringBuffer_EnsureCapacity_(I)V -> (*vmGlobals->StringBufferEnsureCapacity)
	483796877UL ,  // java/lang/StringBuffer_SetLength_(I)V -> (*vmGlobals->StringBufferSetLength)
	483825906UL ,  // java/lang/StringBuffer_Append_(Ljava/lang/String;)Ljava/lang/StringBuffer; -> (*vmGlobals->appendString)
	483792994UL ,  // java/lang/StringBuffer_Append_([C)Ljava/lang/StringBuffer; -> (*vmGlobals->appendCharArray)
	483802340UL ,  // java/lang/StringBuffer_Append_([CII)Ljava/lang/StringBuffer; -> (*vmGlobals->appendCharArrayOffLen)
	483787169UL ,  // java/lang/StringBuffer_Append_(C)Ljava/lang/StringBuffer; -> (*vmGlobals->appendChar)
	483787553UL ,  // java/lang/StringBuffer_Append_(I)Ljava/lang/StringBuffer; -> (*vmGlobals->appendInt)
	483787361UL ,  // java/lang/StringBuffer_Append_(F)Ljava/lang/StringBuffer; -> (*vmGlobals->appendFloat)
	483787233UL ,  // java/lang/StringBuffer_Append_(D)Ljava/lang/StringBuffer; -> (*vmGlobals->appendDouble)
	2253423068UL,  // java/lang/String_ValueOf_(D)Ljava/lang/String; -> (*vmGlobals->valueOfDouble)
	2253423004UL,  // java/lang/String_ValueOf_(C)Ljava/lang/String; -> ConvertToString_Char
	2253423388UL,  // java/lang/String_ValueOf_(I)Ljava/lang/String; -> ConvertToString_Int
	2253423196UL,  // java/lang/String_ValueOf_(F)Ljava/lang/String; -> ConvertToString_Float
	2253429651UL,  // java/lang/String_CopyChars_([CI[CII)Z -> (*vmGlobals->StringCopyChars)
	2253444895UL,  // java/lang/String_ToUpperCase_()Ljava/lang/String; -> (*vmGlobals->StringToUpperCase)
	2253444703UL,  // java/lang/String_ToLowerCase_()Ljava/lang/String; -> (*vmGlobals->StringToLowerCase)
	2253417883UL,  // java/lang/String_Equals_(Ljava/lang/Object;)Z -> (*vmGlobals->StringEquals)
	2253437214UL,  // java/lang/String_CompareTo_(Ljava/lang/String;)I -> (*vmGlobals->StringCompareTo)
	2253450251UL,  // java/lang/String_IndexOf_(I)I -> (*vmGlobals->indexOfChar)
	2253454924UL,  // java/lang/String_IndexOf_(II)I -> (*vmGlobals->indexOfCharStart)
	2253427741UL,  // java/lang/String_IndexOf_(Ljava/lang/String;I)I -> (*vmGlobals->StringIndexOf)
	2253450827UL,  // java/lang/String_HashCode_()I -> (*vmGlobals->StringHashCode)
	2119183522UL,  // java/lang/Object_ToStringNative_()Ljava/lang/String; -> (*vmGlobals->ObjectToStringNative)
	2119199131UL,  // java/lang/Object_GetClass_()Ljava/lang/Class; -> (*vmGlobals->ObjectGetClass)
	3532867608UL,  // java/lang/Throwable_PrintStackTraceNative_()V -> (*vmGlobals->ThrowablePrintStackTraceNative)
	1712322092UL,  // java/lang/Class_ForName_(Ljava/lang/String;)Ljava/lang/Class; -> (*vmGlobals->ClassForName)
	1712313375UL,  // java/lang/Class_NewInstance_()Ljava/lang/Object; -> (*vmGlobals->ClassNewInstance)
	1712312095UL,  // java/lang/Class_IsInstance_(Ljava/lang/Object;)Z -> (*vmGlobals->ClassIsInstance)
	2182092503UL,  // waba/fx/Graphics_DrawDottedCursor_(IIII)V
	2182123060UL,  // waba/fx/Graphics_CopyImageRect_(Lwaba/fx/Image;IIIIILwaba/fx/Color;Z)V
	2182087522UL,  // waba/fx/Graphics_CopyScreen_(Lwaba/fx/ISurface;III)V
	2182087406UL,  // waba/fx/Graphics_DrawImage_(Lwaba/fx/Image;IIILwaba/fx/Color;Z)V -> GraphicsDrawImageSprite
	2182111963UL,  // waba/fx/Graphics_Create_(Lwaba/fx/ISurface;)V
	2182123986UL,  // waba/fx/Graphics_DrawEllipse_(IIII)V
	2182123538UL,  // waba/fx/Graphics_FillEllipse_(IIII)V
	2182100111UL,  // waba/fx/Graphics_DrawArc_(IIIFF)V
	2182100623UL,  // waba/fx/Graphics_DrawPie_(IIIFF)V
	2182100175UL,  // waba/fx/Graphics_FillPie_(IIIFF)V
	2182104986UL,  // waba/fx/Graphics_DrawEllipticalArc_(IIIIFF)V
	2182105498UL,  // waba/fx/Graphics_DrawEllipticalPie_(IIIIFF)V
	2182105050UL,  // waba/fx/Graphics_FillEllipticalPie_(IIIIFF)V
	2182111376UL,  // waba/fx/Graphics_DrawCircle_(III)V
	2182110928UL,  // waba/fx/Graphics_FillCircle_(III)V
	2182093709UL,  // waba/fx/Graphics_GetPixel_(II)I
	2182095309UL,  // waba/fx/Graphics_SetPixel_(II)V
	2182109776UL,  // waba/fx/Graphics_EraseRect_(IIII)V
	2182103119UL,  // waba/fx/Graphics_DrawLine_(IIII)V
	2182103950UL,  // waba/fx/Graphics_ClearScreen_()V
	2182104271UL,  // waba/fx/Graphics_DrawDots_(IIII)V
	2182118865UL,  // waba/fx/Graphics_DrawCursor_(IIII)V
	2182100440UL,  // waba/fx/Graphics_DrawCursorOutline_(IIII)V
	2182103503UL,  // waba/fx/Graphics_DrawRect_(IIII)V
	2182103055UL,  // waba/fx/Graphics_FillRect_(IIII)V
	2182142677UL,  // waba/fx/Graphics_DrawDottedRect_(IIII)V
	2182132179UL,  // waba/fx/Graphics_FillPolygon_([I[II)V
	2182132627UL,  // waba/fx/Graphics_DrawPolygon_([I[II)V
	2182139028UL,  // waba/fx/Graphics_DrawPolyline_([I[II)V
	2182090851UL,  // waba/fx/Graphics_GetAnglePoint_(IIIIF)Lwaba/fx/Coord;
	2182138655UL,  // waba/fx/Graphics_DrawText_(Ljava/lang/String;II)V
	2182115089UL,  // waba/fx/Graphics_DrawText_([CIIII)V -> GraphicsDrawTextChars
	2182093592UL,  // waba/fx/Graphics_DrawHatchedRect_(IIIIZZ)V
	2182093144UL,  // waba/fx/Graphics_FillHatchedRect_(IIIIZZ)V
	2182141461UL,  // waba/fx/Graphics_DrawRoundRect_(IIIII)V
	2182141013UL,  // waba/fx/Graphics_FillRoundRect_(IIIII)V
	2182096846UL,  // waba/fx/Graphics_SetClip_(IIII)V
	2182102117UL,  // waba/fx/Graphics_GetClip_(Lwaba/fx/Rect;)Lwaba/fx/Rect;
	2182139861UL,  // waba/fx/Graphics_Clip_(Lwaba/fx/Rect;)Z
	2182090124UL,  // waba/fx/Graphics_ClearClip_()V
	2182103502UL,  // waba/fx/Graphics_Translate_(II)V
	2182143327UL,  // waba/fx/Graphics_GetTranslation_()Lwaba/fx/Coord;
	2182088099UL,  // waba/fx/Graphics_CopyRect_(Lwaba/fx/ISurface;IIIIII)V
	2182124103UL,  // waba/fx/Graphics_Free_()V
	2182133470UL,  // waba/fx/Graphics_SetForeColor_(Lwaba/fx/Color;)V
	2182131742UL,  // waba/fx/Graphics_SetBackColor_(Lwaba/fx/Color;)V
	2182146423UL,  // waba/fx/Graphics_CollisionDetect_(Lwaba/fx/ISurface;IILwaba/fx/Image;II)Z
	2182138964UL,  // waba/fx/Graphics__applyPalette_([III)V
	2182151224UL,  // waba/fx/Graphics_DrawHighLightFrame_(IIIILwaba/fx/Color;Lwaba/fx/Color;Z)V
	2182094808UL,  // waba/fx/Graphics_SetFont_(Lwaba/fx/Font;)V
	51719907UL  ,  // waba/io/NativeMethods_FileSetAttributes_(Lwaba/io/File;I)V
	51713634UL  ,  // waba/io/NativeMethods_FileGetAttributes_(Lwaba/io/File;)I
	51765932UL  ,  // waba/io/NativeMethods_FileSetTime_(Lwaba/io/File;BLwaba/sys/Time;)V
	51759659UL  ,  // waba/io/NativeMethods_FileGetTime_(Lwaba/io/File;B)Lwaba/sys/Time;
	51749394UL  ,  // waba/io/NativeMethods_FileIsAvailable_()Z
	51718511UL  ,  // waba/io/NativeMethods_FileCreate_(Lwaba/io/File;Ljava/lang/String;II)V
	51749778UL  ,  // waba/io/NativeMethods_FileIsInserted_(I)Z
	51725914UL  ,  // waba/io/NativeMethods_FileClose_(Lwaba/io/File;)Z -> FileClose_
	51750302UL  ,  // waba/io/NativeMethods_FileCreateDir_(Lwaba/io/File;)Z
	51731867UL  ,  // waba/io/NativeMethods_FileDelete_(Lwaba/io/File;)Z -> FileDelete_
	51734747UL  ,  // waba/io/NativeMethods_FileExists_(Lwaba/io/File;)Z
	51737436UL  ,  // waba/io/NativeMethods_FileGetSize_(Lwaba/io/File;)I
	51724186UL  ,  // waba/io/NativeMethods_FileIsDir_(Lwaba/io/File;)Z
	51731547UL  ,  // waba/io/NativeMethods_FileIsOpen_(Lwaba/io/File;)Z
	51729648UL  ,  // waba/io/NativeMethods_FileListFiles_(Lwaba/io/File;)[Ljava/lang/String;
	51769634UL  ,  // waba/io/NativeMethods_FileReadBytes_(Lwaba/io/File;[BII)I
	51709677UL  ,  // waba/io/NativeMethods_FileRename_(Lwaba/io/File;Ljava/lang/String;)Z
	51737244UL  ,  // waba/io/NativeMethods_FileSetPos_(Lwaba/io/File;I)Z
	51713251UL  ,  // waba/io/NativeMethods_FileWriteBytes_(Lwaba/io/File;[BII)I
	76900827UL  ,  // waba/fx/NativeMethods_FontCreate_(Lwaba/fx/Font;)V
	76928361UL  ,  // waba/fx/NativeMethods_FontMetricsCreate_(Lwaba/fx/FontMetrics;)V
	76903792UL  ,  // waba/fx/NativeMethods_FontMetricsGetCharWidth_(Lwaba/fx/FontMetrics;C)I
	76879489UL  ,  // waba/fx/NativeMethods_FontMetricsGetTextWidth_(Lwaba/fx/FontMetrics;Ljava/lang/String;)I -> FontMetricsGetTextWidth_String
	76921459UL  ,  // waba/fx/NativeMethods_FontMetricsGetTextWidth_(Lwaba/fx/FontMetrics;[CII)I
	76888301UL  ,  // waba/fx/NativeMethods_PaletteGetCurrentPalette_(Lwaba/fx/Palette;)[I
	76910557UL  ,  // waba/fx/NativeMethods_ImageCreate_(Lwaba/fx/Image;)V
	76874360UL  ,  // waba/fx/NativeMethods_ImageGetScaledInstance_(Lwaba/fx/Image;II)Lwaba/fx/Image;
	76874477UL  ,  // waba/fx/NativeMethods_ImageLoad_(Lwaba/fx/Image;Ljava/lang/String;)V
	76914910UL  ,  // waba/fx/NativeMethods_ImageParse_(Lwaba/fx/Image;[B)V
	76897115UL  ,  // waba/fx/NativeMethods_ImageFree_(Lwaba/fx/Image;)V
	76912233UL  ,  // waba/fx/NativeMethods_ImageChangeColors_(Lwaba/fx/Image;IIIIII)V
	76892901UL  ,  // waba/fx/NativeMethods_ImageGetPixelRow_(Lwaba/fx/Image;[BI)V
	76877454UL  ,  // waba/fx/NativeMethods_ImageGetRotatedScaledInstance_(Lwaba/fx/Image;IILwaba/fx/Color;)Lwaba/fx/Image;
	76894267UL  ,  // waba/fx/NativeMethods_ImageGetTouchedUpInstance_(Lwaba/fx/Image;BB)Lwaba/fx/Image;
	76878604UL  ,  // waba/fx/NativeMethods_SoundBeep_()V
	76889614UL  ,  // waba/fx/NativeMethods_SoundTone_(II)V
	76922963UL  ,  // waba/fx/NativeMethods_SoundSetEnabled_(Z)V
	76888163UL  ,  // waba/fx/NativeMethods_SoundClipPlay_(Lwaba/fx/SoundClip;)Z
	76886627UL  ,  // waba/fx/NativeMethods_SoundClipFree_(Lwaba/fx/SoundClip;)V
	51772076UL  ,  // waba/io/NativeMethods_SerialPortCreate_(Lwaba/io/SerialPort;IIIII)V
	51742950UL  ,  // waba/io/NativeMethods_SerialPortClose_(Lwaba/io/SerialPort;)Z
	51748583UL  ,  // waba/io/NativeMethods_SerialPortIsOpen_(Lwaba/io/SerialPort;)Z
	51743536UL  ,  // waba/io/NativeMethods_SerialPortSetFlowControl_(Lwaba/io/SerialPort;Z)Z
	51741040UL  ,  // waba/io/NativeMethods_SerialPortSetReadTimeout_(Lwaba/io/SerialPort;I)Z
	51721134UL  ,  // waba/io/NativeMethods_SerialPortReadBytes_(Lwaba/io/SerialPort;[BII)I
	51764650UL  ,  // waba/io/NativeMethods_SerialPortReadCheck_(Lwaba/io/SerialPort;)I
	51730287UL  ,  // waba/io/NativeMethods_SerialPortWriteBytes_(Lwaba/io/SerialPort;[BII)I
	51754100UL  ,  // waba/io/NativeMethods_SocketCreate_(Lwaba/io/Socket;Ljava/lang/String;IIZ)V
	51755742UL  ,  // waba/io/NativeMethods_SocketClose_(Lwaba/io/Socket;)Z
	51761375UL  ,  // waba/io/NativeMethods_SocketIsOpen_(Lwaba/io/Socket;)Z
	51753832UL  ,  // waba/io/NativeMethods_SocketSetReadTimeout_(Lwaba/io/Socket;I)Z
	51733926UL  ,  // waba/io/NativeMethods_SocketReadBytes_(Lwaba/io/Socket;[BII)I
	51743079UL  ,  // waba/io/NativeMethods_SocketWriteBytes_(Lwaba/io/Socket;[BII)I
	51761043UL  ,  // waba/io/NativeMethods_SocketDisconnect_()Z
	51754164UL  ,  // waba/io/NativeMethods_CatalogCreate_(Lwaba/io/Catalog;Ljava/lang/String;I)V
	51750003UL  ,  // waba/io/NativeMethods_CatalogRename_(Lwaba/io/Catalog;Ljava/lang/String;)Z
	51727973UL  ,  // waba/io/NativeMethods_CatalogAddRecord_(Lwaba/io/Catalog;I)I
	51732646UL  ,  // waba/io/NativeMethods_CatalogAddRecord_(Lwaba/io/Catalog;II)I -> CatalogAddRecord2
	51752168UL  ,  // waba/io/NativeMethods_CatalogResizeRecord_(Lwaba/io/Catalog;I)Z
	51766240UL  ,  // waba/io/NativeMethods_CatalogClose_(Lwaba/io/Catalog;)Z
	51772193UL  ,  // waba/io/NativeMethods_CatalogDelete_(Lwaba/io/Catalog;)Z
	51765482UL  ,  // waba/io/NativeMethods_CatalogListCatalogs_(II)[Ljava/lang/String;
	51745511UL  ,  // waba/io/NativeMethods_CatalogDeleteRecord_(Lwaba/io/Catalog;)Z
	51758121UL  ,  // waba/io/NativeMethods_CatalogGetRecordCount_(Lwaba/io/Catalog;)I
	51751080UL  ,  // waba/io/NativeMethods_CatalogGetRecordSize_(Lwaba/io/Catalog;)I
	51771873UL  ,  // waba/io/NativeMethods_CatalogIsOpen_(Lwaba/io/Catalog;)Z
	51750888UL  ,  // waba/io/NativeMethods_CatalogSetRecordPos_(Lwaba/io/Catalog;I)Z
	51744424UL  ,  // waba/io/NativeMethods_CatalogReadBytes_(Lwaba/io/Catalog;[BII)I
	51731429UL  ,  // waba/io/NativeMethods_CatalogSkipBytes_(Lwaba/io/Catalog;I)I
	51753577UL  ,  // waba/io/NativeMethods_CatalogWriteBytes_(Lwaba/io/Catalog;[BII)I
	51767531UL  ,  // waba/io/NativeMethods_CatalogInspectRecord_(Lwaba/io/Catalog;[BI)I
	51731503UL  ,  // waba/io/NativeMethods_CatalogGetRecordAttributes_(Lwaba/io/Catalog;I)B
	51737776UL  ,  // waba/io/NativeMethods_CatalogSetRecordAttributes_(Lwaba/io/Catalog;IB)V
	51753960UL  ,  // waba/io/NativeMethods_CatalogGetAttributes_(Lwaba/io/Catalog;)I
	51760233UL  ,  // waba/io/NativeMethods_CatalogSetAttributes_(Lwaba/io/Catalog;I)V
	51744359UL  ,  // waba/io/NativeMethods_CatalogGetRecordPos_(Lwaba/io/Catalog;)I
	51772001UL  ,  // waba/io/NativeMethods_CatalogCreate_(Lwaba/io/Catalog;)V -> CatalogCreate_NoFunc
	618057257UL ,  // waba/sys/NativeMethods_ConvertUnsigned2hex_(II)Ljava/lang/String;
	618006550UL ,  // waba/sys/NativeMethods_ConvertToLowerCase_(C)C
	618006742UL ,  // waba/sys/NativeMethods_ConvertToUpperCase_(C)C
	618007137UL ,  // waba/sys/NativeMethods_ConvertToInt_(Ljava/lang/String;)I
	618028004UL ,  // waba/sys/NativeMethods_ConvertToString_(C)Ljava/lang/String; -> ConvertToString_Char
	618015127UL ,  // waba/sys/NativeMethods_ConvertToIntBitwise_(F)I
	618028121UL ,  // waba/sys/NativeMethods_ConvertToFloatBitwise_(I)F
	618028196UL ,  // waba/sys/NativeMethods_ConvertToString_(F)Ljava/lang/String; -> ConvertToString_Float
	618028388UL ,  // waba/sys/NativeMethods_ConvertToString_(I)Ljava/lang/String; -> ConvertToString_Int
	618032741UL ,  // waba/sys/NativeMethods_ConvertToString_(DI)Ljava/lang/String; -> ConvertToString_Double
	618026276UL ,  // waba/sys/NativeMethods_ConvertToDouble_(Ljava/lang/String;)D
	618032869UL ,  // waba/sys/NativeMethods_ConvertToString_(FI)Ljava/lang/String; -> ConvertToString_Float_Int
	618019939UL ,  // waba/sys/NativeMethods_ConvertToFloat_(Ljava/lang/String;)F
	618005878UL ,  // waba/sys/NativeMethods_ConvertToString_(Ljava/lang/String;I)Ljava/lang/String; -> ConvertToString_String_Int
	618039323UL ,  // waba/sys/NativeMethods_ConvertDoubleToLongBits_(D)J
	618052701UL ,  // waba/sys/NativeMethods_ConvertDoubleFromLongBits_(J)D
	618038812UL ,  // waba/sys/NativeMethods_TimeCreate_(Lwaba/sys/Time;)V
	618054751UL ,  // waba/sys/NativeMethods_VmGetFile_(Ljava/lang/String;)[B
	618022040UL ,  // waba/sys/NativeMethods_VmGetRemainingBattery_()I
	618053877UL ,  // waba/sys/NativeMethods_VmCopyArray_(Ljava/lang/Object;ILjava/lang/Object;II)Z -> (*vmGlobals->copyArray)
	618040401UL ,  // waba/sys/NativeMethods_VmGetTimeStamp_()I
	618065440UL ,  // waba/sys/NativeMethods_VmSetTimeStamp_(Lwaba/sys/Time;)V
	618047634UL ,  // waba/sys/NativeMethods_VmExitAndReboot_()V
	618038535UL ,  // waba/sys/NativeMethods_VmGC_()V
	618019183UL ,  // waba/sys/NativeMethods_VmExec_(Ljava/lang/String;Ljava/lang/String;IZ)I
	618069398UL ,  // waba/sys/NativeMethods_VmSetDeviceAutoOff_(I)I
	618066699UL ,  // waba/sys/NativeMethods_VmSleep_(I)V
	618019352UL ,  // waba/sys/NativeMethods_VmGetDeviceFreeMemory_()I
	618031257UL ,  // waba/sys/NativeMethods_VmInterceptSystemKeys_(I)V
	618038364UL ,  // waba/sys/NativeMethods_VmDebug_(Ljava/lang/String;)V
	618026340UL ,  // waba/sys/NativeMethods_VmClipboardCopy_(Ljava/lang/String;)V
	618027108UL ,  // waba/sys/NativeMethods_VmClipboardPaste_()Ljava/lang/String;
	618026532UL ,  // waba/sys/NativeMethods_VmAttachLibrary_(Ljava/lang/String;)Z
	618065898UL ,  // waba/sys/NativeMethods_VmAttachNativeLibrary_(Ljava/lang/String;)Z
	618030361UL ,  // waba/sys/NativeMethods_VmGetSystemKeysPressed_()I
	76929631UL  ,  // waba/ui/NativeMethods_WindowCreate_(Lwaba/ui/Window;)V
	76876395UL  ,  // waba/ui/NativeMethods_WindowSetStatePosition_(Lwaba/ui/Window;II)V
	76913895UL  ,  // waba/ui/NativeMethods_MainWindowCreate_(Lwaba/ui/MainWindow;)V
	76917202UL  ,  // waba/ui/NativeMethods_MainWindowExit_(I)V
	76931038UL  ,  // waba/ui/NativeMethods_MainWindowSetTimerInterval_(I)V
	76881260UL  ,  // waba/ui/NativeMethods_MainWindowGetCommandLine_()Ljava/lang/String;
	76916380UL  ,  // waba/ui/NativeMethods_MainWindowSetThreadCount_(I)V
	76923539UL  ,  // waba/ui/NativeMethods_EventIsAvailable_()Z
	76928403UL  ,  // waba/ui/NativeMethods_WindowPumpEvents_()V
	76873877UL  ,  // waba/ui/NativeMethods_WindowUpdateScreen_()V
	0  // Must end with 0
};

// build at runtime the nativeFuncs array
NativeFunc nativeFuncs[201];
void buildNativeFuncArray()
{
	nativeFuncs[  0] = (*vmGlobals->StringBufferEnsureCapacity);
	nativeFuncs[  1] = (*vmGlobals->StringBufferSetLength);
	nativeFuncs[  2] = (*vmGlobals->appendString);
	nativeFuncs[  3] = (*vmGlobals->appendCharArray);
	nativeFuncs[  4] = (*vmGlobals->appendCharArrayOffLen);
	nativeFuncs[  5] = (*vmGlobals->appendChar);
	nativeFuncs[  6] = (*vmGlobals->appendInt);
	nativeFuncs[  7] = (*vmGlobals->appendFloat);
	nativeFuncs[  8] = (*vmGlobals->appendDouble);
	nativeFuncs[  9] = (*vmGlobals->valueOfDouble);
	nativeFuncs[ 10] = ConvertToString_Char;
	nativeFuncs[ 11] = ConvertToString_Int;
	nativeFuncs[ 12] = ConvertToString_Float;
	nativeFuncs[ 13] = (*vmGlobals->StringCopyChars);
	nativeFuncs[ 14] = (*vmGlobals->StringToUpperCase);
	nativeFuncs[ 15] = (*vmGlobals->StringToLowerCase);
	nativeFuncs[ 16] = (*vmGlobals->StringEquals);
	nativeFuncs[ 17] = (*vmGlobals->StringCompareTo);
	nativeFuncs[ 18] = (*vmGlobals->indexOfChar);
	nativeFuncs[ 19] = (*vmGlobals->indexOfCharStart);
	nativeFuncs[ 20] = (*vmGlobals->StringIndexOf);
	nativeFuncs[ 21] = (*vmGlobals->StringHashCode);
	nativeFuncs[ 22] = (*vmGlobals->ObjectToStringNative);
	nativeFuncs[ 23] = (*vmGlobals->ObjectGetClass);
	nativeFuncs[ 24] = (*vmGlobals->ThrowablePrintStackTraceNative);
	nativeFuncs[ 25] = (*vmGlobals->ClassForName);
	nativeFuncs[ 26] = (*vmGlobals->ClassNewInstance);
	nativeFuncs[ 27] = (*vmGlobals->ClassIsInstance);
	nativeFuncs[ 28] = GraphicsDrawDottedCursor;
	nativeFuncs[ 29] = GraphicsCopyImageRect;
	nativeFuncs[ 30] = GraphicsCopyScreen;
	nativeFuncs[ 31] = GraphicsDrawImageSprite;
	nativeFuncs[ 32] = GraphicsCreate;
	nativeFuncs[ 33] = GraphicsDrawEllipse;
	nativeFuncs[ 34] = GraphicsFillEllipse;
	nativeFuncs[ 35] = GraphicsDrawArc;
	nativeFuncs[ 36] = GraphicsDrawPie;
	nativeFuncs[ 37] = GraphicsFillPie;
	nativeFuncs[ 38] = GraphicsDrawEllipticalArc;
	nativeFuncs[ 39] = GraphicsDrawEllipticalPie;
	nativeFuncs[ 40] = GraphicsFillEllipticalPie;
	nativeFuncs[ 41] = GraphicsDrawCircle;
	nativeFuncs[ 42] = GraphicsFillCircle;
	nativeFuncs[ 43] = GraphicsGetPixel;
	nativeFuncs[ 44] = GraphicsSetPixel;
	nativeFuncs[ 45] = GraphicsEraseRect;
	nativeFuncs[ 46] = GraphicsDrawLine;
	nativeFuncs[ 47] = GraphicsClearScreen;
	nativeFuncs[ 48] = GraphicsDrawDots;
	nativeFuncs[ 49] = GraphicsDrawCursor;
	nativeFuncs[ 50] = GraphicsDrawCursorOutline;
	nativeFuncs[ 51] = GraphicsDrawRect;
	nativeFuncs[ 52] = GraphicsFillRect;
	nativeFuncs[ 53] = GraphicsDrawDottedRect;
	nativeFuncs[ 54] = GraphicsFillPolygon;
	nativeFuncs[ 55] = GraphicsDrawPolygon;
	nativeFuncs[ 56] = GraphicsDrawPolyline;
	nativeFuncs[ 57] = GraphicsGetAnglePoint;
	nativeFuncs[ 58] = GraphicsDrawText;
	nativeFuncs[ 59] = GraphicsDrawTextChars;
	nativeFuncs[ 60] = GraphicsDrawHatchedRect;
	nativeFuncs[ 61] = GraphicsFillHatchedRect;
	nativeFuncs[ 62] = GraphicsDrawRoundRect;
	nativeFuncs[ 63] = GraphicsFillRoundRect;
	nativeFuncs[ 64] = GraphicsSetClip;
	nativeFuncs[ 65] = GraphicsGetClip;
	nativeFuncs[ 66] = GraphicsClip;
	nativeFuncs[ 67] = GraphicsClearClip;
	nativeFuncs[ 68] = GraphicsTranslate;
	nativeFuncs[ 69] = GraphicsGetTranslation;
	nativeFuncs[ 70] = GraphicsCopyRect;
	nativeFuncs[ 71] = GraphicsFree;
	nativeFuncs[ 72] = GraphicsSetForeColor;
	nativeFuncs[ 73] = GraphicsSetBackColor;
	nativeFuncs[ 74] = GraphicsCollisionDetect;
	nativeFuncs[ 75] = Graphics_applyPalette;
	nativeFuncs[ 76] = GraphicsDrawHighLightFrame;
	nativeFuncs[ 77] = GraphicsSetFont;
	nativeFuncs[ 78] = FileSetAttributes;
	nativeFuncs[ 79] = FileGetAttributes;
	nativeFuncs[ 80] = FileSetTime;
	nativeFuncs[ 81] = FileGetTime;
	nativeFuncs[ 82] = FileIsAvailable;
	nativeFuncs[ 83] = FileCreate;
	nativeFuncs[ 84] = FileIsInserted;
	nativeFuncs[ 85] = FileClose_;
	nativeFuncs[ 86] = FileCreateDir;
	nativeFuncs[ 87] = FileDelete_;
	nativeFuncs[ 88] = FileExists;
	nativeFuncs[ 89] = FileGetSize;
	nativeFuncs[ 90] = FileIsDir;
	nativeFuncs[ 91] = FileIsOpen;
	nativeFuncs[ 92] = FileListFiles;
	nativeFuncs[ 93] = FileReadBytes;
	nativeFuncs[ 94] = FileRename;
	nativeFuncs[ 95] = FileSetPos;
	nativeFuncs[ 96] = FileWriteBytes;
	nativeFuncs[ 97] = FontCreate;
	nativeFuncs[ 98] = FontMetricsCreate;
	nativeFuncs[ 99] = FontMetricsGetCharWidth;
	nativeFuncs[100] = FontMetricsGetTextWidth_String;
	nativeFuncs[101] = FontMetricsGetTextWidth;
	nativeFuncs[102] = PaletteGetCurrentPalette;
	nativeFuncs[103] = ImageCreate;
	nativeFuncs[104] = ImageGetScaledInstance;
	nativeFuncs[105] = ImageLoad;
	nativeFuncs[106] = ImageParse;
	nativeFuncs[107] = ImageFree;
	nativeFuncs[108] = ImageChangeColors;
	nativeFuncs[109] = ImageGetPixelRow;
	nativeFuncs[110] = ImageGetRotatedScaledInstance;
	nativeFuncs[111] = ImageGetTouchedUpInstance;
	nativeFuncs[112] = SoundBeep;
	nativeFuncs[113] = SoundTone;
	nativeFuncs[114] = SoundSetEnabled;
	nativeFuncs[115] = SoundClipPlay;
	nativeFuncs[116] = SoundClipFree;
	nativeFuncs[117] = SerialPortCreate;
	nativeFuncs[118] = SerialPortClose;
	nativeFuncs[119] = SerialPortIsOpen;
	nativeFuncs[120] = SerialPortSetFlowControl;
	nativeFuncs[121] = SerialPortSetReadTimeout;
	nativeFuncs[122] = SerialPortReadBytes;
	nativeFuncs[123] = SerialPortReadCheck;
	nativeFuncs[124] = SerialPortWriteBytes;
	nativeFuncs[125] = SocketCreate;
	nativeFuncs[126] = SocketClose;
	nativeFuncs[127] = SocketIsOpen;
	nativeFuncs[128] = SocketSetReadTimeout;
	nativeFuncs[129] = SocketReadBytes;
	nativeFuncs[130] = SocketWriteBytes;
	nativeFuncs[131] = SocketDisconnect;
	nativeFuncs[132] = CatalogCreate;
	nativeFuncs[133] = CatalogRename;
	nativeFuncs[134] = CatalogAddRecord;
	nativeFuncs[135] = CatalogAddRecord2;
	nativeFuncs[136] = CatalogResizeRecord;
	nativeFuncs[137] = CatalogClose;
	nativeFuncs[138] = CatalogDelete;
	nativeFuncs[139] = CatalogListCatalogs;
	nativeFuncs[140] = CatalogDeleteRecord;
	nativeFuncs[141] = CatalogGetRecordCount;
	nativeFuncs[142] = CatalogGetRecordSize;
	nativeFuncs[143] = CatalogIsOpen;
	nativeFuncs[144] = CatalogSetRecordPos;
	nativeFuncs[145] = CatalogReadBytes;
	nativeFuncs[146] = CatalogSkipBytes;
	nativeFuncs[147] = CatalogWriteBytes;
	nativeFuncs[148] = CatalogInspectRecord;
	nativeFuncs[149] = CatalogGetRecordAttributes;
	nativeFuncs[150] = CatalogSetRecordAttributes;
	nativeFuncs[151] = CatalogGetAttributes;
	nativeFuncs[152] = CatalogSetAttributes;
	nativeFuncs[153] = CatalogGetRecordPos;
	nativeFuncs[154] = CatalogCreate_NoFunc;
	nativeFuncs[155] = ConvertUnsigned2hex;
	nativeFuncs[156] = ConvertToLowerCase;
	nativeFuncs[157] = ConvertToUpperCase;
	nativeFuncs[158] = ConvertToInt;
	nativeFuncs[159] = ConvertToString_Char;
	nativeFuncs[160] = ConvertToIntBitwise;
	nativeFuncs[161] = ConvertToFloatBitwise;
	nativeFuncs[162] = ConvertToString_Float;
	nativeFuncs[163] = ConvertToString_Int;
	nativeFuncs[164] = ConvertToString_Double;
	nativeFuncs[165] = ConvertToDouble;
	nativeFuncs[166] = ConvertToString_Float_Int;
	nativeFuncs[167] = ConvertToFloat;
	nativeFuncs[168] = ConvertToString_String_Int;
	nativeFuncs[169] = ConvertDoubleToLongBits;
	nativeFuncs[170] = ConvertDoubleFromLongBits;
	nativeFuncs[171] = TimeCreate;
	nativeFuncs[172] = VmGetFile;
	nativeFuncs[173] = VmGetRemainingBattery;
	nativeFuncs[174] = (*vmGlobals->copyArray);
	nativeFuncs[175] = VmGetTimeStamp;
	nativeFuncs[176] = VmSetTimeStamp;
	nativeFuncs[177] = VmExitAndReboot;
	nativeFuncs[178] = VmGC;
	nativeFuncs[179] = VmExec;
	nativeFuncs[180] = VmSetDeviceAutoOff;
	nativeFuncs[181] = VmSleep;
	nativeFuncs[182] = VmGetDeviceFreeMemory;
	nativeFuncs[183] = VmInterceptSystemKeys;
	nativeFuncs[184] = VmDebug;
	nativeFuncs[185] = VmClipboardCopy;
	nativeFuncs[186] = VmClipboardPaste;
	nativeFuncs[187] = VmAttachLibrary;
	nativeFuncs[188] = VmAttachNativeLibrary;
	nativeFuncs[189] = VmGetSystemKeysPressed;
	nativeFuncs[190] = WindowCreate;
	nativeFuncs[191] = WindowSetStatePosition;
	nativeFuncs[192] = MainWindowCreate;
	nativeFuncs[193] = MainWindowExit;
	nativeFuncs[194] = MainWindowSetTimerInterval;
	nativeFuncs[195] = MainWindowGetCommandLine;
	nativeFuncs[196] = MainWindowSetThreadCount;
	nativeFuncs[197] = EventIsAvailable;
	nativeFuncs[198] = WindowPumpEvents;
	nativeFuncs[199] = WindowUpdateScreen;
	nativeFuncs[200] = 0;  // Must end with 0
}

Err callPublic(int index, uint32 *ret, uint32 *args)
{
   switch (index)
   {
      case PUBLIC_GraphicsStop     : graphicsStop();            break;
      case PUBLIC_GraphicsInit     : if (!graphicsInit()) return (Err)-999; break;
      case PUBLIC_GraphicsSwap     : graphicsSwap((int)*args);  break;
      case PUBLIC_FreeGrResources  : freeGrResources();         break;
      case PUBLIC_DrawHourGlass    : drawHourGlass((int)*args); break;
#ifndef PALMOS
      case PUBLIC_InitMainGrHandle : initMainGrHandle();        break; // not used on PalmOS
#endif
      case PUBLIC_GraphicsDestroy  : GraphicsDestroy((WObject)args);   break;
      case PUBLIC_ImageDestroy     : ImageDestroy((WObject)args);      break;
      case PUBLIC_CatalogDestroy   : CatalogDestroy((WObject)args);    break;
      case PUBLIC_SocketDestroy    : SocketDestroy((WObject)args);     break;
      case PUBLIC_SerialPortDestroy: SerialPortDestroy((WObject)args); break;
      case PUBLIC_FileDestroy      : FileDestroy((WObject)args);       break;
#ifdef PALMOS
      case PUBLIC_GraphicsReset    : graphicsReset();           break;
#endif
      case PUBLIC_GraphicsStartImage: graphicsStartImage(ret, args);   break;
      case PUBLIC_GraphicsEndImage:   graphicsEndImage((GrHandle)args[0]); break;
   }
   return 0;
}

#ifndef WINCE
Err handleEvent(EventPtr eventP)
{
   return SWErrEventNotHandled;
}
// called when this lib is about to be closed. The library must free any allocated resources.
void onClose()
{
   // better put serial deactivations here?
}
#endif

ClassHook classHooks[] =
{
	{  NULL, /*"waba/fx/Graphics"  ,*/ PUBLIC_GraphicsDestroy,   0},  // guich@200: from 11 to 7
	{  NULL, /*"waba/fx/Font",      */ -1,                       0},
	{  NULL, /*"waba/fx/Image",     */ PUBLIC_ImageDestroy,      0},
	{  NULL, /*"waba/io/Catalog",   */ PUBLIC_CatalogDestroy,    0}, // guich@200b2: corrected a corruption (was 7)
	{  NULL, /*"waba/io/Socket",    */ PUBLIC_SocketDestroy,     0},
	{  NULL, /*"waba/io/SerialPort",*/ PUBLIC_SerialPortDestroy, 0}, // guich@200b3: changed from 2 to 5 (mathias adds)
   {  NULL, /*"waba/io/File",      */ PUBLIC_FileDestroy,       0}, // guich@350_19
   {  NULL, 0, 0}, // fedoniouk@320_21: must be like this for VC(actually for any compiler) as it doesn't set className to NULL in this case.
	NULL
};

// Called by the system to get the ClassHooks in this custom lib. Must return NULL if no hooks are defined
ClassHook *getClassHooks()
{
   classHooks[0].className = "waba/fx/Graphics";
   classHooks[1].className = "waba/fx/Font";
   classHooks[2].className = "waba/fx/Image";
   classHooks[3].className = "waba/io/Catalog";
   classHooks[4].className = "waba/io/Socket";
   classHooks[5].className = "waba/io/SerialPort";
   classHooks[6].className = "waba/io/File";
   return classHooks;
}
/*************************************************************/

// NOTE: If you subclass a class with an object destroy function, you must
// explicity call your superclasses object destroy function.

#ifndef PALMOS
void swapInt32(int32 *n)
{
   byte *ptr = (byte *)n; // start of int32
   int32 n2 = *n;  // copy of n
   byte *ptr2 = ((byte *)(&n2))+3; // end of int32

   *ptr++ = *ptr2--;
   *ptr++ = *ptr2--;
   *ptr++ = *ptr2--;
   *ptr++ = *ptr2--;
}
void swapInt16(int16 *n)
{
   byte *ptr = (byte *)n; // start of int16
   int16 n2 = *n;  // copy of n
   byte *ptr2 = ((byte *)(&n2))+1; // end of int16

   *ptr++ = *ptr2--;
   *ptr++ = *ptr2--;
}
#endif

#if defined(PALMOS)
	#include "nmpalm_wabafx.c"
	#include "nmpalm_wabaui.c"
	#include "nmpalm_wabasys.c"
	#include "nmpalm_wabaio.c"
#elif defined(WIN32)
	#include "ce\nmwin32_wabafx.c"
	#include "ce\nmwin32_wabaui.c"
	#include "ce\nmwin32_wabasys.c"
	#include "ce\nmwin32_wabaio.c"
   #include "ce\SWLibStatic.c"
#endif
