/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  Copyright (C) 1998 Ken Krugler                                               *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

/* History
 *	18 Jul 98	kwk	New today from 3Com's SWLib.
 *	19 Jul 98	kwk	Now use space at end of libr resource to calc globals
 *							size and copy over initialization values.
 */

// Before including SWLib.h, define BUILDING_SW_LIB so that when
// we make calls to the library's routines from this file, we generate
// direct jumps to the routines, versus system library trap dispatching.

// superwaba.h now includes SWLib.h
#define BUILDING_SW_LIB
#include "superwaba.h"

// **********************************
// I would like to thank  Ken Krugler
// Without his article "PalmOS Shared
// Library Globals", my goals couldnt
// be achieved.                 Guich
// **********************************

// If we're building a real library for the device, then make sure that
// our installation routine is actually __Startup__, which the Metrowerks
// linker will ensure gets called when the PalmOS is loading our library.
// Note that with CW Pro 3 you can specify the entry point in the 68K
// target prefs panel, thus this isn't really necessary, but not everybody
// is using CW Pro 3 yet.

#if EMULATION_LEVEL == EMULATION_NONE
   #define SWLibInstall		__Startup__
#endif

/********************************************************************
 * Private macros
 ********************************************************************/

// If we're built for the device, then we need to provide for
// setting up and restoring register A4 on entry to/exit from
// the library routines. Otherwise we're being built for the
// simulator, and thus are running as part of a normal 68K Mac
// application (better not mess with A4 in that case).

	#define EnterLibrary_(refnum) long oldA4 = SetCurA4(refnum);
	#define ExitLibrary_() RestoreA4(oldA4);
/********************************************************************
 * Global variables
 ********************************************************************/

/********************************************************************
 * Private functions
 ********************************************************************/

	long	SetCurA4(UInt inRefnum);
	void	RestoreA4(long inSavedA4);
	long	SetA4(long:__D0):__D0 = 0xC18C;
	long	GetCurrentA4(void) = 0x200C;
// The following two routines are referenced by the linker, and thus
// must have standard 'C' names & be externally visible, even though
// they only get called from SWLibDispatch.cp

	Err SWLibInstall(UInt inRefnum, SysLibTblEntryPtr inLibTablePtr);

	long SetCurrentA4(void);

/********************************************************************
 * Private variables
 ********************************************************************/

// This array is used to map a hash value to a corresponding native function.
// It must remain sorted by hash value because a binary search is performed
// to find a method by its hash value. There is a small chance of collision
// if two hashes match and if one occurs, the function name should be changed
// to avoid collision. To prevent users from creating invalid native methods
// that hash to a valid value, native methods could be restricted to a specific
// set of classes (this is probably not necessary since any verifier probably
// wouldn't allow native methods to get by anyway).
// guich: since i introduced a hash table, this table dont has anymore to stay sorted. :-))

// guich@102 - implements a hash table, whoose key is a int32 and value can be anything.
// ps: HASHSIZE is constant. increase it if necessary.

struct htElem
{
    uint32 key;
    NativeFunc value;
    struct htElem *next;
};

VoidHand hashTableH;
static struct htElem **hashTable;  // guich@400_33

struct VMGlobals *vmGlobals; // setted by the SWInvokeNative

uint16 libRefNum; // our library reference number

/********************************************************************
 * Private functions
 ********************************************************************/
#define htHash(key) (uint16)((key) % HASHSIZE)

/* return the htElem found, else NULL if key doesn't exist */
static struct htElem *htGet(uint32 key)
{
   struct htElem *np;
   for (np = hashTable[htHash(key)]; np; np = np->next)
      if (np->key == key)
         return np;
   return NULL;
}
/* return struct containing value or NULL if unable to store */
static struct htElem *htPut(uint32 key, NativeFunc value)
{
   struct htElem *np;
   unsigned hashval;
   if ((np = htGet(key)) == NULL) // duplicated keys not allowed
   {
      np = (struct htElem *) (*vmGlobals->xmalloc) (sizeof(struct htElem));
      if (np != NULL)
      {
          hashval = htHash(key);
          /* point to last w/ this hash */
          np->next = hashTable[hashval];
          np->key = key;
          np->value = value;
          /* attach to hashtab array */
          hashTable[hashval] = np;
      }
   }
   else
      ErrFatalDisplayIf(1, (*vmGlobals->int2str)("**** duplicated key in hashtable ****",key));
   return np;
}

static void disposeHashtable()
{
   // dispose the hashtable
   int i;
   for (i = 0; i < HASHSIZE; i++)
   {
      struct htElem *np;
      np = hashTable[i];
      while (np)
      {
         struct htElem *next;
         next = np->next;
         xfree(np);
         np = next;
      }
      hashTable[i] = NULL;
   }
   MemHandleUnlock(hashTableH);
   MemHandleFree(hashTableH);
}

/************************************************************
 *
 *  FUNCTION: SWLibOpen
 *
 *  DESCRIPTION:	Opens the SW library.
 *
 *						This function must be called before any other SW Library functions,
 *						with the exception of SWLibGetLibAPIVersion.
 *
 *						If SWLibOpen fails, do not call any other SW library API functions.
 *						If SWLibOpen succeeds, call SWLibClose when you are done using
 *						the library to enable it to release critical system resources.
 *
 *  LIBRARY DEVELOPER NOTES:
 *
 *						The library's "open" and "close" functions should *not* take an excessive
 *						amount of time to complete.  If the processing time for either of these
 *						is lengthy, consider creating additional library API function(s) to handle
 *						the time-consuming chores.
 *
 *
 *  PARAMETERS:	inRefnum				-- SW library reference number returned by SysLibLoad()
 *										   	   or SysLibFind().
 *						outClientContext	-- pointer to variable for returning client context.  The
 *												   The client context is used to maintain client-specific
 *													data for multiple client support.  The
 *												   value returned here will be used as a parameter for
 *												   other SW library functions which require a client
 *												   context.
 *
 *  CALLED BY:		anyone who wants to use this library
 *
 *  RETURNS:		0							-- no error
 *						SWErrMemory		-- not enough memory to initialize
 *
 *						*outClientContext will be set to client context on success, or zero on error.
 *
 *  CREATED:	5/20/97
 *
 *************************************************************/

Err SWLibOpen(UInt inRefnum, void *inGlobalsPtr)
{
   EnterLibrary_(inRefnum)
   ClassHook *libClassHooks;
   int i;
   ErrFatalDisplayIf(inRefnum == sysInvalidRefNum,"Library not properly loaded!");
   // 1. associate the vmGlobals struct with the given one
   vmGlobals = (struct VMGlobals *)inGlobalsPtr;
   // 2. install our hooks
   libClassHooks = getClassHooks();
   if (libClassHooks)
   {
      int hookLibCount = (*vmGlobals->arrayCount)((char *)libClassHooks,sizeof(ClassHook));
      int insertIndex;
      vmGlobals->classHooks = (ClassHook *)(*vmGlobals->expandArray)((char *)vmGlobals->classHooks,sizeof(ClassHook),hookLibCount,&insertIndex);
      if (insertIndex == -1) return SWErrMemory;
      for (i =0; i < hookLibCount; i++)
         libClassHooks[i].libRefnum = inRefnum;

      xmemmove (vmGlobals->classHooks+insertIndex,libClassHooks, hookLibCount*sizeof(ClassHook));
   }
   libRefNum = inRefnum;
   buildNativeFuncArray();
   i = sizeof(struct htElem *) * HASHSIZE;
   hashTableH = MemHandleNew(i);
   ErrFatalDisplayIf(!hashTableH,"Not enough memory to allocate hashes");
   hashTable = MemHandleLock(hashTableH); // guich@400_33: we can't use xmalloc because when this library gets unloaded, the object heap is already freed.
   xmemzero(hashTable,i);
   for (i=0; nativeHashes[i]; i++)
      htPut(nativeHashes[i], nativeFuncs[i]);
	ExitLibrary_()
	return 0;
}


/************************************************************
 *
 *  FUNCTION: SWLibClose
 *
 *  DESCRIPTION:	Closes the SW libary, frees client context.
 *
 *						***IMPORTANT***
 *						May be called only if SWLibOpen succeeded.
 *
 *						If other applications still have the library open, decrements
 *						the reference count and returns SWErrStillOpen.
 *
 *
 *  LIBRARY DEVELOPER NOTES:
 *
 *						The library's "open" and "close" functions should *not* take an excessive
 *						amount of time to complete.  If the processing time for either of these
 *						is lengthy, consider creating additional library API function(s) to handle
 *						the time-consuming chores.
 *
 *
 *  PARAMETERS:	inRefnum				-- SW library reference number returned by SysLibLoad()
 *												   or SysLibFind().
 *						inClientContext	-- client context
 *
 *  CALLED BY:		Whoever wants to close the SW library
 *
 *  RETURNS:		0							-- no error
 *						SWErrStillOpen	-- library is still open by others (no error)
 *
 *  CREATED:	5/20/97
 *
 *************************************************************/
Err SWLibClose(UInt inRefnum)
{
   EnterLibrary_(inRefnum)
   VoidHand globalsH;
   VoidHand codeH;
   VoidPtr globalsP;
   onClose();
   disposeHashtable();
   // unlock the globalsP
   globalsP = SysLibTblEntry(inRefnum)->globalsP;
   globalsP = (byte *)globalsP - sizeof(VoidHand); // decrement the pointer
   codeH = *((VoidHand *)globalsP); // get codeH
   globalsP = (byte *)globalsP - sizeof(VoidHand); // decrement the pointer
   globalsH = *((VoidHand *)globalsP); // get globalsH
   MemHandleUnlock(globalsH);
   MemHandleFree(globalsH); // guich@200b4_81: forgot to free the handles
   MemHandleUnlock(codeH); // guich@200b4_48: unlock the codeH (note: this one is freed by the Operating System)
   ExitLibrary_()
   return 0;
}


/************************************************************
 *
 *  FUNCTION: SWLibSleep
 *
 *  DESCRIPTION:	Handles system sleep notification.
 *
 *						***IMPORTANT***
 *						This notification function is called from a system interrupt.
 *						It is only allowed to use system services which are interrupt-
 *						safe.  Presently, this is limited to EvtEnqueueKey, SysDisableInts,
 *						SysRestoreStatus.  Because it is called from an interrupt,
 *						it must *not* take a long time to complete to preserve system
 *						integrity.  The intention is to allow system-level libraries
 *						to disable hardware components to conserve power while the system
 *						is asleep.
 *
 *  PARAMETERS:	inRefnum		-- SW library reference number
 *
 *  CALLED BY:		System
 *
 *  RETURNS:		0						-- no error
 *
 *  CREATED:	5/20/97
 *
 *************************************************************/
Err SWLibSleep(UInt inRefnum)
{
   return 0;
}

/************************************************************
 *
 *  FUNCTION: SWLibWake
 *
 *  DESCRIPTION:	Handles system wake notification
 *
 *						***IMPORTANT***
 *						This notification function is called from a system interrupt.
 *						It is only allowed to use system services which are interrupt-
 *						safe.  Presently, this is limited to EvtEnqueueKey, SysDisableInts,
 *						SysRestoreStatus.  Because it is called from an interrupt,
 *						it must *not* take a long time to complete to preserve system
 *						integrity.  The intention is to allow system-level libraries
 *						to enable hardware components which were disabled when the system
 *						went to sleep.
 *
 *  PARAMETERS:	inRefnum		-- SW library reference number
 *
 *  CALLED BY:	System
 *
 *  RETURNS:	0						-- no error
 *
 *************************************************************/
Err SWLibWake(UInt inRefnum)
{
   return 0;
}

/************************************************************
 *
 *  FUNCTION: SWLibGetLibAPIVersion
 *
 *  DESCRIPTION:	Get our library API version.  The SW library does not
 *						need to be "opened" to call SWLibGetLibAPIVersion.
 *
 *  PARAMETERS:	inRefnum		-- SW library reference number
 *						dwVerP		-- pointer to variable for storing the version number
 *
 *  CALLED BY:		Anyone wishing to get our library API version
 *
 *  RETURNS:		0						-- no error
 *
 *************************************************************/
Err SWLibGetLibAPIVersion(UInt inRefnum, DWordPtr dwVerP)
{
   // Error-check our parameters
   ErrFatalDisplayIf(dwVerP == NULL, "null pointer argument");
   *dwVerP = prvSWLibVersion;
   return 0;
}

/*************************************************************/
Err SWGetNativeFunc(UInt uRefNum, int32 hash, NativeFunc *func)
{
   EnterLibrary_(uRefNum)
   Err err = 0;
   struct htElem *elem = htGet(hash);
   if (elem && elem->value)
      *func = elem->value;
   else
      err = SWErrMethodNotFound;
   ExitLibrary_()
   return err;
}
/*************************************************************/
Err SWInvokeNativeFunc(UInt uRefNum, NativeFunc func, Var stack[])
{
   EnterLibrary_(uRefNum)
   func(stack);
   ExitLibrary_()
   return 0;
}
/*************************************************************/
Err SWInvokeNative(UInt uRefNum, int32 hash, Var stack[])
{
   EnterLibrary_(uRefNum)
   Err err = 0;
   struct htElem *elem = htGet(hash);
   if (elem && elem->value)
      elem->value(stack);
   else
   {
      char buf[60];
      xstrprintf(buf,"SWInvokeNative could not find method with hash %lu.",(uint32)hash);
   	(*vmGlobals->debug)(buf);
   	(*vmGlobals->debug)("Check if all required native libraries are installed.");
      err = SWErrMethodNotFound;
   }
   ExitLibrary_()
   return err;
}
/*************************************************************/
Err SWCallPublic(UInt uRefNum, int index, uint32 *ret, uint32 *args)
{
   EnterLibrary_(uRefNum)
   Err err = callPublic(index, ret, args);
   ExitLibrary_()
   return err;
}
/*************************************************************/
Err SWHandleEvent(UInt uRefNum, EventPtr eventP)
{
   EnterLibrary_(uRefNum)
   Err err = handleEvent(eventP);
   ExitLibrary_()
   return err;
}

/***********************************************************************
 * Private functions
 ***********************************************************************/

static long GetGlobalsOffset(void);

static Ptr asm SWLibDispatchTable(void);

/***********************************************************************
 * SWLibInstall
 *
 * This routine gets called when the PalmOS loads our library. During
 * loading, the PalmOS opens our database, finds the 'libr'=0 resource,
 * locks it down, and makes a subroutine call to a routine which it
 * assumes is located at the very beginning of the libr=0 resource.
 * Normally the Metrowerks linker sets up this code resource such that
 * the beginning of the resource contains a jump to the routine called
 * "__Startup__", thus the preceeding #define which renames this function
 * such that everything links properly.
 *
 * This routine needs to set inLibTablePtr->dispatchTblP with the pointer
 * to our dispatch table (see SWLibDispatchTable below). For this
 * sample, it also needs to allocate & save off a handle to our globals.
 ***********************************************************************/
Err SWLibInstall(UInt inRefnum, SysLibTblEntryPtr inLibTablePtr)
{
   #pragma unused(inRefnum)
   VoidHand codeH;
   VoidPtr codeP;
   ULong codeSize;
   ULong dataSize;
   VoidHand globalsH;
   VoidPtr globalsP;

   inLibTablePtr->globalsP = 0;

   // Generate a bogus call to SetCurrentA4() so that it doesn't get
   // dead-code stripped.

   SetCurrentA4();

   // If we're building for the device, then we want to allocate a new
   // memory block which will contain our globals. If that succeeds,
   // then we need to set its owner to zero so that it doesn't get
   // auto-disposed when the current app quits. Note that we could
   // delay this allocation until the first open call, but since
   // the size of the globals record should be small, this isn't a
   // big deal.

   // We need to figure out (a) how big our 'libr' resource is, and (b)
   // its starting address, so that we know how much space to allocate
   // for the globals, and where to copy from. Since our database is
   // open (otherwise we couldn't have gotten loaded), we can easily
   // get a handle to our libr resource.

   codeH = DmGet1Resource('libr', 0);
   ErrFatalDisplayIf(codeH == NULL, "Can't load library code resource");
   codeP = MemHandleLock(codeH);

   codeSize = MemHandleSize(codeH);
   dataSize = codeSize - GetGlobalsOffset();

   globalsH = MemHandleNew(dataSize+sizeof(VoidHand)*2); // guich@200b4_48: stick into globalsH the codeH handle
   if (globalsH == NULL)
   {
   	MemHandleUnlock(codeH);
   	return memErrNotEnoughSpace;
   }

   MemHandleSetOwner(globalsH, 0);

   // Now copy over the globals data from the end of the libr resource,
   // since it might contain pre-initialized values (statically initialized
   // globals).

   globalsP = MemHandleLock(globalsH);
   *((VoidHand *)globalsP) = globalsH; // store globalsH
   globalsP = (VoidPtr)(((byte *)globalsP)+sizeof(VoidHand)); // increment the pointer
   *((VoidHand *)globalsP) = codeH;    // store the codeH
   globalsP = (VoidPtr)(((byte *)globalsP)+sizeof(VoidHand)); // increment the pointer

   MemMove(globalsP, (BytePtr)codeP + GetGlobalsOffset(), dataSize);
   inLibTablePtr->globalsP = globalsP;

   // Install pointer to our dispatch table
   inLibTablePtr->dispatchTblP = (Ptr*)SWLibDispatchTable();

   return(0);
} // SWLibInstall


/***********************************************************************
 * SWLibDispatchTable
 *
 * This routine needs to return a pointer to the library dispatch table,
 * which consists of N 16-bit offsets to the library name, the standard
 * library routines, and any custom library routines.
 ***********************************************************************/

// A little magic to make it easier to create the dispatch table, since
// Metrowerks doesn't support pc-relative address calculations in its
// inline 68K assembler.

#define	Dispatch1_(x)	dc.w	Goto ## x + ((1 + (sysLibTrap ## x) - sysLibTrapOpen) * 2)
#define	Dispatch2_(x)	dc.w	Goto ## x + ((1 + (SWLibTrap ## x) - sysLibTrapOpen) * 2)

static Ptr asm SWLibDispatchTable(void)
{
	lea		@Table,a0								// Load pointer to table.
	rts													// Exit with it

@Table:
	// Offset to library name
	DC.W		@Name

	// Offsets to standard library routines.
	Dispatch1_(Open)
	Dispatch1_(Close)
	Dispatch1_(Sleep)
	Dispatch1_(Wake)

	// Offsets to custom library routines.
	Dispatch2_(SWLibGetLibAPIVersion)
	Dispatch2_(SWInvokeNative)
	Dispatch2_(SWCallPublic)
	Dispatch2_(SWHandleEvent)
	Dispatch2_(SWGetNativeFunc)
	Dispatch2_(SWInvokeNativeFunc)

GotoOpen:
	jmp		SWLibOpen

GotoClose:
	jmp		SWLibClose

GotoSleep:
	jmp		SWLibSleep

GotoWake:
	jmp		SWLibWake

GotoSWLibGetLibAPIVersion:
	jmp		SWLibGetLibAPIVersion

GotoSWInvokeNative:
	jmp		SWInvokeNative

GotoSWCallPublic:
	jmp		SWCallPublic

GotoSWHandleEvent:
	jmp		SWHandleEvent

GotoSWGetNativeFunc:
   jmp      SWGetNativeFunc

GotoSWInvokeNativeFunc:
   jmp      SWInvokeNativeFunc

@Name:
	DC.B		SWLibName							// This name identifies the library.  Apps
															// can pass it to SysLibFind() to check if the
															// library is already installed
}

/************************************************************
 * SetCurrentA4
 *
 * This is a bogus routine used to fool the Metrowerks linker
 * into stuffing the offset of the globals area (at the end
 * of the resource) into a location where we can find it. We
 * know that the linker uses offset 6 into a routine named
 * SetCurrentA4 as the 'magic' location, so give it what it
 * wants.
 *************************************************************/

asm long SetCurrentA4(void)
{
   rts
   nop
   nop

   __globalsOffset:
   dc.l	0

   entry	static GetGlobalsOffset

   move.l	__globalsOffset,d0
   rts
}

/************************************************************
 * SetCurA4
 *
 * Take the passed library refnum, and get our globals ptr.
 * Set A4 to this pointer, and return back the previous value
 * for A4, so that it can be saved in the caller's local stack
 * frame and restored on exit.
 *
 * WARNING!!! This routine CANNOT be named SetCurrentA4, as
 * the MW linker looks for a routine with that name, and stuffs
 * a long at offset 6 into the routine's code. This long is
 * the offset into the resource where the globals would start
 * for a MacOS code resource with globals, and is used by the
 * 'standard' SetCurrentA4 routine (for MacOS 68K apps) to set
 * A4 correctly. Weird.
 *************************************************************/

inline long SetCurA4(UInt inRefnum)
{
   return SetA4((long)SysLibTblEntry(inRefnum)->globalsP);
} // SetCurA4


/************************************************************
 * RestoreA4
 *
 * Unlock our globals, and restore A4 to what it was before our
 * library routine called SetCurA4
 *************************************************************/

inline void RestoreA4(long inSavedA4)
{
   //GetCurrentA4();
   SetA4(inSavedA4);
} // RestoreA4
