/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

// guich@200: I really wont like to see this code running in another JAVA virtual machine for the PalmOS. if you want do so, please ask me before.

// guich@340: NOTE: THIS CODE FOR PALM OS 5 ASSUMES DISPLAYS WITH AT LEAST 8BPP

static ULong oldBPP;
static int usePalmOSFuncs=1; // use internal routines?
static int32 lastForeColor = -1;
static int32 lastTextColor = -1;
#define WINSETFORECOLOR(color) if (lastForeColor != color) {lastForeColor=color; WinSetForeColor(color);}
#define WINSETTEXTCOLOR(color) if (lastTextColor != color) {lastTextColor=color; WinSetTextColor(color);}
static void drawHLineNOCT_ (int x, int y, int w, int32 color);
static void setPixelNOCT_ (int x, int y, int32 color);
static uint32 txtBufSize=0;
static PointType *txtPixels;
static int graphicsActive = 0;

////////////////////////////////////////////////////////////////////////////
//                        PALM COLOR PRIMITIVES                           //
////////////////////////////////////////////////////////////////////////////
// note: for optimization, colors in the primitive routines are always treated by their indexes, not by rgb values.

#define COLOR_HASHSIZE 255
struct htElemColor
{
    uint32 key;
    int32 value;
    struct htElemColor *next;
};

static struct htElemColor **colorHashTable; // guich@400_33        /* The storage array for the methods */
#define htHash_Color(key) (unsigned)((key) & COLOR_HASHSIZE)

/* return the htElemColor found, else NULL if key doesn't exist */
static struct htElemColor *htGetColor(uint32 key)
{
   struct htElemColor *np;
   for (np = colorHashTable[htHash_Color(key)]; np; np = np->next)
      if (np->key == key)
         return np;
   return NULL;
}
/* return struct containing value or NULL if unable to store */
static struct htElemColor *htPutColor(uint32 key, int32 value)
{
   struct htElemColor *np;
   unsigned hashval;
   if ((np = htGetColor(key)) == NULL) // duplicated keys not allowed
   {
      np = (struct htElemColor *) (*vmGlobals->xmalloc) (sizeof(struct htElemColor));
      if (np != NULL)
      {
          hashval = htHash_Color(key);
          /* point to last w/ this hash */
          np->next = colorHashTable[hashval];
          np->key = key;
          np->value = value;
          /* attach to hashtab array */
          colorHashTable[hashval] = np;
      }
   }
   return np;
}
static void disposeColorHashtable(byte freeTable)
{
   // dispose the hashtable
   int i;
   for (i = 0; i <= COLOR_HASHSIZE; i++)
   {
      struct htElemColor *np;
      np = colorHashTable[i];
      while (np)
      {
         struct htElemColor *next;
         next = np->next;
         xfree(np);
         np = next;
      }
      colorHashTable[i] = NULL;
   }
   if (freeTable) xfree(colorHashTable);
}

int mustSwap = 0;
static IndexedColorType rgb2index(int r, int g, int b)
{
	RGBColorType rgb;
	WinHandle old;
	IndexedColorType idx;
	rgb.r = r;
	rgb.g = g;
	rgb.b = b;
   if (mustSwap) old = WinSetDrawWindow(GRHANDLE_WinHandle(mainGrHandle)); // guich@200b4_160
	idx = WinRGBToIndex(&rgb);
   if (mustSwap) WinSetDrawWindow(old);
   return idx;
}
// this method returns the index in the current palette, not the color
static inline int32 getColor_(int red, int green, int blue)
{
   if (isColor)
   {
      IndexedColorType idx;
  	   uint32 key = ((int32)red << 16) | ((int32)green << 8) | blue;
   	struct htElemColor *elem = htGetColor(key);
      if (elem != NULL)
         return elem->value;
      idx = rgb2index(red,green,blue);
      htPutColor(key,idx);
      return idx;
   }
   else
   {
      int lum = (red+green+blue)/3;//(30 * red + 59 * green + 11 * blue) / 100; // converts to luminance
      return 15-lum/16;
   }
}
////////////////////////////////////////////////////////////////////////////
static void getBitmapRow_(GrHandle h, int x, int y, int w, Pixel *row)
{
   if (GRHANDLE_IsBitmap(h))
      xmemmove(row,GRHANDLE_MemAddr(h) + GRHANDLE_WidthInBPP(h)*(y) + x, w);
   else
   {
	   WinHandle oldH = WinSetDrawWindow((WinHandle)GRHANDLE_WinHandle(h));
 	   while (w--) *row++ = WinGetPixel(x++,y);
	   WinSetDrawWindow(oldH);
	}
}
////////////////////////////////////////////////////////////////////////////
static void drawScanLine(int dstX, int dstY, int width, Pixel *pixels)
{
   byte lastColor = pixels[0];
   int lastI=0, i=1;
   for (; i < width; i++)
      if (pixels[i] != lastColor)
      {
         drawHLineNOCT_(dstX+lastI,dstY,i-lastI,(int)lastColor);
         lastI = i;
         lastColor = pixels[i];
      }
   drawHLineNOCT_(dstX+lastI,dstY,i-lastI,(int)lastColor);
}

////////////////////////////////////////////////////////////////////////////
static int32 *lastPal;
static int lastPalStart=-2,lastPalEnd=-1; // guich@230_18
static void setPalette(int32 *pal, int start, int end)
{
   int len = end-start+1;
   if (isColor)
   {
      if (start != lastPalStart || end != lastPalEnd || MemCmp(pal+start,lastPal+start,len*4) != 0) // guich@230_18
      {
         WinHandle old = WinSetDrawWindow(GRHANDLE_WinHandle(mainGrHandle)); // guich@200b4_22
   	   WinPalette(winPaletteSet,start,len,(RGBColorType *)pal);
   	   WinSetDrawWindow(old);
         if (len > 16 || start < 200) disposeColorHashtable(0); // cached colors are now invalid.
         lastPal = pal;
         lastPalStart = start;
         lastPalEnd = end;
      }
	}
	lastForeColor = lastTextColor = -1; // guich@421_22: colors no more valid
   xmemmove(palette+start,pal,4*len); // copy this palette to the global palette
}

static uint32 rgb888_to_rgb565(uint32 k) // guich@421_33
{
   if (k <= 255)
   {
      k = palette[k];
      k = ((k >> 3) & 0x1F) | ((k >> 5)  & 0x7E0) | ((k >> 8) & 0xF800);
   }
   return k;
}

// cuts the area x,y,width,height of the bitmap bmp with dimensions bmpW,bmpH to the (current active) screen location dstX,dstY
static void drawBitmap_(GrHandle h, int x, int y, int width, int height, int dstX, int dstY, int drawOp, int32 backColor, int32 foreColor, int doClip)
{
   int bmpW = GRHANDLE_Width(h);
   int bmpH = GRHANDLE_Height(h);
   int originalWidth = width;
   int originalHeight = height;
   RectangleType r1;
   WinDrawOperation op;
   int i;
   BitmapType *bmpt=NULL;

   dstX += TRANSX;
   dstY += TRANSY;
   if (!doClip)
   {
      /*
      | Although doClip is not set, we make sure that the area of the bitmap
      | that we want to copy IS inside its area
      | If doClip is set, then the clipping computation below ensures
      | the sanity of the operation (and is quite as fast).
      | DoClip is important for game sprites where programmers ensure that the
      | sprite will never pass screen boundaries.
      */
      if ((x <= -width) || (x >= bmpW) || (y <= -height) || (y >= bmpH))
         return;
   }
   else
   {
      /* clip the source rectangle to the source surface */
      if (x < 0)
      {
         width += x;
         dstX -= x;
         x = 0;
      }
      i = bmpW - x;
      if (width > i)
         width = i;
      if (y < 0)
      {
         height += y;
         dstY -= y;
         y = 0;
      }
      i = bmpH - y;
      if (height > i)
         height = i;

      /* clip the destination rectangle against the clip rectangle */
      if (dstX < CLIPX1)
      {
         i = CLIPX1 - dstX;
         dstX = CLIPX1;
         x += i;
         width -= i;
      }
      if ((dstX + width) > CLIPX2)
         width = CLIPX2 - dstX;
      if (dstY < CLIPY1)
      {
         i = CLIPY1 - dstY;
         dstY = CLIPY1;
         y += i;
         height -= i;
      }
      if ((dstY + height) > CLIPY2)
         height = CLIPY2 - dstY;

      /* check the validity */
      if ((width <= 0) || (height <= 0))
         return;
   }

   makeRect(x,y,width,height,r1);
   if (usePalmOSFuncs && GRHANDLE_Palette(h)) // guich@200b3: the palette is not set back; otherwise, bye bye bitmap colors
      setPalette(GRHANDLE_Palette(h),0,GRHANDLE_PaletteSize(h)-1);

   switch (drawOp)
   {
      case WIN_ERASE:        op = winErase; break;
      case WIN_MASK:         op = winMask; break;
      case WIN_INVERT:       op = winInvert; break;
      case WIN_OVERLAY:      op = winOverlay; break;
      case WIN_PAINT_INVERSE:op = winPaintInverse; break;
      case WIN_REPLACE_COLOR: // guich@200b4: since palm doenst have this kind of operation, i have to do it by myself
         {
            int i,j;
            int *bmpRow = (*vmGlobals->xmalloc)(2*width);
            int *bmpPt;
            if (!bmpRow) return;
            for (j = 0; j < height; j++)
            {
               // get the entire row
               WinSetDrawWindow((WinHandle)GRHANDLE_WinHandle(h));
               bmpPt = bmpRow;
               for (i = 0; i < width; i++) *bmpPt++ = WinGetPixel(x+i,y+j);
               bmpPt = bmpRow;
               WinSetDrawWindow((WinHandle)GRHANDLE_WinHandle(activeGrHandle));
               // change the pixels for that row
               for (i = 0; i < width; i++)
                  if (*bmpPt++ != backColor)
                     setPixelNOCT_(dstX+i,dstY+j,foreColor);
            }
            xfree(bmpRow);
         }
         return;
      case WIN_SWAP_COLORS:
         WinSetBackColor(backColor);
         WINSETFORECOLOR(foreColor); // guich@421_46: changed to the macro
         op = winPaint; // first we have to paint, and after that swap colors
         break;
      case WIN_SPRITE:
      {
         //if (bmpPt != backColor) screenPt = bmpPt;
         bmpt = (BitmapType *)GRHANDLE_BmpHandle(h);
         if (bmpt)
         {
            uint32 transp = backColor;
            if (displayBPP > 8 && BmpGetBitDepth(bmpt) > 8)
               transp = rgb888_to_rgb565(transp); // guich@421_33: in 16bpp, convert the color to 565
            if (vmGlobals->isHighDensity || vmGlobals->romVersion >= 0x05000000) // guich@421_13: fixed odd message from simulator
               BmpSetTransparentValue(bmpt,transp);
            else
               (*vmGlobals->BmpGlueSetTransparentValue)(bmpt,transp);
         }
         // no break here!
      }
      case WIN_PAINT:
      default:
         op = winPaint;
         break;
   }

   WinCopyRectangle((WinHandle)GRHANDLE_WinHandle(h),(WinHandle)GRHANDLE_WinHandle(activeGrHandle), &r1,dstX,dstY, op);
   if (drawOp == WIN_SWAP_COLORS) // now we swap the colors
   {
      makeRect(dstX,dstY,width,height,r1);
      WinSetDrawMode(winSwap);
      WinPaintRectangle(&r1,0);
      WinSetDrawMode(winPaint);
   } else
	if (bmpt)
	{
      if (vmGlobals->isHighDensity || vmGlobals->romVersion >= 0x05000000)
         BmpSetTransparentValue(bmpt,(vmGlobals->isHighDensity && vmGlobals->romVersion >= 0x05000000)?kTransparencyNone:255);
      else
   	   (*vmGlobals->BmpGlueSetTransparentValue)(bmpt, 255/*kTransparencyNone*/); // turn off transparency - guich@401_5: fixed problem in TW checking for rom version. - guich@421_13: fixed odd message from simulator
   }
}
// copies the screen width
static void drawScreen_(GrHandle hSrc, GrHandle hDest, int ySrc, int yDest, int height)
{
   RectangleType r1;
   int width = GRHANDLE_Width(hSrc);
   makeRect(0,ySrc,width,height,r1);
   if (width == GRHANDLE_Width(hDest))  // do a basic check
   {
      WinCopyRectangle((WinHandle)GRHANDLE_WinHandle(hSrc),(WinHandle)GRHANDLE_WinHandle(hDest), &r1, 0, yDest, winPaint);
   }
}
////////////////////////////////////////////////////////////////////////////
//                          COLOR PRIMITIVES                            //
////////////////////////////////////////////////////////////////////////////
// gets the bit value of the pixel, using the current translation. returns -1 if error (out of clip bounds)
static int32 getPixel_ (int x, int y)
{
   //RGBColorType rgb;
   //WinGetPixelRGB(x,y,&rgb);
   //return *((int32 *)&rgb) & 0xFFFFFFL;
   return WinGetPixel(x,y);
}
////////////////////////////////////////////////////////////////////////////
// sets the bit value, but does not translates or clip rect
static void setPixelNOCT_ (int x, int y, int32 color)
{
	WINSETFORECOLOR(color);
   WinDrawPixel(x,y);
}
////////////////////////////////////////////////////////////////////////////
// sets the bit value if inside the upper clip bounds
static void setPixelXM_ (int x, int y, int32 index)
{
   if (x < CLIPX2)
   {
   	WINSETFORECOLOR(index);
      WinDrawPixel(x,y);
   }
}
////////////////////////////////////////////////////////////////////////////
// sets the bit value if inside the clip bounds, using the current translation
static void setPixel_ (int x, int y, int32 color)
{
   x += TRANSX;
   y += TRANSY;
   if (CLIPX1 <= x && x < CLIPX2 && CLIPY1 <= y && y < CLIPY2)
   {
   	WINSETFORECOLOR(color);
      WinDrawPixel(x,y);
   }
}
////////////////////////////////////////////////////////////////////////////
// draws a line from x to x+w-1, without translate or clip rect
static void drawHLineNOCT_ (int x, int y, int w, int32 color)
{
	WINSETFORECOLOR(color);
   WinDrawLine(x,y,x+w-1,y);
}
////////////////////////////////////////////////////////////////////////////
// fills a rect from x,y to x+w-1,y+h-1, without translate or clip rect
static void fillRectNOCT_ (int x, int y, int w, int h, int32 color)
{
   RectangleType rt;
	WINSETFORECOLOR(color);
	rt.topLeft.x = x;
	rt.topLeft.y = y;
	rt.extent.x = w;
	rt.extent.y = h;
 	WinDrawRectangle(&rt,0);
}
////////////////////////////////////////////////////////////////////////////
// draws a line from x to x+w-1
static void drawHLine_ (int x, int y, int w, int32 color)
{
   x += TRANSX;
   y += TRANSY;
   if (CLIPY1 <= y && y < CLIPY2 && CLIPX1 <= (x+w) && x < CLIPX2) // line must lie inside y clip bounds, must not end before clip x1 and must not start after clip x2
   {
      if (x < CLIPX1) {w -= CLIPX1-x; x = CLIPX1;} // line start before clip x1
      if ((x+w) > CLIPX2) w = CLIPX2-x; // line stops after clip x2
   	WINSETFORECOLOR(color);
      WinDrawLine(x,y,x+w-1,y);
   }
}
////////////////////////////////////////////////////////////////////////////
// draws a line from y to y+h-1
static void drawVLine_ (int x, int y, int h, int32 color)
{
   x += TRANSX;
   y += TRANSY;
   if (CLIPX1 <= x && x < CLIPX2 && CLIPY1 <= (y+h) && y < CLIPY2) // line must lie inside x clip bounds, must not end before clip y1 and must not start after clip y2
   {
      if (y < CLIPY1) {h -= CLIPY1-y; y = CLIPY1;} // line start before clip x1
      if ((y+h) > CLIPY2) h = CLIPY2-y; // line stops after clip x2
   	WINSETFORECOLOR(color);
      WinDrawLine(x,y,x,y+h-1);
   }
}
////////////////////////////////////////////////////////////////////////////
//        Routines that can use the underlying OS to draw faster          //
////////////////////////////////////////////////////////////////////////////
static void eraseRect(int x, int y, int w, int h, int cfrom, int cto)
{
   if (translateAndClip(&x,&y,&w,&h))
   {
      RectangleType r;
      WinSetBackColor(cto);
      WINSETFORECOLOR(cfrom); // guich@421_46: changed to the macro
      makeRect(x,y,w,h,r);
      WinSetDrawMode(winSwap);
      WinPaintRectangle(&r,0);
      WinSetDrawMode(winPaint);
   }
}
////////////////////////////////////////////////////////////////////////////
static void drawCursor(int x, int y, int w, int h)
{
   if (translateAndClip(&x,&y,&w,&h))
   {
      RectangleType rt;
      rt.topLeft.x = x;
      rt.topLeft.y = y;
      rt.extent.x = w;
      rt.extent.y = h;
      WINSETFORECOLOR(isColor?0xFF:0xF); // guich@420_24: fixed problem in zire21 cursor
      WinSetDrawMode(winInvert);
      WinPaintRectangle(&rt,0);
      WinSetDrawMode(winPaint);
   }
}
////////////////////////////////////////////////////////////////////////////
static void drawCursorOutline(int x, int y, int w, int h)
{
   if (translateAndClip(&x,&y,&w,&h))
   {
      int y2 = y + h -1;
      int x2 = x + w -1;
      WINSETFORECOLOR(isColor?0xFF:0xF); // guich@420_24: fixed problem in zire21 cursor
      WinSetDrawMode(winInvert);
      WinPaintLine(x,y,x2,y);
      WinPaintLine(x2,y,x2,y2);
      WinPaintLine(x,y2,x2,y2);
      WinPaintLine(x,y,x,y2);
      WinSetDrawMode(winPaint);
   }
}
////////////////////////////////////////////////////////////////////////////
static void drawDottedCursor(int x, int y, int w, int h)
{
   int x2,y2,xx,yy;
   x += TRANSX;
   y += TRANSY;
   x2 = x + w -1;
   y2 = y + h -1;
   xx = x;
   yy = y;
   WINSETFORECOLOR(isColor?0xFF:0xF); // guich@552_17: now using WinPaintPixel instead of invPixelNOCT
   WinSetDrawMode(winInvert);
   // TODO a better clipping algotithm
   // vertical lines
   for (; h > 0; h-=2,yy+=2)
      if (CLIPY1 <= yy && yy < CLIPY2)
      {         
         if (CLIPX1 <= xx && xx < CLIPX2)
            WinPaintPixel(x,yy);
         if (CLIPX1 <= x2 && x2 < CLIPX2)
            WinPaintPixel(x2,yy);
      }
   // horizontal lines         
   for (; w > 0; w-=2,xx+=2)
      if (CLIPX1 <= xx && xx < CLIPX2) 
      {
         if (CLIPY1 <= y && y < CLIPY2)
            WinPaintPixel(xx,y);
         if (CLIPY1 <= y2 && y2 < CLIPY2)
            WinPaintPixel(xx,y2);
      }
   WinSetDrawMode(winPaint);
}
////////////////////////////////////////////////////////////////////////////
void drawRect(int x, int y, int w, int h, int32 c);
void drawRect(int x, int y, int w, int h, int32 c)
{
   drawHLine_(x,y,w,c);
   drawHLine_(x,y+h-1,w,c);
   drawVLine_(x,y,h,c);
   drawVLine_(x+w-1,y,h,c);
}
////////////////////////////////////////////////////////////////////////////
void fillRect(int x, int y, int w, int h, int32 c);
void fillRect(int x, int y, int w, int h, int32 c)
{
   x += TRANSX;
   y += TRANSY;

  // if (CLIPY1 <= (y+h) && y < CLIPY2 && CLIPX1 <= (x+w) && x < CLIPX2) // line must lie inside y clip bounds, must not end before clip x1 and must not start after clip x2
   {
      if (x < CLIPX1) {w -= CLIPX1-x; x = CLIPX1;} // line start before clip x1
      if ((x+w) > CLIPX2) w = CLIPX2-x; // line stops after clip x2
      if (y < CLIPY1) {h -= CLIPY1-y; y = CLIPY1;} // line start before clip y1
      if ((y+h) > CLIPY2) h = CLIPY2-y; // line stops after clip y2

      if (h > 0 && w > 0)
         fillRectNOCT_(x,y,w,h,c); // guich@300_51
   }
}

static struct CharBits cbits;
static void drawTextCNOT_userFont(uint16 *text, uint32 len, int x, int y, int32 color);

static void drawTextCNOT_sysFont(uint16 *text, uint16 len, int16 x, int16 y,
   int16 firstChar,
   int16 lastChar
   )
{
   int32 index;
   uint16 c;

   while (len--)
   {
      c = *text++;
      index = (int32)c - (int32)firstChar;
      ////////////// Char validation
      if (c == 13)  // guich@341_9: fixed <enter> problems with MultiEdit
         ;
      else
      if (index < 0 || index > lastChar)
      {
         // guich@320_50: fix support for invalid chars - index = numberOfChars; // MCS index
         struct UserFont *uf = currentUserFont;
         x += UF_MCS_WIDTH; // guich@401_12: just skip 4 pixels
      }
      else
      {
         WinPaintChar(c,x,y);
         x += FntWCharWidth(c);
      }
   }
}

static void drawText(uint16 *text, uint32 len, int x, int y, int32 color)
{
   struct UserFont *uf = currentUserFont;
   if (!uf->recH && vmGlobals->isHighDensity) // internal font?
   {
      RectangleType before,now;
      if (FntGetFont() != uf->fid)
         FntSetFont(uf->fid);
      WinGetClip(&before);
      now.topLeft.x = CLIPX1;
      now.topLeft.y = CLIPY1;
      now.extent.x = CLIPX2-now.topLeft.x;
      now.extent.y = CLIPY2-now.topLeft.y;
      WinSetClip(&now);
      WINSETTEXTCOLOR(color);
      WinSetDrawMode(winOverlay);
      drawTextCNOT_sysFont(text, (uint16)len, (int16)(x + TRANSX), (int16)(y + TRANSY),
         uf->fontP->firstChar, uf->fontP->lastChar - uf->fontP->firstChar);
      WinSetClip(&before);
      WinSetDrawMode(winPaint);
   }
   else
      drawTextCNOT_userFont(text, len, x, y, color);
}
////////////////////////////////////////////////////////////////////////////
//                             COMMON CODE                                //
// The routines presented here are not called intensively, so they may    //
// simply check the right code to call with an 'if'                       //
////////////////////////////////////////////////////////////////////////////
int scaleImage(GrHandle srcG, uint16 width, uint16 height, GrHandle dstG, uint16 newWidth, uint16 newHeight) // guich@420_14
{
   GrHandle last = lastGrHandle;

   int32 h = newHeight << 1;
   int32 hi = height << 1;
   int32 hf = height / h;
   int32 wf = 0;
   int32 w = newWidth << 1;
   int32 wi = width << 1;
   int32 x,y,oldIY=-1;
   Pixel *p;
   byte lowMemoryMode = 0;

   Pixel *buf = (Pixel *)(*vmGlobals->xmalloc)(sizeof(Pixel)*(int32)width*(int32)height);
   if (buf == NULL) // guich@401_30: now if there's no memory to allocate the full buffer, use a slower routine
   {
      lowMemoryMode = 1;
      buf = (Pixel *)(*vmGlobals->xmalloc)(sizeof(Pixel)*(int32)width);
      if (buf == NULL)
         return 0;
   }
   GR_ENTER // guich@400_59
   if (lowMemoryMode)
   {
      for (y=0; y < newHeight; y++, hf += hi)
      {
         int32 iy = hf/h;
         if (iy != oldIY) // guich@402_44: only scan if needed
         {
            oldIY = iy;
            SetDrawWindow(srcG);
            p = buf;
            for (x = 0; x < width; x++)
               *p++ = (*getPixel)(x,iy); // guich@402_43: replaced y by iy
            // now set all pixels
            SetDrawWindow(dstG);
         }
         wf = width/w;
         for (x=0; x < newWidth; x++,wf+=wi)
            (*setPixel)(x,y,buf[wf/w]);
      }
   }
   else
   {
      // first get all pixels
      SetDrawWindow(srcG);
      p = buf;
      for (y = 0; y < height; y++)
         for (x = 0; x < width; x++)
            *p++ = (*getPixel)(x,y);
      // now set all pixels
      SetDrawWindow(dstG);
      for (y=0; y < newHeight; y++, hf += hi)
      {
         int32 iy = hf/h;
         wf = width/w;
         p = &buf[iy*width]; // guich@401_30: changed to int32
         for (x=0; x < newWidth; x++,wf+=wi)
            (*setPixel)(x,y,p[wf/w]);
      }
   }
   GR_EXIT
   xfree(buf);
   if (last != NULL) SetDrawWindow(last); // guich@421_69: check if last is null
   return 1;
}

extern uint32 *colorTable;
int read8bppImage(int32 width, int32 height, byte *in);

int read8bppImage(int32 width, int32 height, byte *in) // guich@420_15
{
   BitmapType *bmp;
   BitmapTypeV3 *bmpv3;
   Err err;
   byte *data;
   int32 i,scanline,scanline2;

   // first, we create a 8bpp bitmap
   GR_ENTER
   bmp = BmpCreate(width,height,8,NULL,&err);
   if (err != errNone)
      return 0;

   if (vmGlobals->isHighDensity)
     	bmpv3 = BmpCreateBitmapV3(bmp,MaxScreenX==160?kDensityLow:kDensityDouble,BmpGetBits(bmp),0);
   else
      bmpv3 = (BitmapTypeV3*)bmp;

  	data = BmpGetBits((BitmapPtr)bmpv3);
   scanline = ((width + 3)>>2)<<2;
   scanline2 = ((width + 1)>>1)<<1;
   // now we fastly move the original bmp to the bitmap
   for (i=0; i < height; i++) // note: we must always allocate a buffer bc the bmp is stored upsidedown (who was the "genius" who invented this???)
      xmemmove(data + (height-i-1)*scanline2, in + i * scanline, width);

   // and draw the bmp into the target
   WinDrawBitmap((BitmapPtr)bmpv3,0,0);
   // free resources and return
   BmpDelete((BitmapPtr)bmpv3);
   if ((void*)bmpv3 != (void*)bmp)
      BmpDelete(bmp);
   GR_EXIT

   return 1;
}

static void fillInRow(GrHandle h, byte *row, int y)
{
   int w = GRHANDLE_Width(h);
   int x =0;
   WinHandle oldH = WinSetDrawWindow((WinHandle)GRHANDLE_WinHandle(h));
   while (w--)
      *row++ = WinGetPixel(x++,y);
   WinSetDrawWindow(oldH);
}

static void clearScreen(int32 color)
{
   fillRectNOCT_(0,0,MaxScreenX,MaxScreenY,color);
}
static void imageChangeColors(GrHandle h, int width, int height, int32 colorFrom, int32 colorTo)
{
   WinHandle oldH = WinSetDrawWindow((WinHandle)GRHANDLE_WinHandle(h));
   int x,y;
   int colorFromIndex = colorFrom;
   WINSETFORECOLOR(colorTo);
   for (y = 0; y < height; y++)
      for (x = 0; x < width; x++)
         if ((WinGetPixel(x,y)) == colorFromIndex)
            WinDrawPixel(x,y);
   WinSetDrawWindow(oldH);
}
static void drawTextCNOT_userFont(uint16 *text, uint32 len, int x, int y, int32 color)
{
   static byte ands8[8] = {0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};
   struct UserFont *uf = currentUserFont;
   struct CharBits *bits = &cbits;
   byte isUff = IS_UNICODE_FONT(uf);
   int32 rowWidthInBytes;
   int32 numberOfChars;
   byte *bitmapTable;
   int16 *bitIndexTable;
   int32 offset;
   int32 index;
   byte width,w;
   byte *start,*current;
   byte startBit,currentBit;
   byte *ands;
   int xx,yy;
   uint16 pixelCount;
   byte height = (byte)uf->fontP->fRectHeight,h;
   uint32 bufSize = sizeof(PointType)*height*uf->fontP->fRectWidth;
   uint16 firstChar,lastChar;

   // speedup
   if (!isUff)
   {
      rowWidthInBytes = uf->rowWidthInBytes;
      numberOfChars   = uf->numberOfChars;
      bitmapTable     = uf->bitmapTable;
      bitIndexTable   = uf->bitIndexTable;
      firstChar       = uf->fontP->firstChar;
      lastChar        = uf->fontP->lastChar - firstChar; // guich@340_46: sub firstChar
   }
   WINSETFORECOLOR(color);

   if (txtBufSize < bufSize)
   {
      txtBufSize = bufSize;
      txtPixels = (*vmGlobals->xrealloc)(txtPixels,&txtBufSize);
   }

   while (len--)
   {
      if (*text == 13)  // guich@341_9: fixed <enter> problems with MultiEdit
         goto next;
      if (isUff)
      {
         uffSetCharBits(*text, uf, bits);  // pgr@402_50
         width = bits->width;
         if (width == 0)
            goto skip;
         bitmapTable = bits->bitmapTable;
         rowWidthInBytes = bits->rowWIB;
         offset = bits->offset;
      }
		else
		{
         index = (int32)*text - firstChar;
         if (index < 0 || index > lastChar)
            goto skip;
         // Get the source x coordinate and width of the character
         offset    = bitIndexTable[index];
         width     = (byte)(bitIndexTable[index+1] - offset);
         if (width == 0) // guich@340_46: show the MCS char if no width (glyph not present)
         {
skip:
            width = UF_MCS_WIDTH;
            goto next;
         }
		}
      start     = bitmapTable + (offset >> 3);
      startBit  = offset & 7;

      yy = y + TRANSY;
      pixelCount = 0;
      // draws the char
      for (h = height; h--; start += rowWidthInBytes) // draw each row
      {
         //if (CLIPY1 <= yy && yy < CLIPY2) guich@330_53: HPC doesnt seem to like the optimizations made with this code
         if (yy >= CLIPY2)
            break;
         if (yy < CLIPY2)
         {
            xx = x + TRANSX;
            current = start;
            currentBit = startBit;
            ands = ands8+currentBit;
            for (w = width; w--;xx++)  // draw each pixel
            {
               if (xx >= CLIPX2)
                  break;
               //if (CLIPX1 <= xx && xx < CLIPX2)
               if (CLIPX1 <= xx)
               {
                  if (*current & *ands++)
                  {
                     PointType *pt = &txtPixels[pixelCount++];
                     pt->x = xx; //WinDrawPixel(xx,yy);
                     pt->y = yy;
                  }
               }
               else
                  ands++;
               if (++currentBit == 8) // finished this byte?
               {
                  currentBit = 0; // reset counter
                  ands = ands8;   // reset test bit pointer
                  current++;      // inc current byte
               }
            }
         }
         yy++;
      }
      if (pixelCount) WinPaintPixels(pixelCount, txtPixels);
next:
      x += width;
      text++;
   }
}
static void eraseWindow(WinHandle wh, int width, int height)
{
   int y=0;
   WinHandle oldwh = &vmGlobals->wabaFrmP->window;
   WinSetDrawWindow(wh);
   WINSETFORECOLOR(rgb2index(0xFF,0xFF,0xFF));
   while (height--)
   {
      WinDrawLine(0,y,width-1,y);
      y++;
   }
   WinSetDrawWindow(oldwh);
}
void graphicsStart()
{
	if (!graphicsActive && !vmGlobals->errorWinDisplayed)
	{
   	Word error;
      if (vmGlobals->isHighDensity)
         WinSetCoordinateSystem(kCoordinatesNative);
      mainGrHandle = CreateOffscreenWindow(0, MaxScreenX,MaxScreenY,0,&error);
      if (error != 0)
      {
         (*vmGlobals->debug)("could not start graphics mode");
         return;
      }
      SetDrawWindow(mainGrHandle);
      eraseWindow(GRHANDLE_WinHandle(mainGrHandle), GRHANDLE_Width(mainGrHandle),GRHANDLE_Height(mainGrHandle));
   }
   graphicsActive = 1;
}
void graphicsSwap(int on)
{
   if (on && !graphicsActive && !vmGlobals->errorWinDisplayed)
      graphicsActive = 1;
   else
   if (!on && graphicsActive)
      graphicsActive = 0;
}
void graphicsReset() // guich@300_60: called when another form is set active: the screen address may had changed - this is only called on PalmOS >= 3.5
{
   WinHandle wh;
   byte *scr;
   byte *mainScr;
   if (vmGlobals->isHighDensity)
      WinSetCoordinateSystem(kCoordinatesNative);
   wh = (WinHandle)GRHANDLE_WinHandle(mainGrHandle);
   scr = (byte *)BmpGetBits(WinGetBitmap(wh));
   mainScr = GRHANDLE_MemAddr(mainGrHandle);

   GRHANDLE_MemAddr(mainGrHandle) = scr;
   if (wh == GRHANDLE_WinHandle(activeGrHandle))
      activeScr = GRHANDLE_MemAddr(mainGrHandle);
   SemaphoreReserve(1);
   (*vmGlobals->executePostEvent)(-6000,0,0,0,0,0);
   SemaphoreRelease(1);
}

// the following variables must be set in this method: MaxScreenX, MaxScreenY, displayBPP, isColor, maxColors
void graphicsSetup()
{
   Coord cx,cy;
   Err err;
   uint32 version;
   uint32 currentDepth;
   // guich@421_: if there's a dynamic input area, close it before the program loads
/*   todo or not todo
   err = FtrGet(pinCreator, pinFtrAPIVersion, &version);
   if (!err && version)
      PINSetInputAreaState(pinInputAreaClosed);*/

   // enable native 320x320 resolution
   err = FtrGet(sysFtrCreator, sysFtrNumWinVersion, &version);
   if (err == 0 && version >= 4) // guich@402_34
      vmGlobals->isHighDensity = 1;
   if (vmGlobals->isHighDensity)
      WinSetCoordinateSystem(kCoordinatesNative); // guich@402_34
   WinGetDisplayExtent(&cx,&cy);
   displayBPP = 2;      // bits per pixel for the display
   maxColors = 4;
   MaxScreenX = cx;
   MaxScreenY = cy;
   isColor = 0;

   usePalmOSFuncs = (vmGlobals->romVersion >= 0x03500000)?1:0;
   palette = (int32*)(*vmGlobals->xmalloc)(1024);
   if (!palette)
      return;
   xmemmove(palette,systemPalette,4*256); // copy this palette to the global palette
   colorHashTable = (struct htElemColor **)((*vmGlobals->xmalloc)(sizeof(struct htElemColor)*(COLOR_HASHSIZE+1)));
   if (!colorHashTable)
      return;
   xmemzero(colorHashTable,sizeof(struct htElemColor)*(COLOR_HASHSIZE+1));

   if (usePalmOSFuncs)
   {
      UInt32 depth,pow2,width,height;
      int bpp;
      Boolean enableColor;

      // get the screen sizes and deph
	   WinScreenMode(winScreenModeGet, &width, &height, &currentDepth, 0);
	   depth = currentDepth;
      //	guich@420_10: use the other information, so we use full height for 480 res devices   MaxScreenX = width; MaxScreenY = height;
	   oldBPP = depth;

	   // get if color is supported or only grayscale
	   WinScreenMode(winScreenModeGetSupportsColor, 0, 0, &depth, &enableColor);
	   isColor = (int)enableColor;
      // get the supported depths
      WinScreenMode(winScreenModeGetSupportedDepths, 0, 0, &depth, 0);

      if (width == 160 && height == 160 && depth == 3) // guich@200b4_101: mono devices must use internal routines.
         usePalmOSFuncs = 0;
      else
      {
   	   // an array of bits is returned. scans for the maximum depth.
   	   pow2 = (UInt32)1 << (UInt32)31;
   	   bpp = 32;
   	   while (pow2)
   	      if (pow2 & depth)
   	      {
   	         displayBPP = bpp;
   	         maxColors = pow2*2;
   	         break;
   	      }
   	      else
   	      {
   	         pow2 /= 2;
   	         bpp -= 1; // available modes expected: 32,24,16,8,4,2
   	      }
   	}
	}

   //guich@421_33: fixed problems in 16bpp - guich@200b3: for some reason, the palette is not being correctly applied in 16bpp. so, we narrow down the bpp.
   if (!vmGlobals->wants16bpp && displayBPP > 8)
   {
      displayBPP = 8;
      maxColors = 256;
   }
	if (currentDepth != displayBPP) // guich@421_33: zire71 supports 16 but may be in a lesser bpp
   	WinScreenMode(winScreenModeSet,0,0,&displayBPP,0);
	vmGlobals->screenBPP = displayBPP; // save it for use in other routines

	vmGlobals->maxColors = maxColors;		// used in settings
	vmGlobals->screenWidth = MaxScreenX;
	vmGlobals->screenHeight = MaxScreenY;
}
////////////////////////////////////////////////////////////////////////////
// called by the system to let the port register the primitives
static void registerPrimitives(GrHandle h)
{
   // under WIN we have only one set of functions
   getBitmapRow      = &getBitmapRow_;
   drawVLine         = &drawVLine_;
   drawHLine         = &drawHLine_;
   setPixel          = &setPixel_;
   getPixel          = &getPixel_;
   setPixelXM        = &setPixelXM_;
   setPixelNOCT      = &setPixelNOCT_;
   drawHLineNOCT     = &drawHLineNOCT_;
   getColor          = &getColor_;
   drawBitmap        = &drawBitmap_;
   //eraseVLineNOCT    = &eraseVLineNOCT_;
   fillRectNOCT      = &fillRectNOCT_;
   drawScreen        = &drawScreen_;
}
static int getWidthInBPP(int width)
{
   return ((int)(width+1)/2) * 2; // always 8 bpp , odd width
}
static byte *allocWinMemory(int32 width, int32 height, VoidPtr *ptr1, VoidPtr *ptr2, VoidPtr *ptr3, int accepts8bpp) // guich@421a_67: added last parameter
{
   Err error;
   int bpp = displayBPP;
   if (vmGlobals->isHighDensity)
   {
   	*ptr3 = BmpCreate(width,height,bpp,NULL,&error);
   	if (*ptr3)
      	*ptr1 = BmpCreateBitmapV3(*ptr3,MaxScreenX==160?kDensityLow:kDensityDouble,BmpGetBits(*ptr3),0);
   	if (accepts8bpp && error && bpp == 16) //sysErrNoFreeResource : guich@421_48: try to alloc an image with less bpp - better than just return with error
	   {
	      error = 0;
      	*ptr3 = BmpCreate(width,height,8,NULL,&error);
      	if (*ptr3)
         	*ptr1 = BmpCreateBitmapV3(*ptr3,MaxScreenX==160?kDensityLow:kDensityDouble,BmpGetBits(*ptr3),0);
	   }
	   if (error) // ok, no memory! go back
	   {
	      if (*ptr3) // guich@421_47
	         BmpDelete(*ptr3);
	      *ptr3 = NULL;
         return NULL;
      }
      *ptr2 = WinCreateBitmapWindow((BitmapType *)*ptr1, &error);
   	if (error)
   	{
	      if (*ptr1) // guich@421_47
	         BmpDelete(*ptr1);
	      if (*ptr3) // guich@421_47
	         BmpDelete(*ptr3);
	      *ptr3 = NULL;
	      *ptr1 = NULL;
         return NULL;
      }
   	return BmpGetBits((BitmapType *)*ptr1);
   }
   else
   {
      Err error;
      int bpp = displayBPP;
 		*ptr1 = BmpCreate(width,height,bpp,NULL,&error);
		if (error) //sysErrNoFreeResource
	      return NULL;
  		*ptr2 = WinCreateBitmapWindow((BitmapType *)*ptr1, &error);
		if (error)
	      return NULL;
		return BmpGetBits((BitmapType *)*ptr1);
   }
}
static void FreeGrHandle(GrHandle h);
static void DestroyOffscreenWindow(GrHandle h)
{
   if (h && GRHANDLE_MemAddr(h))
   {
	   if (GRHANDLE_Palette(h))
	      xfree(GRHANDLE_Palette(h));
	   if (GRHANDLE_WinHandle(h))
	      WinDeleteWindow((WinHandle)GRHANDLE_WinHandle(h),false);
	   if (GRHANDLE_BmpHandle(h))
	      BmpDelete((BitmapType *)GRHANDLE_BmpHandle(h));
      if (GRHANDLE_IsBitmap(h))
         BmpDelete((BitmapType *)GRHANDLE_BmpHandleAux(h));
      if (!GRHANDLE_IsBitmap(h) && (!mainGrHandle || (GRHANDLE_MemAddr(h) != GRHANDLE_MemAddr(mainGrHandle))))
         xfree(GRHANDLE_MemAddr(h));
      FreeGrHandle(h);
   }
}
// if allocWinMem is true, a new "screen" will be allocated. if is false, same "screen" of MainWin is used, but with its own clip and translate
GrHandle CreateOffscreenWindow(WObject obj, int width, int height, int allocWinMem, Word *error)
{
   GrHandle h;
   VoidPtr ptr1=NULL,ptr2=NULL,ptr3=NULL;
   byte *bmp=NULL;
   *error = 1;

   if (obj && (*vmGlobals->pushObject)(obj) == -1) return NULL; // GC can be called on createOffscreenWindow
   if (allocWinMem)
      bmp = allocWinMemory(width,height,&ptr1,&ptr2,&ptr3, allocWinMem == 1); // even if the bitmaps are not handled directly, something different of null must be returned here - guich@421a_67: if allocWinMem is 2 (called by graphicsStartImage), we won't accept 8bpp images if no room for 16bpp ones.
   *error = 1;
   if (!bmp && allocWinMem)
   {
      (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "Not enough memory for bit map"); // guich@200b4_78
      goto freeAndExit;
   }
   h = (GrHandle)(*vmGlobals->xmalloc)(sizeof(GraphicsContext));
   xmemzero(h,sizeof(GraphicsContext)); // @t
   if (!h)
   {
      (*vmGlobals->VmQuickError)(ERR_CantAllocateMemory, "Not enough memory for GraphicsContext"); // guich@200b4_78
      goto freeAndExit;
   }
   GRHANDLE_WidthInBPP(h)   = getWidthInBPP(width); // saves the size in bytes
   GRHANDLE_BmpHandle(h)    = ptr1;
   GRHANDLE_BmpHandleAux(h) = ptr3;
   GRHANDLE_WinHandle(h)    = allocWinMem?ptr2:&vmGlobals->wabaFrmP->window;
   GRHANDLE_MemAddr(h)      = allocWinMem?bmp:(byte *)BmpGetBits(WinGetBitmap(GRHANDLE_WinHandle(h)));
   GRHANDLE_TransX(h)       = 0;
   GRHANDLE_TransY(h)       = 0;
   GRHANDLE_Width(h)        = width;
   GRHANDLE_Height(h)   	 = height;
   GRHANDLE_ClipX1(h)       = 0;
   GRHANDLE_ClipY1(h)       = 0;
   GRHANDLE_ClipX2(h)       = width;
   GRHANDLE_ClipY2(h)       = height;
   GRHANDLE_PaletteSize(h)  = 0;
   GRHANDLE_Palette(h)      = NULL;
   *error = 0;
   if (obj) (*vmGlobals->popObject)(obj);
   return h;
freeAndExit:
   if (obj) (*vmGlobals->popObject)(obj);
   if (bmp) xfree(bmp);
   return NULL;
}
void SetDrawWindow(GrHandle h)
{
   WinHandle wh;
   if (!h) return; // guich@421_69: also check here and return in case of problems.
   if (!activeGrHandle)
   {
      currentUserFont = vmGlobals->allFonts+0;
      registerPrimitives(h);
   }
   wh = (WinHandle)GRHANDLE_WinHandle(h);
   if (wh)
      WinSetDrawWindow(wh);
   activeGrHandle = h;
   transXY = MAKE32(GRHANDLE_TransX(h),GRHANDLE_TransY(h));
   clipX=MAKE32(GRHANDLE_ClipX1(h),GRHANDLE_ClipX2(h));
   clipY=MAKE32(GRHANDLE_ClipY1(h),GRHANDLE_ClipY2(h));

   // guich@300_47: after a new form was called, the main address may change, so we make sure we always use it
   activeScr = GRHANDLE_MemAddr(h);
   maxX = GRHANDLE_Width(h);
   maxY = GRHANDLE_Height(h);
   widthInBPP = GRHANDLE_WidthInBPP(h);
   mustSwap = 0;//displayBPP == 8 && !BmpColortableSize(WinGetBitmap(wh)); // guich@200b4_160
   lastForeColor = lastTextColor = -1;
}
////////////////////////////////////////////////////////////////////////////
void graphicsStop()
{
   if (graphicsActive)
   {
     	WinScreenMode(winScreenModeSet, 0, 0, &oldBPP, 0);
    	xfree(mainGrHandle); // guich@200b4_61: free this mainGrHandle too.
      xfree(palette);
    	disposeColorHashtable(1);
      graphicsActive = 0;
      xfree(txtPixels);
   }
   xfree(systemPalette); // guich@421_20
   releaseAllFonts();
}
