/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  Copyright (C) 1998, 1999 Wabasoft <www.wabasoft.com>, All Rights Reserved    *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/

/*

Welcome to the (Super)Waba VM source. This is the platform independent code for the
interpreter, class parser and garbage collector. I like the code myself (of
course I wrote it, so that's not saying much) and hopefully it doesn't seem
too complicated. Actually, I'm hoping it reads rather easily given what it
does.

Have a blast!

Rick Wild


Guilherme Campos Hazan (Guich) notes:

This code had changed a lot since Waba 1.0b. It is still easy to understand,
but some things may look quite strange, because i changed many (hundreds!)
parts of the code to improve performance.

*/

//#define ENABLE_TRACE


#define __DEFINE_FUNCTIONS
#define BUILDING_SW_STATIC
#include "superwaba.h"

#ifdef WIN32
 #include "ce\SWLibStatic.h"
 #define sysInvalidRefNum 0
#endif

#ifdef PALMOS
 #pragma opt_common_subs off // this must be off to make switchPC work!
#endif

#ifndef __SWNATIVE_CUSTOM_LIB_H
 You had included the wrong CustomLib.h file. Be sure it is the SWNatives\CustomLib.h
#endif

struct VMGlobals *vmGlobals;

static uint16 SWLibRefNum = sysInvalidRefNum;	// library reference number

//
// TYPES AND METHODS
//

#include "opcodes.h"

//
// Error Handling
//

static char *errorMessages[] =
{
    "Sanity",
    "Incompatible device",
    "Can't access SuperWaba classes",
    "Can't access app classes",
    "Can't allocate memory",
    "Out of class memory",
    "Out of object memory",
    "Native stack overflow",
    "Native stack underflow",
    "Stack overflow",

    "Bad class",
    "Bad opcode",
    "Can't find class",
    "Can't find method",
    "Can't find field",
    "Null object access",
    "Null array access",
    "",
    "Divide by zero",
    "",
    "Class too large",

    "Can't allocate object memory",
    "Bad Utf String",
    "No fonts found. Make sure [H,M,L]SW*.pdb font files are located under SuperWaba default directory",
    "Method called on array/Missing class",
    "Can't find native method implementation",
    "Uncaught exception",
    "Class in wrong package",
    "Can't initialize application",
};

// guich@200 all defines moved to superwaba.h

//
// protected function prototypes
//

#ifdef PALMOS
#define debug16 debug
#else
static void debug16(char *buf);
#endif
void dumpStackTrace();
static byte compact();
static byte growStack(byte **stackStart, byte **stackPtr, byte **stackEnd, byte **ptr1, byte **ptr2, byte **ptr3, uint32 *stackSizeInBytes);
static void VmInit();
static void VmError(uint16 errNum, WClass *iclass, UtfString *desc1, UtfString *desc2);
static WObject VmStartApp(char *className);
static void VmStopApp(WObject mainWinObj);
static void VmFree();
static int createVirtualMethodMap(WClass *wclass);
static int initObjectHeap();
static int createChunk(uint32 chunkSize);
static void FreeObjectHeap();
static void markObject(WObject obj);
static void sweep();
static void setClassHooks(WClass *wclass);
int64*	__rt_shrs64(int64 *result,int64 left,short count);
int64*	__rt_shru64(int64 *result,int64 left,short count);
short __rt_cmps64(int64 left,int64 right);
static WClassField *getFieldByIndex(WClass *wclass inA0, uint16 fieldIndex inD0) inD0;
static void getUtfString2(WClass *wclass inA1, uint16 idx inD0, UtfString *s inA0); // Don't inline
_inline_ Var constantToVar2(WClass *wclass, uint16 idx);
// guich@230_14
static WClass *outOfMemoryClass;
// nmpalm_b.c
uchar *lockWarpRec(char *path, uint16 pathLen, uint32 *size);
void unlockWarpRec(uchar *ptr, uint16 pathLen);
uint32 getDeviceFreeMem();
static void deleteChunks();
#ifdef ENABLE_TRACE
static void trace(WClass *iclass, WClassMethod *imethod);
#endif

//
// global vars
//

int GCCreated = 0; // used by xmalloc
static int vmInitialized = 0;

// virtual machine stack
#define VM_STACK_COUNT (((int32)vmStackPtr-(int32)vmStackStart)>>2)  // guich@550: avoid ldiv
static Var *vmStackStart;
static Var *vmStackEnd;
static Var *vmStackPtr;
static uint32 vmStackSizeInBytes;
static Var *exceptionStackPtr; // guich@330_22

// native method stack
#define NM_STACK_COUNT (((int32)nmStackPtr-(int32)nmStackStart)>>2) // guich@550: avoid ldiv
static WObject *nmStackStart;
static WObject *nmStackEnd;
static WObject *nmStackPtr;
static uint32 nmStackSizeInBytes;

static uint32 classHeapSize;
static uint32 classHeapUsed;
static uint32 totalClassHeap; // guich@341_

// note: objectHeapSize is given by heap.memSize

// keep these prime numbers for best distribution and keep them a sequence of 1's (255 = 1111111b)
#define CLASS_HASH_SIZE 255
#define OVERRIDE_HASH_SIZE 127
uint16 methodHash[OVERRIDE_HASH_SIZE+1];

// class heap
static uchar *classHeap;
static WClass *classHashList[CLASS_HASH_SIZE+1]; // guich@400_10: if the hashcode is FF, FF % FF = FF, thus, the array must range 0-FF, not 0-FE

// error status
static ErrorStatus vmStatus;

// pointer to String class (for performance)
static WClass *stringClass;

// support for exceptions
static WClass *throwableClass;
static WObject thrownException = 0;
static WObject firstException = 0; // guich@340_1
static char
   *NullPointerEx  = "java/lang/NullPointerException",
   *ArrayIndexEx   = "java/lang/ArrayIndexOutOfBoundsException",
   *DivideByZeroEx = "java/lang/ArithmeticException",
   *RuntimeEx      = "java/lang/RuntimeException",
   *OutOfMemoryEx  = "java/lang/OutOfMemoryError",
   *ClassCastEx    = "java/lang/ClassCastException"
   ;

/////
//
// OS SPECIFIC AREA CONTAINING RUNNERS
//

#if defined(PALMOS)
#include "nmpalm_bridge.c"
#elif defined(WIN32)
#include "ce\nmwin32_bridge.c"
#endif


UtfString getClassName(WObject obj)
{
	return ((WClass*)WOBJ_class(obj))->className;
}

extern uint32 emptyChunks;
void gc_() // called when the system is out of memory. call gc and release memory to the system by deleting unused chunks
{
   gc();
   if (emptyChunks > 0)
      deleteChunks();
}
//
// Method Execution
//

//
// This is the interpreter code. Each method call pushes a frame on the
// stack. The structure of a stack frame is:
//
// local var 1
// local var 2
// local var N
// local stack 1
// local stack 2
// local stack N
//
// when a method is called, the following is pushed on the stack before
// the next frame:
//
// wclass
// method
// pc
// var
// stack
//
// NOTE: You can, of course, increase the performance of the interpreter
// in a number of ways. I've found a good book on assembly optimization
// to be:
//
// Inner Loops - A sourcebook for fast 32-bit software development
// by Rick Booth
//

// guich@200 - transforms 2 consecutive vars in one double; the first var stores the lower 32 bits and the 2nd var stores the upper 32 bits of the double
/*_inline_ void copy64(Var *src, Var *dest)
{
   *dest = *src;
   *(dest+1) = *(src+1);
} */

/** If, after compile, POSE issue:
"just executed an illegal or unknown machine language instruction.
The opcode executed was 0x20A (or anything else)"
The problem is that CW changed the register used by the next function,
when optimizing the code, or that the displacement to the jump table has
changed.
Note: if in the code appears
 move.b    (a3),d7
 moveq     #0,d3
 move.b    d7,d3
 move.w    d3,d0
you must change the code to make the d7 disappear, otherwise, the VM will
halt. The d7 will be used in 1 other case statement: op_return
This can be fixed by ensuring that
 #pragma opt_common_subs off // this must be off to make switchPC work!
 is off

*/

#if defined(PALMOS)
   // hard code the main switch of executeMethod.
   static _inline_ asm void switchPC(uint16 pc: __D0)
   {
   // note: CW8 compiles wrong this code
   // moveq     #0,d0    -- generated by CW in function call
   // move.b    (a3),d0  -- generated by CW in function call
      add  d0,d0          // d0 *= 2
      dc.w 0x303B, 0x001A // move.w   26(pc,d0.w),d0   -- get displacement value
      dc.w 0x4EFB, 0x0016 // jmp      22(pc,d0.w)      -- jump to case statement
   }
#elif defined(WINCE)
   #define switchPC(pc)
#else
/* win32 makes this...
0042418A  mov         edx,dword ptr [pc]
0042418D  movzx       eax,byte ptr [edx]
00424190  mov         dword ptr [ebp-234h],eax
00424196  cmp         dword ptr [ebp-234h],0FFh
004241A0  ja          $L97934+11h (427563h)
004241A6  mov         ecx,dword ptr [ebp-234h]
004241AC  movzx       edx,byte ptr [ecx+427E14h]
004241B3  jmp         dword ptr [edx*4+427B9Ch]

__asm
    { // ...but it can be turned into this - this improves the VM code in 10%
      mov   edx,dword ptr [pc]
      movzx ecx,byte ptr [edx]
      movzx edx,byte ptr [ecx+427D9Fh] // note: these displacement values are the
      jmp   dword ptr [edx*4+427B27h]  // same ones found with disassembly
    }
*/
   #define switchPC(pc)

#endif

#define STACKSAVESIZE 4

Var * executeMethod(WClass *wclass, WClassMethod *method, Var params[], uint32 numParams)
{
    Var *var;
    Var *stack;
    uchar *pc=0;
    uchar *oldpc=0;
    Var *baseFramePtr=NULL;
    Var savedStack[2]; // guich@500_5: save the old values here bc they will be overridden

    Comp64 c,c1,c2; // note: don't make these local variables, otherwise the code will alloc +700 bytes for the stack!
    uint16 classIndex;
    int32 ndim, key, low, mid, high, npairs;
    float32 f;
    uchar *npc;
    int comp;
    UtfString className;
    WClass *target;
    uint16 fieldIndex;
    WClassField *field;
    Var vtemp;
#ifdef PALMOS_AUTO_RESET // guich@402_22: avoid the auto-reset "feature" of Zire71 by releasing the semaphore
    #define RESET_COUNTER 100
    int avoidResetCounter = RESET_COUNTER;
    #define AVOID_AUTO_RESET if (avoidResetCounter-- == 0) \
    { \
      avoidResetCounter = RESET_COUNTER; \
      SemaphoreRelease(1); \
      SemaphoreReserve(1); \
   }
   #define AVOID_AUTO_RESET_NOW \
   { \
      SemaphoreRelease(1); \
      SemaphoreReserve(1); \
   }
#else
   #define AVOID_AUTO_RESET
   #define AVOID_AUTO_RESET_NOW
#endif

    byte returnedValueCount=0; // guich@500_5: now we don't use the vmGlobals to store the returned value; instead, we use...
    Var returnedValue[2];      // this Var array, which is pushed into the object routine.

    // the variables wclass, method, var, stack, and pc need to be
    // pushed and restored when calling methods using "goto methoinvoke"

    // also note that this method does recurse when we hit static class
    // initializers which execute within a class load operation and this
    // is why we exit when we keep trace of the baseFramePtr.

    if (METH_isNative(method)) // nopt removing
        goto end; // can't execute native methods directly
    //if (method->code.codeAttr == NULL)
    //    return; // method has no code attribute - compiler is broken - guich@200b4r5: test is now made on the method load

    // guich@320_15: since METH_localsStack2 max value is 256, i changed vmStackEnd to be vmStackEnd-300
    //if (vmStackPtr + STACKSAVESIZE + METH_localsStack2(method) >= vmStackEnd) // VERY IMPORTANT NOTE: if we use the pointers directly here to be passed by reference, the compiler will not be able to optimize them, thus making the code larger and MUCH SLOWER!
    if (vmStackPtr > vmStackEnd)
    {
       Var *_vmStackStart = vmStackStart;
       Var *_vmStackEnd = vmStackEnd;
       Var *_vmStackPtr = vmStackPtr;
       if (!growStack((byte **)&_vmStackStart, (byte **)&_vmStackPtr, (byte **)&_vmStackEnd, NULL, NULL, NULL, &vmStackSizeInBytes))
          goto stack_overflow_error;
       vmStackStart = _vmStackStart;
       vmStackEnd = _vmStackEnd - 300;
       vmStackPtr = _vmStackPtr;
    }
    baseFramePtr = vmStackPtr;

    // push params into local vars of frame
    if (numParams > 0)
       xfer32((byte *)params,(byte *)vmStackPtr,numParams);

methodinvoke:
#ifdef ENABLE_TRACE
      trace(wclass,method);
#endif

    // push active stack frame:
    //
    // local var 1
    // ...
    // local var N
    // local stack 1
    // ...
    // local stack N
    // method pointer
    // class pointer
    var = vmStackPtr;
    vmStackPtr += method->maxLocals;
    stack = vmStackPtr;
    vmStackPtr += method->maxStack;
    (vmStackPtr++)->refValue = method;
    (vmStackPtr++)->refValue = wclass;

    pc = method->codeStart;//METH_code(method);

    // guich@200b4: important: this code has been highly optimized with CW. It expects
    // some order in post-incrementing and pre-decrementing. The order is: the pre/post must be
    // performed first in the line, and after the = operator must have priority. So, be aware
    // that your compiler must generate almost the same operation order that i'm expecting.
    // nopt = not optimizable

step:
    switchPC(*pc); // optimize the case statement. This results in a 15% global VM speed improvement
    switch (*pc)
    {
        case OP_nop:
            pc++;
            break;
        case OP_aconst_null:
            (*stack++).obj = 0;
            pc++;
            break;
        case OP_iconst_m1:
            (*stack++).intValue = -1;
            pc++;
            break;
        case OP_iconst_0:
            (*stack++).intValue = 0;
            pc++;
            break;
        case OP_iconst_1:
            (*stack++).intValue = 1;
            pc++;
            break;
        case OP_iconst_2:
            (*stack++).intValue = 2;
            pc++;
            break;
        case OP_iconst_3:
            (*stack++).intValue = 3;
            pc++;
            break;
        case OP_iconst_4:
            (*stack++).intValue = 4;
            pc++;
            break;
        case OP_iconst_5:
            (*stack++).intValue = 5;
            pc++;
            break;
        case OP_fconst_0:
            (*stack++).floatValue = 0.0f;
            pc++;
            break;
        case OP_fconst_1:
            (*stack++).floatValue = 1.0f;
            pc++;
            break;
        case OP_fconst_2:
            (*stack++).floatValue = 2.0f;
            pc++;
            break;
        case OP_bipush:
            (*stack++).intValue = ((char *)pc)[1];
            pc += 2;
            break;
        case OP_sipush:
            (*stack++).intValue = getInt16(pc+1);
            pc += 3;
            break;
        case OP_ldc:
            *stack++ = constantToVar2(wclass, (uint16)pc[1]);
            pc += 2;
            break;
        case OP_ldc_w:
            *stack++ = constantToVar2(wclass, getUInt16(pc+1));
            pc += 3;
            break;
        case OP_iload:
        case OP_fload:
        case OP_aload:
            AVOID_AUTO_RESET
            *stack++ = var[pc[1]];
            pc += 2;
            break;
        case OP_iload_0: // guich@200b4: clustered in a different way
        case OP_fload_0:
        case OP_aload_0:
            AVOID_AUTO_RESET
            *stack++ = var[0];
            pc++;
            break;
        case OP_iload_1:
        case OP_fload_1:
        case OP_aload_1:
            AVOID_AUTO_RESET
            *stack++ = var[1];
            pc++;
            break;
        case OP_iload_2:
        case OP_fload_2:
        case OP_aload_2:
            AVOID_AUTO_RESET
            *stack++ = var[2];
            pc++;
            break;
        case OP_iload_3:
        case OP_fload_3:
        case OP_aload_3:
            AVOID_AUTO_RESET
            *stack++ = var[3];
            pc++;
            break;
        case OP_iaload:
        {
            WObject obj;
            uint32 i;
            stack--; // nopt
            obj = stack[-1].obj;
            if (obj == 0) goto null_array_error;
            i = stack[0].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj)) // guich@400_16: making i uint32, when it become < 0 it will surely be greater than len
               goto index_range_error;
            stack[-1].intValue = ((int32 *)WOBJ_arrayStartP(obj))[i];
            pc++;
            break;
        }
        case OP_saload:
        {
            WObject obj;
            uint32 i;
            stack--; // nopt
            obj = stack[-1].obj;
            if (obj == 0) goto null_array_error;
            i = stack[0].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            stack[-1].intValue = (int32)(((int16 *)WOBJ_arrayStartP(obj))[i]);
            pc++;
            break;
        }
        case OP_faload:
        {
            WObject obj;
            uint32 i;
            stack--; // nopt
            obj = stack[-1].obj;
            if (obj == 0) goto null_array_error;
            i = stack[0].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            stack[-1].floatValue = ((float32 *)WOBJ_arrayStartP(obj))[i];
            pc++;
            break;
        }
        case OP_aaload:
        {
            WObject obj;
            uint32 i;
            stack--; // nopt
            obj = stack[-1].obj;
            if (obj == 0) goto null_array_error;
            i = stack[0].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            stack[-1].obj = ((WObject *)WOBJ_arrayStartP(obj))[i];
            pc++;
            break;
        }
        case OP_baload:
        {
            WObject obj;
            uint32 i;
            stack--; // nopt
            obj = stack[-1].obj;
            if (obj == 0) goto null_array_error;
            i = stack[0].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            stack[-1].intValue = (int32)(((char *)WOBJ_arrayStartP(obj))[i]);
            pc++;
            break;
        }
        case OP_caload:
        {
            WObject obj;
            uint32 i;
            stack--; // nopt
            obj = stack[-1].obj;
            if (obj == 0) goto null_array_error;
            i = stack[0].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            stack[-1].intValue = (int32)(((uint16 *)WOBJ_arrayStartP(obj))[i]);
            pc++;
            break;
        }
        case OP_astore:
        case OP_istore:
        case OP_fstore:
            var[pc[1]] = *--stack;
            pc += 2;
            break;
        case OP_istore_0:
        case OP_fstore_0:
        case OP_astore_0:
            var[0] = *--stack;
            pc++;
            break;
        case OP_istore_1:
        case OP_fstore_1:
        case OP_astore_1:
            var[1] = *--stack;
            pc++;
            break;
        case OP_istore_2:
        case OP_fstore_2:
        case OP_astore_2:
            var[2] = *--stack;
            pc++;
            break;
        case OP_istore_3:
        case OP_fstore_3:
        case OP_astore_3:
            var[3] = *--stack;
            pc++;
            break;
        case OP_iastore:
        {
            WObject obj;
            uint32 i;
            stack -= 3;
            obj = stack[0].obj;
            if (obj == 0) goto null_array_error;
            i = stack[1].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            ((int32 *)WOBJ_arrayStartP(obj))[i] = stack[2].intValue;
            pc++;
            break;
        }
        case OP_sastore:
        {
            WObject obj;
            uint32 i;
            stack -= 3;
            obj = stack[0].obj;
            if (obj == 0) goto null_array_error;
            i = stack[1].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            ((int16 *)WOBJ_arrayStartP(obj))[i] = (int16)stack[2].intValue;
            pc++;
            break;
        }
        case OP_fastore:
        {
            WObject obj;
            uint32 i;
            stack -= 3;
            obj = stack[0].obj;
            if (obj == 0) goto null_array_error;
            i = stack[1].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            ((float32 *)WOBJ_arrayStartP(obj))[i] = stack[2].floatValue;
            pc++;
            break;
        }
        case OP_aastore:
        {
            WObject obj;
            uint32 i;
            stack -= 3;
            obj = stack[0].obj;
            if (obj == 0) goto null_array_error;
            i = stack[1].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            ((WObject *)WOBJ_arrayStartP(obj))[i] = stack[2].obj;
            pc++;
            break;
        }
        case OP_bastore:
        {
            WObject obj;
            uint32 i;
            stack -= 3;
            obj = stack[0].obj;
            if (obj == 0) goto null_array_error;
            i = stack[1].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            ((char *)WOBJ_arrayStartP(obj))[i] = (char)stack[2].intValue;
            pc++;
            break;
        }
        case OP_castore:
        {
            WObject obj;
            uint32 i;
            stack -= 3;
            obj = stack[0].obj;
            if (obj == 0) goto null_array_error;
            i = stack[1].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            ((uint16 *)WOBJ_arrayStartP(obj))[i] = (uint16)stack[2].intValue;
            pc++;
            break;
        }
        case OP_pop:
            stack--;
            pc++;
            break;
        case OP_pop2:
            stack -= 2;
            pc++;
            break;
        case OP_dup:
            (*stack++) = stack[-1];
            pc++;
            break;
        case OP_dup_x1:
            stack[0] = stack[-1];
            stack[-1] = stack[-2];
            stack[-2] = stack[0];
            stack++;
            pc++;
            break;
        case OP_dup_x2:
            stack[0] = stack[-1];
            stack[-1] = stack[-2];
            stack[-2] = stack[-3];
            stack[-3] = stack[0];
            stack++;
            pc++;
            break;
        case OP_dup2:
            stack[1] = stack[-1];
            stack[0] = stack[-2];
            stack += 2;
            pc++;
            break;
        case OP_dup2_x1:
            stack[1] = stack[-1];
            stack[0] = stack[-2];
            stack[-1] = stack[-3];
            stack[-2] = stack[1];
            stack[-3] = stack[0];
            stack += 2;
            pc++;
            break;
        case OP_dup2_x2:
            stack[1] = stack[-1];
            stack[0] = stack[-2];
            stack[-1] = stack[-3];
            stack[-2] = stack[-4];
            stack[-3] = stack[1];
            stack[-4] = stack[0];
            stack += 2;
            pc++;
            break;
        case OP_swap:
            vtemp = stack[-2];
            stack[-2] = stack[-1];
            stack[-1] = vtemp;
            pc++;
            break;
        case OP_iadd:
#ifdef PALMOS
            stack[-1].intValue += (*--stack).intValue;
#else
            stack--;
            stack[-1].intValue += (*stack).intValue;
#endif
            pc++;
            break;
        case OP_fadd:
            stack--; // nopt
            stack[-1].floatValue += stack[0].floatValue;
            pc++;
            break;
        case OP_isub:
#ifdef PALMOS
            stack[-1].intValue -= (*--stack).intValue;
#else
            stack--;
            stack[-1].intValue -= (*stack).intValue;
#endif
            pc++;
            break;
        case OP_fsub:
            stack--; // nopt
            stack[-1].floatValue -= stack[0].floatValue;
            pc++;
            break;
        case OP_imul:
            stack--; // nopt
            stack[-1].intValue *= stack[0].intValue;
            pc++;
            break;
        case OP_fmul:
            stack--; // nopt
            stack[-1].floatValue *= stack[0].floatValue;
            pc++;
            break;
        case OP_idiv:
            if ((*--stack).intValue == 0)
                goto div_by_zero_error;
            stack[-1].intValue /= stack[0].intValue;
            pc++;
            break;
        case OP_fdiv:
            if ((*--stack).floatValue == 0.0f)
                goto div_by_zero_error;
            stack[-1].floatValue /= stack[0].floatValue;
            pc++;
            break;
        case OP_irem:
            if ((*--stack).intValue == 0)
                goto div_by_zero_error;
            stack[-1].intValue %= stack[0].intValue;
            pc++;
            break;
        case OP_frem:
            if ((*--stack).floatValue == 0.0f)
                goto div_by_zero_error;
            f = stack[-1].floatValue / stack[0].floatValue;
            f = (float32)((int32)f);
            f *= stack[0].floatValue;
            stack[-1].floatValue -= f;
            pc++;
            break;
        case OP_ineg:
            stack[-1].intValue = - stack[-1].intValue;
            pc++;
            break;
        case OP_fneg:
            stack[-1].floatValue = - stack[-1].floatValue;
            pc++;
            break;
        case OP_ishl:
            stack--; // nopt
            stack[-1].intValue <<= stack[0].intValue & 0x1F;
            pc++;
            break;
        case OP_ishr:
            stack--; // nopt
            stack[-1].intValue >>= stack[0].intValue & 0x1F;
            pc++;
            break;
        case OP_iushr:
        {
            int32 i,key;
            key = (*--stack).intValue;
            i = (*--stack).intValue;
            stack[0].intValue = ushr((uint32)i, key & 0x1F);
            stack++;
            pc++;
            break;
        }
        case OP_iand:
#ifdef PALMOS
            stack[-1].intValue &= (*--stack).intValue;
#else
            stack--;
            stack[-1].intValue &= (*stack).intValue;
#endif
            pc++;
            break;
        case OP_ior:
#ifdef PALMOS
            stack[-1].intValue |= (*--stack).intValue;
#else
            stack--;
            stack[-1].intValue |= (*stack).intValue;
#endif
            pc++;
            break;
        case OP_ixor:
#ifdef PALMOS
            stack[-1].intValue ^= (*--stack).intValue;
#else
            stack--;
            stack[-1].intValue ^= (*stack).intValue;
#endif
            pc++;
            break;
        case OP_iinc:
            var[pc[1]].intValue += (signed char)pc[2];
            pc += 3;
            break;
        case OP_i2f:
            stack[-1].floatValue = (float32)stack[-1].intValue;
            pc++;
            break;
        case OP_f2i:
            f = stack[-1].floatValue;
            stack[-1].intValue = (f > 2147483647.0)?0x7FFFFFFF:(f < -2147483648.0)?0x80000000:(int32)f;
            pc++;
            break;
        case OP_i2b: // peter@330_16: fixed a wrong convertion
            stack[-1].intValue = (signed char)stack[-1].intValue;
            pc++;
            break;
        case OP_i2c:
            stack[-1].intValue = (uint16)stack[-1].intValue;
            pc++;
            break;
        case OP_i2s:
            stack[-1].intValue = (int16)stack[-1].intValue;
            pc++;
            break;
        case OP_fcmpl:
        case OP_fcmpg:
            // NOTE: NaN values are currently not supported - NaN in either
            // value should return 1 or 0 depending on the opcode
            stack--;
            f = stack[-1].floatValue - stack[0].floatValue;
            stack[-1].intValue = (f > 0.0f)?1:(f < 0.0f)?-1:0;
            pc++;
            break;
        case OP_ifeq:
            if ((*--stack).intValue == 0) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_ifne:
            if ((*--stack).intValue != 0) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_iflt:
            if ((*--stack).intValue < 0) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_ifge:
            if ((*--stack).intValue >= 0) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_ifgt:
            if ((*--stack).intValue > 0) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_ifle:
            if ((*--stack).intValue <= 0) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_if_icmpeq:
            stack -= 2;
            if (stack[0].intValue == stack[1].intValue) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_if_icmpne:
            stack -= 2;
            if (stack[0].intValue != stack[1].intValue) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_if_icmplt:
            stack -= 2;
            if (stack[0].intValue < stack[1].intValue) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_if_icmpge:
            stack -= 2;
            if (stack[0].intValue >= stack[1].intValue) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_if_icmpgt:
            stack -= 2;
            if (stack[0].intValue > stack[1].intValue) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_if_icmple:
            stack -= 2;
            if (stack[0].intValue <= stack[1].intValue) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_if_acmpeq:
            stack -= 2;
            if (stack[0].obj == stack[1].obj) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_if_acmpne:
            stack -= 2;
            if (stack[0].obj != stack[1].obj) // nopt
                pc += getInt16(pc+1);
            else
                pc += 3;
            break;
        case OP_goto:
            pc += getInt16(pc+1);
            break;
        case OP_jsr:
            (*stack++).pc = pc + 3;
            pc += getInt16(pc+1);
            break;
        case OP_ret:
            pc = var[pc[1]].pc;
            break;
        case OP_tableswitch:
            key = (*--stack).intValue;
            npc = pc + 1;
            npc += 4 + ((4 - ((npc - /*METH_code(method)*/method->codeStart) & 3)) & 3);
            low = getInt32(npc);
            npc += 4;
            high = getInt32(npc);
            npc += 4;
            pc += getInt32((key < low || key > high) ? (npc-12) : (&npc[(key - low) * 4])); // guich@400_16
            break;
        case OP_lookupswitch:
        {
            int32 i;
            key = (*--stack).intValue;
            npc = pc + 1;
            npc += (4 - ((npc - /*METH_code(method)*/method->codeStart) & 3)) & 3;
            //defaultOff = getInt32(npc); - guich@300: now only read if necessary
            npc += 4;
            npairs = getInt32(npc);
            npc += 4;

            // binary search
            if (npairs > 0)
            {
                low = 0;
                high = npairs;
                while (1)
                {
                    mid = (high + low) / 2;
                    i = getInt32(npc + (mid * 8));
                    if (key == i)
                    {
                        pc += getInt32(npc + (mid * 8) + 4); // found
                        break;
                    }
                    if (mid == low)
                    {
                        pc += getInt32(npc-8); // not found
                        break;
                    }
                    if (key < i)
                        high = mid;
                    else
                        low = mid;
                }
            }
            else
                pc += getInt32(npc-8); // no pairs
            break;
        }
        case OP_ireturn:
        case OP_freturn:
        case OP_areturn:
            returnedValue[0] = stack[-1];
            returnedValueCount = 1;
            goto methodreturn;
        case OP_return:
            returnedValueCount = 0;
            goto methodreturn;
        case OP_getfield:
        {
            WObject obj;
            fieldIndex = getUInt16(pc+1);
            field = (WClassField *)CONST_binded(wclass, fieldIndex);
            if (!field)
            {
               field = getFieldByIndex(wclass, fieldIndex);
               if (!field)
                  goto error;
            }
            obj = (*--stack).obj;
            if (obj == 0)
                goto null_obj_error;

            *stack++ = WOBJ_var(obj, field->var.varOffset);
            if (FIELD_is64(field)) // guich@200 handle 64bit numbers
               *stack++ = WOBJ_var(obj, field->var2.varOffset);
            pc += 3;
            break;
		  }
        case OP_putfield:
        {
            WObject obj;
            fieldIndex = getUInt16(pc+1);
            field = (WClassField *)CONST_binded(wclass, fieldIndex);
            if (!field)
            {
               field = getFieldByIndex(wclass, fieldIndex);
               if (!field)
                  goto error;
            }
            if (FIELD_is64(field)) // guich@200 handle 64bit numbers
            {
	            stack -= 3;
	            obj = stack[0].obj;
	            if (obj == 0)
	                goto null_obj_error;
	            WOBJ_var(obj, field->var.varOffset) = stack[1];
	            WOBJ_var(obj, field->var2.varOffset) = stack[2];
            }
            else
            {
	            stack -= 2;
	            obj = stack[0].obj;
	            if (obj == 0)
	                goto null_obj_error;
	            WOBJ_var(obj, field->var.varOffset) = stack[1];
	         }
            pc += 3;
            break;
        }
        case OP_getstatic:
            fieldIndex = getUInt16(pc+1);
            field = (WClassField *)CONST_binded(wclass, fieldIndex);
            if (!field)
            {
               field = getFieldByIndex(wclass, fieldIndex);
               if (!field)
                  goto error;
            }
            if (FIELD_is64(field)) // guich@200 handle 64bit numbers
	            *stack++ = field->var2.staticVar;
            *stack++ = field->var.staticVar;
            pc += 3;
            break;
        case OP_putstatic:
            fieldIndex = getUInt16(pc+1);
            field = (WClassField *)CONST_binded(wclass, fieldIndex);
            if (!field)
            {
               field = getFieldByIndex(wclass, fieldIndex);
               if (!field)
                  goto error;
            }
          	field->var.staticVar = *--stack;
            if (FIELD_is64(field)) // guich@200 handle 64bit numbers
            	field->var2.staticVar = *--stack;
            pc += 3;
            break;
        case OP_notused:
           break;
        case OP_invokevirtual:
        case OP_invokespecial:
        case OP_invokestatic:
        case OP_invokeinterface:
            {
            int32 i; // note: don't move these locals from here!
            uint32 methodIndex;
            WObject obj=0;
            UtfString methodName, methodDesc;
            int16 methodMapNum;
            uint16 nameAndTypeIndex;
            WClassMethod *_method; // 'method' must not be passed by reference, so we create an alias
            WClass *_iclass=NULL;

            oldpc = pc; // guich@241_6
            methodIndex = getUInt16(pc+1); // method definition in the constant pool
            _iclass = (WClass *)CONST_binded(wclass, methodIndex); // guich@400_16
            if (_iclass)
                _method = CONST_bindedMeth(wclass,methodIndex); // guich@200b4_183. - don't move this from here, bc wclass may change below
            else
            {
                classIndex = CONST_classIndex(wclass, methodIndex);
                _iclass = getClassByIndex(wclass, classIndex,1);
                if (_iclass == NULL)
                {
                   if (thrownException)
                      goto throw_exception; // guich@320_7
                   else
                      goto method_called_on_array_error; // guich@200 - trap a null pointer exception that occured when a zero length array object was created and was tried to be displayed.
                }
                nameAndTypeIndex = CONST_nameAndTypeIndex(wclass, methodIndex);
                getUtfString2(wclass, CONST_nameIndex(wclass, nameAndTypeIndex), &methodName);
                getUtfString2(wclass, CONST_typeIndex(wclass, nameAndTypeIndex), &methodDesc);
                methodMapNum = getMethodMapNum(_iclass, &methodName, &methodDesc, (byte)((*pc == OP_invokevirtual) ? CLASS_SEARCH_ALL : CLASS_SEARCH_THISCLASS), &_method, NULL);
                if (methodMapNum < 0)
                   goto methoderror;
                // quickbind for methods - bind the constant to the method num and class index if
                // the class is not an interface. If the class is an interface, we
                // need the name and desc of the method later, so we can't bind over
                // the constant that contains the nameAndTypeIndex
                if (!WCLASS_isInterface(_iclass))
                {
                   if (*pc == OP_invokevirtual)
                      _iclass = (WClass*)methodMapNum; // guich@400_16
                   CONST_binded(wclass, methodIndex) = (uint32)_iclass; // guich@400_16
                   // guich@200b4_183
                   CONST_bindedMeth(wclass,methodIndex) = _method;
                }
            }
            numParams = _method->numParams;

            // FIX - double-check sometime, can we do an interface invoke on
            // a static method? if so, we should check the method for static,
            // not the invoke type here
            if (*pc != OP_invokestatic)
            {
               // inc param count and get object reference
               numParams++;
               stack -= numParams;

               obj = stack[0].obj;
               if (obj == 0) // (!)
                  goto null_obj_error;
               method = _method; // Now its safe to do the assign

               if (*pc == OP_invokevirtual) // nopt to a case - note: 80% of the calls are of this type
               {
                  if ((uint32)_iclass >= 0xFFFF) // guich@400_56: make sure the class isnt already bound.
                  {
                     // this is a case where a method was previously called using
                     // invokespecial but now is called using invokevirtual.
                     // It occurs in the following situation:
                     // control.setRect(...) // invoke special
                     // super.setRect(...)   // invoke virtual
                     classIndex = CONST_classIndex(wclass, methodIndex);
                     _iclass = getClassByIndex(wclass, classIndex,1);
                     if (_iclass == NULL)
                     {
                        if (thrownException)
                           goto throw_exception; // guich@320_7
                        else
                           goto method_called_on_array_error; // guich@200 - trap a null pointer exception that occured when a zero length array object was created and was tried to be displayed.
                     }
                     nameAndTypeIndex = CONST_nameAndTypeIndex(wclass, methodIndex);
                     getUtfString2(wclass, CONST_nameIndex(wclass, nameAndTypeIndex), &methodName);
                     getUtfString2(wclass, CONST_typeIndex(wclass, nameAndTypeIndex), &methodDesc);
                     methodMapNum = getMethodMapNum(_iclass, &methodName, &methodDesc, (byte)CLASS_SEARCH_ALL, &_method, NULL);
                     if (methodMapNum < 0)
                        goto methoderror;
                     method = _method;
                     //wclass = _iclass;
                     _iclass = (WClass *)methodMapNum;
                     CONST_binded(wclass, methodIndex) = (uint32)_iclass; // guich@400_82
                  }

                  wclass = WOBJ_class(obj);
                  if (wclass == NULL)
                     goto method_called_on_array_error; // guich@200 - trap a null pointer exception that occured when a zero length array object was created and was tried to be displayed.

                  if ((uint16)_iclass < wclass->vMethodMap.mapSize) // guich@400_16: _iclass now stores methodMapNum
                  {
                     // inherited or overridden method
                     VMapValue *mapValue = &wclass->vMethodMap.mapValues[(uint16)_iclass];
                     wclass = mapValue->c;
                     method = mapValue->m;
                  }
               }
               else
               if (*pc == OP_invokespecial)
               {
                  if ((uint32)_iclass > 0xFFFF) // guich@400_79
                     wclass = _iclass;
                  else
                  {
                     // this is the opposite rare case, where a method was
                     // previously called using invokevirtual but now is called
                     // using invokespecial.
                     // It usually occurs in the following situation:
                     // super.setRect(...)   // invoke virtual
                     // control.setRect(...) // invoke special
                     classIndex = CONST_classIndex(wclass, methodIndex);
                     _iclass = getClassByIndex(wclass, classIndex,1);
                     if (_iclass == NULL)
                     {
                        if (thrownException)
                           goto throw_exception; // guich@320_7
                        else
                           goto method_called_on_array_error; // guich@200 - trap a null pointer exception that occured when a zero length array object was created and was tried to be displayed.
                     }
                     nameAndTypeIndex = CONST_nameAndTypeIndex(wclass, methodIndex);
                     getUtfString2(wclass, CONST_nameIndex(wclass, nameAndTypeIndex), &methodName);
                     getUtfString2(wclass, CONST_typeIndex(wclass, nameAndTypeIndex), &methodDesc);
                     methodMapNum = getMethodMapNum(_iclass, &methodName, &methodDesc, (byte)CLASS_SEARCH_ALL, &_method, &_iclass);
                     if (methodMapNum < 0)
                        goto methoderror;
                     CONST_binded(wclass, methodIndex) = (uint32)_iclass; // guich@400_82
                     method = _method;
                     wclass = _iclass;
                  }
               }
               else
               //if (*pc == OP_invokeinterface)
               {
                  wclass = (WClass *)WOBJ_class(obj);
                  if (wclass == NULL)
                     goto method_called_on_array_error; // guich@200 - trap a null pointer exception that occured when a zero length array object was created and was tried to be displayed.
                  if (getMethodMapNum(wclass, &methodName, &methodDesc, CLASS_SEARCH_ALL, &_method, &_iclass) < 0)
                     goto methoderror;
                  method = _method;
                  wclass = _iclass;
                  pc += 2;
               }
   			}
   			else
   			{
               method = _method;
               wclass = _iclass;
   			   stack -= numParams;
   			}
            pc += 3;
            // push return stack frame:
            //
            // program counter pointer
            // local var pointer
            // local stack pointer

            // return stack frame plus native method active frame
            // guich@320_15: since METH_localsStack2 max value is 256, i changed vmStackEnd to be vmStackEnd-300 and avoid this big sum
            //if ((isNative? (vmStackPtr + STACKSAVESIZE + numParams + 3) : (vmStackPtr + STACKSAVESIZE + METH_localsStack2(imethod))) >= vmStackEnd)
            if (vmStackPtr > vmStackEnd)
            {
               Var *_vmStackStart = vmStackStart;
               Var *_vmStackEnd = vmStackEnd;
               Var *_vmStackPtr = vmStackPtr;
               Var *_var = var;
               Var *_stack = stack;
               Var *_baseFramePtr = baseFramePtr;
               if (!growStack((byte **)&_vmStackStart, (byte **)&_vmStackPtr, (byte **)&_vmStackEnd, (byte **)&_baseFramePtr, (byte **)&_var, (byte **)&_stack, &vmStackSizeInBytes))
                  goto stack_overflow_error;
               vmStackStart = _vmStackStart;
               vmStackEnd = _vmStackEnd - 300;
               vmStackPtr = _vmStackPtr;
               var = _var;
               stack = _stack;
               baseFramePtr = _baseFramePtr;
            }

#ifdef ENABLE_TRACE
                trace(wclass,method);
#endif

            (vmStackPtr++)->oldpc = oldpc; // guich@241_6
            (vmStackPtr++)->pc = pc;
            (vmStackPtr++)->intValue = ((byte *)var - (byte *)vmStackStart); // store relative values since the address can change
            (vmStackPtr++)->intValue = ((byte *)stack - (byte *)vmStackStart); // would be stack-numParams

            // execute native method
            if (METH_isNative(method))
            {
                Err err=0;
                returnedValueCount = (byte)method->returnedValueCount;
                if (returnedValueCount) // guich@500_5: save the stack-1 and stack-2, bc now the returned values are stored there.
                   copy64(stack-2, savedStack);

					 // guich@400_16: removed the active frame (not sure what was it for)
#ifdef PALMOS
                if (method->libRefNum) // already binded?
                {
                   SWInvokeNativeFunc(method->libRefNum, method->code.nativeFunc, stack);
                }
                else
                {
                   // guich@200b4: (stack+1) note: now all native classes have a "ClassHolder", so we skip that class ref in the stack to maintain compatibility with older native methods
                   // guich@200 - if the native func returns long or double, it sets the returnedValueCount to 2 and the stack[-2]
                   // 1. search in main lib
                   err = SWInvokeNative(SWLibRefNum, method->code.hash, stack);
                   // 2. if not found, search in the attached native libraries
                   if (err == SWErrMethodNotFound && vmGlobals->nativeLibs)
                      for (i = 0; vmGlobals->nativeLibs[i].ref; i++)
                         if ((err = SWInvokeNative(vmGlobals->nativeLibs[i].ref, method->code.hash, stack)) == 0) // native method handled
                         {
                            NativeFunc natFunc=NULL;
                            if (SWGetNativeFunc(vmGlobals->nativeLibs[i].ref, method->code.hash, &natFunc) == 0) // guich@401_6: fixed dynamic bind
                            {
                               method->libRefNum = (byte)vmGlobals->nativeLibs[i].ref; // guich@400_16: bind it so that in the next try we are faster
                               method->code.nativeFunc = natFunc;
                            }
                            break;
                         }
                   if (err) // no one handles this method?
                      goto methoderror; // dhodges@421_36: throws the right message "method not found"
                }
#else
                if (method->libRefNum) // already binded?
                {
                   if (method->libRefNum == SWLibRefNum)
                      SWInvokeNativeFuncStatic(method->code.nativeFunc, stack);
                   else
                      (*vmGlobals->nativeLibs[method->libRefNum-1]->SWInvokeNativeFunc)(method->code.nativeFunc, stack);
                }
                else
                {
                   // guich@200 - if the native func returns long or double, it sets the pushReturnedValue to 2 and the returnedValue2 var
                   // 1. in WinCE there's no chance of the methods not be bounded
                   // 2. if not found, search in the attached native libraries
                   if (err == SWErrMethodNotFound && vmGlobals->nativeLibs)
                      for (i = 0; vmGlobals->nativeLibs[i]; i++)
                         if ((err = (*vmGlobals->nativeLibs[i]->SWInvokeNative)(method->code.hash, stack)) == 0) // native method handled
                         {
                            NativeFunc natFunc=NULL;
                            if ((*vmGlobals->nativeLibs[i]->SWGetNativeFunc)(method->code.hash, &natFunc) == 0) // guich@401_6: fixed dynamic bind
                            {
                               method->libRefNum = (byte)(i+1); // guich@400_16: bind it so that in the next try we are faster
                               method->code.nativeFunc = natFunc;
                            }
                            break;
                         }
                   if (err) // no one handles this method?
                      goto methoderror; // dhodges@421_36: throws the right message "method not found"
                }
#endif
                if (vmStatus.errNum != 0)
                   goto error; // error occured during native method
                // guich@500_5: now the returned values are pushed into the stack as stack[-1] and stack[-2] (for double/long)
                if (returnedValueCount)
                {
                   // get the values pushed by the native routine
                   copy64(stack-2, returnedValue);
                   // restore old values of the stack
                   copy64(savedStack, stack-2);
                }
                goto methodreturn;
            }
            // push params into local vars of next frame - guich@500_5: native methods don't need this, so we only push for java methods
            if (numParams > 0)
               xfer32((byte *)stack,(byte *)vmStackPtr,numParams);
            goto methodinvoke;
methoderror:
            VmError(ERR_CantFindMethod, wclass, &methodName, &methodDesc);
            goto error;
            }
        case OP_new:
        {
        		WClass *iclass;
            classIndex = getUInt16(pc+1);
            if ((iclass=(WClass *)CONST_binded(wclass, classIndex)) == 0)
            {
                iclass = getClassByIndex(wclass, classIndex,1);
                if (iclass == NULL)
                {
                   if (thrownException)
                      goto throw_exception; // guich@320_7
                   else
                      goto method_called_on_array_error; // guich@200 - trap a null pointer exception that occured when a zero length array object was created and was tried to be displayed.
                }
            }
            (*stack++).obj = createObject(iclass);
            pc += 3;
            break;
        }
        case OP_newarray:
            stack[-1].obj = createArrayObject((int32)pc[1], stack[-1].intValue);
            pc += 2;
            break;
        case OP_anewarray:
            stack[-1].obj = createArrayObject(1, stack[-1].intValue);
            pc += 3;
            break;
        case OP_arraylength:
        {
            WObject obj;
            obj = stack[-1].obj;
            if (obj == 0)
                goto null_array_error;
            stack[-1].intValue = WOBJ_arrayLen(obj);
            pc++;
            break;
        }
        case OP_instanceof:
        case OP_checkcast:
        {
            WObject obj;
            obj = stack[-1].obj;
            if (obj == 0)
            {
                if (*pc == OP_instanceof)
                    stack[-1].intValue = 0;
                pc += 3;
                break;
            }
            classIndex = getUInt16(pc+1);
            if ((target=(WClass *)CONST_binded(wclass, classIndex)) == NULL)
                target = getClassByIndex(wclass, classIndex,0);
            if (target)
            {
                comp = (WOBJ_class(obj) == target) || compatible(WOBJ_class(obj), target); // target is not array - guich@402_31: added first quick test
                if (!comp) className = target->className; // guich@220: only get the classname if needed
            }
            else
            {
                // target is either array or target class was not found
                // if either of these cases is true, the index couldn't be
                // bound to a pointer, so we still have a pointer into the
                // constant pool and can use the string reference in the constant
                className = getUtfString(wclass, CONST_nameIndex(wclass, classIndex));
                if (className.len > 1 && className.str[0] == '[')
                   comp = compatibleArray(obj, className); // target is array
                else
                   goto error; // guich@200 - trap a null pointer exception that occured when a zero length array object was created and was tried to be displayed.
            }
            if (*pc == OP_checkcast)
            {
                if (!comp)
                   goto class_cast_error;
            }
            else
                stack[-1].intValue = comp;
            pc += 3;
            break;
        }
        case OP_wide: // note: this case is very slow. It needs to be optimized. instead of do a table with displacements, the CW compiler does 12 subq/beq's (but i never found a code that used OP_Wide)
            switch (*++pc)
            {
                case OP_iload:
                case OP_fload:
                case OP_aload:
                    (*stack++) = var[getUInt16(pc+1)];
                    pc += 3;
                    break;
                case OP_dload:
                case OP_lload:
			           copy64(&var[getUInt16(pc+1)],stack);
                    stack+=2;
                    pc += 3;
                    break;
                case OP_astore:
                case OP_istore:
                case OP_fstore:
                    var[getUInt16(pc+1)] = *--stack;
                    pc += 3;
                    break;
                case OP_dstore:
                case OP_lstore:
			           stack-=2;
			           copy64(stack,&var[getUInt16(pc+1)]);
                    pc += 3;
                    break;
                case OP_iinc:
                    var[getUInt16(pc+1)].intValue += getInt16(&pc[3]);
                    pc += 5;
                    break;
                case OP_ret:
//                    drawNumber(99);
                    pc = var[getUInt16(pc+1)].pc;
                    break;
            }
            break;
        case OP_multianewarray:
            classIndex = getUInt16(pc+1);
            // since arrays do not have associated classes which could be bound
            // to the class constant, we can safely access the name string in
            // the constant
            className = getUtfString(wclass, CONST_nameIndex(wclass, classIndex));
            ndim = (int32)pc[3];
            stack -= ndim;
            (*stack++).obj = createMultiArray(ndim, &className.str[1], stack);
            pc += 4;
            break;
        case OP_ifnull:
            if ((*--stack).obj == 0)
               pc += getInt16(pc+1);
            else
               pc += 3;
            break;
        case OP_ifnonnull:
            if ((*--stack).obj != 0)
               pc += getInt16(pc+1);
            else
               pc += 3;
            break;
        case OP_goto_w:
            pc += getInt32(pc+1);
            break;
        case OP_jsr_w:
            (*stack++).pc = pc + 5;
            pc += getInt32(pc+1);
            break;
        case OP_monitorenter: // unsupported
            stack--;
            pc++;
            break;
        case OP_monitorexit: // unsupported
            stack--;
            pc++;
            break;
//////////////////////////////////////////////////////
// guich@200 - added support for 64 bits types
        case OP_dconst_0:
            c.d = (double)0; // @v 0.0d
            copy64((Var*)&c,stack);
            stack+=2;
            pc++;
            break;
        case OP_dconst_1:
            c.d = (double)1; // @v 1.0d
            copy64((Var*)&c,stack);
            stack+=2;
            pc++;
            break;
        case OP_dload:
        case OP_lload:
            copy64(&var[pc[1]],stack);
            AVOID_AUTO_RESET
            stack+=2;
            pc += 2;
            break;
        case OP_dload_0:
        case OP_dload_1:
        case OP_dload_2:
        case OP_dload_3:
            copy64(&var[*pc - OP_dload_0],stack);
            stack+=2;
            pc++;
            break;
        case OP_daload:
        case OP_laload:
        {
            WObject obj;
            uint32 i;
            stack -= 2;
            obj = stack[0].obj;
            if (obj == 0) goto null_array_error;
            i = stack[1].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            copy64((WOBJ_arrayStartP(obj))+(i*2),stack);
            stack += 2;
            pc++;
            break;
        }
        case OP_dastore:
        case OP_lastore:
        {
            WObject obj;
            uint32 i;
            stack -= 4;
            obj = stack[0].obj;
            if (obj == 0) goto null_array_error;
            i = stack[1].intValue;
            if (i >= (uint32)WOBJ_arrayLenP(obj))
               goto index_range_error;
            copy64(stack+2,(WOBJ_arrayStartP(obj))+(i*2));
            pc++;
            break;
        }
        case OP_dstore:
        case OP_lstore:
            stack-=2;
            copy64(stack,&var[pc[1]]);
            pc += 2;
            break;
        case OP_dstore_0:
        case OP_dstore_1:
        case OP_dstore_2:
        case OP_dstore_3:
            stack -= 2;
            copy64(stack,&var[*pc - OP_dstore_0]);
            pc++;
            break;
        case OP_dadd:
            stack -= 4;
            copy64(stack+2,(Var*)&c1);
            copy64(stack,  (Var*)&c2);
            c.d = c1.d + c2.d;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_dsub:
            stack -= 4;
            copy64(stack+2,(Var*)&c1);
            copy64(stack,  (Var*)&c2);
            c.d = c2.d - c1.d;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_dmul:
            stack -= 4;
            copy64(stack+2,(Var*)&c1);
            copy64(stack,  (Var*)&c2);
            c.d = c2.d * c1.d;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_ddiv:
            stack -= 4;
            copy64(stack+2,(Var*)&c1);
            copy64(stack,  (Var*)&c2);
            if (c1.d == (double)0) // @v 0.0d
                goto div_by_zero_error;
            c.d = c2.d / c1.d;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_drem:
            stack -= 4;
            copy64(stack+2,(Var*)&c1);
            copy64(stack,  (Var*)&c2);
            if (c1.d == (double)0) // @v 0.0d
                goto div_by_zero_error;
            c.d = c2.d - ((double)((int64)(c2.d / c1.d))) * c1.d; // take the integer part of the division
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_dneg:
            stack -= 2;
            copy64(stack,(Var*)&c);
            c.d = -c.d;
            copy64((Var*)&c.d,stack);
            stack += 2;
            pc++;
            break;
        case OP_i2d:
            stack--;
            c.d = (double)stack[0].intValue;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_f2d:
            stack--;
            c.d = (double)stack[0].floatValue;
            copy64((Var*)&c,stack);
            stack+=2;
            pc++;
            break;
        case OP_d2i:
            stack-=2;
            copy64(stack,(Var*)&c);
            (*stack++).intValue = _d_dtoi(c.fd);
            pc++;
            break;
        case OP_d2f:
            stack-=2;
            copy64(stack,(Var*)&c);
#ifdef WIN32
            (*stack++).floatValue = _d_dtof(c.fd); // guich@310_15
#else
            (*stack++).intValue = _d_dtof(c.fd);
#endif
            pc++;
            break;
        case OP_dcmpl:
        case OP_dcmpg:
            // NOTE: NaN values are currently not supported - NaN in either
            // value should return 1 or 0 depending on the opcode
            stack-=4;
            copy64(stack,  (Var*)&c1);
            copy64(stack+2,(Var*)&c2);
            switch (_d_cmp(c1.fd,c2.fd))
            {
               case flpGreater: (*stack++).intValue =  1; break;
               case flpLess   : (*stack++).intValue = -1; break;
               default:         (*stack++).intValue =  0; break;
            }
            pc++;
            break;
        case OP_dreturn:
        case OP_lreturn:
            copy64(stack-2,returnedValue); // guich@500_5
            //returnedValue[0] = stack[-2];
            //returnedValue[1] = stack[-1];
            returnedValueCount = 2;
            goto methodreturn;
        case OP_ldc2_w:
            c.d = CONST_double(wclass, getUInt16(pc+1));
            copy64((Var*)&c,stack);
            stack+=2;
            pc += 3;
            break;
/// long 64 bits support
        case OP_lconst_0:
            c.ll = 0;
            copy64((Var*)&c,stack);
            stack+=2;
            pc++;
            break;
        case OP_lconst_1:
            c.ll = 1;
            copy64((Var*)&c,stack);
            stack+=2;
            pc++;
            break;
        case OP_lload_0:
        case OP_lload_1:
        case OP_lload_2:
        case OP_lload_3:
            copy64(&var[*pc - OP_lload_0],stack);
            stack+=2;
            pc++;
            break;
        case OP_lstore_0:
        case OP_lstore_1:
        case OP_lstore_2:
        case OP_lstore_3:
            stack -= 2;
            copy64(stack,&var[*pc - OP_lstore_0]);
            pc++;
            break;
        case OP_ladd:
            stack -= 4;
            copy64(stack+2,(Var*)&c1);
            copy64(stack,  (Var*)&c2);
            c.ll = c1.ll + c2.ll;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_lsub:
            stack -= 4;
            copy64(stack+2,(Var*)&c1);
            copy64(stack,  (Var*)&c2);
	         c.ll = c2.ll - c1.ll;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_lmul:
            stack -= 4;
            copy64(stack+2,(Var*)&c1);
            copy64(stack,  (Var*)&c2);
	         c.ll = c1.ll * c2.ll;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_ldiv:
            stack -= 4;
            copy64(stack+2,(Var*)&c1);
            copy64(stack,  (Var*)&c2);
            if (c1.ll == 0)
                goto div_by_zero_error;
	         c.ll = c2.ll / c1.ll;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_lrem:
            stack -= 4;
            copy64(stack+2,(Var*)&c1);
            copy64(stack,  (Var*)&c2);
            if (c1.ll == 0)
                goto div_by_zero_error;
	         c.ll = c2.ll % c1.ll;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_lneg:
            stack -= 2;
            copy64(stack,  (Var*)&c);
            c.ll = -c.ll;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_lshl:
            stack -= 3;
            copy64(stack,  (Var*)&c);
            c.ll = c.ll << stack[2].intValue;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_lshr:
            stack -= 3;
            copy64(stack,  (Var*)&c);
            __rt_shrs64(&c.ll,c.ll,(short)stack[2].intValue);
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_lushr:
            stack -= 3;
            copy64(stack,  (Var*)&c);
            __rt_shru64(&c.ll,c.ll,(short)stack[2].intValue); // guich@200b4_199: fixed wrong calculation for >>>
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_land:
            stack -= 4;
            copy64(stack,(Var*)&c1);
            copy64(stack+2,(Var*)&c2);
            c.ll = c1.ll & c2.ll;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_lor:
            stack -= 4;
            copy64(stack,(Var*)&c1);
            copy64(stack+2,(Var*)&c2);
            c.ll = c1.ll | c2.ll;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_lxor:
            stack -= 4;
            copy64(stack,(Var*)&c1);
            copy64(stack+2,(Var*)&c2);
            c.ll = c1.ll ^ c2.ll;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_i2l:
            stack -= 1;
            c.ll = (int64)stack[0].intValue;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_l2i:
            stack -= 2;
            copy64(stack,(Var*)&c);
            (*stack++).intValue = (int32)c.ll;
            pc++;
            break;
        case OP_l2f:
            stack -= 2;
            copy64(stack,(Var*)&c);
            (*stack++).floatValue = (float32)c.ll;
            pc++;
            break;
        case OP_f2l:
            stack -= 1;
            c.ll = (int64)stack[0].floatValue;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_l2d:
            stack -= 2;
            copy64(stack,(Var*)&c);
            c.d = (double)c.ll;
            copy64((Var*)&c,stack);
            stack+=2;
            pc++;
            break;
        case OP_d2l:
            stack -= 2;
            copy64(stack,(Var*)&c);
            c.ll = (int64)c.d;
            copy64((Var*)&c,stack);
            stack += 2;
            pc++;
            break;
        case OP_lcmp:
            stack-=4;
            copy64(stack,(Var*)&c1);
            copy64(stack+2,(Var*)&c2);
            (*stack++).intValue = __rt_cmps64(c1.ll,c2.ll);
            pc++;
            break;
 		  case OP_athrow:
   		   oldpc = pc; // guich@241_6
   			if (stack[-1].obj == 0)
	   		{
		   		stack--;
			   	throwException(NullPointerEx,"");//,&stack); // NullPointerException
   			}
	   		thrownException = stack[-1].obj;
throw_exception:
   			if (exceptionStackPtr == NULL)
   			   exceptionStackPtr = vmStackPtr;
		   	stack = var;
			   stack += method->maxLocals;
   			(*stack++).obj = thrownException;
	   		goto handle_exception;

        // NOTE: this is the full list of unsupported opcodes. Adding all
        // these cases here does not cause the VM executable code to be any
        // larger, it just makes sure that the compiler uses a jump table
        // with no spaces in it to make sure performance is as good as we
        // can get (tested under Codewarrior for PalmOS).
        case OP_breakpoint: // 202
        // can't use default!
#ifdef PALMOS
   #define BOOSTVAR var++; break; // guich@300_54: those var++ are used just to make the hit count of 'var' higher and make it be stored in a better register (A4 instead of D4)
#else										 // note: in PalmOS the break statement above *can't* be removed; note also that this will cause an invalid op-code to just be ignored; since this is a extremely rare situation, it doesnt matter
   #define BOOSTVAR
#endif
        case 203: BOOSTVAR
  		  case 204: BOOSTVAR
        case 205: BOOSTVAR
        case 206: BOOSTVAR
        case 207: BOOSTVAR
        case 208: BOOSTVAR
        case 209: BOOSTVAR
        case 210: BOOSTVAR
        case 211: BOOSTVAR
        case 212: BOOSTVAR
        case 213: BOOSTVAR
        case 214: BOOSTVAR
        case 215: BOOSTVAR
        case 216: BOOSTVAR
        case 217: BOOSTVAR
        case 218: BOOSTVAR
        case 219: BOOSTVAR
        case 220: BOOSTVAR
        case 221: BOOSTVAR
        case 222: BOOSTVAR
        case 223: BOOSTVAR
        case 224: BOOSTVAR
        case 225: BOOSTVAR
        case 226: BOOSTVAR
        case 227: BOOSTVAR
        case 228: BOOSTVAR
        case 229: BOOSTVAR
        case 230: BOOSTVAR
        case 231: BOOSTVAR
        case 232: BOOSTVAR
        case 233: BOOSTVAR
        case 234: BOOSTVAR
        case 235: BOOSTVAR
        case 236: BOOSTVAR
        case 237: BOOSTVAR
        case 238: BOOSTVAR
        case 239: BOOSTVAR
        case 240: BOOSTVAR
        case 241: BOOSTVAR
        case 242: BOOSTVAR
        case 243: BOOSTVAR
        case 244: BOOSTVAR
        case 245: BOOSTVAR
        case 246: BOOSTVAR
        case 247: BOOSTVAR
        case 248: BOOSTVAR
        case 249: BOOSTVAR
        case 250: BOOSTVAR
        case 251: BOOSTVAR
        case 252: BOOSTVAR
        case 253: BOOSTVAR
        case 254: BOOSTVAR
        case 255:
            VmQuickError(ERR_BadOpcode, NULL);
            goto error;
     }
    goto step;

handle_exception:
	{
		WObject wasThrown = thrownException;
		uint32 where = oldpc-method->codeStart/*METH_code(method)*/;
		uint32 i;
		thrownException = 0;
		for (i = 0; i< method->numHandlers; i++)
		{
			WExceptionHandler * eh = method->handlers+i;
			if (eh->startPc <= where && eh->endPc >= where)
			{
				if (eh->catchType != 0)
				{
					WClass *iclass = getClassByIndex(wclass, eh->catchType,0);
					if (iclass == NULL || !compatible(WOBJ_class(wasThrown), iclass))
					   continue;
				}
				pc = method->codeStart + eh->handlerPc;
            firstException = 0; // guich@500_6: once the exception is catched, there's no need to keep this.
				goto step;
			}
		}
		thrownException = wasThrown;
		goto methodreturn;
	}

stack_overflow_error:
   VmQuickError(ERR_StackOverflow, NULL);
   goto error;
method_called_on_array_error:
   VmQuickError(ERR_MethodCalledOnArray, NULL);
   goto error;
class_cast_error:
   oldpc = pc; // guich@241_6
   throwException(ClassCastEx,NULL);
   goto throw_exception;
null_obj_error:
   oldpc = pc; // guich@241_6
   throwException(NullPointerEx,NULL);
   goto throw_exception;
div_by_zero_error:
   oldpc = pc; // guich@241_6
   throwException(DivideByZeroEx,NULL);
   goto throw_exception;
index_range_error:
   oldpc = pc; // guich@241_6
   throwException(ArrayIndexEx,NULL);
   goto throw_exception;
null_array_error:
   oldpc = pc; // guich@241_6
   throwException(NullPointerEx,NULL);
   goto throw_exception;
error:
   oldpc = pc; // guich@241_6
   vmStackPtr = baseFramePtr;
   goto end;

methodreturn:
    // pop frame and restore state
    vmStackPtr -= METH_localsStack2(method);

    if (vmStackPtr == baseFramePtr)
    {
       // fully completed execution
       if (thrownException) // unhandled exception?
       {
          UtfString utf;
          WClass *wclass;
          WObject obj;
          if (firstException == NULL)
            firstException = thrownException;
          obj = firstException[1].obj;
          // get the class name
          wclass = WOBJ_class(firstException); // guich@340_1: use firstException instead of thrownException
          // get the message
          if (obj)
             utf = stringToUtf(obj, STU_USE_STATIC | STU_NULL_TERMINATE);
          else
             utf.len = 0;
          if (!(wclass->className.len == 25 && xstrncmp(wclass->className.str,"waba/sys/AppExitException",25) == 0)) // guich@402_30: avoid dumping when just an AppExitException
             VmError(ERR_UncaughtException,wclass,&utf,NULL);
          else
             thrownException = NULL; // guich@402_42: fixed problem of onExit not being called when a popupBlockingModal was used and the user switch out the app.
          // andy@340_35: Added so that a native method can inspect the exception
          vmStackPtr->obj = thrownException;
       }
       // andy@340_35: This entire 'else if' added so that native methods can have access to return value.
       else
       if (returnedValueCount)
       {
          vmStackPtr[0] = returnedValue[0];
          if (returnedValueCount == 2)
             vmStackPtr[1] = returnedValue[1];
       }
       firstException = exceptionStackPtr = NULL;
       goto end;
    }
    stack = (Var *)((--vmStackPtr)->intValue + (byte *)vmStackStart);
    if (returnedValueCount && !thrownException) // (**)
    {
       *stack++ = returnedValue[0];
       if (returnedValueCount == 2)
          *stack++ = returnedValue[1];
    }
    var = (Var *)((--vmStackPtr)->intValue + (byte *)vmStackStart);
    pc = (--vmStackPtr)->pc;
    oldpc = (--vmStackPtr)->oldpc; // guich@241_6
    wclass = (WClass *)(vmStackPtr - 1)->refValue;
    method = (WClassMethod *)(vmStackPtr - 2)->refValue;
 	 if (thrownException) // guich@240_14: catch uncaught Exceptions thrown from onPaint method
 	 {
	    (*stack++).obj = thrownException; // guich@400_16: i can place this here bc stack is not changed since (**)
       if (firstException == NULL) // guich@340_1
 	       firstException = thrownException;
		 goto throw_exception;
	 }
    goto step;
end:
    return vmStackPtr;
}

//////////////////////////////////////////////////////////////////////////////////////
//
// public functions
//

static void _xfree(void *p)
{
   xfree_(p);
}

static byte growStack(byte **stackStart, byte **stackPtr, byte **StackEnd, byte **ptr1, byte **ptr2, byte **ptr3, uint32 *stackSizeInBytes)
{
   byte *ptr;

   //debug(int2str("growing stack - message for debug purposes. ",*stackSizeInBytes));
   *stackSizeInBytes += 512;
   ptr = xrealloc(*stackStart,stackSizeInBytes);
   if (!ptr) return 0;
   *stackPtr = ptr + (*stackPtr - *stackStart); // keep stackPtr at the same position but at a new base address
   *StackEnd = ptr + *stackSizeInBytes-4; // guich@400_87: sub 4 bytes
   // keep given pointers synchronized
   if (ptr1) *ptr1 = ptr + (*ptr1 - *stackStart);
   if (ptr2) *ptr2 = ptr + (*ptr2 - *stackStart);
   if (ptr3) *ptr3 = ptr + (*ptr3 - *stackStart);
   *stackStart = ptr;
   return 1;
}

// guich@200b4: count how many elements have until the last NULL element
int arrayCount(char *array, int itemSize)
{
   int count = 0;
   int size = 0;
   int i,onlyZeros;
   if (array)
      while (1)
      {
         onlyZeros = 1;
         for (i = 0; i < itemSize; i++)
            if (array[size+i]) // some number in the struct?
            {
               onlyZeros = 0;
               break;
            }
         if (onlyZeros)
            break;
         else
         {
            count++;
            size += itemSize;
         }
      }
   return count;
}

// guich@200b4: expand an array. the array must end with a NULL element. Returns -1 in case of error, otherwise the index to the first free element
char *expandArray(char *array, int itemSize, int increment, int *insertIndex)
{
   int32 oldCount = arrayCount(array, itemSize);
   int32 newCount = oldCount + 1 + increment;
   uint32 newSize  = newCount * itemSize;
   char *newArray = xmalloc(newSize);
   if (newArray == NULL)
   {
      VmQuickError(ERR_CantAllocateMemory,NULL);
      *insertIndex = -1;
      return array;
   }
   xmemzero(newArray,newSize);
   xmemmove(newArray, array, oldCount* itemSize);
   xfree(array);

   *insertIndex = oldCount;
   return newArray;
}

static void VmInit()
{
    uint32 i;

    if (vmInitialized)
        return;

    // NOTE: ordering is important here. The first thing we
    // need to do is initialize the global variables so if we
    // return not fully initialized, a VmFree() call will still
    // operate correctly. Also, its important not to statically
    // initialize them since VmInit() can be called after VmFree()
    // and if they were just statically intialized, they wouldn't
    // get reset.
    vmStackSizeInBytes = 16*1024-6;
    vmStackStart = NULL;
    vmStackPtr = 0;

    nmStackSizeInBytes = 1024-6;
    nmStackStart = NULL;
    nmStackPtr = 0;

    classHeap = NULL;
    classHeapSize = 32L*1024L-128L;
    classHeapUsed = 0;
    for (i = 0; i <= CLASS_HASH_SIZE; i++)
        classHashList[i] = NULL;
    xmemzero((uchar *)&vmStatus, sizeof(vmStatus));

    // guich@200b3: no more need for "extra" memory block bc now we allocate memory in the storage heap
    // allocate stacks and init
    vmStackStart = (Var *)xmalloc(vmStackSizeInBytes);
    nmStackStart = (WObject *)xmalloc(nmStackSizeInBytes);
    classHeap = (uchar *)xmalloc(classHeapSize);
    if (vmStackStart == NULL || nmStackStart == NULL || classHeap == NULL)
    {
        if (vmStackStart != NULL)
            xfree(vmStackStart);
        if (nmStackStart != NULL)
            xfree(nmStackStart);
        if (classHeap != NULL)
            xfree(classHeap);
        VmQuickError(ERR_CantAllocateMemory, NULL);
        return;
    }

    // zero out memory areas
    xmemzero((uchar *)vmStackStart, vmStackSizeInBytes);
    xmemzero((uchar *)nmStackStart, nmStackSizeInBytes);
    xmemzero((uchar *)classHeap, classHeapSize);
    vmStackPtr = vmStackStart;
    vmStackEnd = (Var *)((byte *)vmStackStart + vmStackSizeInBytes - 300); // guich@400_87: -300
    nmStackPtr = nmStackStart;
    nmStackEnd = (WObject *)((byte *)nmStackStart + nmStackSizeInBytes - 4);  // guich@400_87: -4

    if (!initObjectHeap())
    {
        VmQuickError(ERR_CantAllocateObjectMemory, NULL);
        return;
    }
    vmInitialized = 1;
}

// copies one or two UtfStrings (by concatenation) into a character
// buffer (null terminated) suitable for output
void printToBuf(char *buf, int maxbuf, UtfString *s1, UtfString *s2)
{
    int len, len2;

    len = 0;
    if (s1 != NULL && s1->str != NULL && s1->len > 0)
    {
        len = s1->len;
        if (len >= maxbuf)
            len = maxbuf - 1;
        xstrncpy(buf, s1->str, len);
    }
    if (s2 != NULL && s2->str != NULL && s2->len > 0)
    {
        len2 = s2->len;
        if (len2 + len >= maxbuf)
            len2 = maxbuf - len - 1;
        xstrncpy(&buf[len], s2->str, len2);
        len += len2;
    }
    buf[len] = 0;
}


void dumpStackTrace()
{
	WClass *wclass;
	WClassMethod *method;
	UtfString *className, methodName, methodDesc;
   Var *stackPtr;
   char *buf,*c;

	if (!exceptionStackPtr)
		return;
   debug("Stack trace:");
	stackPtr = exceptionStackPtr;

   while (stackPtr >= vmStackPtr) // guich@504_9: replaced vmStackStart by vmStackPtr
	{
  	   //debug(int2str("dumping stack at",(int32)stackPtr));
		wclass = (WClass *)(--stackPtr)->refValue;
		method = (WClassMethod *)(--stackPtr)->refValue;
		if (method <= (WClassMethod*)0xFFFF || wclass <= (WClass*)0xFFFF || (method->maxLocals + method->maxStack + (uint32)2) != method->localsStack2) // guich@340_1: this ugly test condition is because i could not find a good way to find the correct point to stop the stack from going up. (i debugged trying to find this for more than 8 hours). So, we just keep looking for an invalid pointer
		   break;
		className = &wclass->className;
		methodName = getUtfString(wclass, METH_nameIndex(method));
		methodDesc = getUtfString(wclass, METH_descIndex(method));
      buf = (char *)xmalloc(className->len + methodName.len + methodDesc.len + 3);
      if (!buf) return;
      c = buf;
      xstrncpy(c, className->str, className->len); c += className->len;
      *c++ = '.';
      xstrncpy(c, methodName.str, methodName.len); c += methodName.len;
      xstrncpy(c, methodDesc.str, methodDesc.len); c += methodDesc.len;
      *c++ = 0;
      debug(buf);

		/*if (METH_isNative(method))
			stackPtr -= (--stackPtr)->intValue;
	   else*/
      stackPtr -= method->maxLocals + method->maxStack + STACKSAVESIZE; // guich@330_22
      xfree(buf);
	}
	debug("---------------------------------");
	exceptionStackPtr = NULL;
}

static void VmError(uint16 errNum, WClass *iclass, UtfString *desc1, UtfString *desc2)
{
    WClass *wclass;
    WClassMethod *method;
    UtfString methodName, methodDesc;

    // NOTE: Don't overwrite an existing error since it may be the
    // root cause of this error.
    if (vmStatus.errNum != 0)
        return;
    vmStatus.errNum = errNum;

    // get current class and method off stack
    if (VM_STACK_COUNT > 0 && errNum != 12)
    {
        wclass = (WClass *)(vmStackPtr - 1)->refValue;
        method = (WClassMethod *)(vmStackPtr - 2)->refValue;
    }
    else
    {
        wclass = 0;
        method = 0;
    }
    if ((uint32)wclass < 0xFFFF || (uint32)method < 0xFFFF)
    {
       wclass = 0;
       method = 0;
    }

    // output class and method name
    if (wclass)
        printToBuf(vmStatus.className, sizeof(vmStatus.className), &wclass->className, NULL);
    if (method)
    {
        methodName = getUtfString(wclass, METH_nameIndex(method));
        methodDesc = getUtfString(wclass, METH_descIndex(method));
        printToBuf(vmStatus.methodName, sizeof(vmStatus.methodName), &methodName, &methodDesc);
    }

    // output additional error arguments (target class, desc, etc.)
    if (iclass)
        printToBuf(vmStatus.arg1, sizeof(vmStatus.arg1), &iclass->className, NULL);
    printToBuf(vmStatus.arg2, sizeof(vmStatus.arg2), desc1, desc2);

    dumpStackTrace(); // guich@300_23: now for all platforms
}

void VmQuickError(uint16 errNum, char *str)
{
   if (str)
      switch (errNum)
      {
         case ERR_CantAllocateMemory:
         case ERR_OutOfClassMem:
         case ERR_OutOfObjectMem:
            throwException(OutOfMemoryEx, str);
            return;
         case ERR_CantFindClass:
            throwException("java/lang/RuntimeException", str);
            return;
         case ERR_NullObjectAccess:
         case ERR_NullArrayAccess:
            throwException(NullPointerEx, str);
            return;
         case ERR_IndexOutOfRange:
            throwException(ArrayIndexEx, str);
            return;
         case ERR_DivideByZero:
            throwException(DivideByZeroEx, str);
            return;
         case ERR_ClassCastException:
            throwException(ClassCastEx,str);
            return;
      }
    VmError(errNum, NULL, NULL, NULL);
}

static void VmFree()
{
    if (!vmInitialized)
        return;
    if (vmStackStart)
       xfree(vmStackStart);
    if (nmStackStart)
       xfree(nmStackStart);
    if (classHeap)
       xfree(classHeap);
    vmInitialized = 0;
}

static WObject VmStartApp(char *className)
{
    WObject mainWinObj;
    WClass *wclass, *vclass, *mainWinClass;
    WClassMethod *method;
    Var params[1];
#ifndef WIN32
    uint32 dummy = 1;
#endif

    if (!vmInitialized)
        return 0;

    stringClass = getClass(createUtfString("java/lang/String"),1);
    if (stringClass == NULL)
        return 0;

    VmGetSettings();     // guich: get the preferences before anything! the user may use them in the onStart method - guich@420_1: moved to here (before mainclass load) to avoid nullpointerexceptions when initializing static colors

    // reserve memory for the outOfMemory exception object
 	 outOfMemoryClass = getClass(createUtfString(OutOfMemoryEx),1); // guich@230_14 - guich@504_7: moved to here

    wclass = getClass(createUtfString(className),1);
    if (wclass == NULL)
        return 0;
    // make sure its a MainWindow class
    mainWinClass = getClass(createUtfString("waba/ui/MainWindow"),1);
    if (!compatible(wclass, mainWinClass))
    {
       debug("The main class must extend waba.ui.MainWindow");
       return 0;
    }
    // create MainWindow object
    mainWinObj = createObject(wclass);
    if (mainWinObj == 0)
        return 0;
    // NOTE: main window object is pushed on native stack to prevent it being GC'd
    if (pushObject(mainWinObj) == -1)  // make sure it doesn't get GC'd
        return 0;
    params[0].obj = mainWinObj;
    // call MainWindow constructor

#ifndef WIN32 // in win32 we do not have a valid screen until the MainWindow construction (next line below)
    SWCallPublic(SWLibRefNum, PUBLIC_DrawHourGlass,0,&dummy);
#endif

    method = getMethod(wclass, createUtfString("<init>"), createUtfString("()V"), NULL);
    if (method != NULL)
        executeMethod(wclass, method, params, 1);
    // call onStart()
    method = getMethod(wclass, createUtfString("appStarting"), createUtfString("()V"), &vclass); // guich@200b4_107: changed call to appStarting.
    if (method != NULL)
        executeMethod(vclass, method, params, 1);
    return mainWinObj;
}

byte callMethod(WObject classObj, char *methodName, char *methodDesc, Var *params, int paramCount)
{
    WClass *wclass, *vclass;
    WClassMethod *method;
    byte canCall;

    wclass = WOBJ_class(classObj);
    method = getMethod(wclass, createUtfString(methodName), createUtfString(methodDesc), &vclass);

    canCall = (byte)(method != NULL && (METH_accessFlags(method) & ACCESS_PRIVATE) == 0); // guich@340_31: now check only if its not private.
    if (canCall)
       executeMethod(vclass, method, params, paramCount);
    else
       {debug("can't call method"); debug(wclass->className.str); debug(methodName); debug(methodDesc); debug(method?"method is private":"method not found");}
    return canCall;
}

static void VmStopApp(WObject mainWinObj)
{
   Var params[1];
   if (!vmInitialized || mainWinObj == 0)
      return;
   params[0].obj = mainWinObj;
   callMethod(mainWinObj,"appEnding","()V",params,1);
   VmStoreSettings(); // store only after called onExit!
   if (vmGlobals->dumpMemStats)
      debugMemory();
}

//
// Class Loader
//

// guich@200b2: try to optimize the vm - instead of comparing strings, compare ULongs
static uint32 getHash(UtfString *name, UtfString *desc)
{
   uint32 i = 0;
   int k;
   char *ac;
   ac = name->str;
   k = name->len;
   while (k--) i = (i << 5) - i + *ac++; // i*31 = i * 32 - i
   ac = desc->str;
   k = desc->len;
   while (k--) i = (i << 5) - i + *ac++;
   return i;
}

static uint32 genHashCode(UtfString *name)
{
    uint32 value=0; // guich@200b4: optimized
    int l = name->len;
    char *c = name->str;

    while (l--)
        value += *c++;
    value = (value << 6) + name->len;
    return value;
}

static uchar *allocClassPart(uint32 size inD0) // don't inline!
{
    uchar *p;

    // align to 4 byte boundry
    size = (size + 3) & ~3;
    if (classHeapUsed + size > classHeapSize)
    {
        // guich@200b4_94: allocate class heap as needed. Theres no problem in overiding the pointer, since all memory is released at end
        uint32 temp = classHeapUsed;
        classHeap = xrealloc(classHeap,&temp); // dont waste memory!
        totalClassHeap += classHeapUsed;
        classHeapUsed = 0;
        if (size > classHeapSize) // maybe size is greater than the default classHeapUsed?
           classHeapSize = size+512;
        classHeap = (uchar *)xmalloc(classHeapSize);
        if (classHeap == NULL)
        {
           VmQuickError(ERR_OutOfClassMem,"Out of class memory");
           return NULL;
        }
        xmemzero((uchar *)classHeap, classHeapSize);
    }
    p = &classHeap[classHeapUsed];
    classHeapUsed += size;
    return p;
}

WClass *getClass(UtfString className, byte throwEx)
{
    WClass *wclass, *superClass;
    WClassMethod *method;
    uint16 i, n, classIndex,maxMethodIdx=0,maxFieldClassIdx=0;
    uint32 classHash, size;
    uchar *p;
    char *sp = className.str;

    if (className.len > 10 && sp[className.len-4] == '4')
       className.len -= 4; // guich@200b4_151: tweak it; instead of create NativeMethods4JDK, create NativeMethods

    // look for class in hash list
    classHash = genHashCode(&className) & CLASS_HASH_SIZE;
    wclass = classHashList[classHash];
    while (wclass != NULL)
    {
        if (className.len == wclass->className.len &&
            !xstrncmp(className.str, wclass->className.str, className.len))
            break;
        wclass = wclass->nextClass;
    }
    if (wclass != NULL)
        return wclass;

    p = nativeLoadClass(className, NULL);
    if (p == NULL)
    {
        if (throwEx)
        {
           char *printfbuf = xmalloc(className.len+20); // guich@350_30
           xmemzero(printfbuf,className.len+20);
           xstrncpy(printfbuf,"Can't find class ",17); // guich@421_34: the xstrprintf was giving "odd address" problems
           xstrncpy(printfbuf+17,className.str,className.len);
           VmQuickError(ERR_CantFindClass, printfbuf);
           xfree(printfbuf);
        }
        return NULL;
    }

    // NOTE: The garbage collector may run at any time during the
    // loading process so we need to make sure the class is valid
    // whenever an allocation may occur (static class vars, <clinit>, etc)
    // The various int variables will be initialzed to zero since
    // the entire class areas is zeroed out to start with.
    wclass = (WClass *)allocClassPart(sizeof(struct WClassStruct));
    if (wclass == NULL)
        return NULL;

    wclass->byteRep = p;

    // parse constants
    p += 6;
    wclass->is12 = (byte)(getUInt16(p) >= 0x002E); // guich@310_14
    p += 2;
    wclass->numConstants = n = getUInt16(p);
    p += 2;
    if (n != 0)
    {
        size = sizeof(ConsOffset) * n;
        wclass->constantOffsets = (ConsOffset *)allocClassPart(size);
        if (wclass->constantOffsets == NULL)
        {
            wclass->numConstants = 0;
            return NULL;
        }
        wclass->constantOffsets--; // guich@200b4: just to avoid doing idx-1 all the times

        for (i = 1; i < n; i++)
        {
           wclass->constantOffsets[i] = (ConsOffset)p;
           p++;
           switch (CONST_tag(wclass, i)) // guich@400_16: moved to here since now we compute other things; also, ordered and inserted blank cases to let the compiler optimize it to a jump table
           {
				  case 0:
				     break;
              case CONSTANT_Utf8: // 1
                 p += 2;
                 p += CONST_utfLen(wclass, i);
                 break;
              case 2:
                 break;
              case CONSTANT_Integer: // 3
              case CONSTANT_Float: // 4
                 p += 4;
                 break;
              case CONSTANT_Long: // 5
              case CONSTANT_Double: // 6
                 p += 8;
                 // after a long or double, next entry does not exist
                 i++;
                 break;
              case CONSTANT_Class: // 7
                 maxFieldClassIdx = i;
                 p += 2;
                 break;
              case CONSTANT_String: // 8
                 p += 2;
                 break;
              case CONSTANT_Fieldref: // 9
                 maxFieldClassIdx = i;
                 p += 4;
                 break;
              case CONSTANT_Methodref: // 10
              case CONSTANT_InterfaceMethodref: // 11
                 maxMethodIdx = i;
                 p += 4;
                 break;
              case CONSTANT_NameAndType: // 12
                 p += 4;
                 break;
           }
        }
        maxMethodIdx++; // not sure if absolutely necessary
        maxFieldClassIdx+=2; // idem
        // note for the following memory optimizations: if we could reorder the constant
        // pool placing references before the rest, we could decrease the memory usage
        // guich@400_29: instead of allocate numConstants for the binded methods,
        // allocate just the maximum method index in the constant pool
        size = sizeof(WClassMethod*) * maxMethodIdx;
        wclass->bindedMethods = (WClassMethod **)allocClassPart(size);
        if (wclass->bindedMethods == NULL)
        {
            wclass->numConstants = 0;
            return NULL;
        }
        // guich@400_29: instead of allocate numConstants for the binded contants,
        // allocate just the maximum field/class/method index in the constant pool
        size = sizeof(BindedConst) * (maxMethodIdx>maxFieldClassIdx?maxMethodIdx:maxFieldClassIdx);
        wclass->bindedConstants = (BindedConst *)allocClassPart(size);
        if (wclass->bindedConstants == NULL)
        {
            wclass->numConstants = 0;
            return NULL;
        }
        wclass->bindedConstants--; // guich@200b4: just to avoid doing idx-1 all the times
    }

    // second attribute section
    wclass->attrib2 = p;
    p += 6;

    // assign class name
    wclass->className = getUtfString(wclass, CONST_nameIndex(wclass, WCLASS_thisClass(wclass)));

    // guich@240_10: fix when tour.Tour is located in Tour (and not tour/Tour)
    if (wclass->className.len != className.len)
    {
       VmError(ERR_ClassInWrongPackage,wclass,NULL,NULL);
       releaseClass(className, wclass->byteRep);
       return NULL;
    }

    // NOTE: add class to class list here so garbage collector can
    // find it during the loading process if it needs to collect.
    wclass->nextClass = classHashList[classHash];
    classHashList[classHash] = wclass;

    // load superclasses (recursive) here so we can resolve var
    // and method offsets in one pass
    superClass = NULL;
    classIndex = WCLASS_superClass(wclass);
    if (classIndex != 0)
    {
        superClass = getClassByIndex(wclass, classIndex,1);
        if (superClass == NULL)
            return NULL; // can't find superclass
        // fill in superclasses table
        n = superClass->numSuperClasses + 1;
        size = n * sizeof(WClass *);
        wclass->superClasses = (WClass **)allocClassPart(size);
        if (wclass->superClasses == NULL)
           return NULL;
        size = (n - 1) * sizeof(WClass *);
        xmemmove(wclass->superClasses, superClass->superClasses, size);
        wclass->superClasses[n - 1] = superClass;
        wclass->numSuperClasses = n;

        // inherit num of superclass variables to start
        wclass->numVars = superClass->numVars;
    }

    // skip past interfaces
    n = getUInt16(p);
    p += 2 + (n * 2);

    // parse fields
    wclass->numFields = n = getUInt16(p); // guich@400_16: store in n for fast access
    p += 2;
    if (n != 0)
    {
       size = sizeof(WClassField) * n;
       wclass->fields = (WClassField *)allocClassPart(size);
       if (wclass->fields == NULL)
       {
          wclass->numFields = 0;
          return NULL;
       }
       for (i = 0; i < n; i++)
          p = loadClassField(wclass, &wclass->fields[i], p);
    }

    // parse methods
    wclass->numMethods = n = getUInt16(p);
    p += 2;
    size = sizeof(WClassMethod) * n;
    if (size != 0)
    {
        wclass->methods = (WClassMethod *)allocClassPart(size);
        if (wclass->methods == NULL)
        {
            wclass->numMethods = 0;
            return NULL;
        }
        for (i = 0; i < n; i++)
            if ((p = loadClassMethod(wclass, &wclass->methods[i], p)) == NULL)
               return NULL; // guich@200final_2: correct when the class was invalid and null is returned
    }

    // sort the methods and create the virtual method map for fast lookup
    if (createVirtualMethodMap(wclass) < 0)
        return NULL;

    // skip final attributes section

    // set hooks (before class init which might create/free objects of this type)
    setClassHooks(wclass);

    // if our superclass has a destroy func, we inherit it. If not, ours overrides
    // our base classes destroy func (the native destroy func should call the
    // superclasses)
    if (superClass != NULL && !WCLASS_hasDestroyFunc(wclass) && WCLASS_hasDestroyFunc(superClass))
    {
       wclass->objDestroyFuncNum = superClass->objDestroyFuncNum;
       wclass->objDestroyFuncLibRefnum = superClass->objDestroyFuncLibRefnum;
    }
    wclass->isInterface = (byte)((WCLASS_accessFlags(wclass) & ACCESS_INTERFACE) > 0);
    wclass->isAbstract = (byte)((WCLASS_accessFlags(wclass) & ACCESS_ABSTRACT) > 0);

    // call static class initializer method if present
    method = getMethod(wclass, createUtfString("<clinit>"), createUtfString("()V"), NULL);
    if (method != NULL)
        executeMethod(wclass, method, NULL, 0);
 	 if (thrownException != 0)
 	 {
 		 WObject thrown = thrownException;
		 classHashList[classHash] = wclass->nextClass;
		 if (WOBJ_class(thrown) != outOfMemoryClass)
		 {
			 WObject ex;
			 thrownException = 0;
			 ex = throwException("java/lang/RuntimeException","ExceptionInInitializerError");
			 if (WOBJ_class(ex) != outOfMemoryClass)
				 ex[3].obj = thrown;
		 }
		 return NULL;
	 }
    return wclass;
}

static int createVirtualMethodMap(WClass *wclass)
{
    WClass *superClass;
    VirtualMethodMap *superVMap;
    uint32 size;
    uint16 i, n, nLow, nHigh, numSuperClasses;
    WClassMethod *method;
    VirtualMethodMap *vMap;
    UtfString name, desc;
    WClassMethod tempMethod;

    // This method is responsible for filling in the VirtualMethodMap
    // structure to allow fast method lookup.
    //
    // It also sorts the method table so new virtual methods appear first,
    // overridden virtual methods second and non-virtual methods last.
    //
    // The method map contains the list of this classes superclasses
    // as well as an array that maps virtual method indicies to
    // class and method indicies to handle overridden methods.
    vMap = &wclass->vMethodMap;

    // sort methods so virtual appear first and non-virtual last
    n = wclass->numMethods;
    nLow = 0;
    nHigh = 0;
    while (nLow + nHigh < n)
    {
        method = &wclass->methods[nLow];
        // NOTE: the virtual section will not include <init> methods since
        // they should not be inherited
        if (((METH_accessFlags(method) & (ACCESS_PRIVATE | ACCESS_STATIC)) == 0) && !method->isInit)
            nLow++;
        else
        {
            nHigh++;
            // swap non-virtual to bottom
            tempMethod = *method;
            *method = wclass->methods[n - nHigh];
            wclass->methods[n - nHigh] = tempMethod;
        }
    }
    vMap->numVirtualMethods = nLow;

    numSuperClasses = wclass->numSuperClasses;
    if (numSuperClasses == 0)
    {
        // Object class - no superclass map to inherit and no inherited methods
        // to override
        return 0;
    }

    superClass = wclass->superClasses[numSuperClasses - 1];
    superVMap = &superClass->vMethodMap;

    // create method map by copying superclass method map and inheriting
    // superclass virtual methods
    vMap->mapSize = superVMap->mapSize + superVMap->numVirtualMethods;
    size = vMap->mapSize * sizeof(VMapValue);
    vMap->mapValues = (VMapValue *)allocClassPart(size);
    if (vMap->mapValues == NULL)
        return -1;
    size = superVMap->mapSize * sizeof(VMapValue);
    xmemmove(vMap->mapValues, superVMap->mapValues, size);
    //add superclass #'s + method numbers into second portion
    n = 0;
    for (i = superVMap->mapSize; i < vMap->mapSize; i++)
    {
        vMap->mapValues[i].c = superClass; // guich@400_16: store the pointer instead the index
        vMap->mapValues[i].m = &superClass->methods[n++];
    }

    // generate hash table of inherited methods allowing fast check
    // for overriden methods
    xmemzero(methodHash, sizeof(methodHash));
    for (i = vMap->mapSize; i > 0; i--)
       methodHash[((WClassMethod*)vMap->mapValues[i-1].m)->hash & OVERRIDE_HASH_SIZE] = i;
    // in virtual method section, determine overrides and move overrides to
    // bottom of virtual section
    n = vMap->numVirtualMethods;
    nLow = 0;
    nHigh = 0;
    while (nLow + nHigh < n)
    {
        int32 overrideIndex;
        method = &wclass->methods[nLow];
        // look in hash table first
        overrideIndex = methodHash[method->hash & OVERRIDE_HASH_SIZE];
        if (!overrideIndex)
        {
            nLow++;
            continue;
        }
        overrideIndex--;
        if (getMethodByMapNum(wclass, NULL, (uint16)overrideIndex)->hash == method->hash)
            ; // found it from hash
        else
        {
           name = getUtfString(wclass, METH_nameIndex(method));
           desc = getUtfString(wclass, METH_descIndex(method));
           overrideIndex = getMethodMapNum(wclass, &name, &desc, CLASS_SEARCH_INHERITED, NULL, NULL);
        }
        if (overrideIndex == -1)
            nLow++;
        else
        {
            // override - swap overridden method to bottom and add to map
            nHigh++;
            tempMethod = wclass->methods[nLow];
            wclass->methods[nLow] = wclass->methods[n - nHigh];
            wclass->methods[n - nHigh] = tempMethod;
            vMap->mapValues[overrideIndex].c = wclass; // guich@400_16: store the pointer instead the index
            vMap->mapValues[overrideIndex].m = &wclass->methods[n - nHigh];
        }
    }
    vMap->numVirtualMethods -= nHigh;

//#define NEVER
#ifdef NEVER
    // NOTE: This is some code you can run to see what the
    // internal class structure/virtual method map looks like
    {
    VMapValue mapValue;
    int ii, methodIndex, superIndex;
    WClass *iclass;
    UtfString mname, mdesc;
    UtfString className;
    int i;
    static char buf[120];

#ifdef __WINDOWSCE
    AllocConsole();
#endif
    debug("className: ");
    className = wclass->className;
    for (i = 0; i < className.len; i++)
    {
       xstrprintf(buf,"%c", className.str[i]);
       debug(buf);
    }
    xstrprintf(buf,"- nSuperClasses=%d inherited=%d new=%d total=%d",wclass->numSuperClasses, vMap->mapSize, vMap->numVirtualMethods,wclass->numMethods);
    debug(buf);
    debug("- FULL METHOD MAP");

    for (ii = 0; ii < vMap->mapSize + wclass->numMethods; ii++)
    {
        if (ii < vMap->mapSize)
        {
            mapValue = vMap->mapValues[ii];
            superIndex = mapValue.classNum;
            if (superIndex < wclass->numSuperClasses)
                iclass = wclass->superClasses[superIndex];
            else
                iclass = wclass;
            methodIndex = mapValue.methodNum;
            xstrprintf(buf,"- inherited %d/%d %d ", superIndex, wclass->numSuperClasses, methodIndex);
            debug(buf);
        }
        else
        {
            iclass = wclass;
            methodIndex = ii - vMap->mapSize;
            xstrprintf(buf,"- this class %d ", methodIndex);
            debug(buf);
        }
        method = &iclass->methods[methodIndex];
        if (methodIndex >= iclass->numMethods)
            debug("*************************************");
        mname = getUtfString(iclass, METH_nameIndex(method));
        mdesc = getUtfString(iclass, METH_descIndex(method));
        xstrprintf(buf,"- [%d] %s %s", ii, mname.str, mdesc.str);
        debug(buf);
    }
    }
#endif

    return 0;
}

uchar *loadClassField(WClass *wclass, WClassField *field, uchar *p)
{
    uint32 bytesCount; // guich@200b4: optimized
    uint16 attrCount, nameIndex;
    UtfString attrName;
    UtfString fieldName,fieldDesc;

    field->header = p;
    fieldDesc = getUtfString(wclass, FIELD_descIndex(field)); // guich@200 - added support for 64 bit vars.
    fieldName = getUtfString(wclass, FIELD_nameIndex(field)); // guich@200 - added support for 64 bit vars.
    field->hash = getHash(&fieldName,&fieldDesc);
    if (fieldDesc.str[0] == 'D' || fieldDesc.str[0] == 'J') // mlb@200b3 : fixed support for double and longs
       field->is64 = 1;
    if (FIELD_accessFlags(field) & ACCESS_STATIC)
       field->isStatic = 1;
	 else
    {
       // compute offset of this field's variable in the object
       field->var.varOffset = wclass->numVars++;
	    if (FIELD_is64(field))
	       field->var2.varOffset = wclass->numVars++;
    }

    p += 6;
    attrCount = getUInt16(p);
    p += 2;

    while (attrCount--)
    {
        nameIndex = getUInt16(p);
        attrName = getUtfString(wclass, nameIndex);
        p += 2;
        bytesCount = getUInt32(p);
        p += 4;
        if (FIELD_isStatic(field) && attrName.len == 13 && bytesCount == 2 &&
            xstrncmp(attrName.str, "ConstantValue", 13) == 0)
        {
            if (!FIELD_is64(field))
               field->var.staticVar = constantToVar(wclass, getUInt16(p));
            else
	         {
               Comp64 fcd;
               fcd.d = CONST_double(wclass, getUInt16(p));
               field->var.staticVar.half64 = fcd.ul[0];
               field->var2.staticVar.half64 = fcd.ul[1];
	         }
        }
        else
            ; // MS Java has COM_MapsTo field attributes which we skip
        p += bytesCount;
    }
    return p;
}

_inline_ Var constantToVar2(WClass *wclass, uint16 idx) // guich@320_15: good performance improvement!
{
    Var v;
    if (CONST_tag(wclass, idx) == CONSTANT_String) // guich@241: optimized the switch to 2 ifs only
       v.obj = createStringFromUtf(getUtfString(wclass, CONST_stringIndex(wclass, idx))); // guich@200b2: a constant string already is unicode; there is no need to convert to unicode
    else // CONSTANT_Integer and CONSTANT_Float
       v.intValue = CONST_integer(wclass, idx);
    return v;
}

Var constantToVar(WClass *wclass inA0, uint16 idx inD0) // Don't inline!
{
    Var v;
    if (CONST_tag(wclass, idx) == CONSTANT_String) // guich@241: optimized the switch to 2 ifs only
       v.obj = createStringFromUtf(getUtfString(wclass, CONST_stringIndex(wclass, idx))); // guich@200b2: a constant string already is unicode; there is no need to convert to unicode
    else // CONSTANT_Integer and CONSTANT_Float
       v.intValue = CONST_integer(wclass, idx);
    return v;
}

uchar *loadClassMethod(WClass *wclass, WClassMethod *method, uchar *p)
{
    uint32 bytesCount, codeCount,num; // guich@200b4: optimized
    uint16 attrCount,attrCount0,attrNameIndex, numHandlers, numAttributes;
    int32 numParams;
    byte hasAttributes; // native methods don't have attributes (guich@241: but they may have exceptions as attributes).
    uchar *attrStart;
    UtfString attrName, methodName, methodDesc;
    uint16 exceptionsCount = 0;
    char retType;

    method->header = p;
    p += 6;
    attrCount0 = attrCount = getInt16(p);
    p += 2;
    method->code.codeAttr = NULL;
    while (attrCount--)
    {
        attrStart = p;
        attrNameIndex = getUInt16(p);
        p += 2;
        attrName = getUtfString(wclass, attrNameIndex);
        bytesCount = getUInt32(p);
        p += 4;
  		  if (attrName.len == 10 && !xstrncmp(attrName.str,"Exceptions",10))
  		  {
			  p += bytesCount;
			  exceptionsCount++;
			  continue;
		  }
        if (attrName.len != 4 || xstrncmp(attrName.str, "Code", 4))
        {
            p += bytesCount;
            continue;
        }
        // Code Attribute
        method->code.codeAttr = attrStart;
        p += 4;
        codeCount = getUInt32(p);
        p += 4 + codeCount;

        // skip handlers and attributes
        numHandlers = getUInt16(p);
  		  method->numHandlers = numHandlers; // guich@230_14
  		  method->handlers = NULL;           // guich@230_14
        p += 2;
        // guich@230_14 p += numHandlers*8;
		  if (numHandlers != 0) // guich@230_14
		  {
		     int j;
			  method->handlers = (WExceptionHandler *)allocClassPart(sizeof(struct WExceptionHandlerStruct)*numHandlers);
			  for (j = 0; j < numHandlers; j++, p += 8)
			  {
				  WExceptionHandler *eh = method->handlers+j;
				  eh->startPc = getUInt16(p);
				  eh->endPc = getUInt16(p+2);
				  eh->handlerPc = getUInt16(p+4);
				  eh->catchType = getUInt16(p+6);
			  }
		  }

        numAttributes = getUInt16(p);
        p += 2;
        while (numAttributes--)
        {
            p += 2;
            p += getUInt32(p) + 4;
        }
    }

    // this may be used in the VmError
    methodDesc = getUtfString(wclass, METH_descIndex(method));
    methodName = getUtfString(wclass, METH_nameIndex(method));
    method->hash = getHash(&methodName,&methodDesc); // guich@200b2
    hasAttributes = (byte)((attrCount0-exceptionsCount) > 0);
    if (method->code.codeAttr == 0 && hasAttributes) // TODO maybe this is a case of classes with abstract methods that implements an interface? check later
    {
       VmError(ERR_BadClass, wclass, &methodName, &methodDesc); // guich@200final_2: a special situation caused by code from Z. Sachen, probably due to a compiler error.
       return NULL;
    }
    if (method->code.codeAttr)
    {
       method->maxStack = (byte)method->code.codeAttr[7]; // (byte)METH_maxStack(method);
       method->maxLocals = (byte)method->code.codeAttr[9]; // (byte)METH_maxLocals(method);
    }
    num = hasAttributes ? (method->maxLocals + method->maxStack + (uint32)2) : 0;
    method->localsStack2 = (byte)num;
    if (num > 255)
       debug("Too much locals and stack!!! please contact guich@superwaba.com.br");
    // determine numParams, isInit and returnedValueCount
    numParams = countMethodParams(methodDesc);
    if (numParams < 0)
        numParams = 0; // error
    method->numParams = (uint16)numParams;
    if (methodName.len > 2 && methodName.str[0] == '<' && methodName.str[1] == 'i')
        method->isInit = 1;
    else
        method->isInit = 0;
    if ((retType=methodDesc.str[methodDesc.len - 1]) != 'V')
        method->returnedValueCount = (retType=='D' || retType=='J') ? 2 : 1; // guich@500_5: store if the returned value is 64bits - 'D'Double or 'J'Long
    else
        method->returnedValueCount = 0;

    // resolve native functions
    if (METH_accessFlags(method) & ACCESS_NATIVE)
    {
       uint32 classHash, methodHash;
       Err err;
       NativeFunc natFunc=NULL;
       uint16 libFound = SWLibRefNum;
       int i;

       method->isNative = 1;
       classHash = LOWER16(genHashCode(&wclass->className));
       methodHash = LOWER16(genHashCode(&methodName) + genHashCode(&methodDesc));
       method->code.hash = (classHash << 16) + methodHash;

       // guich@200b4_92: try to bind to the address of the native in the libs
       // first, in the default lib, then in the other native libraries.
#ifdef PALMOS
       err = SWGetNativeFunc(SWLibRefNum, method->code.hash, &natFunc);
       if (err == SWErrMethodNotFound && vmGlobals->nativeLibs)
       {
          libFound = 0; // guich@241_1: fixed bug if method was not found and the hashcodes was not correctly generated
          for (i = 0; vmGlobals->nativeLibs[i].ref; i++)
             if ((err = SWGetNativeFunc (vmGlobals->nativeLibs[i].ref, method->code.hash, &natFunc)) == 0) // native method handled
             {
                libFound = vmGlobals->nativeLibs[i].ref;
                break;
             }
       }
       if (natFunc != NULL && libFound < 32) // note: libRefNum will usually be between 3 and 10 for PalmOS
       {
          method->code.nativeFunc = natFunc;
          method->libRefNum = (byte)libFound;
       }
#else
       err = SWGetNativeFuncStatic(method->code.hash, &natFunc);
       if (err == SWErrMethodNotFound && vmGlobals->nativeLibs)
       {
          libFound = 0; // guich@241_1: fixed bug if method was not found and the hashcodes was not correctly generated
          for (i = 0; vmGlobals->nativeLibs[i]; i++)
             if ((err = (*vmGlobals->nativeLibs[i]->SWGetNativeFunc)(method->code.hash, &natFunc)) == 0) // native method handled
             {
                libFound = (uint16)(i+1); // must not start from 0!
                break;
             }
       }
       // else dhodges@421_36: commented
       if (err == SWErrMethodNotFound) // In WinCE the standard library is static, so it MUST be present
       {
          char *buf;
          debug(int2str("native implementation not found. Hash",method->code.hash));
          buf = xmalloc(wclass->className.len+methodName.len+2);
          xstrncpy(buf,wclass->className.str,wclass->className.len);
          buf[wclass->className.len] = '.';
          xstrncpy(&buf[wclass->className.len+1],methodName.str,methodName.len);
          buf[wclass->className.len+methodName.len+1] = 0;
          debug(buf);
          xfree(buf);
       }

       if (natFunc != NULL)
       {
          method->code.nativeFunc = natFunc;
          method->libRefNum = (byte)libFound;
       }
#endif
    }
    else
    if (!(METH_accessFlags(method) & ACCESS_ABSTRACT) && method->code.codeAttr == NULL) // not an interface/abstract method and no code available?
       VmError(ERR_BadClass, wclass, &methodName, &methodDesc);

    method->codeStart = METH_code(method);
    if (method->isInit && wclass->numSuperClasses == 1 && numHandlers == 0) // guich@400_16: check if this constructor call Object constructor. If yes, skip this code bc in SW it does nothing (just waste time) - except for Object.<init> itself - guich@401_22: only for classes that does not throws an exception at the constructor
    {
       byte *pc = method->codeStart;
       if (*(pc+1) == OP_invokespecial) // aload_0  -  invokespecial #6 <Method void Object()>
       {
          UtfString callingClassName;
          // check if we're calling Object.init (we may be calling this.init)
          uint16 idx = getUInt16(pc+2); // idx points to the method class index in the constant pool
          idx = CONST_classIndex(wclass,idx); // idx now points to the name index in the constant pool
          idx = CONST_nameIndex(wclass,idx); // idx now points to the String that specifies the class name
          getUtfString2(wclass,idx,&callingClassName);
          if (callingClassName.len == 16 && xstrncmp(callingClassName.str,"java/lang/Object",16) == 0) // ok, we're really calling the Object.init
             method->codeStart += 4; // skip this construction call
       }
    }
    return p;
}

int32 countMethodParams(UtfString desc)
{
    uint32 n;
    char *c;

    c = desc.str;
    if (*c++ != '(')
        return -1;
    n = 0;
    while (1)
    {
        switch (*c)
        {
            case 'B':
            case 'C':
            case 'F':
            case 'I':
            case 'S':
            case 'Z':
                n++;
                c++;
                break;
            case 'D':
            case 'J':
                n+=2; // push 2 consecutive vars
                c++;
                break;
            case 'L':
                c++;
                while (*c++ != ';')
                   ;
                n++;
                break;
				case '[':
				    // mlb@200b3 : fixed support for double and longs
					 n++; //Arrays are always one parameter.
					 while (*c == '[') c++; //Skip past array chars.
					 if (*c == 'L')
						 while (*c++ != ';') ; //Skip past object.
					 else
						 c++; //Skip past primitive type.
					 break; //Now resume counting.
            case ')':
                return n;
            default:
                return -1;
        }
    }
}

// guich@230_14: exception functions

WObject throwException(char *exceptionClass inA0, char *message inA1)
{
	WClass *ec;
	if (thrownException) // guich@500_14: if there was already an exception thrown, don't throw another one.
	   return thrownException;
	ec = getClass(createUtfString(exceptionClass),1);
	if (ec == NULL) return 0;
	if (thrownException == NULL)
	   firstException = NULL; // guich@400_85
	thrownException = createObject(ec);
	if (thrownException && message && pushObject(thrownException) != -1)
	{
		thrownException[1].obj = createStringFromUtf(createUtfString(message));
		popObject();
	}
	//if (ec != outOfMemoryClass) fillStackTrace(ex,0);
	return thrownException;
}

//
// Utility Routines
//

char *int2str(char *msg, int32 ii)
{
   static char buff[70];
   xstrprintf(buff, "%s: %ld", msg, ii); // guich@321_11
   return &buff[0];
}

//
// UtfString Routines
//

UtfString createUtfString(char *buf)
{
   UtfString s;

   s.str = buf;
   s.len = xstrlen(buf);
   return s;
}

UtfString getUtfString(WClass *wclass inA0, uint16 idx inD0)
{
   UtfString s;

   if (idx >= 1 && CONST_tag(wclass, idx) == CONSTANT_Utf8) // no performance difference at all if using this if or not!
   {
      s.str = (char *)CONST_utfStr(wclass, idx);
      s.len = CONST_utfLen(wclass, idx);
   }
   else
   {
      s.str = "";
      s.len = 0;
   }
   return s;
}

static void getUtfString2(WClass *wclass inA1, uint16 idx inD0, UtfString *s inA0) // Don't inline
{
   s->str = (char *)CONST_utfStr(wclass, idx);
   s->len = CONST_utfLen(wclass, idx);
}

//
// Object Routines
//

WObject createStringObject()
{
   return createObject(stringClass);
}

_inline_ WObject createObject(WClass *wclass)
{
   WObject obj=NULL;

   if (wclass != NULL && !WCLASS_isAbstract(wclass)) // can't be array, interface or abstract class
   {
      obj = allocObject(WCLASS_objectSize(wclass));
      if (obj != NULL)
         WOBJ_class(obj) = wclass;
   }
   return obj;
}

byte typesSizeShl[12] = {0,2,2,0,0,1,2,3,0,1,2,3}; // guich@200b4r5: code optimized
_inline_ int32 arraySize(int32 type, int32 len)
{
   // align to 4 byte boundry
#ifdef ARRAY_BIG_BUT_FAST
   return (3 * sizeof(Var) + (((len << typesSizeShl[type])) + 3) & ~3); // guich@400_15: now an array needs only 2 more Vars instead of 3
#else
   return (2 * sizeof(Var) + (((len << typesSizeShl[type])) + 3) & ~3); // guich@400_15: now an array needs only 2 more Vars instead of 3
#endif
}

_inline_ WObject createArrayObject(int32 type inD0, int32 len inD1)
{
   WObject obj=0;

   obj = allocObject(arraySize(type, len));
   if (len >= 0 && obj != 0)
   {
      // pointer to class is NULL for arrays
      //WOBJ_class(obj) = NULL; - already zeroed
      WOBJ_arrayType(obj) = type;
      WOBJ_arrayLen(obj) = len;
   }
   return obj;
}

uint16 arrayType(char c)
{
   static byte arrayTypes[] = {0,8,5,7,0,6,0,0,10,11,0,1,0,0,0,0,0,0,9,0,0,0,0,0,0,4};
   return (c != '[') ? arrayTypes[c-'A'] : 2;
/*   switch(c) - very inefficient switch
   {
      case '[': return 2;  // array
      case 'B': return 8;  // byte
      case 'C': return 5;  // char
      case 'D': return 7;  // double
      case 'F': return 6;  // float
      case 'I': return 10; // int
      case 'J': return 11; // long
      case 'L': return 1;  // object
      case 'S': return 9;  // short
      case 'Z': return 4;  // boolean
   }*/
}

WObject createMultiArray(int32 ndim, char *desc, Var *sizes)
{
    WObject arrayObj, subArray, *itemStart;
    int32 len, type;

    len = sizes[0].intValue;
    type = arrayType(desc[0]);
    arrayObj = createArrayObject(type, len);
    if (len < 0 || !arrayObj)
        return 0;
    if (type != 2 || ndim <= 1)
        return arrayObj;
    // NOTE: it is acceptable to push only the "upper"
    // array objects and not the most derived subarrays
    // because if the array is only half filled and we gc,
    // the portion that is filled will still be found since
    // its container was pushed
    if (pushObject(arrayObj) == -1)
        return 0;
    // create subarray (recursive)
    itemStart = (WObject *)WOBJ_arrayStart(arrayObj); // guich@400_16: no more need to recompute the offset
    while (len--)
    {
        subArray = createMultiArray(ndim - 1, desc + 1, sizes + 1);
        *itemStart++ = subArray;
    }
    popObject();
    return arrayObj;
}

//
// Method Routines
//

int16 getMethodMapNum(WClass *wclass, UtfString *name, UtfString *desc, byte searchType, WClassMethod **imethod, WClass **iclass)
{
   uint16 start, count, i;
   uint32 hash;

   WClassMethod *method;

   if (searchType == CLASS_SEARCH_ALL)
   {
      start = 0;
      count = wclass->vMethodMap.mapSize + wclass->numMethods;
   }
   else if (searchType == CLASS_SEARCH_INHERITED)
   {
      start = 0;
      count = wclass->vMethodMap.mapSize;
   }
   else // CLASS_SEARCH_THISCLASS
   {
      start = wclass->vMethodMap.mapSize;
      count = wclass->numMethods;
   }
   hash = getHash(name,desc);
   for (i = start; count--; i++)
   {
  	   method = getMethodByMapNum(wclass, iclass, i);
      if (method->hash == hash)
      {
         if (imethod)
            *imethod = method;
         return i;
      }
   }
   return -1;
}

WClassMethod *getMethodByMapNum(WClass *wclass inA0, WClass **vclass inA1, uint16 mapNum inD0)
{
   VMapValue *mapValue;
   int16 methodIndex;
   WClassMethod *method;

   methodIndex = mapNum - wclass->vMethodMap.mapSize;
   if (methodIndex < 0)
   {
      // inherited or overridden method
      mapValue = &wclass->vMethodMap.mapValues[mapNum];
      wclass = mapValue->c;
      method = mapValue->m;
   }
   else method = &wclass->methods[methodIndex];

   if (vclass != NULL)
      *vclass = wclass;
   return method;
}

WClassMethod *getMethod(WClass *wclass, UtfString name, UtfString desc, WClass **vclass)
{
    WClassMethod *method;
    if (getMethodMapNum(wclass, &name, &desc, (byte)((vclass != NULL)?CLASS_SEARCH_ALL:CLASS_SEARCH_THISCLASS), &method, vclass) < 0)
       method = NULL;
    return method;
}

// return 1 if two classes are compatible (if wclass is compatible
// with target). this function is not valid for checking to see if
// two arrays are compatible (see compatibleArray()).
// see page 135 of the book by Meyers and Downing for the basic algorithm
int compatible(WClass *source inA0, WClass *target inA1)
{
    int targetIsInterface=0;
    uint32 i, n,nn;

    if (source && target)  // source or target is array?
    {
       if (source == target) return 1; // guich@402_31: quick test
       if (WCLASS_isInterface(target))
           targetIsInterface = 1;
       n = source->numSuperClasses;
       while (1)
       {
           if (targetIsInterface)
           {
               nn = WCLASS_numInterfaces(source);
               i = 0;
               while (nn--)
               {
                   uint16 classIndex;
                   WClass *interfaceClass;

                   classIndex = WCLASS_interfaceIndex(source, i);
                   if ((interfaceClass = (WClass *)CONST_binded(source, classIndex)) == NULL)
                      interfaceClass = getClassByIndex(source, classIndex,0);
                   i++; // guich@400_36: DON'T REMOVE THIS FROM HERE!

                   // NOTE: Either one of the interfaces in the source class can
                   // equal the target interface or one of the interfaces
                   // in the target interface (class) can equal one of the
                   // interfaces in the source class for the two to be compatible
                   if (interfaceClass == target || compatible(interfaceClass, target))
                      return 1;
               }
           }
           else
           if (source == target)
              return 1;
           if (n == 0)
              break;
           // look in superclass
           source = source->superClasses[--n];
       }
    }
    return 0;
}

_inline_ int compatibleArray(WObject obj, UtfString arrayName)
{
   // NOTE: this isn't a full check to see if the arrays
   // are the same type. Any two arrays of objects (or other
   // arrays since they are objects) will test equal here.
   return WOBJ_class(obj) == NULL && WOBJ_arrayType(obj) == arrayType(arrayName.str[1]);
}

//
// Memory Management
//

// Rick Wild notes for Waba 1.0:
// Here's the garbage collector. I implemented the mark and sweep below
// after testing out a few different ones and reading:
//
// Garbage Collection, Algorithms for Automatic Dynamic Memory Management
// by Richard Jones and Rafael Lins
//
// which is an excellent book. Also, this collector has gone through a
// lot of testing. It runs when the system is completely out of memory
// which can happen at any time... for example during class loading.
//

// Guilherme C. Hazan notes for SuperWaba 2.0 beta 4 release 9:
// The GC was completely changed, to allow more than 64kb to be allocated
// for Objects.

// Guilherme C. Hazan notes for SuperWaba 4.0:
// The GC was again enhanced in 3 points:
// 1. In the old GC, there was a chance that a number be confused with an
// object, thus resulting in crash or in memory leaks. This is now fixed
// 2. Now the extra information needed for an Object shrinked from 16 bytes
// to 8 bytes. This means that more objects can be allocated per chunk, and
// also means that your program will need less memory to run.
// 3. Unused chunks are now freed. When the objectsUsed counter reaches 0,
// its memory is returned to the OS.

// Class A
// {
//    int a;
//    float b;
//    int []ii;
//    Object c;
//
// Is stored in mem as
// Var[-1] = ObjectInfo			     } Extra
// Var[0]  = ptr to class A		  } Information
// Var[1]  = int a
// Var[2]  = float b
// Var[3]  = int array ii
// Var[4]  = Object c
//
// Note that 16 bit and 8 bit primitive types are stored as 32 bit.

// Terminology: a Chunk is a memory block that contains objects inside it.

// The memory allocated for class A in Object Memory will be 5*4 = 20 bytes.
// The ObjectInfo is used during GC. It contains the offset from the Object
// to the start of the Chunk. During the MARK stage, it may be replaced by
// a temporary stack where the objects are marked.

// Object's size can be calculated by: 8+4*n, where n, >= 0, is the number
// of fields.

// Next are the details of how the GC works:

// . First, a big chunk (DEFAULT_CHUNK_SIZE = 16L*1024L-128L) is created. It
// is less 128 to make possible PalmOS allocate 4 big chunks in a same
// 64k segment. Chunks stores Objects in sequence inside it.

// . Each chunk contains a bitmap of addresses which is used to determine
// a valid object. IE, a valid object has a bit set in the corresponding
// address of the bitmap. The offset Object field (Var[-1]) points to the
// address of each chunk.

// . The chunks are stored in an array. Array index 0
// stores chunks with 8 bytes, index 1 store chunks with 16 bytes,
// and so on. The big chunk is placed at FREE_CHUNK_MAX_INDEX.

// . Each array entry is a linked list of free chunks. The pointer is placed
// in Var[0] since, in a freed object, that entry is not being used.

// . in the allocObject routine,
//   . it first try to find a free chunk at the index computed based in the
//     size of the chunk being allocated (preferred index). If there are no
//     free chunks of this size, it tries to allocate it at 32.
//   . before GC is called, we verify how often it is being called; if
//     necessary, we allocate one more chunk. This timed-allocation will only
//     be alowed again after a chunk is freed.
//   . If this also fails, it runs GC and try again in the (preferred index)
//     and in index FREE_CHUNK_MAX_INDEX.
//   . If it fails again, it compacts the memory (by joining consecutive
//     chunks) and tries again.
//   . If it fails once more, it allocs one more chunk (big
//     enough to hold the requested size) and try again.
//   . If all fail, it throws an OutOfMemoryError.

// Otherwise, if success, it sets the Object's address bit in the bitmap of
// allocated objects. If a chunk located at index FREE_CHUNKS_MAX_INDEX was
// used, shrink its size and "promote" it to another index if necessary

// . The GC is implemented as a Mark and Sweep gc. In the mark stage, all
// objects in the two stacks and all static class members are marked "in use".
// In the sweep stage, it checks every object in the Object bitmap to
// see if they are marked; if not marked, it is freed by moving to the free
// chunks array with a computed index based in his size.

// . There's a crucial requirement for this GC to work: an allocated chunk
// of memory CANNOT be located in an address below 64kb. If this happens,
// the GC will fail. Also, no object can have more than 64kb of size, even
// if the OS allows it.

// Thanks to Peter Dickerson, who gave me some good insights from his
// experience with Jump's GC.

//// Parameters that can affect the GC behaviour ////
#define DEFAULT_CHUNK_SIZE (16L*1024L-128L) // note: changing this to 32k is NOT a good idea
#define FREE_CHUNK_MAX_INDEX 8 // The freeChunks is an array that stores the chunks of memory available to be allocated. total allocated = FREE_CHUNK_MAX_INDEX+1
#define TIMER_TIMED_CHUNK 2000 // the minimum timming where a new chunk will be allocated bc too much memory is being requested at once (default was 500)

////                                             ////

// a chunk has at least 8 bytes
#define CHUNK_INFO_SIZE (1 * sizeof(WObject))
#define size2freeChunksIndex(s) ( ((s) > (FREE_CHUNK_MAX_INDEX*4+CHUNK_INFO_SIZE)) ? FREE_CHUNK_MAX_INDEX : (((s)-CHUNK_INFO_SIZE-1)/4) ) // changed by FrankDiebolt@320
WObject freeChunks[FREE_CHUNK_MAX_INDEX+1];
static uint32 firstValidChunk = 0xFFFFFFFF;
static uint32 lastValidChunk;
static uint32 firstValidObj,lastValidObj;
static byte eightBit[] = {128,64,32,16,8,4,2,1};

#define NUMBER_OF_OBJECTS_IN_CHUNK(size) ((size)>>2) // multiple of 4

// this structure represents a Chunk
typedef struct
{
   byte *bitmap;       // the bitmap that indicates if a memory position inside this chunk is or not a valid object
   int32 usedObjects;  // number of objects used in this chunk. If == 0, this chunk is completely free
   uint16 size;        // chunk's size
   uint16 bitmapSize;  // maximum index for the bitmap table
   uint32 limit;        // ptr + size
   // objects[]        // the objects
} Chunk, *ChunkPtr;

#define CHUNK_INUSE(c) (c->size > 0)

// This is the information that is stored before each object
typedef struct // NOPT
{
   uint16 freedSize; // for a freed object, it will store the size of the chunk. for an object in use, it will be null
   uint16 offset;    // now an Object can't have more than 64kb, because he will mean "marked" if this info gets anything above 64kb.
} ObjectInfo;

// The freeSize/infoOffset contents will determine the Object state:
// ssoo: size,offset = object collected (in Free chunks list), isValid return false
// 0000: freedNext   = object marked, not yet collected, isValid return false
// 00oo: offset only = object not marked, isValid return true

#define WOBJ_infoValue(o)  (o-1)->intValue
#define WOBJ_info(o)       ((ObjectInfo *)(o-1))
#define WOBJ_infoOffset(o) WOBJ_info(o)->offset

// fields used when object is free
#define WOBJ_freedNext(o)   o[0].refValue // used when chaining objects in the freeChunks array. points to next allocated WObject in the free list - reuses the WClass place
#define WOBJ_freedSize(o)   WOBJ_info(o)->freedSize // free objects stores the chunk's size in the mark field

// fields used in the mark stage
#define WOBJ_nextMarked(o) ((o)[-1]).obj // used in the mark stage. used to store the object scanned. It replaces the WOBJ_info by an address that is surely greater than 64kb. The offset will be further reintroduced here

#define WOBJ_MARKED(o)        (WOBJ_infoValue(o) == 0) // 0000
#define WOBJ_FREED(o)         (WOBJ_freedSize(o) != 0)
#define WOBJ_IS_INFO_VALID(o) (WOBJ_freedSize(o) == 0 && WOBJ_infoOffset(o) != 0)

// suppose the chunk consists of:
// 16 bytes of header
// 32000 bytes of object area
// The first WObject allocated there will have offset 16

// in the defines below: c = ChunkPtr, o = WObject, i = index
#define OBJS_OFFSET (CHUNK_INFO_SIZE + sizeof(Chunk))
#define OBJ2CHUNK(o) ((ChunkPtr)((byte*)o - WOBJ_infoOffset(o))) // given an object, return the pointer to its chunk
#ifdef PALMOS
 #define USEDOBJ2CHUNK(o) ((ChunkPtr)((byte*)o - WOBJ_infoValue(o))) // valid only for objects in use. Since freedSize is unused, we can safely access the whole 32bit address as being the chunk offset
#else
 #define USEDOBJ2CHUNK OBJ2CHUNK // Windows doesnt like this optimization
#endif
#define OBJ2INDEX(c,o) (((uint32)o - OBJS_OFFSET - (uint32)c)>>2) // given a chunk and an object, returns the index of the object in the bitmap
#define NEXTOBJ2INDEX(c,o,size) (((uint32)o - OBJS_OFFSET + size - (uint32)c)>>2) // given a chunk and an object, returns the index of the object in the bitmap
#define INDEX2OFFSET(i) (uint16)(OBJS_OFFSET + ((i)<<2))
#define INDEX2OBJ(c,i) (WObject)((byte*)c + INDEX2OFFSET(i))// given a bit index in the bitmap table, return the address of the WObject
#define _BIT_INDEX(i) _eightBit[i & 7]
#define BIT_INDEX(i) eightBit[i & 7]
#define BYTE_INDEX(bitmap,i) bitmap[(i)>>3]
#define IS_OBJ_ACTIVE(c,obj) ((BYTE_INDEX(c->bitmap,OBJ2INDEX(c,obj)) & _BIT_INDEX(OBJ2INDEX(c,obj))) != 0)

#ifdef ARM
   #define ADDR_MASK 3 // arm must access doubleword values in doubleword address
#else
   #define ADDR_MASK 1
#endif

// used in memory stats
int32 calledGC;	 // number of times the GC was called
uint32 elapsedGC;  // total time used in the GC
uint32 totalAlloc; // currently total bytes allocated
uint32 maxAlloc;	 // max bytes allocated (chunks may be returned to the OS)
uint32 totalUsed;	 // how many bytes, from the total allocated, are being used
uint32 elapsedCompact; // total time used in the compact
uint32 calledCompact;  // number of times called compact
uint32 elapsedDelete; // total time used in the deleteChunk
uint32 chunksCreated; // total chunks created
uint32 chunksDeleted; // total chunks deleted
uint32 timedChunksCreated; // total timed chunks created
uint32 lastCalled; // last time a new Chunk was requested
uint32 compactWorked; // times that an object was allocated after a compact operation
uint32 deleteChunksCount; // times chunks were deleted
uint32 emptyChunks; // number of currently empty chunks. computed in the sweep stage

// Ordered ChunkPtr array manipulation
ChunkPtr *chunksArray;
int chunksCount;
int chunksArrayMax;

static ChunkPtr chunksArrayFindNearest(WObject p);
static void chunksArrayRemove(ChunkPtr p);
static int chunksArrayAdd(ChunkPtr p);
static int chunksArrayFindPos(ChunkPtr p);
static void checkChunks();

/** a valid object...
. must be between allocated ranges
. must reside in a address % 4 (or % 2)
. its chunk exists in the chunks tree. We verify if it belongs to the first chunk, since this occurs often
. the info block must not be marked nor freed - note: putting this before the chunk test leads to GPF under win32 in some circunstances
. finally, the chunk's bitmap for this object is active
*/
#define WOBJ_IS_VALID(o) ((uint32)o >= firstValidObj && (uint32)o < lastValidObj && \
                         ((uint32)o & ADDR_MASK) == 0 && \
                         (probableChunk=(((uint32)o <= ((ChunkPtr)firstValidChunk)->limit)? (ChunkPtr)firstValidChunk : chunksArrayFindNearest(o))) != NULL && \
                         WOBJ_IS_INFO_VALID(o) && \
                         USEDOBJ2CHUNK(o) == probableChunk && IS_OBJ_ACTIVE(probableChunk, o)) // 100% reliable

/*int32 inside1,inside2,inside3,inside4,inside5,inside6;
_inline_ static byte WOBJ_IS_VALID(WObject o)
{
   byte *_eightBit = eightBit;
   ChunkPtr probableChunk;
   inside1++; // 1620598 - number of "objects" that gets here
   if ((uint32)o >= firstValidObj && (uint32)o < lastValidObj && ((uint32)o & ADDR_MASK) == 0)
   {
      inside3++; // 170769
      if ((uint32)o <= ((ChunkPtr)firstValidChunk)->limit)
         probableChunk = (ChunkPtr)firstValidChunk;
      else
         probableChunk = chunksArrayFindNearest(o);
      if (probableChunk != NULL)
      {
         inside2++; // 459693
         if (WOBJ_IS_INFO_VALID(o)) // not freed or collected?
         {
            inside4++; // 170769
            if (probableChunk == USEDOBJ2CHUNK(o)) // does the chunks match?
            {
               inside5++; // 170745
               if (IS_OBJ_ACTIVE(probableChunk, o))
               {
                  inside6++; // 170745 - from here you see that this last test is not completely necessary
                  return 1;
               }
            }
         }
      }
   }
   return 0;
} */


static int chunksArrayAdd(ChunkPtr p)
{
   if (chunksCount == chunksArrayMax) // expand array if necessary
   {
      ChunkPtr *temp;
      uint32 newSize;
      chunksArrayMax += 50;
      newSize = chunksArrayMax*sizeof(ChunkPtr);
      temp = vmGlobals->xrealloc(chunksArray, &newSize);
      if (!temp)
         return 0;
      chunksArray = temp;
   }
   if (chunksCount == 0) // empty?
      chunksArray[0] = p;
   else
   {
      int pos = chunksArrayFindPos(p); // before reallocing, we must find the insert pos
      if (pos < 0)
         pos = 0;
      if (pos < chunksCount)
         xmemmove((byte*)&chunksArray[pos+1],(byte*)&chunksArray[pos],(chunksCount-pos)<<2);
      chunksArray[pos] = p;
   }
   chunksCount++;
   //checkChunks();
   return 1;
}

// Initialize a object heap. It is called many times to allocate a new chunk
// note: chunkSize *must* be above DEFAULT_CHUNK_SIZE. This parameter must be used only
// if the size required is ABOVE DEFAULT_CHUNK_SIZE.
static int createChunk(uint32 chunkSize) // guich@400_14: initObjectHeap is now renamed to createChunk
{
   WObject obj;
   ChunkPtr chunk;
   uint32 nObj;
   byte singleObj = (byte)(chunkSize != 0);

   chunksCreated++;
   if (singleObj)
      chunkSize += OBJS_OFFSET;
   else
      chunkSize = DEFAULT_CHUNK_SIZE;

   // 1. allocate memory for the big chunk
   chunk = (Chunk *)vmGlobals->xmalloc(chunkSize);
   if (chunk == NULL) return 0;
   xzero32((byte*)chunk, chunkSize >> 2);

   // 2. setup internal chunk fields
   chunk->size = (uint16)chunkSize;
   chunk->limit = (int32)chunk+chunkSize;

   nObj = NUMBER_OF_OBJECTS_IN_CHUNK(chunkSize-OBJS_OFFSET); // there must be space for an extra bit in the bitmaps table. and the bitmap table must be 4 byte aligned.
   // now that we computed how many bytes the bitmap table has, what is left
   chunk->bitmapSize = (uint16)(((nObj>>3) & ~3) + 8); // guich@402_12: adds 8 bytes instead of 4.
   chunk->bitmap = vmGlobals->xmalloc(chunk->bitmapSize);
   if (chunk->bitmap == NULL)
   {
      _xfree(chunk);
      return 0;
   }

   // 3. insert the chunk in the chunks tree.
   if (!chunksArrayAdd(chunk)) // guich@402_13: exchanged 2 with 3, because a gc can be called by the malloc routine, what would lead to free this chunk before using it
   {
      _xfree(chunk);
      _xfree(chunk->bitmap);
      chunk->bitmap = 0;
      return 0;
   }

   if (chunk->bitmapSize == 0)
      chunk->bitmapSize = (uint16)(((nObj>>3) & ~3) + 8); // guich@402_12: adds 8 bytes instead of 4.
   xzero32((byte*)chunk->bitmap, chunk->bitmapSize>>2);

   // 4. convert the chunk to object
   obj = (WObject)((byte*)chunk + OBJS_OFFSET); // chunk -> WObject
   WOBJ_infoOffset(obj) = OBJS_OFFSET;
   WOBJ_freedSize(obj) = (uint16)(chunkSize - sizeof(Chunk));

   BYTE_INDEX(chunk->bitmap,0) |= BIT_INDEX(0); // activate first bit

   // 5. update the global variables
   if ((uint32)chunk > lastValidChunk)
   {
      lastValidChunk = (uint32)chunk;
      lastValidObj   = lastValidChunk+chunkSize;
   }
   if ((uint32)chunk < firstValidChunk)
   {
      firstValidChunk = (uint32)chunk;
      firstValidObj   = firstValidChunk + sizeof(Chunk);
   }
   totalAlloc += chunkSize/*+chunk->bitmapSize*/;
   if (totalAlloc > maxAlloc) maxAlloc = totalAlloc;

   // 6. add the chunk to the appropriate freeChunk index
   WOBJ_freedNext(obj) = freeChunks[FREE_CHUNK_MAX_INDEX];
   freeChunks[FREE_CHUNK_MAX_INDEX] = obj;

   return 1;
}

static void freeChunk(ChunkPtr chunk);
void FreeObjectHeap()
{
   int i = chunksCount;
   while (--i >= 0)
      freeChunk(chunksArray[i]);
   _xfree(chunksArray);
}

static void freeChunk(ChunkPtr chunk)
{
   if (chunk->usedObjects > 0)
   {
      WClass *wclass;
      WObject obj;
      // Search for all active objects in the bitmap table
      byte *bmp = chunk->bitmap;
      int16 max = chunk->bitmapSize;
      int16 index=0,byteIndex=0,bitIndex=1;

      // goto's sometimes are a very efficient statement
findMoreBytes:
      // find the next byte with any bit set
      while (max > 0 && *bmp == 0)
      {
         max--;
         byteIndex++;
         bmp++;
      }
      if (max > 0) // found?
      {
         int16 b = *bmp;
findMoreBits:
         // find, inside this byte, the next bit set
         while (bitIndex < 8 && (b & eightBit[bitIndex]) == 0)
            bitIndex++;
         // if no more bits in this byte, test next bytes
         if (bitIndex == 8)
         {
            bitIndex = 0;
            max--;
            byteIndex++;
            bmp++;
            goto findMoreBytes;
         }
         // found another bit
         obj = INDEX2OBJ(chunk,index);
         if (!WOBJ_FREED(obj))
         {
            if (WOBJ_IS_INFO_VALID(obj) && (wclass = WOBJ_class(obj)) != NULL && WCLASS_hasDestroyFunc(wclass))
#ifdef PALMOS
               SWCallPublic(wclass->objDestroyFuncLibRefnum, wclass->objDestroyFuncNum, NULL, (uint32 *)obj);
#else
               if (wclass->objDestroyFuncLibRefnum == SWLibRefNum)
                  SWCallPublicStatic(wclass->objDestroyFuncNum, NULL, (uint32 *)obj);
               else
               if (vmGlobals->nativeLibs != NULL)
                  (*vmGlobals->nativeLibs[wclass->objDestroyFuncLibRefnum-1]->SWCallPublic)(wclass->objDestroyFuncNum, NULL, (uint32 *)obj);
#endif
         }
         index = bitIndex+(byteIndex<<3);
         bitIndex++;
         goto findMoreBits;
      }
      // no more bits found
   }
   _xfree(chunk->bitmap);
   xzero32((byte*)chunk,chunk->size>>2);
   _xfree(chunk);
}

static int initObjectHeap() // guich@400_14: now this method is used to initialize some things
{
   return GCCreated = createChunk(0);
}

static void removeObjectsFromUnusedChunks()
{
   int index;
   WObject lastObj,obj;
   for (index=0; index <= FREE_CHUNK_MAX_INDEX; index++)
      if ((obj=freeChunks[index]) != NULL)
      {
         lastObj = NULL;
         while (obj != NULL)
         {
            if (OBJ2CHUNK(obj)->usedObjects == 0)
            {
               // skip all adjacent free objects of this chunk
               do
               {
                  obj = WOBJ_freedNext(obj);
               } while (obj && OBJ2CHUNK(obj)->usedObjects == 0);

               if (lastObj == NULL)
                  freeChunks[index] = obj;
               else
                  WOBJ_freedNext(lastObj) = obj;
            }
            lastObj = obj;
            if (obj != NULL)
               obj = WOBJ_freedNext(obj);
         }
      }
}

static void deleteUnsusedChunks()
{
   int i = chunksCount;
   while (--i >= 0)
      if (chunksArray[i]->usedObjects == 0)
      {
         ChunkPtr chunk = chunksArray[i],temp;
         chunksDeleted++;
         totalAlloc -= chunk->size;
         // free the chunk
         //xzero32(chunk->bitmap,chunk->bitmapSize>>2); - not necessary
         _xfree(chunk->bitmap); // must be after the chunk is zeroed
         xzero32((byte*)chunk,chunk->size>>2); // must be last
         temp = chunk;
         _xfree(chunk); // guich@402_14
         chunksArrayRemove(temp);
      }
}

static void deleteChunks()
{
   uint32 elapsed = -getTimeStamp(); // get the dif
   deleteChunksCount++;

   AVOID_AUTO_RESET_NOW
   removeObjectsFromUnusedChunks();
   deleteUnsusedChunks();
   // recompute first/lastValicChunk
   firstValidChunk = (uint32)chunksArray[0];
   firstValidObj   = firstValidChunk + sizeof(Chunk);
   lastValidChunk = (uint32)chunksArray[chunksCount-1];
   lastValidObj   = lastValidChunk + ((ChunkPtr)lastValidChunk)->size;

   elapsed += getTimeStamp();
   elapsedDelete += elapsed;
}

static void debugMemory()
{
   debug("Memory Statistics for");
   debug(vmGlobals->className);
   if (calledGC == 0)
      debug("GC was never called");
   else
   {
      debug(int2str("Times GC was called",calledGC));
      debug(int2str("Total GC time",elapsedGC));
      debug(int2str("Times Compact was called",calledCompact));
      debug(int2str("Total Compact time",elapsedCompact));
      debug(int2str("Times Compact succeed",compactWorked));
      debug(int2str("Times Delete was called",deleteChunksCount));
      debug(int2str("Total Delete time",elapsedDelete));
   }
   debug(int2str("Total Obj Allocated",totalAlloc));
   debug(int2str("Max Obj Allocated",maxAlloc)); // guich@400_14
   debug(int2str("Currently used",totalUsed));
   debug(int2str("Total Class mem",(totalClassHeap + classHeapUsed))); // guich@341_5
   debug(int2str("Chunks created",chunksCreated));
   debug(int2str("Timed chunks created",timedChunksCreated));
   debug(int2str("Chunks deleted",chunksDeleted));
   debug("Calling GC...");
   gc_();
   debug(int2str("Total Obj Allocated",totalAlloc));
   debug(int2str("Currently used",totalUsed));
   debug(int2str("Chunks deleted",chunksDeleted));
   tone(1300,10);
}

static WObject allocInFreeChunk(uint32 size)
{
   WObject last=NULL;
   ChunkPtr chunk;
   WObject obj;
   uint32 oldSize;
   uint16 idx;
   int index;

   // 1. given the size, compute the index to the array of sweeped objects
   index = size2freeChunksIndex(size);

   // 2. allocate at that index a chunk of the same size, if available
   if (index < FREE_CHUNK_MAX_INDEX && (obj=freeChunks[index]) != NULL)
   {
      chunk = OBJ2CHUNK(obj);
      // 2.1. yes. "activate" it
      idx = (uint16)(NEXTOBJ2INDEX(chunk,obj,size));
      BYTE_INDEX(chunk->bitmap,idx) |= BIT_INDEX(idx); // set where this object ends
      // 2.2. remove it from the list of free chunks
      freeChunks[index] = WOBJ_freedNext(obj);
      WOBJ_freedNext(obj) = 0; // reset the pointer (used to store WClass) - must be zeroed only after used in the line above.
      chunk->usedObjects++;
      // 2.3. "allocate" it
      WOBJ_freedSize(obj) = 0; // reset the size bc it will be used for mark further on
      totalUsed += size;
      return obj;
   }
   // 3. chunk could not be allocated from the sweeped chunks. Try at the big chunk now
   index = FREE_CHUNK_MAX_INDEX;
   for (obj = freeChunks[index]; obj != NULL; last = obj, obj = WOBJ_freedNext(obj))
      if ( (oldSize=WOBJ_freedSize(obj)) >= size) // can this chunk hold the new object?
      {
         // 3.1. yes. "allocate" it
         WObject found = obj;
         if ((oldSize - size) < 8)
            size = oldSize; // note: suppose the big 32k was eaten until there was only 188 bytes left. Then, a 184 bytes chunk is allocated. The 4 bytes remaining will not fit to any other chunk. So, if this happen, allocate the rest of the big chunk so it can be fully reclaimed later.
         chunk = OBJ2CHUNK(found);
         chunk->usedObjects++;

         idx = (uint16)(NEXTOBJ2INDEX(chunk,found,size));
         BYTE_INDEX(chunk->bitmap,idx) |= BIT_INDEX(idx);

         WOBJ_freedSize(found) = 0; // clear the info upper bits

         // 3.2. remove it from the list of free chunks
         if (last)
            WOBJ_freedNext(last) = WOBJ_freedNext(found);
         else
            freeChunks[index] = WOBJ_freedNext(found);
         WOBJ_freedNext(found) = 0;

         // 3.3. does the chunk still have space?
         if (size != oldSize)
         {
            // 3.3.1. prepare the chunk with the space left from the creation of the object
            obj = (WObject)((byte *)found + size);
            // 3.3.2. compute the object's new size
            WOBJ_infoOffset(obj) = (uint16)(WOBJ_infoOffset(found)+size);
            oldSize -= size;
            WOBJ_freedSize(obj) = (uint16)oldSize;
            // 3.3.3. the index may change if the size get smaller than the current free chunk index can hold.
            index = size2freeChunksIndex(oldSize);
            // 3.3.4. add it to the new freeChunks list
            WOBJ_freedNext(obj) = freeChunks[index];
            freeChunks[index] = obj;
         }
         totalUsed += size;
         return found;
      }
   return NULL;
}

// NOTE: size passed must be 4 byte aligned (see arraySize())
WObject allocObject(int32 size)
{
   WObject obj = NULL;
   uint32 dif,now;

   // 1. check parameters
   if (size <= 0)
      return NULL;

   if (size > 65514) // trying to allocate something greater than the maximum heap size? - guich@400_14: reintroduced the limit for all platforms - guich@504_10: changed 65500 to 65512 (an array has 12 more bytes) - guich@550: changed to 65514
   {
      char buf[70];
      xstrprintf(buf, "Can't allocate %ld bytes, limit is 65500", size);
      VmQuickError(ERR_OutOfObjectMem, buf);
      return NULL;
   }

   // 2. compute the required size
   size += CHUNK_INFO_SIZE;

   // 3. try to allocate it
   if ((obj = allocInFreeChunk(size)) != NULL)
      return obj;
   // 4. if fail, verify if we're calling GC too much. If yes, allocate one more chunk so we can reduce GC calls.
   now = getTimeStamp();
   dif = now-lastCalled;
   lastCalled = now;
   if (0 <= dif && dif < TIMER_TIMED_CHUNK) // if calling too much, allocate one more chunk
   {
      if (vmGlobals->audibleGC) tone(1000,10);
      timedChunksCreated++;
      // this code is duplicated here bc if we fail we don't want to abort the VM, since we can simply call GC and continue
      if (createChunk( (size>DEFAULT_CHUNK_SIZE) ? size : 0)
          && (obj = allocInFreeChunk(size)) != NULL)
         return obj;
   }

   // 4. fail? call GC and try again
   gc();
   if ((obj = allocInFreeChunk(size)) != NULL)
   {
      if (emptyChunks > 4) // if we had allocated too much memory, release it
         deleteChunks();
      return obj;
   }

   // 5. if gc failed, compact the memory and try again
 /*guich@520_5: removed from here   if (compact() && (obj = allocInFreeChunk(size)) != NULL)
   {
      compactWorked++;
      return obj;
   }*/

   // 6. still fail? Delete the possible chunks before creating one more chunk.
   if (emptyChunks > 0)
      deleteChunks();

   // 7. allocate one more chunk
   if (createChunk((size>DEFAULT_CHUNK_SIZE) ? size : 0)
       && (obj = allocInFreeChunk(size)) != NULL)
       return obj;
   // 8. all failed?
   VmQuickError(ERR_OutOfObjectMem, "GC, Compact and deleteChunks failed.");
   return NULL;
}

// mark this object and all the objects this object refers to and all
// objects those objects refer to, etc.
static void markObject(WObject obj)
{
   WClass *wclass;
   WObject *arrayStart, o;
   int32 len;
   byte *_eightBit = eightBit;
   ChunkPtr probableChunk;

   // invalid or null objects never gets here (here, before markInterior)
   WOBJ_nextMarked(obj) = NULL; // this destroys the offset, which will be recomputed in the sweep stage

markInterior:
   wclass = WOBJ_class(obj);
   if (wclass != NULL)
   {
      // object
      len = wclass->numVars;
      while (--len >= 0)
      {
         o = WOBJ_var(obj, len).obj;
         if (WOBJ_IS_VALID(o)) // marked objects won't be valid
         {
            WOBJ_nextMarked(o) = WOBJ_nextMarked(obj);
            WOBJ_nextMarked(obj) = o;
         }
      }
   }
   else
   {
      // array - see if it contains object references
      if (WOBJ_arrayType(obj) <= 2) //type == 1 || type == 2) - can't be 0!
      {
         // for an array of arrays or object array
         arrayStart = (WObject *)WOBJ_arrayStart(obj);
         len = WOBJ_arrayLen(obj);
         while (--len >= 0)
         {
            o = *arrayStart++;
            if (WOBJ_IS_VALID(o))
            {
               WOBJ_nextMarked(o) = WOBJ_nextMarked(obj);
               WOBJ_nextMarked(obj) = o;
            }
         }
      }
   }
   if ((o=WOBJ_nextMarked(obj)) != NULL)
   {
      // Note: we use goto since we want to avoid recursion here
      // since structures like linked links could create deep
      // stack calls
      WOBJ_nextMarked(obj) = NULL;
      obj = o; // pop last object and scan it
      goto markInterior;
   }
}


// NOTE: There are no waba methods that are called when objects are destroyed.
// This is because if a method was called, the object would be on its way to
// being GC'd and if we set another object (or static field) to reference it,
// after the GC, the reference would be stale.
void sweep()
{
   int i = chunksCount;
   ChunkPtr chunk;
   emptyChunks = 0;

	while (--i >= 0)
   {
	   if ((chunk=chunksArray[i])->usedObjects > 0)
      {
         WClass *wclass;
         WObject obj;
         // Search for all active objects in the bitmap table
         byte *bmp = chunk->bitmap;
         int16 max = chunk->bitmapSize;
         int16 freeIndex,index=0,nextIndex,byteIndex=0,bitIndex=1;
         uint16 size;
         AVOID_AUTO_RESET_NOW

         // goto's sometimes are a very efficient statement
   findMoreBytes:
         // find the next byte with any bit set
         while (max > 0 && *bmp == 0)
         {
            max--;
            byteIndex++;
            bmp++;
         }
         if (max > 0) // found?
         {
            int16 b = *bmp;
   findMoreBits:
            // find, inside this byte, the next bit set
            while (bitIndex < 8 && (b & eightBit[bitIndex]) == 0)
               bitIndex++;
            // if no more bits in this byte, test next bytes
            if (bitIndex == 8)
            {
               bitIndex = 0;
               max--;
               byteIndex++;
               bmp++;
               goto findMoreBytes;
            }
            // found another bit
            nextIndex = bitIndex+(byteIndex<<3);
            obj = INDEX2OBJ(chunk,index);

            if (!WOBJ_FREED(obj))
            {
               if (!WOBJ_MARKED(obj))
               {
                  //{char buf[12]; xstrprintf(buf,"%lx",(int32)obj); debug(buf);}
                  // 1. call destroy function if present
                  if ((wclass = WOBJ_class(obj)) != NULL && WCLASS_hasDestroyFunc(wclass))
   #ifdef PALMOS
                     SWCallPublic(wclass->objDestroyFuncLibRefnum, wclass->objDestroyFuncNum, NULL, (uint32 *)obj);
   #else
                     if (wclass->objDestroyFuncLibRefnum == SWLibRefNum)
                        SWCallPublicStatic(wclass->objDestroyFuncNum, NULL, (uint32 *)obj);
                     else
                        (*vmGlobals->nativeLibs[wclass->objDestroyFuncLibRefnum-1]->SWCallPublic)(wclass->objDestroyFuncNum, NULL, (uint32 *)obj);
   #endif
                  // 2. store internal info
                  size = (nextIndex-index)<<2;
                  WOBJ_freedSize(obj) = size; // offset hasnt changed; just store the size
                  // 3. get the index to the freeChunks list
                  freeIndex = size2freeChunksIndex(size);
                  // 4. add the chunk to the appropriate freeChunk index
                  WOBJ_freedNext(obj) = freeChunks[freeIndex];
                  freeChunks[freeIndex] = obj;
                  // 5. zeroes the object
                  if (size > 8) // a minimum object can have 8 bytes. we cannot remove this if because xzero32 does not check if size is 0 (what will corrupt data)
                     xzero32((byte*)obj+4, (size - CHUNK_INFO_SIZE-4) >> 2); // keep WClass, where the fredNext is stored
                  totalUsed -= size;
                  chunk->usedObjects--;
               }
               else
               {
                  // restore the offset - value already 0000 (marked)
                  WOBJ_infoOffset(obj) = INDEX2OFFSET(index);
               }
            }
            index = nextIndex;
            bitIndex++;
            goto findMoreBits;
         }
         // no more bits found
      }
      if (chunk->usedObjects == 0)
         emptyChunks++;
   }
}

// join two consecutive chunks and upgrade it.
static byte compact()
{
   int index, newIndex;
   WObject obj, last, nextObj, next;
   uint16 size, sizeNext,idx;
   int32 joined = 0;
   ChunkPtr chunk;
   uint32 lastValid;
   uint32 elapsed;
   byte *_eightBit = eightBit;

   elapsed = -getTimeStamp();
   calledCompact++;
   AVOID_AUTO_RESET_NOW

   for (index=0; index <= FREE_CHUNK_MAX_INDEX; index++)
      if (freeChunks[index])
      {
         last = NULL;
         for (obj = freeChunks[index]; obj != NULL;)
         {
            if (WOBJ_FREED(obj)) // may have been joined...
            {
               // get chunk info
               chunk = OBJ2CHUNK(obj);
               //if (chunk->usedObjects == 0) goto nextChunk; // no free objects? go to next chunk

               size = WOBJ_freedSize(obj);
               // get the object consecutive to this one
               nextObj = (WObject)((byte *)obj + size);

               lastValid = (uint32)chunk+chunk->size;
               if ((uint32)nextObj >= lastValid)
                  goto nextChunk;
               // if its also not collected...
               if (WOBJ_FREED(nextObj))
               {
                  // ...join all consecutive freed chunks
                  while (WOBJ_FREED(nextObj))
                  {
                     sizeNext = WOBJ_freedSize(nextObj);
						   // disable current object
                     idx = (uint16)(OBJ2INDEX(chunk,nextObj));
                     BYTE_INDEX(chunk->bitmap,idx) ^= _BIT_INDEX(idx);
					      // rest of the object will be zeroed in next stage
					      WOBJ_freedSize(nextObj) = 0;
                     joined++;
                     // compute new size
   						size += sizeNext;
   						// see if we can join next one too
                     nextObj = (WObject)((byte *)obj + size);
                     if ((uint32)nextObj >= lastValid)
                        break;
						}

  						// "create" the new object
 						WOBJ_freedSize(obj) = size;
                  newIndex = size2freeChunksIndex(size);
                  // upgrade the object
                  next = WOBJ_freedNext(obj);
                  // remove it from this freeChunks index
                  if (last)
                     WOBJ_freedNext(last) = next;
                  else
                     freeChunks[index] = next;
                  // add it to the new freeChunks index
                  WOBJ_freedNext(obj) = freeChunks[newIndex];
                  freeChunks[newIndex] = obj;
                  obj = next;
                  continue;
               }
            }
nextChunk:
            last = obj; // last live object.
            obj = WOBJ_freedNext(obj);
         }
      }


   // stage 2: scan the free chunks again and remove/zero the joined objects
   if (joined)
   {
      for (index=0; index <= FREE_CHUNK_MAX_INDEX; index++)
         if (freeChunks[index])
         {
            last = NULL;
            for (obj = freeChunks[index]; obj != NULL;)
               if (!WOBJ_FREED(obj)) // Not free and in free chunk? remove
               {
                  next = WOBJ_freedNext(obj);
                  // remove it from this freeChunks index
                  if (last)
                     WOBJ_freedNext(last) = next;
                  else
                     freeChunks[index] = next;
                  // erase it
				      WOBJ_infoValue(obj) = 0;
				      WOBJ_freedNext(obj) = 0;
                  obj = next;
               }
               else
               {
                  last = obj; // last live object.
                  obj = WOBJ_freedNext(obj);
               }
         }
   }

   elapsed += getTimeStamp(); // get the dif
   elapsedCompact += elapsed;
   return (byte)(joined > 0);
}

void gc()
{
   WClass *wclass;
   WObject obj;
   uint32 i, j, n;
   int32 elapsed;
   byte *_eightBit = eightBit;
   Var *stackStart = vmStackStart;
   ChunkPtr probableChunk;
   if (vmGlobals->audibleGC) tone(2000,10);
   calledGC++;

   AVOID_AUTO_RESET_NOW

   elapsed = -getTimeStamp();

   // mark objects on vm stack
   for (n = VM_STACK_COUNT; n--;)
   {
      obj = (stackStart++)->obj;
      if (WOBJ_IS_VALID(obj))
         markObject(obj);
   }

   // mark objects on native stack
   for (i = 0, n = NM_STACK_COUNT; n--; i++)
   {
      obj = nmStackStart[i];
      if (WOBJ_IS_VALID(obj))
         markObject(obj);
   }

   // mark all static class objects
   for (i = 0, n = CLASS_HASH_SIZE; n--; i++)
   {
      wclass = classHashList[i];
      while (wclass != NULL)
      {
         WClassField *field = wclass->fields;  // guich@200b4: optimized by removing a __lmul__(18)
         j = wclass->numFields;
         while (j--)
         {
            if (FIELD_isStatic(field))
            {
               obj = field->var.staticVar.obj;
               if (WOBJ_IS_VALID(obj))
                  markObject(obj);
            }
            field++;
         }
         wclass = wclass->nextClass;
      }
   }

   sweep();
   elapsed += getTimeStamp(); // get the dif
   elapsedGC += elapsed;
   if (compact()) // guich@520_5: call the compact after a GC.
      compactWorked++;
   if (vmGlobals->audibleGC) tone(2200,10);
}

static void checkChunks()
{
   int i;
   char buf[30];
   for (i = 1; i < chunksCount; i++)
      if (chunksArray[i] < chunksArray[i-1])
      {
         debug(int2str("ERR in",(int32)i));
         for (i = 1; i < chunksCount; i++)
         {
            xstrprintf(buf,"%d: %ld",i,(int32)chunksArray[i]);
            debug(buf);
         }
         break;
      }
}

static int chunksArrayFindPos(ChunkPtr p)
{
   int middle = 0;
   int lower = 0;
   int upper = chunksCount-1;
   int32 offset;
   while (lower <= upper)
   {
      middle = (lower + upper) >> 1;
      offset = (int32)p - (int32)chunksArray[middle];
      if (offset == 0)
         return middle;
      else
      if (offset > 0)
         lower = middle+1;
      else
         upper = middle-1;
   }
   return lower; //(offset>=0) ? lower : upper; - in some cases this was making a wrong insertion. seems that LOWER is always the correct position. or we should check: return (p < chunksArray[lower]) ? (lower-1) : lower
}

static void chunksArrayRemove(ChunkPtr p)
{
   int pos = chunksArrayFindPos(p);
   if (pos < 0)
      pos = 0;
   if (pos < (chunksCount-1))
      xmemmove((byte*)&chunksArray[pos], (byte*)&chunksArray[pos+1], (chunksCount-1-pos)<<2);
   chunksCount--;
}

static ChunkPtr chunksArrayFindNearest(WObject p)
{
   ChunkPtr chunk;
   int lower = 1;
   int upper = chunksCount-1;

   // first chunk already verified
   if (upper < 200) // for low numbers, a loop is faster than a binary search
   {
      for (;lower <= upper; lower++)
      {
         chunk = chunksArray[lower];
         if ((uint32)p >= (uint32)chunk && (uint32)p < chunk->limit) // is object inside chunk?
            return chunk;
      }
   }
   else
   {
      int middle;
      while (lower <= upper)
      {
         middle = (lower + upper) >> 1;
         chunk = chunksArray[middle];
         if ((uint32)p >= (uint32)chunk)
         {
            if ((uint32)p < chunk->limit) // is object inside chunk?
               return chunk;
            else
               lower = middle+1;
         }
         else
            upper = middle-1;
      }
   }
   return NULL;
}

//
// Native Method Stack
//

int pushObject(WObject obj)
{
   // prevent the pushed object from being freed by the garbage
   // collector. Used in native methods and with code calling
   // the VM. For example, if you have the following code
   //
   // obj1 = createObject(...);
   // obj2 = createObject(...);
   //
   // or..
   //
   // obj1 = createObject(...);
   // m = getMethod(..)
   //
   // since the second statement can cause a memory allocation
   // resulting in garbage collection (in the latter a class
   // load that allocates static class variables), obj1
   // would normally be freed. Pushing obj1 onto the "stack"
   // (which is a stack for this purpose) prevents that
   //
   // the code above should be change to:
   //
   // obj1 = createObject(...);
   // pushObject(obj1);
   // obj2 = createObject(...);
   // pushObject(obj2);
   // ..
   // if (popObject() != obj2)
   //   ..error..
   // if (popObject() != obj1)
   //   ..error..
   //

   // NOTE: Running out of Native Stack space can cause serious program
   // failure if any code doesn't check the return code of pushObject().
   // Any code that does a pushObject() should check for failure and if
   // failure occurs, then abort.
   if (nmStackPtr >= nmStackEnd)
   {
      // guich@200b4r5: grow the stack
      if (!growStack((byte **)&nmStackStart, (byte **)&nmStackPtr, (byte **)&nmStackEnd, NULL, NULL, NULL, &nmStackSizeInBytes))
      {
         VmQuickError(ERR_NativeStackOverflow, NULL);
         return -1;
      }
   }
   *nmStackPtr++ = obj;
   return 0;
}

WObject popObject()
{
   if (nmStackPtr == nmStackStart)
   {
      VmQuickError(ERR_NativeStackUnderflow, NULL);
      return 0;
   }
   return *--nmStackPtr;
}

//
// Native Methods and Hooks
//

// Hooks are used for objects that access system resources. Classes
// that are "hooked" may allocate extra variables so system resource
// pointers can be stored in the object. All hooked objects
// have an object destroy function that is called just before they
// are garbage collected allowing system resources to be deallocated.
static void setClassHooks(WClass *wclass)
{
   UtfString *className;
   ClassHook *hook;
   uint16 nameLen;

   // NOTE: Like native methods, we could hash the hook class names into
   // a value if we make sure that the only time we'd check an object for
   // hooks is if it was in the waba package. This would make lookup
   // faster and take up less space. If the hook table is small, though,
   // it doesn't make much difference.
   className = &wclass->className;
   if (className->len < 6)
      return; // all hook classes are >= 6 character names
   for (hook = vmGlobals->classHooks; hook->className; hook++)
   {
      if (className->str[5] != hook->className[5])
         continue; // quick check to see if we should compare at all
      nameLen = (uint16)xstrlen(hook->className);
      if (className->len == nameLen &&
         !xstrncmp(className->str, hook->className, nameLen))
      {
         wclass->objDestroyFuncNum = hook->destroyFuncNum;
         wclass->objDestroyFuncLibRefnum = hook->libRefnum;
         return;
      }
   }
}
////////////////////////////////////////////////////////////////////////////
// modified by guich to correctly handle international char support
// BETTER UNICODE TO UTF8 CONVERSION ROUTINES - THANKS TO SEAN LUKE and Pierre G. Richard

// Returns the number of unicode characters in the block of bytes of utf8
// up to length len, or the inputted len if the UTF8 is bad.

static uint16 NumUnicodeCharsInUtf8(char *utf8, uint16 len) // guich@402_31: now inputs and returns uint16
{
	int ptr=0;
	uint16 i;
	for (i = 0; i < len; i++, ptr++)
	{
		byte ch = utf8[i];
		if ((ch & 0x80) == 0)
		   ;
		else
		if ((ch & 0xE0) == 0xC0)
			i++;
		else
		if ((ch & 0xF0) == 0xE0)
			i+=2;
		else
			return len; // guich@402_31: now instead of return BAD_UTF in case of bad utf, we just return here the len

	}
	return (i==len)? ptr : len; // prichard@400_52
}
////////////////////////////////////////////////////////////////////////////
WObject createString(char *buf) // guich@502_8: new implementation, to get rid of NumUnicodeCharsInUtf8
{
    WObject obj, charArrayObj;
    uint16 *charStart;
    byte *in;      // guich@502_7: must use unsigned char! Assume char('') = -25. Converting to uint16 would give 0xFFFF-25 instead of 0xFF-25.
    uint16 len = xstrlen(buf);
    charArrayObj = createArrayObject(5, len);
    if (!charArrayObj || pushObject(charArrayObj) == -1)
       return 0;
    charStart = (uint16 *)WOBJ_arrayStart(charArrayObj);
    in = (byte*)buf; // convert to unsigned
    while (*in)
       *charStart++ = *in++;
    // create String object and set char array
    obj = createObject(stringClass);
    if (obj != 0)
        WOBJ_StringCharArrayObj(obj) = charArrayObj;
    popObject(); // charArrayObj
    return obj;
}

WObject createStringFromCharPtr(char *s, int len) // TODO: maybe this should call createString instead?
{
   UtfString u;
   u.str = s;
   u.len = len;
   return createStringFromUtf(u);
}

// modified by guich to correctly handle international char support
// unicodes are transformed to utf8 and stored as a 16 bits char
// utf8 -> unicode
WObject createStringFromUtf(UtfString s)
{
    WObject obj, charArrayObj;
    uint16 *charStart;
    byte *in;      // guich@502_7: must use unsigned char! Assume char('') = -25. Converting to uint16 would give 0xFFFF-25 instead of 0xFF-25.

	 uint16 len = NumUnicodeCharsInUtf8(s.str,(uint16)s.len); // guich@240_17

    // create and fill char array
    charArrayObj = createArrayObject(5, len);
    if (!charArrayObj || pushObject(charArrayObj) == -1)
        return 0;
    charStart = (uint16 *)WOBJ_arrayStart(charArrayObj);
    in = (byte*)s.str;
    if (len == s.len) // guich@400_16: if there are no unicode characters, just copy the array
       while (len--)
          *charStart++ = *in++;
    else
    	 while (len--)
       {
          uint16 out; // guich@310_21: changed byte to uint16

      	 if ((*in & 0x80) == 0)
       		 out = *in;
      	 else
      	 if ((*in & 0xE0) == 0xC0)
      	 {
      		 out = *in & 0x1F;
      		 out <<= 6;
      		 in++;
      		 out += *in & 0x3F;
      	 }
      	 else
      	 if ((*in & 0xF0) == 0xE0)
      	 {
      		 out = *in & 0x0F;
      		 out <<= 6;
      		 in++;
      		 out += *in & 0x3F;

      		 out <<= 6;
      		 in++;
      		 out += *in & 0x3F;
      	 }
      	 else
      	 	 out = '?';
      	 *charStart++ = out; // changed cast char to byte - guich@310_21: removed cast
      	 in++;
       }
    // create String object and set char array
    obj = createObject(stringClass);
    if (obj != 0)
        WOBJ_StringCharArrayObj(obj) = charArrayObj;
    popObject(); // charArrayObj
    return obj;
}


// NOTE: Only set STU_USE_STATIC if the string is temporary and will not be
// needed before stringToUtf is called again. The flags parameter is a
// combination of the STU constants.
// unicode -> utf8
UtfString stringToUtf(WObject string, byte flags)
{
    static char sbytes[192];
    UtfString s;
    WObject charArray;
    uint16 *charStart;
    uint32 len, extra;
    uint16 *chars;
    char *bytes;
    char *b;
    byte nullTerminate, useStatic, alwaysCreate;

    nullTerminate = flags & STU_NULL_TERMINATE;
    useStatic = flags & STU_USE_STATIC;
    alwaysCreate = flags & STU_CREATE;
    s.str = "";
    s.len = 0;
    if (string == 0)
        goto end; // sorry for the 'goto end's here, but a 'return s' is 48 bytes more expensive
    charArray = WOBJ_StringCharArrayObj(string);
    if (charArray == 0)
        goto end;
    charStart = (uint16 *)WOBJ_arrayStart(charArray);
    len = WOBJ_arrayLen(charArray);
    extra = 0;
    if (nullTerminate)
        extra = 1;
    if (!alwaysCreate && useStatic && ((len + extra) <= sizeof(sbytes)))
        bytes = sbytes;
    else
    {
        WObject byteArray = createArrayObject(8, len + extra);
        if (byteArray == 0)
            goto end;
        bytes = (char *)WOBJ_arrayStart(byteArray);
    }

    chars = (uint16 *)WOBJ_arrayStart(charArray);
    // guich@200b2: convert the unicode chars in utf8
    b = bytes;
    s.len = len;
    while (len--)
       *b++ =(char)*chars++;

    if (nullTerminate)
        *b = 0;
    s.str = (char *)bytes;
end:
    return s;
}

///////// Guich@320_5: added support for unicode characters
UtfString16 stringToUtf16(WObject string)
{
   UtfString16 s;
   WObject charArray;

   s.str = 0;
   s.len = 0;
   if (string != NULL)
   {
      charArray = WOBJ_StringCharArrayObj(string);
      if (charArray != NULL)
      {
         s.str = (uint16 *)WOBJ_arrayStart(charArray);
         s.len = WOBJ_arrayLen(charArray);
      }
   }
   return s;
}

////////////////////////// END OF INTERNATIONAL SUPPORT ///////////////////

#define arrayRangeCheck2(array, start, count) (start >= 0 && (start + count) <= WOBJ_arrayLen(array))

int arrayRangeCheck(WObject array, int32 start, int32 count)
{
   return array != 0 && start >= 0 && count >= 0 && (start + count) <= WOBJ_arrayLen(array);
}

void copyArray(Var stack[])
{
    WObject srcArray, dstArray;
    int32 srcStart, dstStart, len, srcType;
    uchar *srcPtr, *dstPtr;
    int shl;

    srcArray = stack[0].obj;
    srcStart = stack[1].intValue;
    dstArray = stack[2].obj;
    dstStart = stack[3].intValue;
    len = stack[4].intValue;
    if (srcArray != 0 && dstArray != 0 && // guich@200: corrected error that Sean pointed out.
        WOBJ_class(srcArray) == NULL && WOBJ_class(dstArray) == NULL) // ensure both src and dst are arrays
    {
       // NOTE: This is not a full check to see if the two arrays are compatible.
       // Any two arrays of objects are compatible according to this check
       // see also: compatibleArray()
       srcType = WOBJ_arrayType(srcArray);
       if (srcType == WOBJ_arrayType(dstArray))
       {
          // check ranges
          if (len < 0 || arrayRangeCheck2(srcArray, srcStart, len) == 0 ||
                         arrayRangeCheck2(dstArray, dstStart, len) == 0)
              VmQuickError(ERR_IndexOutOfRange,"In copyArray");
          else
          {
             // guich@200b4: put this here to avoid 3 multiplications and improve performance
             shl = typesSizeShl[srcType]; // guich@200b4r5: improve performance by using typeSizeShl (instead of the old switch)
             srcPtr = (uchar *)WOBJ_arrayStart(srcArray) + (srcStart << shl);
             dstPtr = (uchar *)WOBJ_arrayStart(dstArray) + (dstStart << shl);
             len <<= shl;
             xmemmove(dstPtr, srcPtr, len);

             RETURN32(1)
          }
       }
    }
    RETURN32(0)
}

WClassField *getField(WClass *wclass, UtfString *name, UtfString *desc)
{
    WClassField *field;
    uint16 n;
    uint32 fieldHash = getHash(name,desc);

    field = wclass->fields;
    n = wclass->numFields;
    while (n--)
    {
       if (field->hash == fieldHash)
          return field;
       field++;
    }
    if (wclass->is12 && wclass->numSuperClasses > 0) // guich@310_14: added support for jdk 1.2 class file format
    {
       int i;
       for (i = wclass->numSuperClasses-1; i >= 0; i--)
       {
    	    WClass *wc = wclass->superClasses[i];
          field = wc->fields;
          n = wc->numFields;
          while (n--)
          {
             if (field->hash == fieldHash)
                return field;
             field++;
          }
       }
    }
    VmError(ERR_CantFindField, wclass, name, desc);
    return NULL;
}

static WClassField *getFieldByIndex(WClass *wclass inA0, uint16 fieldIndex inD0) inD0 // don't inline this function!
{
    WClassField *field=NULL;
    WClass *targetClass;
    uint16 targetClassIndex, nameAndTypeIndex;
    UtfString fieldName, fieldDesc;

    targetClassIndex = CONST_classIndex(wclass, fieldIndex);
    targetClass = (WClass *)CONST_binded(wclass, targetClassIndex);
    if (targetClass == NULL)
       targetClass = getClassByIndex(wclass, targetClassIndex,1);
    if (targetClass != NULL)
    {
       nameAndTypeIndex = CONST_nameAndTypeIndex(wclass, fieldIndex);
       getUtfString2(wclass, CONST_nameIndex(wclass, nameAndTypeIndex), &fieldName);
       getUtfString2(wclass, CONST_typeIndex(wclass, nameAndTypeIndex), &fieldDesc);
       field = getField(targetClass, &fieldName, &fieldDesc);
       if (field != NULL)
          CONST_binded(wclass, fieldIndex) = (BindedConst)field;
    }
    return field;
}

WClass *getClassByIndex(WClass *wclass inA0, uint16 classIndex inD0, byte throwEx inD1)
{
    UtfString className;
    WClass *targetClass;

    targetClass = (WClass *)CONST_binded(wclass, classIndex);
    if (targetClass == NULL)
    {
       getUtfString2(wclass, CONST_nameIndex(wclass, classIndex), &className);
       if (!(className.len > 1 && className.str[0] == '[')) // arrays have no associated class
       {
          targetClass = getClass(className,throwEx);
          if (targetClass != NULL)
             CONST_binded(wclass, classIndex) = (BindedConst)targetClass;
       }
    }
    return targetClass;
}

#ifdef ENABLE_TRACE
static void trace(WClass *iclass, WClassMethod *imethod)
{
   int j=0,l;
   char *c;
   uint16 idx;
//#ifdef PALMOS
   static char traceBuf[200];
/*#else
   uint16 traceBuf[200];
#endif*/
   UtfString methodName = (imethod==NULL)?createUtfString("new "):getUtfString(iclass, (uint16)((idx=METH_nameIndex(imethod)) < iclass->numConstants ? idx : 0));
   UtfString methodDesc = (imethod==NULL)?createUtfString("")    :getUtfString(iclass, (uint16)((idx=METH_descIndex(imethod)) < iclass->numConstants ? idx : 0));

   // finds the last '/' of the class. ignores before it
   c = iclass->className.str;
   while (*c && *c != '/') c++;
   if (*c == '/') // there exist a '/'?
   {
	   l = iclass->className.len;
	   c = &iclass->className.str[l-1];
	   while (*c != '/' && l > 0) {l--; c--;}
	   if (*c == '/') c++;
	}
   // concats the strings
   for (; *c; j++)
      traceBuf[j] = *c++;
   j--;
   traceBuf[j++] = '.';
   c = methodName.str;
   for (; *c && (j < 99); j++,c++)
      traceBuf[j] = *c;
   j--;
   c = methodDesc.str;
   if (*c)
   {
	   for (; *c && (j < 99); j++)
   	   traceBuf[j] = *c++;
	   c--; j--;
	}
   traceBuf[j++] = '\r';
   traceBuf[j++] = '\n';
   traceBuf[j] = 0;
   debug(traceBuf);
//   if (xstrncmp(traceBuf,"Rect.<init>(IIII)V",18) == 0) // @ t
//      *traceBuf = 0;
}
#endif