/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_Catalog.c,v 1.10 2005/09/25 12:56:34 guich Exp $ */

#include "superwaba.h"
#include "SW_natives.h"
#include "SW_palmDb.h"

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

/* are these definitions required ? not in this implementation file at least.
 * if yes, move them to an include file
 *
#ifdef WINCE
  #define errno GetLastError()
  #define _tmkdir(path) CreateDirectory(path, 0)
#elif PALMOS
#else
  #include <errno.h>
  #include <time.h>
  #include <sys/types.h>
  #include <sys/stat.h>
#endif

#define INVALID_FILEPTR_VALUE (void *)0

#ifdef WIN32
  #include "win32/dirent.h"
  #ifdef WINCE
    #define _trename MoveFile
  #else
    #include <direct.h>
    #include <sys/timeb.h>
    #define stat _stat
  #endif
#elif PALMOS
#else

// POSIX
  #include <sys/time.h>
  #include <unistd.h>
  #include <dirent.h>
  #ifdef UNICODE
  #define _tmkdir >>>PGR: FIXME
  #define _tstat  >>>PGR: FIXME
  #define _tunlink >>>PGR: FIXME
  #define _trename >>>PGR: FIXME
  #define _tchmod >>>PGR: FIXME
  #define _tfopen >>>PGR: FIXME
  #define _trmdir >>>PGR: FIXME
  #else
  #define _tmkdir(path) mkdir(path, S_IRWXU | S_IRWXG | S_IRWXO)
  #define _tstat stat
  #define _tunlink unlink
  #define _trename rename
  #define _tchmod chmod
  #define _tfopen fopen
  #define _trmdir rmdir
  #endif
#endif

*/

#if defined PALMOS

static void CatalogReadWriteBytes(Var stack[], int isRead)
{
   WObject cat, byteArray;
   int32 start, count, pos, recOffset;
   uchar *bytes, *recBytes;

   cat = stack[0].obj;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   pos = WOBJ_CatalogCurRecPos(cat);
   if (pos == -1)
      RETURN32(-1)  // no current record
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
      RETURN32(-1)  // array null or range invalid
   recBytes = WOBJ_CatalogCurRecPtr(cat);
   recOffset = WOBJ_CatalogCurRecOffset(cat);
   if (recOffset + count > WOBJ_CatalogCurRecLen(cat))
      RETURN32(-1)  // past end of record
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   if (isRead)
      xmemmove(&bytes[start], &recBytes[recOffset], count);
   else
   {
      if (DmWrite(recBytes, recOffset, &bytes[start], count) != 0)
         RETURN32(-1)
      WOBJ_CatalogCurRecModified(cat) = 1;
   }
   WOBJ_CatalogCurRecOffset(cat) = recOffset + count;
   RETURN32(count)
}
static void _RecClose(WObject cat)
{
   DmOpenRef dmRef;
   VoidHand recH;
   int32 pos;
   Boolean dirty;

   pos = WOBJ_CatalogCurRecPos(cat);
   if (pos == -1)
      return; // no current record
   recH = WOBJ_CatalogCurRecHandle(cat);
   MemHandleUnlock(recH);
   dmRef = WOBJ_CatalogDmRef(cat);
   if (WOBJ_CatalogCurRecModified(cat) != 0)
   {
      dirty = true;
      WOBJ_CatalogSetBackup(cat) = true; // guich@200
   }
   else
      dirty = false;
   DmReleaseRecord(dmRef, pos, dirty);
   WOBJ_CatalogCurRecPos(cat) = -1;
}

static int32 _CatalogClose(WObject cat)
{
   DmOpenRef dmRef;

   _RecClose(cat); // release any open records
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      return 0;
   // guich@200 - sets the backup bit if the file was modified
   if (WOBJ_CatalogSetBackup(cat))
   {
      LocalID dbID;
      _unused_ UInt cardNo;
      Word attributes;

      DmOpenDatabaseInfo(dmRef, &dbID, NULL, NULL, PALM_CARD(&cardNo) NULL);
      DmDatabaseInfo(PALM_CARD(cardNo) dbID, NULL, &attributes,NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
      attributes |= dmHdrAttrBackup;
      DmSetDatabaseInfo(PALM_CARD(cardNo) dbID, NULL, &attributes,NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
   }
   if (dmRef != vmGlobals->dmRefSystemOut) // guich@200b4_19: only close if is not the memo db
      DmCloseDatabase(dmRef); // guich@200b2: removed from before the last block to correct the "address error" from pose
   WOBJ_CatalogDmRef(cat) = 0;
   return 1;
}

void catalogDestroy(WObject cat)
{
   _CatalogClose(cat);
}

//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogCreate(Catalog obj, String name, int mode)
void CatalogCreate(Var stack[]) // (Lwaba/io/Catalog;Ljava/lang/String;I)V
{
   UtfString s;
   WObject cat, name;
   int32 mode;
   DmOpenRef dmRef;
   ULong dbType, dbCreator;
   byte isMemo=0;

   cat = stack[0].obj;
   name = stack[1].obj;
   mode = stack[2].intValue;
   WOBJ_CatalogDmRef(cat) = 0;
   WOBJ_CatalogCurRecPos(cat) = -1;
   s = (*vmGlobals->stringToUtf)(name, STU_USE_STATIC | STU_NULL_TERMINATE);
   if (s.len == 0)
      return;
   dbCreator = vmGlobals->appCreatorId;
   dbType = 'DATA';
   if (s.len > 10 && s.str[s.len - 10] == '.' && s.str[s.len - 5] == '.')
   {
      isMemo = !xstrncmp(s.str,"MemoDB.memo.DATA",s.len); // are we trying to open the memo db?
      s.str[s.len - 10] = 0;
      dbCreator = READ_BE_UINT32((uchar *)&s.str[s.len - 9]);
      dbType = READ_BE_UINT32((uchar *)&s.str[s.len - 4]);
   }
   // guich@200b4_19: fixed the memo and debug concurrent use
   if (isMemo && vmGlobals->dmRefSystemOut) // memo already opened?
      dmRef = vmGlobals->dmRefSystemOut;
   else
   {
      dmRef = (*vmGlobals->findAndOpenDB)(s.str, dbCreator, dbType, (byte)(mode==Catalog_READ_ONLY), 0);
      if (isMemo && dmRef)
         vmGlobals->dmRefSystemOut = dmRef; // assign to the global pos
   }

   if (mode == Catalog_CREATE && !dmRef)
   {
      if (DmCreateDatabase(PALM_CARD(0) s.str, dbCreator, dbType, false) != 0)
         return;
      dmRef = (*vmGlobals->findAndOpenDB)(s.str, dbCreator, dbType, 0, 0);
      // set the backup bit on the database
      if (dmRef != 0)
      {
         LocalID dbID;
         _unused_ UInt cardNo;
         Word attributes;

         DmOpenDatabaseInfo(dmRef, &dbID, NULL, NULL, PALM_CARD(&cardNo) NULL);
         DmDatabaseInfo(PALM_CARD(cardNo) dbID, NULL, &attributes,
            NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
         attributes |= dmHdrAttrBackup;
         DmSetDatabaseInfo(PALM_CARD(cardNo) dbID, NULL, &attributes,
            NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
      }
   }
   if (dmRef != 0)
      WOBJ_CatalogDmRef(cat) = dmRef;
}
//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean catalogRename(Catalog obj, String newName)
void CatalogRename(Var stack[]) // (Lwaba/io/Catalog;Ljava/lang/String;)Z
{
   WObject cat;
   DmOpenRef dmRef;
   LocalID dbID;
   _unused_ UInt cardNo;
   UtfString s;
   WObject name;
   ULong dbType, dbCreator;
   Err err;
   byte isMemo=0;

   cat = stack[0].obj;
   name = stack[1].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   s = (*vmGlobals->stringToUtf)(name, STU_USE_STATIC | STU_NULL_TERMINATE);
   if (s.len == 0)
      RETURN32(0)
   if (s.len < 10 || s.str[s.len - 10] != '.' || s.str[s.len - 5] != '.')
      RETURN32(0)
   isMemo = !xstrncmp(s.str,"MemoDB.memo.DATA",s.len); // are we trying to open the memo db?
   if (isMemo) // can't rename to the memo database!
      RETURN32(0)

   s.str[s.len - 10] = 0;
   dbCreator = READ_BE_UINT32((uchar *)&s.str[s.len - 9]);
   dbType = READ_BE_UINT32((uchar *)&s.str[s.len - 4]);
   DmOpenDatabaseInfo(dmRef, &dbID, NULL, NULL, PALM_CARD(&cardNo) NULL);
   err = DmSetDatabaseInfo(PALM_CARD(cardNo) dbID, s.str, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &dbType, &dbCreator);
   RETURN32(err == errNone)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogAddRecord(Catalog obj, int size)
void CatalogAddRecord(Var stack[]) // (Lwaba/io/Catalog;I)I
{
   WObject cat;
   DmOpenRef dmRef;
   VoidHand recH;
   VoidPtr recPtr;
   UInt16 pos;
   int32 size;

   cat = stack[0].obj;
   size = stack[1].intValue;
   _RecClose(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)
   pos = DmNumRecords(dmRef);
   recH = DmNewRecord(dmRef, &pos, size);
   if (!recH)
      RETURN32(-1)
   recPtr = MemHandleLock(recH);
   if (!recPtr)
   {
      DmRemoveRecord(dmRef, pos); // guich@400_16: changed from DmDeleteRecord
      RETURN32(-1)
   }
   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recPtr;
   WOBJ_CatalogCurRecOffset(cat) = 0;
   WOBJ_CatalogCurRecLen(cat) = size;
   WOBJ_CatalogCurRecModified(cat) = true;
   RETURN32((int32)pos)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogAddRecord(Catalog obj, int size, int pos)
void CatalogAddRecord2(Var stack[]) // (Lwaba/io/Catalog;II)I
{
   WObject cat;
   DmOpenRef dmRef;
   VoidHand recH;
   VoidPtr recPtr;
   UInt16 pos;
   int32 size;

   cat = stack[0].obj;
   size = stack[1].intValue;
   _RecClose(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)
   pos = stack[2].intValue;//DmNumRecords(dmRef);
   recH = DmNewRecord(dmRef, &pos, size);
   if (!recH)
      RETURN32(-1)
   recPtr = MemHandleLock(recH);
   if (!recPtr)
   {
      DmRemoveRecord(dmRef, pos); // guich@400_16: changed from DmDeleteRecord
      RETURN32(-1)
   }
   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recPtr;
   WOBJ_CatalogCurRecOffset(cat) = 0;
   WOBJ_CatalogCurRecLen(cat) = size;
   WOBJ_CatalogCurRecModified(cat) = true;
   RETURN32((int32)pos)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogResizeRecord(Catalog obj, int size)
void CatalogResizeRecord(Var stack[]) // (Lwaba/io/Catalog;I)Z
{
   WObject cat;
   DmOpenRef dmRef;
   VoidHand recH;
   VoidPtr recPtr;
   int32 pos, size;

   cat = stack[0].obj;
   size = stack[1].intValue;
   if (size <= 0)
      RETURN32(0)
   pos = WOBJ_CatalogCurRecPos(cat);
   if (pos == -1)            
      RETURN32(0)
   _RecClose(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   recH = DmResizeRecord(dmRef, pos, (ULong)size);
   if (recH == NULL)
      RETURN32(0)
   recH = DmGetRecord(dmRef, pos); // guich@310_12: fixed Record Not Locked bug
   if (recH == NULL)
      RETURN32(0)
   recPtr = MemHandleLock(recH);
   if (recPtr == NULL)
      RETURN32(0)
   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recPtr;
   WOBJ_CatalogCurRecLen(cat) = size;
   WOBJ_CatalogCurRecModified(cat) = true;
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogClose(Catalog obj)
void CatalogClose(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat = stack[0].obj;
   RETURN32(_CatalogClose(cat))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogDelete(Catalog obj)
void CatalogDelete(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat;
   DmOpenRef dmRef;
   LocalID localID;
   _unused_ UInt cardNo;

   cat = stack[0].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   DmOpenDatabaseInfo(dmRef, &localID, NULL, NULL, PALM_CARD(&cardNo) NULL);
   _CatalogClose(cat);
   if (DmDeleteDatabase(PALM_CARD(cardNo) localID) != 0)
      RETURN32(0)
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
static void _ULongToChars(ULong l, char *buf)
{
   buf[0] = l >> 24;
   buf[1] = LOWER8(l >> 16);
   buf[2] = LOWER8(l >> 8);
   buf[3] = LOWER8(l);
}
// waba/io/NativeMethods String []catalogListCatalogs(int creatorId, int type)
void CatalogListCatalogs(Var stack[]) // (II)[Ljava/lang/String;
{
   WObject stringArray, *strings;
   DmSearchStateType state;
   Boolean first;
   _unused_ UInt cardNo;
   LocalID dbID;
   ULong dbType, dbCreator;
   int n, len;
   char name[32 + 10];
   int32 type,crtr;

   crtr = stack[0].intValue;
   type = stack[1].intValue;
   n = 0;
   first = true;
   while (DmGetNextDatabaseByTypeCreator(first, &state, type, crtr, false, PALM_CARD(&cardNo) &dbID) == 0)
   {
      first = false;
      n++;
   }
   if (n == 0)
      RETURN32(0)
   stringArray = (*vmGlobals->createArrayObject)(1, n);
   if (stringArray == NULL || (*vmGlobals->pushObject)(stringArray) == -1)
      RETURN32(0)
   strings = (WObject *)WOBJ_arrayStart(stringArray);
   n = 0;
   first = true;
   while (DmGetNextDatabaseByTypeCreator(first, &state,
         type, crtr, false, PALM_CARD(&cardNo) &dbID) == 0)
   {
      first = false;
      name[0] = 0;
      DmDatabaseInfo(PALM_CARD(cardNo) dbID, name, NULL, NULL, NULL, NULL,
         				NULL, NULL, NULL, NULL, &dbType, &dbCreator);
      len = xstrlen(name);
      name[len++] = '.';
      _ULongToChars(dbCreator, &name[len]);
      len += 4;
      name[len++] = '.';
      _ULongToChars(dbType, &name[len]);
      len += 4;
      name[len++] = 0;
      strings[n++] = (*vmGlobals->createString)(name);
   }
   (*vmGlobals->popObject)(); // stringArray
   RETURN32(stringArray)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogDeleteRecord(Catalog obj)
void CatalogDeleteRecord(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat = stack[0].obj;
   int32 pos = WOBJ_CatalogCurRecPos(cat);
   DmOpenRef dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0 || pos == -1)
      RETURN32(0)
   _RecClose(cat);
   // NOTE: Was DmDeleteRecord() but this didn't match with CE
   // and was too difficult to use in actual programs since it
   // didn't delete the record immediately
   RETURN32(DmRemoveRecord(dmRef, pos) == errNone)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetRecordCount(Catalog obj)
void CatalogGetRecordCount(Var stack[]) // (Lwaba/io/Catalog;)I
{
   WObject cat = stack[0].obj;
   DmOpenRef dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   RETURN32((dmRef == 0)?-1:DmNumRecords(dmRef))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetRecordSize(Catalog obj)
void CatalogGetRecordSize(Var stack[]) // (Lwaba/io/Catalog;)I
{
   WObject cat = stack[0].obj;
   int32 pos = WOBJ_CatalogCurRecPos(cat);
   DmOpenRef dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   RETURN32((dmRef == 0 || pos == -1)?-1:WOBJ_CatalogCurRecLen(cat))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogIsOpen(Catalog obj)
void CatalogIsOpen(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat = stack[0].obj;
   DmOpenRef dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   RETURN32(dmRef != 0)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogSetRecordPos(Catalog obj, int pos)
void CatalogSetRecordPos(Var stack[]) // (Lwaba/io/Catalog;I)Z
{
   WObject cat;
   int32 pos, count;
   DmOpenRef dmRef;
   VoidHand recH;
   VoidPtr recPtr;

   cat = stack[0].obj;
   pos = stack[1].intValue;
   _RecClose(cat);
   if (pos < 0)
      RETURN32(0)
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   if (vmGlobals->stateSystemOut == SYSOUT_OPENED && dmRef == vmGlobals->dmRefSystemOut && pos == vmGlobals->posSystemOut) // guich@200b4_19: don't let him use the debug record - guich@200final_8: forgot to test the dmRef
      RETURN32(0)
   count = (int32)DmNumRecords(dmRef);
   if (pos >= count)
      RETURN32(0)
   recH = DmGetRecord(dmRef, pos);
   if (!recH)
      RETURN32(0)
   recPtr = MemHandleLock(recH);
   if (!recPtr)
   {
      DmReleaseRecord(dmRef, pos, false);
      RETURN32(0)
   }
   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recPtr;
   WOBJ_CatalogCurRecOffset(cat) = 0;
   WOBJ_CatalogCurRecLen(cat) = MemHandleSize(recH);
   WOBJ_CatalogCurRecModified(cat) = 0;
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogReadBytes(Catalog obj, byte buf[], int start, int count)
void CatalogReadBytes(Var stack[]) // (Lwaba/io/Catalog;[BII)I
{
   CatalogReadWriteBytes(stack, 1);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogSkipBytes(Catalog obj, int count)
void CatalogSkipBytes(Var stack[]) // (Lwaba/io/Catalog;I)I
{
   int32 offset;
   WObject cat = stack[0].obj;
   int32 count = stack[1].intValue;
   int32 pos = WOBJ_CatalogCurRecPos(cat);
   DmOpenRef dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0 || pos == -1)
      RETURN32(-1)
   offset = WOBJ_CatalogCurRecOffset(cat)+count;
   if (offset < 0 || offset > WOBJ_CatalogCurRecLen(cat)) // guich@210_11
      RETURN32(-1)
   WOBJ_CatalogCurRecOffset(cat) = offset;    // guich@230_23: changed count to offset
   RETURN32(count)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogWriteBytes(Catalog obj, byte buf[], int start, int count)
void CatalogWriteBytes(Var stack[]) // (Lwaba/io/Catalog;[BII)I
{
   CatalogReadWriteBytes(stack, 0);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogInspectRecord(Catalog obj, byte buf[], int recPosition)
void CatalogInspectRecord(Var stack[]) // (Lwaba/io/Catalog;[BI)I
{
   int32 size;
   VoidHand recH;
   VoidPtr recPtr;

   WObject cat = stack[0].obj;
   WObject byteArray = stack[1].obj;
   int32 pos = stack[2].intValue;
   int32 count = WOBJ_arrayLen(byteArray);
   byte* bytes = (uchar *)WOBJ_arrayStart(byteArray);

   // gets the record
   DmOpenRef dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)
   recH = DmQueryRecord(dmRef, pos);
   if (!recH)
      RETURN32(-1)
   recPtr = MemHandleLock(recH);
   size = MemHandleSize(recH);
   if (count > size) count = size;

   // transfers the data. ps: i did not made a direct reference to the record because i dont have where to store the memhandle params.
   xmemmove(bytes, recPtr, count);
   MemHandleUnlock(recH);

   RETURN32(count)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static byte catalogGetRecordAttributes(Catalog obj, int recordPos)
void CatalogGetRecordAttributes(Var stack[]) // (Lwaba/io/Catalog;I)B
{
   UInt16 attr;
   Err err;
   WObject cat = stack[0].obj;
   int32 pos = stack[1].intValue; // guich@400_42: now the user send the record index
   DmOpenRef dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);

   if (dmRef == 0 || pos < 0)
      RETURN32(0)
   err = DmRecordInfo (dmRef, pos, &attr, NULL, NULL);
   RETURN32(err?-1:LOWER8(attr))  // returns byte
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogSetRecordAttributes(Catalog obj, int recordPos, byte attr)
void CatalogSetRecordAttributes(Var stack[]) // (Lwaba/io/Catalog;IB)V
{
   Err err;

   WObject cat = stack[0].obj;
   int32 pos = stack[1].intValue; // guich@400_42
   uint16 attr = (UInt16)stack[2].intValue;
   DmOpenRef dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0 || pos < 0)
      return;
   err = DmSetRecordInfo (dmRef, pos, &attr, NULL);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetAttributes(Catalog obj)
void CatalogGetAttributes(Var stack[]) // (Lwaba/io/Catalog;)I
{
   LocalID dbID;
   _unused_ UInt cardNo;
   Word attributes;

   WObject cat = stack[0].obj;
   WObject dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)

   DmOpenDatabaseInfo(dmRef, &dbID, NULL, NULL, PALM_CARD(&cardNo) NULL);
   DmDatabaseInfo(PALM_CARD(cardNo) dbID, NULL, &attributes, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
   RETURN32(attributes)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogSetAttributes(Catalog obj, int i)
void CatalogSetAttributes(Var stack[]) // (Lwaba/io/Catalog;I)V
{
   LocalID dbID;
   _unused_ UInt cardNo;
   WObject cat = stack[0].obj;
   Word attributes = LOWER16(stack[1].intValue);
   DmOpenRef dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef != 0)
   {
      DmOpenDatabaseInfo(dmRef, &dbID, NULL, NULL, PALM_CARD(&cardNo) NULL);
      DmSetDatabaseInfo(PALM_CARD(cardNo) dbID, NULL, &attributes, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
   }
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetRecordPos(Catalog obj)
void CatalogGetRecordPos(Var stack[]) // (Lwaba/io/Catalog;)I
{
   WObject cat = stack[0].obj; // class object
   RETURN32(WOBJ_CatalogCurRecPos(cat))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogCreate(Catalog obj)
void CatalogCreate_NoFunc(_unused_ Var stack[]) // (Lwaba/io/Catalog;)V
{
   (*vmGlobals->debug)("This Catalog constructor has no effect!");
}



#else // defined PALMOS



////////////////////////////////////////////////////////////////////////////
// PORTING NOTES: ALL CATALOG METHODS ARE PORTED WHEN YOU IMPLEMENT THE
// NEEEDING ROUTINES IN CATALOG.C

static void catalogReadWriteBytes(Var stack[], int isRead)
{
   WObject cat, byteArray;
   DmOpenRef dmRef;
   int32 start, count, pos, recOffset;
   uchar *bytes, *recBytes;

   cat = stack[0].obj;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   pos = WOBJ_CatalogCurRecPos(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)
   if (pos == -1)
      RETURN32(-1) // no current record
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
      RETURN32(-1) // array null or range invalid
   recBytes = WOBJ_CatalogCurRecPtr(cat);
   recOffset = WOBJ_CatalogCurRecOffset(cat);
   if ((recOffset + count) > WOBJ_CatalogCurRecLen(cat))
      RETURN32(-1) // past end of recor
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   if (isRead)
      xmemmove(&bytes[start], &recBytes[recOffset], count);
   else
   {
      if (DmWrite(dmRef, (uint16)pos, recOffset, &bytes[start], count) != 0)
         RETURN32(-1)
      WOBJ_CatalogCurRecModified(cat) = 1;
   }
   WOBJ_CatalogCurRecOffset(cat) = recOffset + count;
   RETURN32(count)
}
static void recClose(WObject cat)
{
   DmOpenRef dmRef;
   int32 pos;
   byte dirty;

   pos = WOBJ_CatalogCurRecPos(cat);
   if (pos == -1)
      return; // no current record
// recH = WOBJ_CatalogCurRecHandle(cat);
// MemHandleUnlock(recH);
   dmRef = WOBJ_CatalogDmRef(cat);
   if (WOBJ_CatalogCurRecModified(cat) != 0)
   {
      dirty = 1;
      WOBJ_CatalogSetBackup(cat) = 1; // guich@200
   }
   else
      dirty = 0;
   DmReleaseRecord(dmRef, (uint16)pos, dirty);
   WOBJ_CatalogCurRecPos(cat) = -1;
}

static int32 catalogClose(WObject cat, byte callDmClose)
{
   DmOpenRef dmRef;

   recClose(cat); // release any open records
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      return 0;
   // guich@200 - sets the backup bit if the file was modified
   if (WOBJ_CatalogSetBackup(cat))
      DmSetBackupAttribute(dmRef);
   if (callDmClose) // guich@200b4_19: only close if is not the memo db
      DmCloseDatabase(dmRef); // guich@200b2: removed from before the last block to correct the "address error" from pose
   WOBJ_CatalogDmRef(cat) = 0;
   return 1;
}

void catalogDestroy(WObject cat)
{
   catalogClose(cat,1);
}

//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogCreate(Catalog obj, String name, int mode)
void CatalogCreate(Var stack[]) // (Lwaba/io/Catalog;Ljava/lang/String;I)V
{
   UtfString s;
   WObject cat, name;
   int32 mode;
   DmOpenRef dmRef;
   uint32 dbType, dbCreator;

   cat = stack[0].obj;
   name = stack[1].obj;
   mode = stack[2].intValue;
   WOBJ_CatalogDmRef(cat) = 0;
   WOBJ_CatalogCurRecPos(cat) = -1;
   s = (*vmGlobals->stringToUtf)(name, STU_USE_STATIC | STU_NULL_TERMINATE);
   if (s.len == 0)
      return;
   dbCreator = vmGlobals->appCreatorId;
   dbType = DATA_TYPE_ID;
   if (s.len > 10 && s.str[s.len - 10] == '.' && s.str[s.len - 5] == '.')
   {
      s.str[s.len - 10] = 0;
      dbCreator = READ_BE_UINT32((uchar *)&s.str[s.len - 9]);
      dbType = READ_BE_UINT32((uchar *)&s.str[s.len - 4]);
   }
   dmRef = (*vmGlobals->findAndOpenDB)(s.str, dbCreator, dbType, 0/*(byte)(mode==Catalog_READ_ONLY)*/,0); // guich@510_31: removed the read_only support. 
                                                                                                        
   if (mode == Catalog_CREATE && !dmRef)                                                                
   {                                                                                                    
      if (DmCreateDatabase(s.str, dbCreator, dbType) != 0)
         return;
      dmRef = (*vmGlobals->findAndOpenDB)(s.str, dbCreator, dbType, 0,0);
      DmSetBackupAttribute(dmRef);
   }
   if (dmRef != 0)
      WOBJ_CatalogDmRef(cat) = dmRef;
}

//////////////////////////////////////////////////////////////////////////
// waba/io/NativeMethods boolean catalogRename(Catalog obj, String newName)
void CatalogRename(Var stack[]) // (Lwaba/io/Catalog;Ljava/lang/String;)Z
{
   WObject cat;
   DmOpenRef dmRef;
   UtfString s;
   WObject name;
   uint32 dbType, dbCreator;
   Err err;
   byte isMemo=0;

   cat = stack[0].obj;
   name = stack[1].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   s = (*vmGlobals->stringToUtf)(name, STU_USE_STATIC | STU_NULL_TERMINATE);
   if (s.len == 0)
      RETURN32(0)
   if (s.len < 10 || s.str[s.len - 10] != '.' || s.str[s.len - 5] != '.')
      RETURN32(0)
   isMemo = (byte)(xstrncmp(s.str,"MemoDB.memo.DATA",s.len) == 0); // are we trying to open the memo db?
   if (isMemo) // can't rename to the memo database!
      RETURN32(0)

   s.str[s.len - 10] = 0;
   dbCreator = READ_BE_UINT32((uchar *)&s.str[s.len - 9]);
   dbType = READ_BE_UINT32((uchar *)&s.str[s.len - 4]);
   err = DmRenameDatabase(dmRef, dbType, dbCreator, s.str);
   RETURN32(err == 0)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogAddRecord(Catalog obj, int size)
void CatalogAddRecord(Var stack[]) // (Lwaba/io/Catalog;I)I
{
   WObject cat;
   DmOpenRef dmRef;
   VoidHand recH;
   uint16 pos;
   int32 size;

   cat = stack[0].obj;
   size = stack[1].intValue;
   recClose(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)
   pos = DmNumRecords(dmRef);
   recH = DmNewRecord(dmRef, &pos, size);
   if (!recH)
      RETURN32(-1)
   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recH;
   WOBJ_CatalogCurRecOffset(cat) = 0;
   WOBJ_CatalogCurRecLen(cat) = size;
   WOBJ_CatalogCurRecModified(cat) = 1;
   RETURN32(pos)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogAddRecord(Catalog obj, int size, int pos)
void CatalogAddRecord2(Var stack[]) // (Lwaba/io/Catalog;II)I
{
   WObject cat;
   DmOpenRef dmRef;
   VoidHand recH;
   uint16 pos;
   int32 size;

   cat = stack[0].obj;
   size = stack[1].intValue;
   recClose(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)
   pos = (uint16)stack[2].intValue;//DmNumRecords(dmRef);
   recH = DmNewRecord(dmRef, &pos, size);
   if (!recH)
      RETURN32(-1)
   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recH;
   WOBJ_CatalogCurRecOffset(cat) = 0;
   WOBJ_CatalogCurRecLen(cat) = size;
   WOBJ_CatalogCurRecModified(cat) = 1;
   RETURN32(pos)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogResizeRecord(Catalog obj, int size)
void CatalogResizeRecord(Var stack[]) // (Lwaba/io/Catalog;I)Z
{
   WObject cat;
   DmOpenRef dmRef;
   VoidHand recH;
   int32 pos, size;

   cat = stack[0].obj;
   size = stack[1].intValue;
   if (size <= 0)
      RETURN32(0)
   pos = WOBJ_CatalogCurRecPos(cat);
   if (pos == -1)
      RETURN32(0)
   //recClose(cat); can't close record here before a resizerecord
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)

   recH = DmResizeRecord(dmRef, (uint16)pos, (uint32)size);
   if (recH == NULL)
      RETURN32(0)

   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recH;
   WOBJ_CatalogCurRecLen(cat) = size;
   WOBJ_CatalogCurRecModified(cat) = 1;
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogClose(Catalog obj)
void CatalogClose(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat;

   cat = stack[0].obj;
   RETURN32(catalogClose(cat,1))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogDelete(Catalog obj)
void CatalogDelete(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat;
   DmOpenRef dmRef;

   cat = stack[0].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   catalogClose(cat,0); // don't let it call DmCloseDatabase bc the delete method needs it open
   RETURN32(DmDeleteDatabase(dmRef) == 0)
}
//////////////////////////////////////////////////////////////////////////
static void ulongToChars(uint32 l, char *buf)
{
   buf[0] = LOWER8(l >> 24);
   buf[1] = LOWER8(l >> 16);
   buf[2] = LOWER8(l >> 8);
   buf[3] = LOWER8(l);
}

typedef struct
{
   int count;
   int index;
   WObject *strings;
   char name[32+10];
   TCHAR uniName[32];
   uint32 type,crtr;
} CLC_Vars;

static byte clcAddFile(TCHAR *fileName, void *userVars)
{
   uint32 dbType, dbCreator;
   int len;
   CLC_Vars *vars = (CLC_Vars *)userVars;
   byte end = (byte)(vars->index == vars->count); // stop reason

   DmGetDatabaseTypeCreatorCount(fileName, &dbType, &dbCreator, NULL);

   if (vars->crtr != 0 && dbCreator != vars->crtr) // guich@330_17
      return end;
   if (vars->type != 0 && dbType != vars->type) // guich@330_17
      return end;

   splitPath(fileName, NULL, vars->uniName);

   len = (*vmGlobals->unicode2ascii)(vars->name, vars->uniName, sizeof(vars->uniName));
   while (len > 0 && vars->name[len-1] != '.') len--; // remove the .pdb

   ulongToChars(dbCreator, &vars->name[len]);
   len += 4;
   vars->name[len++] = '.';
   ulongToChars(dbType, &vars->name[len]);
   len += 4;
   vars->name[len++] = 0;

   vars->strings[vars->index++] = (*vmGlobals->createString)(vars->name);

   return end;
}
// waba/io/NativeMethods String []catalogListCatalogs(int creatorId, int type)
void CatalogListCatalogs(Var stack[]) // (II)[Ljava/lang/String;
{
   WObject stringArray;
   CLC_Vars vars;

   vars.crtr = stack[0].intValue; // guich@330_17
   vars.type = stack[1].intValue;

   // first count how many are out there
   vars.count = DmListDatabases(NULL, NULL, 1);
   if (vars.count <= 0) RETURN32(0)
   stringArray = (*vmGlobals->createArrayObject)(1, vars.count);
   if ((*vmGlobals->pushObject)(stringArray) == -1)
      RETURN32(0)

   vars.index = 0;
   vars.strings = (WObject *)WOBJ_arrayStart(stringArray);
   DmListDatabases(&clcAddFile, &vars, 1);

   (*vmGlobals->popObject)(); // stringArray
   RETURN32(stringArray)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogDeleteRecord(Catalog obj)
void CatalogDeleteRecord(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat;
   int32 pos;
   DmOpenRef dmRef;

   cat = stack[0].obj;
   pos = WOBJ_CatalogCurRecPos(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   if (pos == -1)
      RETURN32(0)
   recClose(cat);
   // NOTE: Was DmDeleteRecord() but this didn't match with CE
   // and was too difficult to use in actual programs since it
   // didn't delete the record immediately
   RETURN32(DmRemoveRecord(dmRef, (uint16)pos) == 0)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetRecordCount(Catalog obj)
void CatalogGetRecordCount(Var stack[]) // (Lwaba/io/Catalog;)I
{
   WObject cat;
   DmOpenRef dmRef;

   cat = stack[0].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   RETURN32((dmRef == 0)?-1:DmNumRecords(dmRef))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetRecordSize(Catalog obj)
void CatalogGetRecordSize(Var stack[]) // (Lwaba/io/Catalog;)I
{
   WObject cat;
   int32 pos;
   DmOpenRef dmRef;

   cat = stack[0].obj;
   pos = WOBJ_CatalogCurRecPos(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   RETURN32((dmRef == 0 || pos == -1)?-1:WOBJ_CatalogCurRecLen(cat))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogIsOpen(Catalog obj)
void CatalogIsOpen(Var stack[]) // (Lwaba/io/Catalog;)Z
{
   WObject cat;
   DmOpenRef dmRef;

   cat = stack[0].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   RETURN32(dmRef != 0)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static boolean catalogSetRecordPos(Catalog obj, int pos)
void CatalogSetRecordPos(Var stack[]) // (Lwaba/io/Catalog;I)Z
{
   WObject cat;
   int32 pos, count;
   DmOpenRef dmRef;
   VoidHand recH;

   cat = stack[0].obj;
   pos = stack[1].intValue;
   recClose(cat);
   if (pos < 0)
      RETURN32(0)
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   count = (int32)DmNumRecords(dmRef);
   if (pos >= count)
      RETURN32(0)
   recH = DmGetRecord(dmRef, (uint16)pos);
   if (!recH)
      RETURN32(0)

   WOBJ_CatalogCurRecPos(cat) = pos;
   WOBJ_CatalogCurRecHandle(cat) = recH;
   WOBJ_CatalogCurRecPtr(cat) = recH;
   WOBJ_CatalogCurRecOffset(cat) = 0;
   WOBJ_CatalogCurRecLen(cat) = DmGetRecordSize(dmRef, (uint16)pos);
   WOBJ_CatalogCurRecModified(cat) = 0;
   RETURN32(1)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogReadBytes(Catalog obj, byte buf[], int start, int count)
void CatalogReadBytes(Var stack[]) // (Lwaba/io/Catalog;[BII)I
{
   catalogReadWriteBytes(stack, 1);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogSkipBytes(Catalog obj, int count)
void CatalogSkipBytes(Var stack[]) // (Lwaba/io/Catalog;I)I
{
   WObject cat;
   int32 count, pos, offset;
   DmOpenRef dmRef;

   cat = stack[0].obj;
   count = stack[1].intValue;
   pos = WOBJ_CatalogCurRecPos(cat);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)
   if (pos == -1)
      RETURN32(-1)
   offset = WOBJ_CatalogCurRecOffset(cat)+count;
   if (offset < 0 || offset > WOBJ_CatalogCurRecLen(cat)) // guich@210_11
      RETURN32(-1)
   WOBJ_CatalogCurRecOffset(cat) = offset;    // guich@230_23: changed count to offset
   RETURN32(count)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogWriteBytes(Catalog obj, byte buf[], int start, int count)
void CatalogWriteBytes(Var stack[]) // (Lwaba/io/Catalog;[BII)I
{
   catalogReadWriteBytes(stack, 0);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogInspectRecord(Catalog obj, byte buf[], int recPosition)
void CatalogInspectRecord(Var stack[]) // (Lwaba/io/Catalog;[BI)I
{
   WObject cat, byteArray;
   int32 count, pos;
   int32 size;
   uchar *bytes;
   DmOpenRef dmRef;
   VoidHand recH;

   cat = stack[0].obj;
   byteArray = stack[1].obj;
   pos = stack[2].intValue;
   count = WOBJ_arrayLen(byteArray);
   bytes = (uchar *)WOBJ_arrayStart(byteArray);

   // gets the record
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(-1)

   size = 0;
   recH = DmQueryRecord(dmRef, (uint16)pos, &size);
   if (!recH)
      RETURN32(-1)
   if (count > size) count = size;

   // transfers the data. ps: i did not made a direct reference to the record because i dont have where to store the memhandle params.
   xmemmove(bytes, recH, count);
   xfree(recH);

   RETURN32(count)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static byte catalogGetRecordAttributes(Catalog obj, int recordPos)
void CatalogGetRecordAttributes(Var stack[]) // (Lwaba/io/Catalog;I)B
{
   WObject cat;
   int32 pos;
   DmOpenRef dmRef;
   byte attr;
   Err err;

   cat = stack[0].obj;
   pos = stack[1].intValue; // guich@400_42
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)
   if (pos < 0)
      RETURN32(0)
   err = DmGetRecordAttributes(dmRef, (uint16)pos, &attr);
   RETURN32(err?-1:attr) // returns byte
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogSetRecordAttributes(Catalog obj, int recordPos, byte attr)
void CatalogSetRecordAttributes(Var stack[]) // (Lwaba/io/Catalog;IB)V
{
   WObject cat;
   int32 pos;
   DmOpenRef dmRef;
   byte attr;

   cat = stack[0].obj;
   pos = stack[1].intValue; // guich@400_42
   attr = (byte)stack[2].intValue;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef != 0 && pos >= 0)
      DmSetRecordAttributes (dmRef, (uint16)pos, attr);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetAttributes(Catalog obj)
void CatalogGetAttributes(Var stack[]) // (Lwaba/io/Catalog;)I
{
   WObject cat;
   DmOpenRef dmRef;
   int16 attributes;

   cat = stack[0].obj;
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef == 0)
      RETURN32(0)

   DmGetAttributes(dmRef, &attributes);
   RETURN32(attributes)
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogSetAttributes(Catalog obj, int i)
void CatalogSetAttributes(Var stack[]) // (Lwaba/io/Catalog;I)V
{
   WObject cat;
   DmOpenRef dmRef;
   int16 attributes;

   cat = stack[0].obj;
   attributes = LOWER16(stack[1].intValue);
   dmRef = (DmOpenRef)WOBJ_CatalogDmRef(cat);
   if (dmRef != 0)
      DmSetAttributes(dmRef, attributes);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static int catalogGetRecordPos(Catalog obj)
void CatalogGetRecordPos(Var stack[]) // (Lwaba/io/Catalog;)I
{
   WObject cat = stack[0].obj; // class object
   RETURN32(WOBJ_CatalogCurRecPos(cat))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/Catalog static void catalogCreate(Catalog obj)
void CatalogCreate_NoFunc(Var stack[]) // (Lwaba/io/Catalog;)V
{
   (*vmGlobals->debug)("The Catalog() constructor has no effect!");
}

/*===========================================================================*/

#endif // defined PALMOS
