/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_Serial.c,v 1.12 2005/07/24 20:54:45 fdie Exp $ */

#include "superwaba.h"
#include "SW_natives.h"
#include "SW_types.h"

#ifndef NULL
#define NULL 0
#endif

#define DEBUGm(x)  //x   // don't override DEBUG

////////////////////////////////////////////////////////////////////////////
// Java interface constants

enum {
   SERIAL_DEFAULT     = 0x0000,
   SERIAL_IRCOMM      = 0x1000, // guich@320_42
   SERIAL_SIR         = 0x1001,
   SERIAL_USB         = 0x1002,
   SERIAL_BLUETOOTH   = 0x1003 // guich@330_34
};

enum {
   SERIAL_PARITY_NONE = 0,
   SERIAL_PARITY_EVEN = 1,
   SERIAL_PARITY_ODD  = 2
};

#ifdef WIN32  // <<<<<< WIN32 only!
#include <stdio.h>  // for sprintf

 // guich@120
static byte queryRegistry(HKEY key, TCHAR *subkey, TCHAR *name, char *buf, uint32 size)
{
   HKEY handle;
   long ret;
   uint32 type;

   ret = RegOpenKeyEx(key,subkey,0,KEY_READ,&handle);
   if (ret == ERROR_SUCCESS) // error or success? :-P
   {
      type = REG_DWORD;
      RegQueryValueEx(handle,name,NULL,(LPDWORD)&type,(byte*)buf,(LPDWORD)&size);
      RegCloseKey(handle);
      return 1;
   }
   return 0;
}

static void serialPortCreate(WObject port, int32 number, int32 baudRate, int32 bits, int32 parity, int32 stopBits, int32 timeOutValue)
{
   TCHAR buf[20];
   HANDLE h;
   COMMTIMEOUTS timeout;
   DCB dcb;
   int32 i32;

   WOBJ_SerialPortRefNum(port) = NULL;
   WOBJ_SerialPortLastErr(port) = 0;

   switch (number) // guich@320_42
   {
      case SERIAL_IRCOMM:
         if (queryRegistry(HKEY_LOCAL_MACHINE, TEXT("Drivers\\BuiltIn\\IrCOMM"), TEXT("Index"), (char *)&i32, 4))
            number = i32;
         else
            number = 3; // maybe an old device?
         break;
      case SERIAL_SIR:
         if (queryRegistry(HKEY_LOCAL_MACHINE, TEXT("Drivers\\BuiltIn\\IrDA"), TEXT("Index"), (char *)&i32, 4))
            number = i32;
         else
            number = 2;
         break;
      case SERIAL_USB:
         if (queryRegistry(HKEY_LOCAL_MACHINE, TEXT("Drivers\\BuiltIn\\SerialUSB"), TEXT("Index"), (char *)&i32, 4))
            number = i32;
         else
            return;
         break;
      case SERIAL_BLUETOOTH:
      case 0:
         if (queryRegistry(HKEY_LOCAL_MACHINE, TEXT("Drivers\\BuiltIn\\Serial"), TEXT("Index"), (char *)&i32, 4))
            number = i32;
         else
            number = 1;
         break;
   }

   #ifdef WINCE
      wsprintf(buf, TEXT("COM%d:"), number);
   #else
      sprintf(buf, "COM%d:", number);
   #endif
   h = CreateFile(buf, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
   if (h == INVALID_HANDLE_VALUE)
   {
      WOBJ_SerialPortLastErr(port) = GetLastError();
      return;
   }

   // set serial timeouts
   GetCommTimeouts(h, &timeout);
   timeout.ReadIntervalTimeout = 0;
   timeout.ReadTotalTimeoutMultiplier = 5;
   timeout.ReadTotalTimeoutConstant = 100;
   timeout.WriteTotalTimeoutMultiplier = 5;
   timeout.WriteTotalTimeoutConstant = 1000;
   SetCommTimeouts(h, &timeout);

   GetCommState(h, &dcb);
   if (baudRate == 0)
      baudRate = 9600;
   dcb.BaudRate = baudRate;
   dcb.ByteSize = (byte)bits;
	dcb.Parity = (parity == SERIAL_PARITY_NONE) ? NOPARITY:(parity == SERIAL_PARITY_EVEN) ? EVENPARITY:ODDPARITY; // guich@450_29: added oddparity
   if (stopBits == 1)
      dcb.StopBits = ONESTOPBIT;
   else if (stopBits == 2)
      dcb.StopBits = TWOSTOPBITS;

   // NOTE: Handshaking is off by default for WindowsCE and on by
   // default for all other Win32 platforms
   #ifdef WINCE
      dcb.fOutxCtsFlow = 0;
      dcb.fRtsControl = RTS_CONTROL_ENABLE;
   #else
      dcb.fOutxCtsFlow = 1;
      dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
   #endif
   dcb.fOutxDsrFlow = 0;
   dcb.fDtrControl = DTR_CONTROL_DISABLE;
   dcb.fDsrSensitivity = 0;
   dcb.fTXContinueOnXoff = 0;
   dcb.fOutX = 0;
   dcb.fInX = 0;
   dcb.fNull = 0;
   dcb.fAbortOnError = 0;
   if (SetCommState(h, &dcb) == 0)
   {
      CloseHandle(h);
      WOBJ_SerialPortLastErr(port) = GetLastError();
      return;
   }
   WOBJ_SerialPortRefNum(port) = h;
}

static byte serialPortClose(WObject port)
{
   HANDLE h;
   BOOL status;

   WOBJ_SerialPortLastErr(port) = 0;
   h = WOBJ_SerialPortRefNum(port);
   if (h == NULL)
      return 0;
   status = CloseHandle(h);
   WOBJ_SerialPortRefNum(port) = NULL;
   if (!status)
   {
      WOBJ_SerialPortLastErr(port) = GetLastError();
      return 0;
   }
   return 1;
}

void serialPortDestroy(WObject port)
{
   serialPortClose(port);
}

static int32 serialPortSetFlowControl(Var stack[])
{
	HANDLE h;
	WObject port;
	int32 flowOn;
	DCB dcb;

	port = stack[0].obj;
   WOBJ_SerialPortLastErr(port) = 0;
	flowOn = stack[1].intValue;
	h = WOBJ_SerialPortRefNum(port);
	if (h == NULL)
		return 0;
	GetCommState(h, &dcb);
	if (flowOn)
	{
		dcb.fOutxCtsFlow = 1;
		dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
	}
	else
	{
		dcb.fOutxCtsFlow = 0;
		dcb.fRtsControl = RTS_CONTROL_ENABLE;
	}
	if (SetCommState(h, &dcb) == 0)
   {
      WOBJ_SerialPortLastErr(port) = GetLastError();
		return 0;
   }
	return 1;
}

static int32 serialPortReadCheck(Var stack[]) // guich@510_10
{
   DWORD dwErrors; // guich@320_41
   COMSTAT Stat;
	WObject port;
   HANDLE h;

	port = stack[0].obj;
   WOBJ_SerialPortLastErr(port) = 0;
	h = WOBJ_SerialPortRefNum(port);
	if (h == NULL)
		return -1;

   if(!ClearCommError(h, &dwErrors, &Stat))
   {
      WOBJ_SerialPortLastErr(port) = dwErrors;
      return -1;
   }
   return Stat.cbInQue;
}

static int32 serialPortSetReadTimeout(Var stack[])
{
	HANDLE h;
	WObject port;
	int32 millis;
	COMMTIMEOUTS timeout;

	port = stack[0].obj;
   WOBJ_SerialPortLastErr(port) = 0;
	millis = stack[1].intValue;
	h = WOBJ_SerialPortRefNum(port);
	if (millis < 0 || h == NULL)
		return 0;
	GetCommTimeouts(h, &timeout);
	if (millis == 0)
	{
		timeout.ReadIntervalTimeout = MAXDWORD;
		timeout.ReadTotalTimeoutMultiplier = 0;
		timeout.ReadTotalTimeoutConstant = 0;
	}
	else
	{
		timeout.ReadIntervalTimeout = 0;
		timeout.ReadTotalTimeoutMultiplier = 5;
		timeout.ReadTotalTimeoutConstant = millis;
	}
	if (SetCommTimeouts(h, &timeout))
   {
      WOBJ_SerialPortLastErr(port) = GetLastError();
		return 0;
   }
	return 1;
}

static int32 serialPortReadWriteBytes(Var stack[], int isRead)
{
   WObject port, byteArray;
   int32 start, count;
   uchar *bytes;
   HANDLE h;
   DWORD numRW;

   port = stack[0].obj;
   WOBJ_SerialPortLastErr(port) = 0;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   h = WOBJ_SerialPortRefNum(port);
   if (h == NULL)
      return -1; // port not open
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
   {
      WOBJ_SerialPortLastErr(port) = -14;
      return -1; // array null or range invalid
   }
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   if (isRead)
      ReadFile(h, (LPVOID)&bytes[start], count, &numRW, NULL);
   else
      WriteFile(h, (LPVOID)&bytes[start], count, &numRW, NULL);

   return numRW;
}

////////////////////////////////////////////////////////////////////////////
//>>>PGR:  Non WIN32 system needs to be implemented
//         See: http://www.easysw.com/~mike/serial/serial.html

#elif defined(POSIX)

#include <features.h>
#include <stdio.h>   /* Standard input/output definitions */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */

static void serialPortCreate(WObject port, int32 number, int32 baudRate, int32 bits, int32 parity, int32 stopBits, int32 timeOutValue)
{
   char dev[16];
   struct termios options;
   int fd;

   WOBJ_SerialPortLastErr(port) = 0;
   WOBJ_SerialPortRefNum(port) = 0;

   // fdie@500 Currently only Linux serial ports are supported
   if (number < 0 || number >= SERIAL_IRCOMM)
   {
      WOBJ_SerialPortLastErr(port) = EPERM;
   }
   sprintf(dev, "/dev/ttyS%d", number);

   fd = open(dev, O_RDWR | O_NOCTTY /*| O_NDELAY*/);
   if (fd <= 0)
   {
      perror(dev);
      WOBJ_SerialPortLastErr(port) = EBADF;
      return;
   }
/*
   if (fcntl(fd, F_SETFL, 0) < 0)
   {
      WOBJ_SerialPortLastErr(port) = errno;
      return;
   }
*/

   // Get the current options for the port...
   if (tcgetattr(fd, &options) < 0)
   {
      WOBJ_SerialPortLastErr(port) = errno;
      return;
   }

#define CASE(x)   case x: baudRate = B##x; break;
   switch (baudRate)
   {
      CASE(50)    CASE(75)    CASE(110)   CASE(134)   CASE(150)   CASE(200)
      CASE(300)   CASE(600)   CASE(1200)  CASE(1800)  CASE(2400)  CASE(4800)
      CASE(9600)  CASE(19200) CASE(38400) CASE(57600) CASE(115200)
      default: baudRate = B0; /* hang up */
   }
   // Set the baud rate
   if (cfsetispeed(&options, baudRate) < 0)
   {
      WOBJ_SerialPortLastErr(port) = errno;
      return;
   }
   if (cfsetospeed(&options, baudRate) < 0)
   {
      WOBJ_SerialPortLastErr(port) = errno;
      return;
   }
   DEBUGm(printf ("set baudRate multiplier = %d\n", baudRate));

   options.c_lflag &= ~ICANON;
   options.c_cc[VMIN] = 0;
   options.c_cc[VTIME] = timeOutValue;
   DEBUGm(printf ("set timeout = %d\n", options.c_cc[VTIME]));

   /* Set character size.  csize variable is an integer holding the desired character size. */
   options.c_cflag &= ~(CSIZE);  /* clear bits used for char size */
   switch (bits)
   {
      case 8:
         options.c_cflag |= CS8;
         break;
      case 7:
         options.c_cflag |= CS7;
         break;
      case 6:
         options.c_cflag |= CS6;
         break;
      case 5:
         options.c_cflag |= CS5;
         break;
      default:
         fprintf(stderr, "wrong number of bits: %d\n", bits);
         WOBJ_SerialPortLastErr(port) = EINVAL;
         return;
   }
   DEBUGm(printf ("%d data bits\n", bits));

   /* set parity -- parity variable can be an enum type */
   switch (parity)
   {
      case SERIAL_PARITY_NONE:
         options.c_cflag &= ~(PARENB);  /* clear parity enable */
         options.c_iflag &= ~(INPCK);   /* disable input parity checking */
         break;
      case SERIAL_PARITY_ODD:
         options.c_cflag |= (PARODD | PARENB);  /* enable parity, set to ODD */
         options.c_iflag |= INPCK;              /* enable input parity checking */
         break;
      case SERIAL_PARITY_EVEN:
         options.c_cflag |= PARENB;     /* enable parity, default is EVEN */
         options.c_cflag &= ~(PARODD);  /* ensure PARODD is clear */
         options.c_iflag |= INPCK;      /* enable input parity checking */
         break;
   }

   if (stopBits == 2)
   {
      options.c_cflag |= CSTOPB;
   }
   else
   {
      options.c_cflag &= ~CSTOPB; // default is 1 stop bit
   }

   // Set the new options for the port...
   if (tcsetattr(fd, TCSADRAIN, &options) < 0)
   {
      WOBJ_SerialPortLastErr(port) = errno;
      return;
   }

   DEBUGm(printf ("serialPortCreate %s succeeded\n", dev));
   WOBJ_SerialPortRefNum(port) = (WObject)fd;
}

static byte serialPortClose(WObject port)
{
   if (port == NULL)
   {
      return WOBJ_SerialPortLastErr(port) = EFAULT;
   }
   int fd = (int)WOBJ_SerialPortRefNum(port);
   if (fd <= 0)
   {
      return WOBJ_SerialPortLastErr(port) = EBADF;
   }
/*
   if (tcdrain(fd) < 0) // wait until output buffer is empty
   {
      return WOBJ_SerialPortLastErr(port) = errno;
   }
*/
   return WOBJ_SerialPortLastErr(port) = ((close(fd) == 0) ? 0:errno);
}

void serialPortDestroy(WObject port)
{
   serialPortClose(port);
}

static int32 serialPortSetFlowControl(Var stack[])
{
   struct termios options;
   int fd;
   WObject port = stack[0].obj;
   int rcts = stack[1].intValue;

   if (port == NULL)
   {
      return WOBJ_SerialPortLastErr(port) = EFAULT;
   }
   WOBJ_SerialPortLastErr(port) = 0;

   fd = (int)WOBJ_SerialPortRefNum(port);
   if (fd <= 0)
   {
      return WOBJ_SerialPortLastErr(port) = EBADF;
   }

   // Get the current options for the port...
   if (tcgetattr(fd, &options) < 0)
   {
      return WOBJ_SerialPortLastErr(port) = errno;
   }

   options.c_cflag &= ~CRTSCTS;
   if (rcts)
   {
      options.c_cflag |= CRTSCTS;
   }

   // Set the new options for the port...
   if (tcsetattr(fd, TCSADRAIN, &options) < 0)
   {
      return WOBJ_SerialPortLastErr(port) = errno;
   }

   DEBUGm(printf ("set flow control on desc=%d succeeded\n", fd));
   return 1;
}

static int32 serialPortReadCheck(Var stack[])
{
   return 0;
}

static int32 serialPortSetReadTimeout(Var stack[])
{
   struct termios options;
   int fd;
   WObject port = stack[0].obj;
   int timeout = stack[1].intValue;

   if (port == NULL)
   {
      return WOBJ_SerialPortLastErr(port) = EFAULT;
   }
   WOBJ_SerialPortLastErr(port) = 0;

   fd = (int)WOBJ_SerialPortRefNum(port);
   if (fd <= 0)
   {
      return WOBJ_SerialPortLastErr(port) = EBADF;
   }

   // Get the current options for the port...
   if (tcgetattr(fd, &options) < 0)
   {
      return WOBJ_SerialPortLastErr(port) = errno;
   }

   //options.c_cc[VMIN] = 0; already set
   options.c_cc[VTIME] = timeout/100; // convert millis to tenth of seconds

   // Set the new options for the port...
   if (tcsetattr(fd, TCSADRAIN, &options) < 0)
   {
      return WOBJ_SerialPortLastErr(port) = errno;
   }

   DEBUGm(printf ("set read timeout(%d) on desc=%d succeeded\n", timeout, fd));
   return 1;
}

static int32 serialPortReadWriteBytes(Var stack[], int isRead)
{
   WObject port, byteArray;
   int32 start, count;
   uchar *bytes;
   int fd, num;

   port = stack[0].obj;
   WOBJ_SerialPortLastErr(port) = 0;

   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;

   fd = (int)WOBJ_SerialPortRefNum(port);
   if (fd <= 0)
   {
      return WOBJ_SerialPortLastErr(port) = EBADF;
   }

   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
   {
      return WOBJ_SerialPortLastErr(port) = -14;   // array null or range invalid
   }
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   if (isRead)
   {
      if ((num = read(fd, &bytes[start], count)) < 0)
      {
         return WOBJ_SerialPortLastErr(port) = errno;
      }
      DEBUGm(printf ("%d bytes read on desc=%d succeeded\n", num, fd));
   }
   else
   {
      if ((num = write(fd, &bytes[start], count)) < 0)
      {
         return WOBJ_SerialPortLastErr(port) = errno;
      }
      DEBUGm(printf ("%d bytes written on desc=%d succeeded\n", num, fd));
   }

   return num;
}

#elif defined(PALMOS)

static void serialPortCreate(WObject port, int32 number, int32 baudRate, int32 bits, int32 parity, int32 stopBits, int32 timeOutValue)
{
   SerSettingsType settings;
   UInt refNum;
   UInt16 sizePtr;
   Err err;

   WOBJ_SerialPortLastErr(port) = 0;

   WOBJ_SerialPortRefNum(port) = -1;
   WOBJ_SerialPortNumber(port) = number;

   // default baud rate
   if (baudRate == 0)
      baudRate = 9600;

   WOBJ_SerialPortNewSerialMgr(port) = true;

   switch (number)
   {
      case SERIAL_SIR:
         // use port serPortIrPort and call IrDA Enable
         UNLOCK_MEMORY; // guich@400_27: released semaphore in all ports to avoid reset in Palm OS 5.
         err = SrmOpen(serPortIrPort, baudRate, &refNum);
         LOCK_MEMORY;
         if (err != 0)
         {
            WOBJ_SerialPortLastErr(port) = err;
            return;
         }
         UNLOCK_MEMORY;
         err = SrmControl(refNum, srmCtlIrDAEnable, 0, 0 );
         LOCK_MEMORY;
         if (err != 0)
         {
            WOBJ_SerialPortLastErr(port) = err;
            return;
         }
         break;
      case SERIAL_IRCOMM:
         // use port 'ircm'
         UNLOCK_MEMORY;
         err = SrmOpen( 'ircm', baudRate, &refNum);
         LOCK_MEMORY;
         if (err != 0)
         {
            WOBJ_SerialPortLastErr(port) = err;
            return;
         }
         break;
      case SERIAL_USB:
         // use port usbLibSrmDriverIDDesktop
         UNLOCK_MEMORY;
         // todo@ usbLibSrmDriverIDDesktop? can't be found in sdk-5r4
         err = 1;//SrmOpen(usbLibSrmDriverIDDesktop, baudRate, &refNum);
         LOCK_MEMORY;
         if (err != 0)
         {
            WOBJ_SerialPortLastErr(port) = err;
            return;
         }
         break;
      case SERIAL_DEFAULT:
         // use port serPortCradlePort
         UNLOCK_MEMORY;
         err = SrmOpen(serPortCradlePort, baudRate, &refNum);
         LOCK_MEMORY;
         if (err != 0)
         {
            WOBJ_SerialPortLastErr(port) = err;
            return;
         }
         break;
      // guich@330_34: bluetooth
      case SERIAL_BLUETOOTH:
         UNLOCK_MEMORY;
         err = SrmOpen('rfcm', baudRate, &refNum); // a dialog appears here
         LOCK_MEMORY;
         // use port specified by user
         if (err != 0)
         {
            WOBJ_SerialPortLastErr(port) = err;
            return;
         }
         break;
      // andrzejjan@330_34: Separated the SerialPort_DEFAULT and default cases to support Bluetooth and other ports
      default:
         UNLOCK_MEMORY;
         err = SrmOpen(number, baudRate, &refNum);
         LOCK_MEMORY;
         // use port specified by user
         if (err != 0)
         {
            WOBJ_SerialPortLastErr(port) = err;
            return;
         }
         break;
   }
   settings.baudRate = baudRate;
   settings.flags = 0;

   // no handshake via SIR
   if (number != SERIAL_SIR)
      settings.flags |= serSettingsFlagRTSAutoM | serSettingsFlagCTSAutoM;
   if (bits == 8)
      settings.flags |= serSettingsFlagBitsPerChar8;
   else if (bits == 7)
      settings.flags |= serSettingsFlagBitsPerChar7;
   else if (bits == 6)
      settings.flags |= serSettingsFlagBitsPerChar6;
   else if (bits == 5)
      settings.flags |= serSettingsFlagBitsPerChar5;
// if (parity != 0)
//    settings.flags |= serSettingsFlagParityEvenM;

   if (parity != 0)
   {
      settings.flags |= serSettingsFlagParityOnM;
      settings.flags |= (parity==1)?serSettingsFlagParityEvenM:serSettingsFlagXonXoffM;
   }

   if (stopBits == 1)
      settings.flags |= serSettingsFlagStopBits1;
   else if (stopBits == 2)
      settings.flags |= serSettingsFlagStopBits2;
   settings.ctsTimeout = timeOutValue;

   // do settings
   sizePtr = sizeof(Int32);
   UNLOCK_MEMORY; // guich@401_39
   err = SrmControl(refNum, srmCtlSetBaudRate,  &settings.baudRate, &sizePtr);
   if (err == 0)
      err = SrmControl(refNum, srmCtlSetFlags,  &settings.flags,  &sizePtr);
   if (err == 0)
      err = SrmControl(refNum, srmCtlSetCtsTimeout,&settings.ctsTimeout,   &sizePtr);
   LOCK_MEMORY;
   if (err)
   {
      WOBJ_SerialPortLastErr(port) = err;
      SrmClose(refNum);
      return;
   }

   WOBJ_SerialPortRefNum(port) = (int32)refNum;
   WOBJ_SerialPortTimeout(port) = (*vmGlobals->millisToTicks)(100);
}

static int32 serialPortClose(WObject port)
{
   UInt refNum;
   int32 portNumber = WOBJ_SerialPortNumber(port);
   int32 ret;

   if (WOBJ_SerialPortRefNum(port) == -1)
      return 0;

   refNum = (UInt)WOBJ_SerialPortRefNum(port);

   WOBJ_SerialPortRefNum(port) = -1;
   UNLOCK_MEMORY; // guich@401_39
   // new serial mgr
   SrmSendWait(refNum);
   // SIR Port
   if (portNumber == SERIAL_SIR )
      SrmControl(refNum, srmCtlIrDADisable, 0, 0);
   SrmSetReceiveBuffer(refNum, NULL, 0); // guich@200b4_184
   ret = SrmClose(refNum) == 0;
   LOCK_MEMORY;
   return ret;
}

void serialPortDestroy(WObject port)
{
   serialPortClose(port);
}

static int32 serialPortSetFlowControl(Var stack[])
{
   WObject port;
   int32 flowOn;
   UInt refNum;
   SerSettingsType settings;
   UInt16 sizePtr;
   int32 v = 0;

   port = stack[0].obj;
   flowOn = stack[1].intValue;

   WOBJ_SerialPortLastErr(port) = 0;

   if (WOBJ_SerialPortRefNum(port) == -1)
   {
      WOBJ_SerialPortLastErr(port) = -10;
      return 0;
   }
   refNum = (UInt)WOBJ_SerialPortRefNum(port);

   if (WOBJ_SerialPortNumber(port) != SERIAL_SIR)
   {
      UNLOCK_MEMORY; // guich@401_39
      sizePtr = sizeof(Int32);
      if (SrmControl(refNum, srmCtlGetFlags, &settings.flags, &sizePtr) != 0)
         goto err;

      if (flowOn == 1)
         settings.flags |= srmSettingsFlagRTSAutoM | srmSettingsFlagCTSAutoM;
      else
         settings.flags &= 0xffff - (srmSettingsFlagRTSAutoM | srmSettingsFlagCTSAutoM);

      if ( SrmControl(refNum, srmCtlSetFlags, &settings.flags, &sizePtr) != 0)
         goto err;
   }
   v = 1;
err:
   LOCK_MEMORY;
   return v;
}

static int32 serialPortReadCheck(Var stack[])
{
   WObject port;
   UInt refNum;
   ULong numBytes;
   Err err;

   port = stack[0].obj;
   if (WOBJ_SerialPortRefNum(port) == -1)
   {
      WOBJ_SerialPortLastErr(port) = -10;
      return -1; // port not open
   }
   refNum = (UInt)WOBJ_SerialPortRefNum(port);
   WOBJ_SerialPortLastErr(port) = 0;

   UNLOCK_MEMORY; // guich@401_39
   err = SrmReceiveCheck(refNum, &numBytes);
   LOCK_MEMORY;
   if (err != 0)  // terryf@200b4_179: changed SrmSendCheck to SrmReceiveCheck
   {
      WOBJ_SerialPortLastErr(port) = err;
      return -1;
   }
   return numBytes;
}

static int32 serialPortSetReadTimeout(Var stack[])
{
   WObject port;
   int32 millis;

   port = stack[0].obj;
   millis = stack[1].intValue;
   WOBJ_SerialPortLastErr(port) = 0;

   if (millis < 0 || WOBJ_SerialPortRefNum(port) == -1)
      return 0;

   WOBJ_SerialPortTimeout(port) = (*vmGlobals->millisToTicks)(millis);
   return 1;
}

static int32 serialPortReadWriteBytes(Var stack[], int isRead)
{
   WObject port, byteArray;
   int32 start, count;
   byte *bytes, *bytes0, *temp;
   ULong numRW;
   UInt refNum;
   Err err = 0;
   VoidHand tempH;
   int32 timeout, portNumber;

   port = stack[0].obj;
   byteArray = stack[1].obj;
   start = stack[2].intValue;
   count = stack[3].intValue;
   WOBJ_SerialPortLastErr(port) = 0;

   if (WOBJ_SerialPortRefNum(port) == -1)
   {
      WOBJ_SerialPortLastErr(port) = err;
      return -1; // port not open
   }
   if ((*vmGlobals->arrayRangeCheck)(byteArray, start, count) == 0)
   {
      WOBJ_SerialPortLastErr(port) = -14;
      return -1; // array null or range invalid
   }
   bytes = (uchar *)WOBJ_arrayStart(byteArray);
   bytes = &bytes[start];
   refNum = (UInt)WOBJ_SerialPortRefNum(port);
   timeout = WOBJ_SerialPortTimeout(port);
   portNumber = WOBJ_SerialPortNumber(port);
   tempH = MemHandleNew(count); // note: we can't use xmalloc bc the buffer would be created in the global memory, which is exactly what we need to avoid reference from now on.
   if (!tempH)
   {
      if (isRead) // guich@512_8: if isRead, try to allocate smaller block sizes so we read less than requested, but read something!
      {
         while (1)
         {
            count -= 512;
            if (count < 256) break;
            tempH = MemHandleNew(count);
            if (tempH) goto continueReading;
         }
      }
      WOBJ_SerialPortLastErr(port) = -99;
      return -1; // no memory to allocate temp buffer
   }

continueReading:
   temp = bytes;
   bytes = MemHandleLock(tempH);
   bytes0 = bytes; // bytes will be incremented later
   if (!isRead)
   {
      xmemmove(bytes,temp,count); // copy our sending string to a internal buffer, because we're going to release the semaphore, and so, we can't access the user's buffer
   }

   UNLOCK_MEMORY; // release it now, bc PalmOS 5 doesnt like the release/reserve/release/reserve/... approach
   if (isRead)
   {
      numRW = SrmReceive(refNum, (VoidPtr)bytes, (ULong)count, timeout, &err);
   }
   else
   {
      if (portNumber == SERIAL_SIR)
      {
         // disable rx via srmcontrol
         SrmControl(refNum, srmCtlRxDisable, 0, 0);
      }
      numRW = SrmSend(refNum, (VoidPtr)bytes, (ULong)count, &err);
      if (portNumber == SERIAL_SIR)
      {
         // wait till tx finished
         SrmSendWait(refNum);
         // re-enable IRDA
         SrmControl(refNum, srmCtlIrDAEnable, 0, 0); // egr@330_54: Need this for using SIR (new Ser.Man.)
         // disable rx via srmcontrol
         SrmControl(refNum, srmCtlRxEnable, 0, 0);
      }
   }
   LOCK_MEMORY;
   if (err != 0)
   {
      WOBJ_SerialPortLastErr(port) = err;
      SrmClearErr(refNum);
   }
   xmemmove(temp, bytes0, numRW);
   MemHandleUnlock(tempH);
   MemHandleFree(tempH);
   return numRW;
}

#else // ???

static void serialPortCreate(WObject port, int32 number, int32 baudRate, int32 bits, int32 parity, int32 stopBits, int32 timeOutValue)
{
   //>>>PGR:TODO fill in (not yet implemented)
}

static byte serialPortClose(WObject port)
{
   //>>>PGR:TODO fill in (not yet implemented)
   return 0;
}

void serialPortDestroy(WObject port)
{
   serialPortClose(port);
}

static int32 serialPortSetFlowControl(Var stack[])
{
   return 0;
}

static int32 serialPortReadCheck(Var stack[])
{
   return 0;
}

static int32 serialPortSetReadTimeout(Var stack[])
{
   return 0;
}

static int32 serialPortReadWriteBytes(Var stack[], int isRead)
{
   //>>>PGR:TODO fill in (not yet implemented)
   return 0;
}

#endif

//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static void serialPortCreate(SerialPort obj, int number, int baudRate, int bits, boolean parity, int stopBits)
void SerialPortCreate(Var stack[]) // (Lwaba/io/SerialPort;IIIZI)V
{
#ifdef PALMOS  // guich@551_5: fixed the timeout
   int32 timeout = 2 * sysTicksPerSecond;
#else
   int32 timeout = 200;
#endif
   serialPortCreate(stack[0].obj, stack[1].intValue,  stack[2].intValue,  stack[3].intValue,  stack[4].intValue,  stack[5].intValue, timeout);
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static boolean serialPortClose(SerialPort obj)
void SerialPortClose(Var stack[]) // (Lwaba/io/SerialPort;)Z
{
   RETURN32(serialPortClose(stack[0].obj))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static boolean serialPortIsOpen(SerialPort obj)
void SerialPortIsOpen(Var stack[]) // (Lwaba/io/SerialPort;)Z
{
   WObject port = stack[0].obj;
   RETURN32((WOBJ_SerialPortRefNum(port) != NULL))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static boolean serialPortSetFlowControl(SerialPort obj, boolean on)
void SerialPortSetFlowControl(Var stack[]) // (Lwaba/io/SerialPort;Z)Z
{
   RETURN32(serialPortSetFlowControl(stack))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static boolean serialPortSetReadTimeout(SerialPort obj, int millis)
void SerialPortSetReadTimeout(Var stack[]) // (Lwaba/io/SerialPort;I)Z
{
   RETURN32(serialPortSetReadTimeout(stack))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static int serialPortReadBytes(SerialPort obj, byte buf[], int start, int count)
void SerialPortReadBytes(Var stack[]) // (Lwaba/io/SerialPort;[BII)I
{
   RETURN32(serialPortReadWriteBytes(stack, 1))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static int serialPortReadCheck(SerialPort obj)
void SerialPortReadCheck(Var stack[]) // (Lwaba/io/SerialPort;)I
{
   RETURN32(serialPortReadCheck(stack))
}
//////////////////////////////////////////////////////////////////////////
// waba/io/SerialPort static int serialPortWriteBytes(SerialPort obj, byte buf[], int start, int count)
void SerialPortWriteBytes(Var stack[]) // (Lwaba/io/SerialPort;[BII)I
{
   RETURN32(serialPortReadWriteBytes(stack, 0))
}

/*===========================================================================*/
