/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_Time.c,v 1.2 2005/01/05 18:10:32 guich Exp $ */

#include "superwaba.h"
#include "SW_natives.h"

#if defined WIN32
  #if !defined(WINCE)
    #include <time.h>
    #include <sys/types.h>
    #include <sys/timeb.h>
  #endif

#elif defined PALMOS

// POSIX
#else
  #include <sys/time.h>
  #include <time.h>
  #include <unistd.h>
#endif

//////////////////////////////////////////////////////////////////////////
// waba/sys/Time static void timeCreate(Time obj)
void TimeCreate(Var stack[]) // (Lwaba/sys/Time;)V
{
   WObject time = stack[0].obj;

   #if defined(WINCE)
     SYSTEMTIME tm;
     GetLocalTime(&tm);
     WOBJ_TimeYear(time) = tm.wYear;
     WOBJ_TimeMonth(time) = tm.wMonth;
     WOBJ_TimeDay(time) = tm.wDay;
     WOBJ_TimeHour(time) = tm.wHour;
     WOBJ_TimeMinute(time) = tm.wMinute;
     WOBJ_TimeSecond(time) = tm.wSecond;
     WOBJ_TimeMillis(time) = tm.wMilliseconds;
  #elif defined PALMOS
     DateTimeType nowTM;
     TimSecondsToDateTime(TimGetSeconds(), &nowTM);
     WOBJ_TimeYear(time) = nowTM.year;
     WOBJ_TimeMonth(time) = nowTM.month;
     WOBJ_TimeDay(time) = nowTM.day;
     WOBJ_TimeHour(time) = nowTM.hour;
     WOBJ_TimeMinute(time) = nowTM.minute;
     WOBJ_TimeSecond(time) = nowTM.second;
     WOBJ_TimeMillis(time) = (*vmGlobals->getTimeStamp)() % 1000; // guich@200b4_139
  #else
     int millis;
     int secsSince;
     struct tm tm;
     #if defined WIN32
        struct _timeb tb;
        _ftime(&tb);
        secsSince = tb.time;
        millis = tb.millitm;
        tm = *(localtime((time_t *)&secsSince));
     #else
        struct timeval tv;
        struct timezone tz;
        gettimeofday(&tv, &tz);
        secsSince = tv.tv_sec;
        millis = tv.tv_usec/1000;
        localtime_r((time_t *)&secsSince, &tm);
     #endif
     WOBJ_TimeYear(time) = tm.tm_year + 1900;
     WOBJ_TimeMonth(time) = tm.tm_mon + 1;
     WOBJ_TimeDay(time) = tm.tm_mday;
     WOBJ_TimeHour(time) = tm.tm_hour;
     WOBJ_TimeMinute(time) = tm.tm_min;
     WOBJ_TimeSecond(time) = tm.tm_sec;
     WOBJ_TimeMillis(time) = millis;
  #endif
}
