/*********************************************************************************
 *  SuperWaba Virtual Machine, version 5                                         *
 *  Copyright (C) 2000-2005 SuperWaba Ltda.                                      *
 *  All Rights Reserved                                                          *
 *                                                                               *
 *  This library and virtual machine is distributed in the hope that it will     *
 *  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                         *
 *                                                                               *
 *  Please see the software license located at /license.txt for more details.    *
 *                                                                               *
 *  You should have received a copy of the License along with this software;     *
 *  if not, write to                                                             *
 *                                                                               *
 *     SuperWaba Ltda.                                                           *
 *     Rua Padre Leonel Franca, 480, sala 27A                                    *
 *     Instituto Genesis - PUC-RIO                                               *
 *     Rio de Janeiro / RJ - Brazil                                              *
 *     Cep: 22451-000                                                            *
 *     E-mail: licensing@superwaba.com.br                                        *
 *********************************************************************************/
/* $Id: SW_Vm.c,v 1.21 2005/11/13 13:24:50 fdie Exp $ */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#if HAVE_STDLIB_H
#include <stdlib.h>
#endif

#if HAVE_STRING_H
#include <string.h>
#endif

#if HAVE_WAITPID
#include <sys/wait.h>
#endif

#include "superwaba.h"
#include "SW_natives.h"
#include "SDL.h"
#include "SW_utils.h"

#if defined(__SYMBIAN32__)
  #include "epoc_dll.h"
  #include "epoc_time.h"
#else
  #define GPA_TYPE const char *
  #if defined(WINCE)
    #define GPA_IDENT(x) TEXT(#x)
  #else
    #define GPA_IDENT(x) #x
  #endif
#endif

#if defined(WIN32) //***

#define DLL_EXT ".dll"
#define DLerror() "unknown"

#elif defined(PALMOS) && defined(__arm__) //***

#include <pealstub.h>
extern void *pealLoadLibrary68K, *pealGetProcAddress68K;

void *LoadLibrary(const char *p)
{
   struct 
   {
      const char *name;
      DmResType resType;
      DmResID resID;
   } 
   __attribute((packed)) load_library_args;

   load_library_args.name = (char*)Swap32(p);
   load_library_args.resType = Swap32('ARMC');
   load_library_args.resID = Swap16(0x1000);
    
   return (void*)(*gCall68KFuncP)(gEmulStateP, pealLoadLibrary68K, 
                            &load_library_args, sizeof(load_library_args));    
}

void *GetProcAddress(const void *p, const char *sym)
{
   struct 
   {
      const void *module;
      const char *func;
   } 
   __attribute((packed)) get_proc_address_args;

   get_proc_address_args.module = (void*)Swap32(p);
   get_proc_address_args.func = (char *)Swap32(sym);

   return (void*)(*gCall68KFuncP)(gEmulStateP, pealGetProcAddress68K, 
                            &get_proc_address_args, sizeof(get_proc_address_args));
}

#define FreeLibrary 
#define DLerror() "unknown"
#define DLL_EXT ""

#elif defined(__SYMBIAN32__) //***

#define DLL_EXT ".dll"
#define DLerror() "unknown"

#elif !defined(PALMOS) //***

  // POSIX
  extern char **environ;
  #include <unistd.h>
  #include <errno.h>
  #include <time.h>
  #include <sys/time.h>
  #include <dlfcn.h>
  #define LoadLibrary(p) dlopen(p, RTLD_LAZY);
  #define GetProcAddress dlsym
  #define FreeLibrary dlclose
  #define DLerror dlerror
  #define DLL_EXT ".so"
#endif

//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static int vmGetTimeStamp()
void VmGetTimeStamp(Var stack[]) // ()I
{
   RETURN32((*vmGlobals->getTimeStamp)())
}

//////////////////////////////////////////////////////////////////////////
// waba/sys/NativeMethods static void vmSetTimeStamp(waba.sys.Time t)
void VmSetTimeStamp(Var stack[]) // (Lwaba/sys/Time;)V
{
#if defined WIN32
   WObject time = stack[0].obj;
   SYSTEMTIME now;
   now.wYear   = WOBJ_TimeYear(time);
   now.wMonth  = WOBJ_TimeMonth(time);
   now.wDay    = WOBJ_TimeDay(time);
   now.wHour   = WOBJ_TimeHour(time);
   now.wMinute = WOBJ_TimeMinute(time);
   now.wSecond = WOBJ_TimeSecond(time);
   now.wMilliseconds = WOBJ_TimeMillis(time);
   SetSystemTime(&now);

#elif defined __SYMBIAN32__
   WObject time = stack[0].obj;

   epoc_time_set(
      WOBJ_TimeYear(time),
      WOBJ_TimeMonth(time),
      WOBJ_TimeDay(time),
      WOBJ_TimeHour(time),
      WOBJ_TimeMinute(time),
      WOBJ_TimeSecond(time),
      WOBJ_TimeMillis(time)
   );

#elif defined PALMOS
   WObject time = stack[0].obj;
   DateTimeType now;
   int32 seconds;
   now.year   = WOBJ_TimeYear(time);
   now.month  = WOBJ_TimeMonth(time);
   now.day    = WOBJ_TimeDay(time);
   now.hour   = WOBJ_TimeHour(time);
   now.minute = WOBJ_TimeMinute(time);
   now.second = WOBJ_TimeSecond(time);
   seconds = TimDateTimeToSeconds(&now);
   TimSetSeconds(seconds);

#elif !defined __ECOS  // POSIX
   WObject time = stack[0].obj;
   struct tm tm;
   struct timeval tv;
   struct timezone tz;

   tm.tm_year = WOBJ_TimeYear(time);
   tm.tm_mon = WOBJ_TimeMonth(time);
   tm.tm_mday = WOBJ_TimeDay(time);
   tm.tm_hour = WOBJ_TimeHour(time);
   tm.tm_min = WOBJ_TimeMinute(time);
   tm.tm_sec = WOBJ_TimeSecond(time);
   // tm.tm_yday = 0;  // irrelevant
   // tm.tm_wday = 0;  // irrelevant
   tm.tm_isdst = -1;   // unavailable
   gettimeofday(&tv, &tz);  // fill-up tz (buggy for Linux)
   tv.tv_sec = mktime(&tm);
   tv.tv_usec = WOBJ_TimeMillis(time) * 1000;
   settimeofday(&tv, &tz);

#endif
}

//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static int vmExec(String command, String args, int launchCode, boolean wait)
void VmExec(Var stack[]) // (Ljava/lang/String;Ljava/lang/String;IZ)I
{
#if defined WIN32
   WObject pathString, argsString;
   UtfString path, args;
   int32 doWait;
   BOOL status;
   int32 ret;
   PROCESS_INFORMATION pInfo;
   DWORD exitCode;
   #ifdef WINCE
     TCHAR pathBuf[MAX_PATH], argsBuf[MAX_PATH];
   #else
     char argsBuf[MAX_PATH*2];
     STARTUPINFO sInfo;
   #endif

   pathString = stack[0].obj;
   argsString = stack[1].obj;
   doWait = stack[3].intValue;
   // NOTE: can't use static here since we call stringToUtf on both
   // path and args
   path = (*vmGlobals->stringToUtf)(pathString, STU_NULL_TERMINATE);
   if (path.len == 0)
      RETURN32(-1)
   args = (*vmGlobals->stringToUtf)(argsString, STU_NULL_TERMINATE | STU_USE_STATIC);
   #ifdef WINCE
      (*vmGlobals->ascii2unicode)(pathBuf, path.str, MAX_PATH);
      (*vmGlobals->ascii2unicode)(argsBuf, args.str, MAX_PATH);
      status = CreateProcess(pathBuf, argsBuf, NULL, NULL, FALSE,
                             0, NULL, NULL, NULL, &pInfo);
   #else
      xmemzero(&sInfo, sizeof(sInfo));
      sInfo.cb = sizeof(sInfo);
      // guich@511_1: concats first with 2nd cmdline
      if (args.len > 0) xstrprintf(argsBuf,"%s %s",path.str,args.str);
      status = CreateProcess(0, (args.len>0)?argsBuf:path.str, NULL, NULL, FALSE, // guich@511_1: changed to use the 2nd parameter only
                             NORMAL_PRIORITY_CLASS, NULL, NULL, &sInfo, &pInfo);
   #endif
   if (!status)
      RETURN32(-1)
   CloseHandle(pInfo.hThread);
   if (doWait)
   {
      WaitForSingleObject(pInfo.hProcess, INFINITE);
      ret = GetExitCodeProcess(pInfo.hProcess, &exitCode);
   }
   else
   {
      ret = 0;
   }
   CloseHandle(pInfo.hProcess);
   RETURN32(ret)

#elif defined PALMOS
   WObject pathString, argsString;
   UtfString path, args;
   LocalID localID;
   Word launchCode;
   CharPtr s;
   int32 doWait; //RKK20000305
   UInt launchFlags; //RKK20000306
   DWord result; //RKK20000305

   pathString = stack[0].obj;
   argsString = stack[1].obj;
   launchCode = (Word)stack[2].intValue;
   doWait = stack[3].intValue; //RKK20000305
   launchFlags = 0; //RKK20000306

   // NOTE: we could use static here since we call stringToUtf twice
   // but before the second call, the path is no longer needed
   path = (*vmGlobals->stringToUtf)(pathString, STU_NULL_TERMINATE);
   if (path.len == 0)
      RETURN32(-1)
   localID = DmFindDatabase(PALM_CARD(0)  path.str);
   if (!localID)
      RETURN32(-1)
   args = (*vmGlobals->stringToUtf)(argsString, STU_NULL_TERMINATE | STU_USE_STATIC);
   s = NULL;
   if (args.len > 0)
   {
      s = MemPtrNew(args.len + 1);
      StrCopy(s, args.str);
      MemPtrSetOwner(s, 0);
   }
   if (launchCode == 0)
      launchCode = sysAppLaunchCmdNormalLaunch;
   if (doWait == 0) //RKK20000305
      RETURN32(SysUIAppSwitch(PALM_CARD(0) localID, launchCode, s))
   else //RKK20000305
      RETURN32((int32)SysAppLaunch(PALM_CARD(0)  localID, launchFlags, launchCode, s, &result)) //RKK20000305

#elif !defined __ECOS && !defined __SYMBIAN32__ // POSIX
   int32 pid;
   int32 ret;
   //UtfString args;
   WObject pathString = stack[0].obj;
   WObject argsString = stack[1].obj;
   int32 doWait = stack[3].intValue;
   UtfString temp = (*vmGlobals->stringToUtf)(pathString, STU_NULL_TERMINATE  | STU_USE_STATIC);

   if (temp.len == 0) {
      RETURN32(-1)
   }

   pid = fork();
   if (pid == -1) {
      RETURN32(-1);
   }else if (pid == 0) {  // thread pid
      char * argv[4];
      char tmpbuf[MAX_PATH];
      char * argMax = tmpbuf + MAX_PATH;
      char * argCur = tmpbuf;

      argCur = boundedCopy(argCur, temp.str, argMax);
      *argCur++ = ' ';
      if (argCur >= argMax) RETURN32(-1)
      temp = (*vmGlobals->stringToUtf)(argsString, STU_NULL_TERMINATE  | STU_USE_STATIC);
      argCur = boundedCopy(argCur, temp.str, argMax);
      argv[0] = "sh";
      argv[1] = "-c";
      argv[2] = tmpbuf;
      argv[3] = 0;
      execve("/bin/sh", argv, environ);
      exit(127);        // exit the fork
   }
   if (!doWait) {
      RETURN32(0);
   }else {
      while (waitpid(pid, &ret, 0) == -1) {
         if (errno != EINTR) RETURN32(-1);
      }
      RETURN32(ret);
   }
#endif
}

//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static void vmGC()
void VmGC(_unused_ Var stack[]) // ()V
{
   (*vmGlobals->gc)();
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static int vmSetDeviceAutoOff(int seconds)
void VmSetDeviceAutoOff(Var stack[]) // (I)I
{
    int32 d=0;
#if defined(WINCE)
    // dave@400_48
    SystemParametersInfo(SPI_GETBATTERYIDLETIMEOUT, (UINT)0, &d, 0);
    SystemParametersInfo(SPI_SETBATTERYIDLETIMEOUT, (UINT)stack[0].intValue, NULL, 0);
#endif
   RETURN32(d)
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static void vmSleep(int millis)
void VmSleep(Var stack[]) // (I)V
{
   SDL_UpdateRect(SDL_GetVideoSurface(), 0, 0, 0, 0);
   SDL_PumpEvents();  // do your home work b4 to go to bed!
   SDL_Delay(stack[0].intValue);
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static int vmGetDeviceFreeMemory()
void VmGetDeviceFreeMemory(Var stack[]) // ()I
{
   int32 ret = 0;
#if defined(WINCE) // dave@400_48
   STORE_INFORMATION   si;
   GetStoreInformation(&si); // Storage memory
   ret = si.dwFreeSize;
#else
   ret = xFreeMemInfo();
#endif
   RETURN32(ret)
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static void vmInterceptSystemKeys(int keys)
void VmInterceptSystemKeys(Var stack[]) // (I)V
{
   (*vmGlobals->unregisterHotkeys)(); // unregister all
   vmGlobals->interceptSystemKey = stack[0].intValue;
   if (vmGlobals->interceptSystemKey)
      (*vmGlobals->registerHotkeys)();
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static void vmDebug(String s)
void VmDebug(Var stack[]) // (Ljava/lang/String;)V
{
   UtfString s;
   s = (*vmGlobals->stringToUtf)(stack[0].obj, STU_USE_STATIC | STU_NULL_TERMINATE);
   (*vmGlobals->debug)(s.str);
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static void vmClipboardCopy(String s)
void VmClipboardCopy(_unused_ Var stack[]) // (Ljava/lang/String;)V
{
#if defined(PALMOS)
    UtfString s = (*vmGlobals->stringToUtf)(stack[0].obj, STU_USE_STATIC | STU_NULL_TERMINATE);
    ClipboardAddItem(clipboardText,s.str, s.len);
#elif defined(SDL) && !defined(WIN32) // guich@510_9: use the win32 implementations
    //>>>PGR:TODO fill in (not yet implemented)    
#else
    HANDLE hClipData;                           /* handle to clip data  */
    LPTSTR lpClipData;                           /* pointer to clip data */
    WObject charArray;
    uint32 i, len;
    uint16 *chars;

    charArray = WOBJ_StringCharArrayObj(stack[0].obj);
    if (charArray == 0)
        return;
    len = WOBJ_arrayLen(charArray);
#ifdef WINCE
    if (!(hClipData = LocalAlloc(LPTR, (len + 1) * sizeof(TCHAR))))
       return;
    lpClipData = (LPTSTR) hClipData;
#else
    if ((hClipData = GlobalAlloc(GMEM_DDESHARE | GMEM_MOVEABLE , (len + 1) * sizeof(TCHAR))) == 0)
       return;
    if ((lpClipData = GlobalLock(hClipData)) == 0)
    {
       GlobalFree(hClipData);
       return;
    }
#endif
    chars = (uint16 *)WOBJ_arrayStart(charArray);
    for (i = 0; i < len; i++)
       // lets keep the clipboard data in the right format for the system
#ifdef WINCE
       lpClipData[i] = chars[i];
#else
       lpClipData[i] = (char) chars[i];
#endif
    lpClipData[i] = 0;

#ifndef WINCE
   GlobalUnlock (hClipData);
#endif

   /* Clear the current contents of the clipboard, and set
   * the data handle to the new string.
   */
   if (OpenClipboard(vmGlobals->mainHWnd))
   {
      EmptyClipboard();
#ifdef WINCE
      if (!SetClipboardData(CF_UNICODETEXT, hClipData))
      {
         LocalFree(hClipData);
#else
      if (!SetClipboardData(CF_TEXT, hClipData))
      {
         GlobalFree(hClipData);
#endif
      }
      CloseClipboard();
   }
   else
   {
#ifdef WINCE
      LocalFree(hClipData);
#else
      GlobalFree(hClipData);
#endif
   }
#endif
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static String vmClipboardPaste()
void VmClipboardPaste(Var stack[]) // ()Ljava/lang/String;
{
#if defined(PALMOS)
   unsigned short len;
   WObject obj;
   VoidHand ch = ClipboardGetItem(clipboardText, &len);
   if (ch && len)
   {
      obj = (*vmGlobals->createStringFromCharPtr)(MemHandleLock(ch),len);
      MemHandleUnlock(ch);
   } else obj = (*vmGlobals->createString)("");
   RETURN32(obj)
#elif defined(SDL) && !defined(WIN32) // guich@510_9
   //>>>PGR:TODO fill in (not yet implemented)
   WObject obj = (*vmGlobals->createString)("");
#else
   HANDLE hClipData;                           /* handle to clip data  */
   LPTSTR lpClipData;                           /* pointer to clip data */
   WObject obj;

   if (OpenClipboard(vmGlobals->mainHWnd))
   {
      /* get text from the clipboard */
#ifdef WINCE
      if (!(hClipData = GetClipboardData(CF_UNICODETEXT)))
      {
#else
      if ((hClipData = GetClipboardData(CF_TEXT)) == 0)
      {
#endif
         CloseClipboard();
         RETURN32((*vmGlobals->createString)(""))
      }
#ifdef WINCE
      lpClipData = (LPTSTR) hClipData;
//    v.obj = createString(lpClipData);
      obj = (*vmGlobals->createStringFromUnicode)(lpClipData, _tcslen(lpClipData));
#else
      if ((lpClipData = GlobalLock(hClipData)) == 0)
      {
         CloseClipboard();
         RETURN32((*vmGlobals->createString)(""))  // dave@400_48
      }
      obj = (*vmGlobals->createString)(lpClipData);
      GlobalUnlock(hClipData);
#endif
      CloseClipboard();
   } else {
      obj = (*vmGlobals->createString)("");
   }
#endif
   RETURN32(obj)
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static boolean vmAttachLibrary(String name)
void VmAttachLibrary(Var stack[]) // (Ljava/lang/String;)Z
{
   DmOpenRef dmref;
   UtfString s = (*vmGlobals->stringToUtf)(stack[0].obj, STU_USE_STATIC | STU_NULL_TERMINATE);
   // use SUPERWABA_CREATOR_ID instead of 'SWAB'
   uint32 creatorId = vmGlobals->appCreatorId?vmGlobals->appCreatorId:SUPERWABA_CREATOR_ID;

   if (s.len > 5 && s.str[s.len-5] == '.') // guich@200b4_117: has a different creator id?
   {
      creatorId = READ_BE_UINT32(&s.str[s.len-4]);
      s.len -= 5;
      s.str[s.len] =0;
   }

   dmref = (*vmGlobals->findAndOpenDB)(s.str, creatorId, SUPERWABA_TYPE_EXT, 1,0);
   if (dmref != NULL)
   {
      // ok, we found the library.
      int32 insertIndex;
      vmGlobals->classDbList = (DmOpenRef *)(*vmGlobals->expandArray)((char *)vmGlobals->classDbList,sizeof(DmOpenRef),1,&insertIndex);
      if (insertIndex != -1)
      {
         vmGlobals->classDbList[insertIndex] = dmref;
         RETURN32(1) // everything ok
      }
   }
   RETURN32(0)
}
#if defined(PALMOS) && !defined(__arm__)
//#ifdef BUILDING_SW_LIB
// #define SW_LIB_TRAP(trapNum)
//#else
// #define SW_LIB_TRAP(trapNum) SYS_TRAP(trapNum)
//#endif
#include "palm/SW_libDynam.h"

#else //elif !defined(__SYMBIAN32__)

static void *loadLib(UtfString lname, int libPrefix)
{
   void * h;
   
#if defined(PALMOS) && defined(__arm__)
   h = LoadLibrary(lname.str);
#else
   TCHAR path[MAX_PATH];

   // append the user path and ".dll/.so" to the library name
#if defined(__SYMBIAN32__)
   int l = (*vmGlobals->ascii2unicode)(path, "\\system\\Apps\\SuperWaba\\", MAX_PATH);
#else
   int l = (*vmGlobals->ascii2unicode)(path, vmGlobals->appPath, MAX_PATH);
#endif

   // try to prefix with "lib" on Linux
   int l2 = libPrefix ? (*vmGlobals->ascii2unicode)(path+l, "lib", MAX_PATH) : 0;
   int l3 = (*vmGlobals->ascii2unicode)(path+l+l2, lname.str, MAX_PATH);
   (*vmGlobals->ascii2unicode)(path+l+l2+l3, DLL_EXT, MAX_PATH);

   // first search in the user path
   h = LoadLibrary(path);
   if (h == NULL) // not found in user path?
   {
      h = LoadLibrary(path+l); // skip the user path
   }
#endif   
   return h;
}

#endif // !defined(__SYMBIAN32__)

//////////////////////////////////////////////////////////////////////////
// waba/sys/Vm static boolean vmAttachNativeLibrary(String name)
void VmAttachNativeLibrary(Var stack[]) // (Ljava/lang/String;)Z
{
#if defined(PALMOS_NATIVES_SYSLIB)
   DmSearchStateType state;
   LocalID dbID;
   UInt cardNo;
   Err err;
   Boolean first=true;
   UtfString s = (*vmGlobals->stringToUtf)(stack[0].obj, STU_USE_STATIC | STU_NULL_TERMINATE);
   char dbName[32];
   ULong creatorId;
   UInt refnum;

    // search for the database name given
   while (1)
   {
      err = DmGetNextDatabaseByTypeCreator(first, &state, SWLibTypeID, 0, false, PALM_CARD(&cardNo) &dbID);
      if (err == dmErrCantFind)
      {
         break;
      }      
      
      if (DmDatabaseInfo(PALM_CARD(cardNo) dbID, dbName, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &creatorId) == 0 && !xstrncmp(dbName,s.str,s.len))
      {
         // ok, we found the library. try to load it
         err = SysLibFind(s.str, &refnum);
         if (err)
         {
            err = SysLibLoad(SWLibTypeID, creatorId, &refnum);
         }
         if (!err) // loaded ok?
         {
            int insertIndex;
            SWLibOpen(refnum, vmGlobals);
            vmGlobals->nativeLibs = (UInt *)(*vmGlobals->expandArray)((char *)vmGlobals->nativeLibs,sizeof(vmGlobals->nativeLibs[0]),1,&insertIndex);
            if (insertIndex == -1)
            {
               RETURN32(0)
            }
            vmGlobals->nativeLibs[insertIndex] = refnum;
            RETURN32(1) // everything ok.
         }
         else
         {
            char buf[128];
            xstrprintf(buf, "Cannot load library %s: error %d", s.str, err);
            (*vmGlobals->debug)(buf);
            RETURN32(0)
         }
         break;
      }
      first = false;
   }
#else //POSIX

   NativeLibPtr lib;
   int32 insertIndex;
   UtfString s = (*vmGlobals->stringToUtf)(stack[0].obj, STU_USE_STATIC | STU_NULL_TERMINATE);

#ifdef linux
   void *h = loadLib(s, true);
#else
   void *h = NULL;   
#endif
   if (h == NULL)
   {
      h = loadLib(s, false);
   }

   if (h == NULL || (lib = (NativeLibPtr) xmalloc(sizeof(NativeLib))) == NULL) { // can't find library? no more memory?
      #if !defined WIN32  // POSIX
      char buf[300];
      xstrprintf(buf, "Cannot load library %s: %s", s.str, DLerror());
      (*vmGlobals->debug)(buf);
      #endif
      RETURN32(0)
   }
   // lib->ref = h;
   lib->SWLibOpen          = (SWLibOpenProcType)GetProcAddress          (h,GPA_IDENT(SWLibOpen));
   lib->SWLibClose         = (SWLibCloseProcType)GetProcAddress         (h,GPA_IDENT(SWLibClose));
   lib->SWCallPublic       = (SWCallPublicProcType)GetProcAddress       (h,GPA_IDENT(SWCallPublic));
   lib->SWHandleEvent      = (SWHandleEventProcType)GetProcAddress      (h,GPA_IDENT(SWHandleEvent));
   lib->SWInvokeNative     = (SWInvokeNativeProcType)GetProcAddress     (h,GPA_IDENT(SWInvokeNative));
   lib->SWGetNativeFunc    = (SWGetNativeFuncProcType)GetProcAddress    (h,GPA_IDENT(SWGetNativeFunc));
   lib->SWInvokeNativeFunc = (SWInvokeNativeFuncProcType)GetProcAddress (h,GPA_IDENT(SWInvokeNativeFunc));
   
   if (lib->SWLibOpen == NULL) // not a SW Native Library?
   {
      char buf[200];
      xstrprintf(buf,"Cannot load library %s: SWLibOpen entry point not found!", s.str);
      (*vmGlobals->debug)(buf);
      xfree(lib);
      RETURN32(0)
   }

   // insert this library in the nativeLibs array
   vmGlobals->nativeLibs = (NativeLibPtr *)(*vmGlobals->expandArray)((char *)vmGlobals->nativeLibs,sizeof(NativeLibPtr),1,&insertIndex);
   if (insertIndex == -1)
   {
      xfree(lib);
      RETURN32(0)
   }

#if defined(PALMOS) && defined(__arm__)
   {
      uint32 *got_ptr = h; // first uin32 field of PealModule is the GOT   
      lib->ref = Swap32(*got_ptr);
      vmGlobals->tempLibInsertIndex = insertIndex+1;
   }
#elif defined(__SYMBIAN32__)
   lib->ref = (LibRefnum)h;
   vmGlobals->tempLibInsertIndex = insertIndex+1;
#else
   lib->ref = insertIndex + 1;
#endif

   // library loaded and methods initialized
   (*lib->SWLibOpen)(lib->ref, vmGlobals);
   vmGlobals->nativeLibs[insertIndex] = lib;

   RETURN32(1) // everything ok.
#endif //POSIX
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/NativeMethods static byte[] vmGetFile(String name)
void VmGetFile(Var stack[]) // (Ljava/lang/String;)[B
{
   UtfString path;
   WObject pathString,obj;
   uint16 pathLen;
   uchar *p;
   uint32 size=0;

   pathString = stack[0].obj;
   path = (*vmGlobals->stringToUtf)(pathString, STU_USE_STATIC);
   if (path.len == 0)
      RETURN32(0)

   pathLen = (uint16)path.len;
   p = (*vmGlobals->lockWarpRec)(path.str, (uint16)path.len, &size);
   if (p == NULL)
      RETURN32(0)
   // create the byte array object
   obj = (*vmGlobals->createArrayObject)(8,size);
   if (obj != NULL)
      xmemmove((byte*)WOBJ_arrayStart(obj),p,size);
   (*vmGlobals->unlockWarpRec)(p, pathLen);
   RETURN32(obj)
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/NativeMethods static void vmExitAndReboot()
void VmExitAndReboot(_unused_ Var stack[]) // ()V
{
   vmGlobals->reboot = 1;
   (*vmGlobals->postStopEvent)(0);
}
//////////////////////////////////////////////////////////////////////////
// waba/sys/NativeMethods int vmGetSystemKeysPressed()
void VmGetSystemKeysPressed(Var stack[]) // ()I
{
   //>>>PGR:TODO fill in (not yet implemented)
#ifndef WIN32
   RETURN32(0)
#else
   /** These defines are from PalmOS. We must make the return
       compatible values */
#define  keyBitPower       0x0001      // Power key
#define  keyBitPageUp      0x0002      // Page-up
#define  keyBitPageDown    0x0004      // Page-down
#define  keyBitHard1       0x0008      // App #1
#define  keyBitHard2       0x0010      // App #2
#define  keyBitHard3       0x0020      // App #3
#define  keyBitHard4       0x0040      // App #4
#define  keyBitCradle      0x0080      // Button on cradle
#define  keyBitAntenna     0x0100      // Antenna "key" <chg 3-31-98 RM>
#define  keyBitContrast    0x0200      // Contrast key
   // specific to WinCE
#define  keyBitPageLeft    0x1000      // left arrow
#define  keyBitPageRight   0x2000      // right arrow
#define  keyBitAction      0x08000000

   int32 cursorLeft  = (GetAsyncKeyState(VK_LEFT)  & 0x8000) != 0;
   int32 cursorRight = (GetAsyncKeyState(VK_RIGHT) & 0x8000) != 0;
   int32 cursorUp    = (GetAsyncKeyState(VK_UP)    & 0x8000) != 0;
   int32 cursorDown  = (GetAsyncKeyState(VK_DOWN)  & 0x8000) != 0;
#ifdef WINCE // guich@400_72
   int32 f1  = (GetAsyncKeyState(193)  & 0x8000) != 0;
   int32 f2  = (GetAsyncKeyState(194)  & 0x8000) != 0;
   int32 f3  = (GetAsyncKeyState(195)  & 0x8000) != 0;
   int32 f4  = (GetAsyncKeyState(196)  & 0x8000) != 0;
   int32 f10  = (GetAsyncKeyState(134)  & 0x8000) != 0;
#else
   int32 f1  = (GetAsyncKeyState(VK_F1)  & 0x8000) != 0;
   int32 f2  = (GetAsyncKeyState(VK_F2)  & 0x8000) != 0;
   int32 f3  = (GetAsyncKeyState(VK_F3)  & 0x8000) != 0;
   int32 f4  = (GetAsyncKeyState(VK_F4)  & 0x8000) != 0;
   int32 f10 = (GetAsyncKeyState(VK_F10) & 0x8000) != 0;
#endif
   RETURN32( (cursorLeft  ? keyBitPageLeft  : 0) |
             (cursorRight ? keyBitPageRight : 0) |
             (cursorUp    ? keyBitPageUp    : 0) |
             (cursorDown  ? keyBitPageDown  : 0) |
             (f1          ? keyBitHard1     : 0) |
             (f2          ? keyBitHard2     : 0) |
             (f3          ? keyBitHard3     : 0) |
             (f4          ? keyBitHard4     : 0) |
             (f10         ? keyBitAction    : 0) )
#endif
}

//////////////////////////////////////////////////////////////////////////
// waba/sys/NativeMethods static int vmGetRemainingBattery()
void VmGetRemainingBattery(Var stack[]) // ()I
{
   int32 ret;
#ifdef WINCE
   SYSTEM_POWER_STATUS_EX p;

   if (GetSystemPowerStatusEx(&p,TRUE))
   {
      if (p.BatteryLifePercent <= 100)
         ret = p.BatteryLifePercent;
      if (p.BackupBatteryLifePercent < 100) // even if not present, it returns 100, so we will only consider values less than 100
      {
         if (p.BatteryLifePercent == 0xFF) // sometimes, right after the device is undocked, the main battery is at an unknown state
            ret = 0;
         ret += p.BackupBatteryLifePercent;
      }
   }
   if (ret > 100)
      ret = 100;
#else
   ret = 100;
#endif
   RETURN32(ret)
}
